////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./command/guile/globalcommandregister.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "util/myassert.h"
#include "util/exceptions.h"

#include "commandimpl.h"
#include "guilecommandcall.h"

#include "expression_operations.h"
#include "graphic/proofsheet.h"
#include "import/opendbxml.h"
#include "import/importflatproof.h"
#include "import/importotter.h"
#include "export/savedbxml.h"
#include "export/exportflatproof.h"
#include "export/exportlatex.h"
#ifdef HAVE_PDFLIB
#include "graphic/exportpdf.h"
#endif
#include "simplesymboltable.h"

#include "globalcommandregister.h"
#include "command/commandnames.h"
#include "menucommands.h"
#include "guiletypes.h"

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class CmdQuit : public CommandImpl
{
  CMD_COMMAND_NAME( CmdQuit, CMD_FNAME_QUIT )

private:
  SCM Execute()
  {
    Shutdown(); // no return!

    return SCM_UNDEFINED;
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdQuit )

////////////////////////////////////////////////////////////////////////////////

class CmdNewDb : public CommandImpl
{
  CMD_COMMAND_NAME( CmdNewDb, CMD_FNAME_NEW_DB )

private:
  SCM Execute()
  {
    if( GetContext()->GetProofWindow() != NULL )
    {
      GraphicView* v = GetContext()->GetProofWindow()->GetView();
      ASSERT( v != NULL );
  
      GraphicProofView *pv = dynamic_cast<GraphicProofView*>( v );
      ASSERT( pv != NULL );
  
      pv->GetProofSheet()->Clear();
  
      GraphicSystem *gs = GetContext()->GetGraphicSystem();
  
      gs->GetMainInputDispatcher()->Lock();
  
      pv->ZoomAll();
  
      gs->GetMainInputDispatcher()->Unlock();
    }

    delete GetContext()->GetProofDatabase();
    GetContext()->SetProofDatabase(
      new ProofLib::ProofDatabase( new SimpleSymbolTable() ) );

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdNewDb )

////////////////////////////////////////////////////////////////////////////////

class CmdOpenDb : public CommandImpl
{
  CMD_COMMAND_NAME( CmdOpenDb, CMD_FNAME_OPEN_DB )

  CMD_ARGUMENT_STR( FileName )

private:
  SCM Execute()
  {
    if( GetFileName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Filename can't be empty.";
      throw e;
    }

    if( GetContext()->GetProofWindow() != NULL )
    {
      GraphicView* v = GetContext()->GetProofWindow()->GetView();
      ASSERT( v != NULL );
  
      GraphicProofView *pv = dynamic_cast<GraphicProofView*>( v );
      ASSERT( pv != NULL );
  
      pv->GetProofSheet()->Clear();
  
      GraphicSystem *gs = GetContext()->GetGraphicSystem();
  
      gs->GetMainInputDispatcher()->Lock();
  
      pv->ZoomAll();
  
      gs->GetMainInputDispatcher()->Unlock();
    }

    delete GetContext()->GetProofDatabase();
    GetContext()->SetProofDatabase(
      new ProofLib::ProofDatabase( new SimpleSymbolTable() ) );

    ProofLib::OpenDbXML *imp =
      ProofLib::OpenDbXML::New( GetContext()->GetProofDatabase() );

    ProofLib::Proof *p = NULL;

    try
    {
      imp->Open( GetCStrFileName() );

      p = imp->Read();

      imp->Close();

      delete imp;
    }
    catch( ... )
    {
      delete imp;
      throw;
    }

    if( p == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "No proof found.";
      throw e;
    }

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return ToSCM( p->GetName() );
  }
};

CMD_DECLARE_ARG_LIST( CmdOpenDb )
CMD_ARG_LIST_ENTRY( FileName, 0 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdSaveDb : public CommandImpl
{
  CMD_COMMAND_NAME( CmdSaveDb, CMD_FNAME_SAVE_DB )

  CMD_ARGUMENT_STR( FileName )

private:
  SCM Execute()
  {
    if( GetFileName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Filename can't be empty.";
      throw e;
    }

    SaveDbXML *exp = SaveDbXML::New();

    try
    {
      exp->Open( GetCStrFileName() );

      exp->Write( GetContext()->GetProofDatabase() );

      exp->Close();

      delete exp;
    }
    catch( ... )
    {
      delete exp;

      throw;
    }

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdSaveDb )
CMD_ARG_LIST_ENTRY( FileName, 0 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

#if 0
class CmdImport : public CommandImpl
{
  CMD_COMMAND_NAME( CmdImport, CMD_FNAME_IMPORT )

  CMD_ARGUMENT_STR( FileName )

private:
  SCM Execute()
  {
    if( GetFileName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Filename can't be empty.";
      throw e;
    }

    ImportLatexProof imp( GetContext()->GetProofDatabase() );

    imp.Open( GetCStrFileName() );

    ProofLib::Proof *p = imp.Read();

    imp.Close();

    if( p == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "No proof found.";
      throw e;
    }

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return ToSCM( p->GetName() );
  }
};

CMD_DECLARE_ARG_LIST( CmdImport )
CMD_ARG_LIST_ENTRY( FileName, 0 )
CMD_END_DECLARE_ARG_LIST
#endif

////////////////////////////////////////////////////////////////////////////////

class CmdImportProofXml : public CommandImpl
{
  CMD_COMMAND_NAME( CmdImportProofXml, CMD_FNAME_IMPORT_PROOF_XML )

  CMD_ARGUMENT_STR( FileName )

private:
  SCM Execute()
  {
    if( GetFileName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Filename can't be empty.";
      throw e;
    }

    ProofLib::ImportFlatProofXML *imp =
      ProofLib::ImportFlatProofXML::New( GetContext()->GetProofDatabase() );

    ProofLib::Proof *p = NULL;

    try
    {
      imp->Open( GetCStrFileName() );

      p = imp->Read();

      imp->Close();

      delete imp;
    }
    catch( ... )
    {
      delete imp;
      throw;
    }

    if( p == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "No proof found.";
      throw e;
    }

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return ToSCM( p->GetName() );
  }
};

CMD_DECLARE_ARG_LIST( CmdImportProofXml )
CMD_ARG_LIST_ENTRY( FileName, 0 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdImportOtter : public CommandImpl
{
  CMD_COMMAND_NAME( CmdImportOtter, CMD_FNAME_IMPORT_OTTER )

  CMD_ARGUMENT_STR( FileName )

private:
  SCM Execute()
  {
    if( GetFileName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Filename can't be empty.";
      throw e;
    }

    ProofLib::ImportOtterProof *imp =
      ProofLib::ImportOtterProof::New( GetContext()->GetProofDatabase() );

    ProofLib::Proof *p = NULL;

    try
    {
      imp->Open( GetCStrFileName() );

      p = imp->Read();

      imp->Close();

      delete imp;
    }
    catch( ... )
    {
      delete imp;
      throw;
    }

    if( p == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "No proof found.";
      throw e;
    }

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return ToSCM( p->GetName() );
  }
};

CMD_DECLARE_ARG_LIST( CmdImportOtter )
CMD_ARG_LIST_ENTRY( FileName, 0 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdExportProofFlat : public CommandImpl
{
  CMD_COMMAND_NAME( CmdExportProofFlat, CMD_FNAME_EXPORT_PROOF_XML )

  CMD_ARGUMENT_STR( Proof )
  CMD_ARGUMENT_STR( FileName )

private:
  SCM Execute()
  {
    if( GetFileName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Filename can't be empty.";
      throw e;
    }
    if( GetProof().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Proof name can't be empty.";
      throw e;
    }


    ExportFlatProofXML *exp = ExportFlatProofXML::New();

    try
    {
      exp->Open( GetCStrFileName() );

      exp->Write( GetContext()->GetProofDatabase()->GetProof( GetProof() ),
                  GetContext()->GetProofDatabase() );

      exp->Close();

      delete exp;
    }
    catch( ... )
    {
      delete exp;

      throw;
    }

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdExportProofFlat )
CMD_ARG_LIST_ENTRY( Proof, 0 )
CMD_ARG_LIST_ENTRY( FileName, 1 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdExportProofLatex : public CommandImpl
{
  CMD_COMMAND_NAME( CmdExportProofLatex, CMD_FNAME_EXPORT_PROOF_LATEX )

  CMD_ARGUMENT_STR( Proof )
  CMD_ARGUMENT_STR( FileName )

private:
  SCM Execute()
  {
    if( GetFileName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Filename can't be empty.";
      throw e;
    }
    if( GetProof().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Proof name can't be empty.";
      throw e;
    }


    ExportLaTeX *exp = ExportLaTeX::New();

    try
    {
      exp->Open( GetCStrFileName() );

      exp->Write( GetContext()->GetProofDatabase()->GetProof( GetProof() ),
                  GetContext()->GetProofDatabase() );

      exp->Close();

      delete exp;
    }
    catch( ... )
    {
      delete exp;

      throw;
    }

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdExportProofLatex )
CMD_ARG_LIST_ENTRY( Proof, 0 )
CMD_ARG_LIST_ENTRY( FileName, 1 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdExportSequentListLatex : public CommandImpl
{
  CMD_COMMAND_NAME( CmdExportSequentListLatex, CMD_FNAME_EXPORT_SEQUENTLIST_LATEX )

  CMD_ARGUMENT_STR( SequentList )
  CMD_ARGUMENT_STR( FileName )

private:
  SCM Execute()
  {
    if( GetFileName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Filename can't be empty.";
      throw e;
    }
    if( GetSequentList().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Proof name can't be empty.";
      throw e;
    }


    ExportLaTeX *exp = ExportLaTeX::New();

    try
    {
      exp->Open( GetCStrFileName() );

      exp->Write( GetContext()->GetProofDatabase()->GetSequentList( GetSequentList() ),
                  GetContext()->GetProofDatabase() );

      exp->Close();

      delete exp;
    }
    catch( ... )
    {
      delete exp;

      throw;
    }

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdExportSequentListLatex )
CMD_ARG_LIST_ENTRY( SequentList, 0 )
CMD_ARG_LIST_ENTRY( FileName, 1 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdExportLatex : public CommandImpl
{
  CMD_COMMAND_NAME( CmdExportLatex, CMD_FNAME_EXPORT_LATEX )

  CMD_ARGUMENT_STR( FileName )

private:
  SCM Execute()
  {
    if( GetFileName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Filename can't be empty.";
      throw e;
    }


    ExportLaTeX *exp = ExportLaTeX::New();

    try
    {
      exp->Open( GetCStrFileName() );

      exp->Write( GetContext()->GetProofDatabase() );

      exp->Close();

      delete exp;
    }
    catch( ... )
    {
      delete exp;

      throw;
    }

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdExportLatex )
CMD_ARG_LIST_ENTRY( FileName, 0 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

#ifdef HAVE_PDFLIB
class CmdExportPdf : public CommandImpl
{
  CMD_COMMAND_NAME( CmdExportPdf, CMD_FNAME_EXPORT_PDF )

//  CMD_ARGUMENT_STR( Proof )
  CMD_ARGUMENT_STR( FileName )

private:
  SCM Execute()
  {
    if( GetFileName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Filename can't be empty.";
      throw e;
    }
/*
    if( GetProof().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Proof name can't be empty.";
      throw e;
    }
*/

    ExportPdf *exp = ExportPdf::New();

    try
    {
      exp->Open( GetCStrFileName() );

      // get view
      GraphicView *v_ = GetContext()->GetGraphicSystem()->
                        GetActiveProofWindow()->GetView();
      GraphicProofView *v = dynamic_cast<GraphicProofView*>(v_);
      ASSERT( v != NULL );

      exp->Write( v->GetProofSheet() );

      exp->Close();

      delete exp;
    }
    catch( ... )
    {
      delete exp;

      throw;
    }

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdExportPdf )
//CMD_ARG_LIST_ENTRY( Proof, 0 )
CMD_ARG_LIST_ENTRY( FileName, 0 )
CMD_END_DECLARE_ARG_LIST
#endif

////////////////////////////////////////////////////////////////////////////////

class CmdListProofs : public CommandImpl
{
  CMD_COMMAND_NAME( CmdListProofs, CMD_FNAME_LISTPROOFS )

private:
  SCM Execute()
  {
    ProofLib::ProofDatabase* db = GetContext()->GetProofDatabase();

    if( db == NULL )
      return SCM_EOL;

    SCM out = SCM_EOL;

    for( size_t i = 0; i < db->GetCount(); i++ )
    {
      ProofLib::Proof *p = db->GetProof(i);
      out = scm_cons( ToSCM( p->GetName() ), out );
    }

    return out;
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdListProofs )

////////////////////////////////////////////////////////////////////////////////

class CmdListSequentLists : public CommandImpl
{
  CMD_COMMAND_NAME( CmdListSequentLists, CMD_FNAME_LISTSEQUENTLISTS )

private:
  SCM Execute()
  {
    ProofLib::ProofDatabase* db = GetContext()->GetProofDatabase();

    if( db == NULL )
      return SCM_EOL;

    SCM out = SCM_EOL;

    for( size_t i = 0; i < db->GetSequentListCount(); i++ )
    {
      ProofLib::SequentList *l = db->GetSequentList(i);
      out = scm_cons( ToSCM( l->GetName() ), out );
    }

    return out;
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdListSequentLists )

////////////////////////////////////////////////////////////////////////////////

class CmdShowSequentList : public CommandImpl
{
  CMD_COMMAND_NAME( CmdShowSequentList, CMD_FNAME_SHOWSEQUENTLIST )

  CMD_ARGUMENT_STR( SequentListName )

private:
  SCM Execute()
  {
    if( GetSequentListName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Sequent list name can't be empty.";
      throw e;
    }

    GraphicSystem *gs = GetContext()->GetGraphicSystem();

    if( GetContext()->GetProofWindow() == NULL )
    {
      ProofSheet *sheet = new ProofSheet( GetContext()->GetProofDatabase() );
      sheet->SetFontManager( gs->GetFontManager() );

      gs->GetMainInputDispatcher()->Lock();

      GraphicWindow *wnd = gs->NewProofWindow( sheet );
      GetContext()->SetProofWindow( wnd );

      gs->GetMainInputDispatcher()->Unlock();
    }

    GraphicView* v = GetContext()->GetProofWindow()->GetView();
    ASSERT( v != NULL );

    GraphicProofView *pv = dynamic_cast<GraphicProofView*>( v );
    ASSERT( pv != NULL );

    ProofLib::Proof *p = pv->GetProofSheet()->SelectSequentList( GetSequentListName() );

    return ( p == NULL ? SCM_BOOL_F : SCM_BOOL_T );
  }
};

CMD_DECLARE_ARG_LIST( CmdShowSequentList )
CMD_ARG_LIST_ENTRY( SequentListName, 0 )
CMD_END_DECLARE_ARG_LIST


////////////////////////////////////////////////////////////////////////////////

class CmdShowProof : public CommandImpl
{
  CMD_COMMAND_NAME( CmdShowProof, CMD_FNAME_SHOWPROOF )

  CMD_ARGUMENT_STR( ProofName )

private:
  SCM Execute()
  {
    if( GetProofName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Proof name can't be empty.";
      throw e;
    }

    GraphicSystem *gs = GetContext()->GetGraphicSystem();

    if( GetContext()->GetProofWindow() == NULL )
    {
      ProofSheet *sheet = new ProofSheet( GetContext()->GetProofDatabase() );
      sheet->SetFontManager( gs->GetFontManager() );

      gs->GetMainInputDispatcher()->Lock();

      GraphicWindow *wnd = gs->NewProofWindow( sheet );
      GetContext()->SetProofWindow( wnd );

      gs->GetMainInputDispatcher()->Unlock();
    }

    GraphicView* v = GetContext()->GetProofWindow()->GetView();
    ASSERT( v != NULL );

    GraphicProofView *pv = dynamic_cast<GraphicProofView*>( v );
    ASSERT( pv != NULL );

    ProofLib::Proof *p = pv->GetProofSheet()->SelectProof( GetProofName() );

    return ( p == NULL ? SCM_BOOL_F : SCM_BOOL_T );
  }
};

CMD_DECLARE_ARG_LIST( CmdShowProof )
CMD_ARG_LIST_ENTRY( ProofName, 0 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdSplitProof : public CommandImpl
{
  CMD_COMMAND_NAME( CmdSplitProof, CMD_FNAME_SPLITPROOF )

  CMD_ARGUMENT_STR( ProofName )
  CMD_ARGUMENT_SEQUENT( Sequent )

private:
  SCM Execute()
  {
    // check for non empty proof name
    if( GetProofName().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Proof name can't be empty.";
      throw e;
    }

    // get view
    GraphicView *v_ = GetContext()->GetGraphicSystem()->
                      GetActiveProofWindow()->GetView();
    GraphicProofView *v = dynamic_cast<GraphicProofView*>(v_);
    ASSERT( v != NULL );

    // check if item exists
    DrawSequentItem *item = GetSequent();

    if( item == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "No such proof link item found.";
      throw e;
    }

    // check if item exists on current proof sheet
    if( !v->GetProofSheet()->Contains( item ) )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected item doesn't exist.";
      throw e;
    }

    // get proof db
    ProofLib::ProofDatabase *pdb = GetContext()->GetProofDatabase();
    ASSERT( pdb != NULL );

    // check that item is not successor of an proof link
    ProofLib::ProofLink *l_ = dynamic_cast<ProofLib::ProofLink*>(
                                item->GetSequent()->GetParent() );

    if( l_ != NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected item is successor of a proof link.";
      throw e;
    }

    pdb->SplitProof( v->GetProofSheet()->GetActiveProof(),
                     item->GetSequent(), GetProofName() );


// next lines should be useless in future
    v->GetProofSheet()->SelectProof(
      v->GetProofSheet()->GetActiveProof()->GetName() );

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif
    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdSplitProof )
CMD_ARG_LIST_ENTRY( ProofName, 0 )
CMD_ARG_LIST_ENTRY( Sequent, 1 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdUnsplitProof : public CommandImpl
{
  CMD_COMMAND_NAME( CmdUnsplitProof, CMD_FNAME_UNSPLITPROOF )

  CMD_ARGUMENT_PROOFLINK( ProofLink )

private:
  SCM Execute()
  {
    if( !IsProofLinkUp() )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected item isn't a proof up link.";
      throw e;
    }

    GraphicView *v_ = GetContext()->GetGraphicSystem()->
                      GetActiveProofWindow()->GetView();
    GraphicProofView *v = dynamic_cast<GraphicProofView*>(v_);
    ASSERT( v != NULL );

    DrawToProofLinkItem *item = GetProofLinkUp();

    if( item == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected item doesn't exist.";
      throw e;
    }

    if( !v->GetProofSheet()->Contains( item ) )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected item doesn't exist.";
      throw e;
    }

    ProofLib::ProofDatabase *pdb = GetContext()->GetProofDatabase();
    ASSERT( pdb != NULL );

    ProofLib::ProofLink *plink = item->GetProofLink();

    pdb->UnsplitProof( plink );

// next lines should be useless in future
    v->GetProofSheet()->SelectProof(
      v->GetProofSheet()->GetActiveProof()->GetName() );

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdUnsplitProof )
CMD_ARG_LIST_ENTRY( ProofLink, 0 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdSubstFormula : public CommandImpl
{
  CMD_COMMAND_NAME( CmdSubstFormula, CMD_FNAME_SUBSTFORMULA )

  CMD_ARGUMENT_STR( FormulaVar )
  CMD_ARGUMENT_SEQUENT( FormulaRef )
  CMD_ARGUMENT_SYM( Scope )

private:
  SCM Execute()
  {
    if( GetFormulaVar().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Formula variable name can't be empty.";
      throw e;
    }

    GraphicView *v_ = GetContext()->GetGraphicSystem()->
                      GetActiveProofWindow()->GetView();
    GraphicProofView *v = dynamic_cast<GraphicProofView*>(v_);
    ASSERT( v != NULL );

    DrawSequentItem *item = GetFormulaRef();

    ASSERT( item != NULL );

    if( !v->GetProofSheet()->Contains( item ) )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected item doesn't exist.";
      throw e;
    }

    ProofLib::Formula *to_subst = dynamic_cast<ProofLib::Formula*>
              ( GetFormulaRefNodePointer().GetDataNode( item->GetSequent() ) );

    if( to_subst == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected item isn't a formula.";
      throw e;
    }

    const ProofLib::Symbol *symbol =
      GetContext()->GetSymbolTable()->
      GetOrAdd( GetFormulaVar(), 0, SYMBOL_TYPE_PREDICATE );

    ASSERT( symbol != NULL );

    ProofLib::FormulaVariable *fv  = ProofLib::FormulaVariable::
      New( GetContext()->GetNodeDatabase(), symbol, to_subst );

    if( fv == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Formula variable '" << GetCStrFormulaVar()
                  << "' is already defined different.";
      throw e;
    }

    ProofLib::OperationReplaceFormula
      rep( GetContext()->GetNodeDatabase(), to_subst, fv );

    if( !strcasecmp( GetScope(), SUBST_HERE ) )
    {
      ProofLib::OperationScope *scope =
        new ProofLib::OperationScope( GetFormulaRefNodePointer(),
                                      ProofLib::NodePointer() );
      rep.SetScope( scope );
      item->GetSequent()->Apply( rep );
      delete scope;
    }
    else if( !strcasecmp( GetScope(), SUBST_IN_SEQUENT ) )
      item->GetSequent()->Apply( rep );
    else if( !strcasecmp( GetScope(), SUBST_ABOVE_SEQUENT ) )
      item->GetSequent()->GetParent()->Apply( rep );
    else if( !strcasecmp( GetScope(), SUBST_IN_PROOF ) )
      v->GetProofSheet()->GetActiveProof()->Apply( rep );
    else
    {
      GenericException *e = new GenericException();
      e->Stream() << "Scope '" << GetScope() << "' is unknown.";
      throw e;
    }

    v->GetProofSheet()->Layout();

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdSubstFormula )
CMD_ARG_LIST_ENTRY( FormulaVar, 0 )
CMD_ARG_LIST_ENTRY( FormulaRef, 1 )
CMD_ARG_LIST_ENTRY( Scope, 2 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdUnsubstFormula : public CommandImpl
{
  CMD_COMMAND_NAME( CmdUnsubstFormula, CMD_FNAME_UNSUBSTFORMULA )

  CMD_ARGUMENT_SEQUENT( FormulaRef )
  CMD_ARGUMENT_SYM( Scope )

private:
  SCM Execute()
  {
    GraphicView *v_ = GetContext()->GetGraphicSystem()->
                      GetActiveProofWindow()->GetView();
    GraphicProofView *v = dynamic_cast<GraphicProofView*>(v_);
    ASSERT( v != NULL );

    DrawSequentItem *item = GetFormulaRef();

    ASSERT( item != NULL );

    if( !v->GetProofSheet()->Contains( item ) )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected item doesn't exist.";
      throw e;
    }

    ProofLib::FormulaVariable *fv =
              dynamic_cast<ProofLib::FormulaVariable*>
              ( GetFormulaRefNodePointer().GetDataNode( item->GetSequent() ) );

    if( fv == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected item isn't a formula variable.";
      throw e;
    }

    ProofLib::OperationReplaceFormula
      rep( GetContext()->GetNodeDatabase(), fv, fv->GetFormula() );

    if( !strcasecmp( GetScope(), SUBST_HERE ) )
    {
      ProofLib::OperationScope *scope =
        new ProofLib::OperationScope( GetFormulaRefNodePointer(),
                                      ProofLib::NodePointer() );
      rep.SetScope( scope );
      item->GetSequent()->Apply( rep );
      delete scope;
    }
    else if( !strcasecmp( GetScope(), SUBST_IN_SEQUENT ) )
      item->GetSequent()->Apply( rep );
    else if( !strcasecmp( GetScope(), SUBST_ABOVE_SEQUENT ) )
      item->GetSequent()->GetParent()->Apply( rep );
    else if( !strcasecmp( GetScope(), SUBST_IN_PROOF ) )
      v->GetProofSheet()->GetActiveProof()->Apply( rep );
    else
    {
      GenericException *e = new GenericException();
      e->Stream() << "Scope '" << GetScope() << "' is unknown.";
      throw e;
    }

    v->GetProofSheet()->Layout();

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdUnsubstFormula )
CMD_ARG_LIST_ENTRY( FormulaRef, 0 )
CMD_ARG_LIST_ENTRY( Scope, 1 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdSubstSElements : public CommandImpl
{
  CMD_COMMAND_NAME( CmdSubstSElements, CMD_FNAME_SUBSTSELEMENTS )

  CMD_ARGUMENT_STR( FormulaListVar )
  CMD_ARGUMENT_SEQUENT( SElementRefStart )
  CMD_ARGUMENT_SEQUENT( SElementRefEnd )
  CMD_ARGUMENT_SYM( Scope )

private:
  SCM Execute()
  {
    if( GetFormulaListVar().length() == 0 )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Formula list variable name can't be empty.";
      throw e;
    }

    GraphicView *v_ = GetContext()->GetGraphicSystem()->
                      GetActiveProofWindow()->GetView();
    GraphicProofView *v = dynamic_cast<GraphicProofView*>(v_);
    ASSERT( v != NULL );

    ProofLib::NodePointer np_start = GetSElementRefStartNodePointer();
    ProofLib::NodePointer np_end = GetSElementRefEndNodePointer();

    DrawSequentItem *item = GetSElementRefStart();

    ASSERT( item != NULL );

    if( !v->GetProofSheet()->Contains( item ) )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected start item doesn't exist.";
      throw e;
    }

    ProofLib::SElement *to_subst_start = dynamic_cast<ProofLib::SElement*>
      ( np_start.GetDataNode( item->GetSequent() ) );

    if( to_subst_start == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected start item isn't a s-formula.";
      throw e;
    }

    if( item != GetSElementRefEnd() )
    {
      GenericException *e = new GenericException();
      e->Stream() <<
        "Selected end item isn't in the same sequent with the start item.";
      throw e;
    }

    ProofLib::SElement *to_subst_end = dynamic_cast<ProofLib::SElement*>
      ( np_end.GetDataNode( item->GetSequent() ) );

    if( to_subst_end == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected end item isn't a s-formula.";
      throw e;
    }

    size_t index_start = np_start.Pop();
    size_t index_end   = np_end.Pop();

    if( np_start != np_end ) // have to be the same now (same side of seq)
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected s-formulas are not on the same side of sequent.";
      throw e;
    }

    if( index_end < index_start )
      std::swap( index_start, index_end );

    ProofLib::SElementList *slist =
      np_start.GetTypedDataNode<ProofLib::SElementList>( item->GetSequent() );

    ASSERT( index_end < slist->GetChildCount() );

    ProofLib::FormulaList list;
    
    for( size_t i = index_start; i <= index_end; i++ ) 
    {
      ProofLib::SFormula *f =
        dynamic_cast<ProofLib::SFormula*>( slist->GetSElement( i ) );

      if( f != NULL )
      {
        list.AddRight( f->GetFormula() );
        continue;
      }

      ProofLib::SFormulaListVariable *fv =
       dynamic_cast<ProofLib::SFormulaListVariable*>( slist->GetSElement( i ) );

      if( fv != NULL )
      {
        list.AddRight(
          (ProofLib::FormulaListElement*) fv->GetFormulaListVariable() );
        continue;
      }

      ASSERT( false );
    }

    const ProofLib::Symbol *symbol =
      GetContext()->GetSymbolTable()->
      GetOrAdd( GetFormulaListVar(), 0, SYMBOL_TYPE_PREDICATE );

    ASSERT( symbol != NULL );

    ProofLib::FormulaListVariable *fv  =
      ProofLib::FormulaListVariable::New(
                      GetContext()->GetNodeDatabase(),
                      symbol, &list );
#if 0
    ProofLib::FormulaListVariable *fv  = GetContext()->GetProofDatabase()->
      NewFormulaListVariable( GetFormulaListVar(), &list );
#endif

    if( fv == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Formula list variable '" << GetCStrFormulaListVar()
                  << "' is already defined different.";
      throw e;
    }

#if 0
    ProofLib::SElementReplacement rep( GetContext()->GetProofDatabase(), fv );

    if( !strcasecmp( GetScope(), SUBST_HERE ) )
    {
      rep.SetReplacePosition( GetSElementRefStartNodePointer() );
      item->GetSequent()->ApplyReplacement( &rep );
    }
    else if( !strcasecmp( GetScope(), SUBST_IN_SEQUENT ) )
      item->GetSequent()->ApplyReplacement( &rep );
    else if( !strcasecmp( GetScope(), SUBST_ABOVE_SEQUENT ) )
      item->GetSequent()->GetParent()->ApplyReplacement( &rep );
    else if( !strcasecmp( GetScope(), SUBST_IN_PROOF ) )
      v->GetProofSheet()->GetActiveProof()->ApplyReplacement( &rep );
    else
    {
      GenericException *e = new GenericException();
      e->Stream() << "Scope '" << GetScope() << "' is unknown.";
      throw e;
    }
#endif
    ProofLib::OperationReplaceFormulaSequence
      rep( GetContext()->GetNodeDatabase(), fv->GetFormulaList(), fv );

    if( !strcasecmp( GetScope(), SUBST_HERE ) )
    {
      ProofLib::OperationScope *scope =
        new ProofLib::OperationScope( GetSElementRefStartNodePointer(),
                                      ProofLib::NodePointer() );
      rep.SetScope( scope );
      item->GetSequent()->Apply( rep );
      delete scope;
    }
    else if( !strcasecmp( GetScope(), SUBST_IN_SEQUENT ) )
      item->GetSequent()->Apply( rep );
    else if( !strcasecmp( GetScope(), SUBST_ABOVE_SEQUENT ) )
      item->GetSequent()->GetParent()->Apply( rep );
    else if( !strcasecmp( GetScope(), SUBST_IN_PROOF ) )
      v->GetProofSheet()->GetActiveProof()->Apply( rep );
    else
    {
      GenericException *e = new GenericException();
      e->Stream() << "Scope '" << GetScope() << "' is unknown.";
      throw e;
    }

    v->GetProofSheet()->Layout();

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdSubstSElements )
CMD_ARG_LIST_ENTRY( FormulaListVar, 0 )
CMD_ARG_LIST_ENTRY( SElementRefStart, 1 )
CMD_ARG_LIST_ENTRY( SElementRefEnd, 2 )
CMD_ARG_LIST_ENTRY( Scope, 3 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdUnsubstSElements : public CommandImpl
{
  CMD_COMMAND_NAME( CmdUnsubstSElements, CMD_FNAME_UNSUBSTSELEMENTS )

  CMD_ARGUMENT_SEQUENT( SFormulaListVarRef )
  CMD_ARGUMENT_SYM( Scope )

private:
  SCM Execute()
  {
    GraphicView *v_ = GetContext()->GetGraphicSystem()->
                      GetActiveProofWindow()->GetView();
    GraphicProofView *v = dynamic_cast<GraphicProofView*>(v_);
    ASSERT( v != NULL );

    DrawSequentItem *item = GetSFormulaListVarRef();

    ASSERT( item != NULL );

    if( !v->GetProofSheet()->Contains( item ) )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected item doesn't exist.";
      throw e;
    }

    ProofLib::SFormulaListVariable *fv =
      dynamic_cast<ProofLib::SFormulaListVariable*>
      ( GetSFormulaListVarRefNodePointer().GetDataNode( item->GetSequent() ) );

    if( fv == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Selected item isn't a s-formula list variable.";
      throw e;
    }

    ProofLib::OperationReplaceFormulaListElementBySequence
      rep( GetContext()->GetNodeDatabase(),
           fv->GetFormulaListVariable(),
           fv->GetFormulaListVariable()->GetFormulaList() );

    if( !strcasecmp( GetScope(), SUBST_HERE ) )
    {
      ProofLib::OperationScope *scope =
        new ProofLib::OperationScope( GetSFormulaListVarRefNodePointer(),
                                      ProofLib::NodePointer() );
      rep.SetScope( scope );
      item->GetSequent()->Apply( rep );
      delete scope;
    }
    else if( !strcasecmp( GetScope(), SUBST_IN_SEQUENT ) )
      item->GetSequent()->Apply( rep );
    else if( !strcasecmp( GetScope(), SUBST_ABOVE_SEQUENT ) )
      item->GetSequent()->GetParent()->Apply( rep );
    else if( !strcasecmp( GetScope(), SUBST_IN_PROOF ) )
      v->GetProofSheet()->GetActiveProof()->Apply( rep );
    else
    {
      GenericException *e = new GenericException();
      e->Stream() << "Scope '" << GetScope() << "' is unknown.";
      throw e;
    }

#if 0
    ProofLib::FormulaListVarReplacement rep =
      ProofLib::FormulaListVarReplacement( GetContext()->GetProofDatabase(),
                                           fv->GetFormulaListVariable() );

    if( !strcasecmp( GetScope(), SUBST_HERE ) )
    {
      rep.SetReplacePosition( GetSFormulaListVarRefNodePointer() );
      item->GetSequent()->ApplyReplacement( &rep );
    }
    else if( !strcasecmp( GetScope(), SUBST_IN_SEQUENT ) )
      item->GetSequent()->ApplyReplacement( &rep );
    else if( !strcasecmp( GetScope(), SUBST_ABOVE_SEQUENT ) )
      item->GetSequent()->GetParent()->ApplyReplacement( &rep );
    else if( !strcasecmp( GetScope(), SUBST_IN_PROOF ) )
      v->GetProofSheet()->GetActiveProof()->ApplyReplacement( &rep );
    else
    {
      GenericException *e = new GenericException();
      e->Stream() << "Scope '" << GetScope() << "' is unknown.";
      throw e;
    }
#endif

    v->GetProofSheet()->Layout();

#ifdef DEBUG
    GetContext()->GetProofDatabase()->DumpStats();
#endif

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdUnsubstSElements )
CMD_ARG_LIST_ENTRY( SFormulaListVarRef, 0 )
CMD_ARG_LIST_ENTRY( Scope, 1 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdSelectProofLink : public CommandImpl
{
  CMD_COMMAND_NAME( CmdSelectProofLink, CMD_FNAME_SELECTPROOFLINK )

private:
  SCM Execute()
  {
    GetContext()->GetCommandProcessor()->BeginUserInput();

    DrawProofLinkItem *item =
      GetContext()->GetGraphicSystem()->
      GetMainInputDispatcher()->SelectProofLinkItem();

    GetContext()->GetCommandProcessor()->EndUserInput();

    if( item == NULL )
      return SCM_UNDEFINED;

    return ScmObjProofLink::Make( item );
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdSelectProofLink )

////////////////////////////////////////////////////////////////////////////////

class CmdSelectFormulaVar : public CommandImpl
{
  CMD_COMMAND_NAME( CmdSelectFormulaVar, CMD_FNAME_SELECTFORMULAVAR )

private:
  SCM Execute()
  {
    GetContext()->GetCommandProcessor()->BeginUserInput();

    DrawSequentItem *item =
      GetContext()->GetGraphicSystem()->
      GetMainInputDispatcher()->SelectFormulaVar();

    GetContext()->GetCommandProcessor()->EndUserInput();

    if( item == NULL )
      return SCM_UNDEFINED;

    return ScmObjSequent::Make( item, item->GetHighlightPart() );
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdSelectFormulaVar )

////////////////////////////////////////////////////////////////////////////////

class CmdSelectFormula : public CommandImpl
{
  CMD_COMMAND_NAME( CmdSelectFormula, CMD_FNAME_SELECTFORMULA )

private:
  SCM Execute()
  {
    GetContext()->GetCommandProcessor()->BeginUserInput();

    DrawSequentItem *item =
      GetContext()->GetGraphicSystem()->
      GetMainInputDispatcher()->SelectFormula();

    GetContext()->GetCommandProcessor()->EndUserInput();

    if( item == NULL )
      return SCM_UNDEFINED;

    return ScmObjSequent::Make( item, item->GetHighlightPart() );
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdSelectFormula )

////////////////////////////////////////////////////////////////////////////////

// TODO: Optional prompt
class CmdSelectSElement : public CommandImpl
{
  CMD_COMMAND_NAME( CmdSelectSElement, CMD_FNAME_SELECTSELEMENT )

private:
  SCM Execute()
  {
    GetContext()->GetCommandProcessor()->BeginUserInput();

    DrawSequentItem *item =
      GetContext()->GetGraphicSystem()->
      GetMainInputDispatcher()->SelectSElement();
  
    GetContext()->GetCommandProcessor()->EndUserInput();

    if( item == NULL )
      return SCM_UNDEFINED;
  
    return ScmObjSequent::Make( item, item->GetHighlightPart() );
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdSelectSElement )

////////////////////////////////////////////////////////////////////////////////

// TODO: Optional prompt
class CmdSelectSFormulaListVar : public CommandImpl
{
  CMD_COMMAND_NAME( CmdSelectSFormulaListVar,
                    CMD_FNAME_SELECTSFORMULALISTVAR )

private:
  SCM Execute()
  {
    GetContext()->GetCommandProcessor()->BeginUserInput();

    DrawSequentItem *item =
      GetContext()->GetGraphicSystem()->
      GetMainInputDispatcher()->SelectSFormulaListVar();
  
    GetContext()->GetCommandProcessor()->EndUserInput();

    if( item == NULL )
      return SCM_UNDEFINED;
  
    return ScmObjSequent::Make( item, item->GetHighlightPart() );
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdSelectSFormulaListVar )

////////////////////////////////////////////////////////////////////////////////

// TODO: Optional prompt
class CmdSelectSequent : public CommandImpl
{
  CMD_COMMAND_NAME( CmdSelectSequent, CMD_FNAME_SELECTSEQUENT )

private:
  SCM Execute()
  {
    GetContext()->GetCommandProcessor()->BeginUserInput();

    DrawSequentItem *item =
      GetContext()->GetGraphicSystem()->
      GetMainInputDispatcher()->SelectSequentItem();
  
    GetContext()->GetCommandProcessor()->EndUserInput();

    if( item == NULL )
      return SCM_UNDEFINED;
  
    return ScmObjSequent::Make( item, ProofLib::NodePointer() );
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdSelectSequent )

////////////////////////////////////////////////////////////////////////////////

// TODO: Optional prompt
class CmdDlgInput : public CommandImpl
{
  CMD_COMMAND_NAME( CmdDlgInput, CMD_FNAME_DLGINPUT )

private:
  SCM Execute()
  {
    USTRING title = L"Input";
 
    GraphicInputDialog *dlg =
      GetContext()->GetGraphicSystem()->NewInputDialog( title );
  
    dlg->SetModal( true );

    GetContext()->GetCommandProcessor()->BeginUserInput();

    dlg->Do();

    GetContext()->GetCommandProcessor()->EndUserInput();

    const char *res = dlg->GetInput();
  
    SCM ret = SCM_UNDEFINED;

    if( res != NULL )
     ret = scm_makfrom0str( strdup(res) );
  
    delete dlg;

    return ret;
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdDlgInput )

////////////////////////////////////////////////////////////////////////////////

class CmdDlgFile : public CommandImpl
{
  CMD_COMMAND_NAME( CmdDlgFile, CMD_FNAME_DLGFILE )

  CMD_ARGUMENT_STR( Title )
  CMD_ARGUMENT_SYM( Type )

private:
  SCM Execute()
  {
    bool openDlg = true;

    if( !strcasecmp( GetType(), DLGFILE_TYPE_OPEN ) )
    {
    }
    else if( !strcasecmp( GetType(), DLGFILE_TYPE_SAVE ) )
    {
      openDlg = false;
    }
    else
    {
      GenericException *e = new GenericException();
      e->Stream() << "Type '" << GetType() << "' is unknown.";
      throw e;
    }
    
    GraphicFileDialog *dlg =
      GetContext()->GetGraphicSystem()->NewFileDialog( GetTitle(), openDlg );
  
    dlg->SetModal( true );

    GetContext()->GetCommandProcessor()->BeginUserInput();

    dlg->Do();

    GetContext()->GetCommandProcessor()->EndUserInput();

    const char *res = dlg->GetFile();

    SCM ret = SCM_UNDEFINED;

    if( res != NULL )
     ret = scm_makfrom0str( res );
  
    delete dlg;

    return ret;
  }
};

CMD_DECLARE_ARG_LIST( CmdDlgFile )
CMD_ARG_LIST_ENTRY( Title, 0 )
CMD_ARG_LIST_ENTRY( Type, 1 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdZoomAll : public CommandImpl
{
  CMD_COMMAND_NAME( CmdZoomAll, CMD_FNAME_ZOOMALL )

private:
  SCM Execute()
  {
    GraphicView *v_ = GetContext()->GetGraphicSystem()->
                      GetActiveProofWindow()->GetView();
    GraphicProofView *v = dynamic_cast<GraphicProofView*>(v_);

    ASSERT( v != NULL );

    GraphicMainInputDispatcher *mi =
      GetContext()->GetGraphicSystem()->GetMainInputDispatcher();

    mi->Lock();
    v->ZoomAll();
    mi->Unlock();

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdZoomAll )

////////////////////////////////////////////////////////////////////////////////

class CmdZoomGetFactor : public CommandImpl
{
  CMD_COMMAND_NAME( CmdZoomGetFactor, CMD_FNAME_GETZOOMFACTOR )

private:
  SCM Execute()
  {
    GraphicView *v_ = GetContext()->GetGraphicSystem()->
                      GetActiveProofWindow()->GetView();
    GraphicProofView *v = dynamic_cast<GraphicProofView*>(v_);

    ASSERT( v != NULL );

    return scm_make_real( v->GetZoomFactor() );
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdZoomGetFactor )

////////////////////////////////////////////////////////////////////////////////

class CmdZoomGetMidpoint : public CommandImpl
{
  CMD_COMMAND_NAME( CmdZoomGetMidpoint, CMD_FNAME_GETZOOMMIDPT )

private:
  SCM Execute()
  {
    GraphicView *v_ = GetContext()->GetGraphicSystem()->
                      GetActiveProofWindow()->GetView();
    GraphicProofView *v = dynamic_cast<GraphicProofView*>(v_);

    ASSERT( v != NULL );

    return scm_cons( scm_make_real( v->GetZoomMidpoint().m_x ),
                     scm_make_real( v->GetZoomMidpoint().m_y ) );
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdZoomGetMidpoint )

////////////////////////////////////////////////////////////////////////////////

class CmdZoomFactor : public CommandImpl
{
  CMD_COMMAND_NAME( CmdZoomFactor, CMD_FNAME_ZOOMFACTOR )

  CMD_ARGUMENT_DBL( Factor )
  CMD_ARGUMENT_DBL( X )
  CMD_ARGUMENT_DBL( Y )

private:
  SCM Execute()
  {
    GraphicView *v_ = GetContext()->GetGraphicSystem()->
                      GetActiveProofWindow()->GetView();
    GraphicProofView *v = dynamic_cast<GraphicProofView*>(v_);

    GraphicMainInputDispatcher *mi =
      GetContext()->GetGraphicSystem()->GetMainInputDispatcher();

    mi->Lock();
    v->SetZoomMidpoint( DrawPoint( GetX(), GetY() ), GetFactor() );
    mi->Unlock();
  
    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdZoomFactor )
CMD_ARG_LIST_ENTRY( Factor, 0 )
CMD_ARG_LIST_ENTRY( X, 1 )
CMD_ARG_LIST_ENTRY( Y, 2 )
CMD_END_DECLARE_ARG_LIST


////////////////////////////////////////////////////////////////////////////////

class CmdBindKey : public CommandImpl
{
  CMD_COMMAND_NAME( CmdBindKey, CMD_FNAME_BINDKEY )

  CMD_ARGUMENT_INT( Key )
  CMD_ARGUMENT_INT( Mods )
  CMD_ARGUMENT_OBJ( Call )

private:
  SCM Execute()
  {
    GraphicMainInputDispatcher *mi =
      GetContext()->GetGraphicSystem()->GetMainInputDispatcher();

    GuileCommandCall *cmd = new GuileCommandCall;

    cmd->SetSCM( GetCall() );

    mi->Lock();
    mi->GetAccelDb()->AddCommand( GetKey(), GetMods(), cmd );
    mi->Unlock();

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdBindKey )
CMD_ARG_LIST_ENTRY( Key, 0 )
CMD_ARG_LIST_ENTRY( Mods, 1 )
CMD_ARG_LIST_ENTRY( Call, 2 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////

class CmdBindButton : public CommandImpl
{
  CMD_COMMAND_NAME( CmdBindButton, CMD_FNAME_BINDBUTTON )

  CMD_ARGUMENT_INT( Button )
  CMD_ARGUMENT_INT( Mods )
  CMD_ARGUMENT_OBJ( Call )

private:
  SCM Execute()
  {
    GraphicMainInputDispatcher *mi =
      GetContext()->GetGraphicSystem()->GetMainInputDispatcher();

    GuileCommandCall *cmd = new GuileCommandCall;

    cmd->SetSCM( GetCall() );

    mi->Lock();
    mi->GetButtonDb()->AddCommand( GetButton(), GetMods(), cmd );
    mi->Unlock();

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_ARG_LIST( CmdBindButton )
CMD_ARG_LIST_ENTRY( Button, 0 )
CMD_ARG_LIST_ENTRY( Mods, 1 )
CMD_ARG_LIST_ENTRY( Call, 2 )
CMD_END_DECLARE_ARG_LIST

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

#ifdef DEBUG
class CmdDEBUGPdbStats : public CommandImpl
{
  CMD_COMMAND_NAME( CmdDEBUGPdbStats, "debug-pdb-stats" )

private:
  SCM Execute()
  {
    GetContext()->GetProofDatabase()->DumpStats();

    return SCM_BOOL_T;
  }
};

CMD_DECLARE_EMPTY_ARG_LIST( CmdDEBUGPdbStats )
#endif

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

void RegisterGlobalCommands()
{
  ScmObjSequent::Init();
  ScmObjProofLink::Init();

  CmdQuit::Register();

  CmdNewDb::Register();
  CmdOpenDb::Register();
  CmdSaveDb::Register();

#if 0
  CmdImport::Register();
#endif
  CmdImportProofXml::Register();
  CmdImportOtter::Register();
  CmdExportProofFlat::Register();
  CmdExportProofLatex::Register();
  CmdExportSequentListLatex::Register();
  CmdExportLatex::Register();
#ifdef HAVE_PDFLIB
  CmdExportPdf::Register();
#endif

  CmdListProofs::Register();
  CmdShowProof::Register();
  CmdSplitProof::Register();
  CmdUnsplitProof::Register();
  CmdSubstFormula::Register();
  CmdUnsubstFormula::Register();
  CmdSubstSElements::Register();
  CmdUnsubstSElements::Register();

  CmdListSequentLists::Register();
  CmdShowSequentList::Register();

  CmdSelectFormulaVar::Register();
  CmdSelectFormula::Register();
  CmdSelectSElement::Register();
  CmdSelectSFormulaListVar::Register();
  CmdSelectSequent::Register();
  CmdSelectProofLink::Register();

  CmdDlgInput::Register();
  CmdDlgFile::Register();

  CmdZoomAll::Register();
  CmdZoomFactor::Register();
  CmdZoomGetFactor::Register();
  CmdZoomGetMidpoint::Register();

  CmdBindKey::Register();
  CmdBindButton::Register();

  RegisterMenuCommands();

#ifdef DEBUG
  CmdDEBUGPdbStats::Register();
#endif
}

