////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/datastring.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <typeinfo>

#include "libproof.h"
#include "datastring.h"

#include <iostream>
#include <sstream>
#ifdef DEBUG
using namespace std;
#endif

////////////////////////////////////////////////////////////////////////////////

DataChar::DataChar()
{
  m_char = 0x0000;
  m_cal = m_italic = false;
  m_sup = m_sub = NULL;
  m_start = 0;
  m_end = 0;
  m_y = 0;
}

#ifdef DEBUG
void DataChar::Dump( ostream &out ) const
{
  if( isprint( m_char ) )
    out << (char) m_char;
  else
  {
    ostream::fmtflags of = out.flags( ostream::hex );
    out << "0x" << (int) m_char;
    out.flags( of );
  }
}
#endif

////////////////////////////////////////////////////////////////////////////////

DataString::DataString()
{
//  m_buffer.resize(1);
}

DataString::DataString( size_t len )
{
  m_buffer.resize( len );
}

DataChar& DataString::operator [] ( size_t i )
{
  ASSERT( i < GetLength() );
  return m_buffer[i];
}

const DataChar& DataString::operator [] ( size_t i ) const
{
  ASSERT( i < GetLength() );
  return m_buffer[i];
}

void DataString::operator += ( const DataChar &c )
{
  size_t index = GetLength();
  m_buffer.resize( index + 1 );
  m_buffer[index] = c;
}

#ifdef DEBUG
void DataString::Dump( ostream &out ) const
{
  for( size_t i = 0; i < GetLength(); i++ )
  {
    m_buffer[i].Dump( out );
  }
}
#endif

////////////////////////////////////////////////////////////////////////////////

DataStream::DataStream( DataString &dest )
  : m_dest( dest )
{
  m_state = Normal;
  m_next_stream = NULL; 
  m_formulastart = true;
  m_termstart = true;
}

DataStream::~DataStream()
{
  if( m_next_stream != NULL )
    delete m_next_stream;
  m_next_stream = NULL;
}

inline void DataStream::Reset()
{
#ifdef DEBUG
  if( m_state != Normal )
    std::cerr << "DataStream::Reset (" << __FILE__ << ":" << __LINE__ 
              << ") state was not normal." << std::endl;
#endif
  m_state = Normal;
  if( m_next_stream != NULL )
    delete m_next_stream;
  m_next_stream = NULL; 
}

inline void DataStream::AddChar( UCHAR c )
{
  m_char.m_np   = m_np;
  m_char.m_char = c;

  size_t index = m_dest.GetLength();
  m_dest += m_char;

  m_char = DataChar();
  m_char.m_italic = m_dest[index].m_italic;
  m_char.m_cal = m_dest[index].m_cal;
}

inline void DataStream::SetItalic( bool on )
{
  m_char.m_italic = on;
}

inline void DataStream::SetCal( bool on )
{
  m_char.m_cal = on;
}

inline DataStream* DataStream::GetSuperscript()
{
  if( m_char.m_sup == NULL )
    m_char.m_sup = new DataString;
  
  DataStream *stream = new DataStream( *(m_char.m_sup) );
  stream->m_np = m_np;
  return stream;
}

inline DataStream* DataStream::GetSubscript()
{
  if( m_char.m_sub == NULL )
    m_char.m_sub = new DataString;
  
  DataStream *stream = new DataStream( *(m_char.m_sub) );
  stream->m_np = m_np;
  return stream;
}

DataStream& DataStream::operator << ( UCHAR c )
{
  size_t index = m_dest.GetLength() - 1;

  switch( m_state )
  {
  case Normal:
    switch( c )
    {
//    case DCHAR_LATEX_BEGIN:
//      m_state = ReadLatexSymbol;
//      break;

    case DCHAR_SUPERSCRIPT:
      if( m_dest[index].m_sup == NULL )
        m_dest[index].m_sup = new DataString;

      m_next_stream = new DataStream( *(m_dest[index].m_sup) );
      m_next_stream->m_np = m_np;

      m_state = StartSuperScript;
      break;

    case DCHAR_SUBSCRIPT:
      if( m_dest[index].m_sub == NULL )
        m_dest[index].m_sub = new DataString;

      m_next_stream = new DataStream( *(m_dest[index].m_sub) );
      m_next_stream->m_np = m_np;

      m_state = StartSubScript;
      break;

    case DCHAR_LEFT_BRAKET:
      m_next_stream = new DataStream( m_dest );
      m_next_stream->m_np = m_np;

      m_state = InSequence;
      break;

    case DCHAR_RIGHT_BRAKET:
      m_state = EndOfSequence;
      break;

    default:
      AddChar( c );
    }
    break;

  case ReadLatexSymbol:
    m_state = Normal;
    break;

  case StartSuperScript:
  case StartSubScript:
    if( c != DCHAR_LEFT_BRAKET )
    {
      (*m_next_stream) << c;
      if( m_next_stream->GetState() == Normal )
        m_state = Normal;
      else
        m_state = WaitForNormal;
    }
    else
      m_state = InSequence;
    break;

  case InSequence:
    (*m_next_stream) << c;
    if( m_next_stream->GetState() == EndOfSequence )
    {
      m_state = Normal;
      delete m_next_stream;
      m_next_stream = NULL;
    }
    break;

  case WaitForNormal:
    (*m_next_stream) << c;
    if( m_next_stream->GetState() == Normal )
    {
      m_state = Normal;
      delete m_next_stream;
      m_next_stream = NULL;
    }
    break;
  
  default:
    ASSERT( false );
  }


  return (*this);
#if 0
  if( m_current_index + 1 >= m_buffer.size() )
    m_buffer.resize( m_current_index + 2 );

  m_char.m_np = m_np;
  m_buffer[m_current_index++].m_char = c;

  m_char.m_italic = m_buffer[m_current_index-1].m_italic;
  m_char.m_cal = m_buffer[m_current_index-1].m_cal;
#endif

#if 0
  m_buffer[m_current_index-1].Dump( std::cout );
  std::cout << ": ";
  m_buffer[m_current_index-1].m_np.Dump( std::cout );
  std::cout << endl;
#endif
}

DataStream& DataStream::operator << ( char c )
{
  /// \todo utf8 decoding?
  (*this) << ( UCHAR ) c;

  return (*this);
#if 0
  if( m_current_index + 1 >= m_buffer.size() )
    m_buffer.resize( m_current_index + 2 );

  m_char.m_np = m_np;
  m_buffer[m_current_index++].m_char = c;

  m_char.m_italic = m_buffer[m_current_index-1].m_italic;
  m_char.m_cal = m_buffer[m_current_index-1].m_cal;
#endif

#if 0
  m_buffer[m_current_index-1].Dump( std::cout );
  std::cout << ": ";
  m_buffer[m_current_index-1].m_np.Dump( std::cout );
  std::cout << endl;
#endif
}

DataStream& DataStream::operator << ( const char *str )
{
  for( size_t i = 0; str[i] != '\0'; i++ )
    (*this) << str[i];

  return *this;
}

DataStream& DataStream::operator << ( const std::string &str )
{
  for( size_t i = 0; i < str.length(); i++ )
    (*this) << str[i];

  return *this;
}

#if 0
size_t DataStream::operator<<( const char *str)
{
  if( str == NULL )
    return 0 ;

  char l[16]; // pfui
  size_t count = 0;

  while( *str && *str != '}' ) 
  {
    if( *str == '{' )
    {
      size_t c1 = (*this) << ++str;
      str += c1;
      count += c1 + 1;
    }
    else if( *str == '\\' )
    {
      str++;
      count++;
      char *p = l;
      while( *str && ( ( *str >= 'a' && *str <= 'z' ) ||
                       ( *str >= 'A' && *str <= 'Z' ) ) )
      {
        *p = *(str++);
        count++;
        p++;
      }
      *p = '\0';
        
      //cout << "lsym: " << l << endl;
      UCHAR c = LatexConvertor.GetUnicode( l );

      if( c != 0 )
        (*this) << c;

      continue;
    }
    else
    {
      (*this) << *(str++);
      count++;
    }
  }
  
  return count;
}
#endif

DataStream& DataStream::operator << ( const USTRING &str )
{
  for( size_t i = 0; i < str.length(); i++ )
    (*this) << str[i];

  return *this;
}

DataStream& DataStream::operator << ( const UCHAR *str )
{
  for( size_t i = 0; str[i] != '\0'; i++ )
    (*this) << str[i];

  return *this;
}

DataStream& DataStream::operator << ( int i )
{
  std::ostringstream str;
  str << i;
  (*this) << str.str().c_str();

  return *this;
}

////////////////////////////////////////////////////////////////////////////////

DataStream& operator << ( DataStream &str, const ProofLib::Variable &e )
{
  str.SetItalic();
  str << e.GetName();
  str.Reset();
  str.SetItalic( false );

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::SecondOrderVariable &e )
{
  str.SetItalic();
  str << e.GetName();
  str.Reset();
  str.SetItalic( false );

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::Constant &e )
{
  str << e.GetName();
  str.Reset();

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::Function &e )
{
  using namespace ProofLib;

  size_t rek = 1;

  Function *fn = (Function*) &e;

  if( e.GetArity() == 1 )
  {
    while( true )
    {
      AbstractTerm *t = fn->GetTerm(0);
  
      Function *fn_ = dynamic_cast<Function*>( t );
  
      if( fn_ != NULL && e.GetSymbol() == fn_->GetSymbol() )
      {
        fn = fn_;
        rek++;
        str.GetNodePointer()->Push( 0 );
      }
      else
        break;
    }
  }

#ifdef PRINT_BINARY_INFIX
#if 0
  if( e.GetArity() == 2 &&
      ( e.GetSymbol()->GetName()[0] == '+' ||
        e.GetSymbol()->GetName()[0] == '-' ||
        e.GetSymbol()->GetName()[0] == '*' ||
        e.GetSymbol()->GetName()[0] == '/' ||
        e.GetSymbol()->GetName()[0] == CHAR_ISIN ||
        e.GetSymbol()->GetName()[0] == CHAR_NOTIN ||
        e.GetSymbol()->GetName()[0] == CHAR_NI ||
        e.GetSymbol()->GetName()[0] == CHAR_MINUS ||
        e.GetSymbol()->GetName()[0] == CHAR_CAP ||
        e.GetSymbol()->GetName()[0] == CHAR_CUP ||
        e.GetSymbol()->GetName()[0] == CHAR_LE ||
        e.GetSymbol()->GetName()[0] == CHAR_GE ||
        e.GetSymbol()->GetName()[0] == CHAR_SUB ||
        e.GetSymbol()->GetName()[0] == CHAR_SUP ||
        e.GetSymbol()->GetName()[0] == CHAR_NSUB ||
        e.GetSymbol()->GetName()[0] == CHAR_SUBE ||
        e.GetSymbol()->GetName()[0] == CHAR_SUPE ||
        e.GetSymbol()->GetName()[0] == CHAR_OPLUS ||
        e.GetSymbol()->GetName()[0] == CHAR_OTIMES ) )
#endif
  if( e.GetSymbol()->IsInfix() )
  {
    bool termstart = str.m_termstart;

    if( !termstart )
      str << "(" << CHAR_THIN_SPACE;

    str.m_termstart = false;

    str.GetNodePointer()->Push(0);
    str << *(e.GetTerm(0));
    str.GetNodePointer()->Pop();

    str << CHAR_THIN_SPACE;

    str.SetItalic();
    str << e.GetName();
    str.Reset();
    str.SetItalic( false );

    str << CHAR_THIN_SPACE;

    str.m_termstart = false;

    str.GetNodePointer()->Push(1);
    str << *(e.GetTerm(1));
    str.GetNodePointer()->Pop();

    if( !termstart )
      str << CHAR_THIN_SPACE << ")";
  }
  else
  {
#endif

  str.SetItalic();
  str << e.GetName();

  if( rek != 1 )
  {
    str << DCHAR_SUPERSCRIPT;
    str << DCHAR_LEFT_BRAKET;
    str << CHAR_THIN_SPACE;
    str << int(rek);

    str << DCHAR_RIGHT_BRAKET;

/*
    DataStream *sup = str.GetSuperscript();
    (*sup) << CHAR_THIN_SPACE;
    (*sup) << (int) rek;

    delete sup;
*/
  }

  str.Reset();
  str.SetItalic( false );

  str << dynamic_cast<const ProofLib::StaticAbstractTermList&>( *fn );

#ifdef PRINT_BINARY_INFIX
  }
#endif

  while( rek > 1 )
  {
    str.GetNodePointer()->Pop();
    rek--;
  }

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::AbstractTerm &e )
{
  using namespace ProofLib;

  try
  {
    const LambdaTerm &a = dynamic_cast<const LambdaTerm&>( e );
    str << a;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const SecondOrderVariable &a = dynamic_cast<const SecondOrderVariable&>( e );
    str << a;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const DefinedSet &a = dynamic_cast<const DefinedSet&>( e );
    str << a;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const Term &a = dynamic_cast<const Term&>( e );
    str << a;
    return str;
  }
  catch( std::bad_cast ) {}

  ASSERT( false );

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::Term &e )
{
  using namespace ProofLib;

  try
  {
    const Variable &a = dynamic_cast<const Variable&>( e );
    str << a;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const Constant &a = dynamic_cast<const Constant&>( e );
    str << a;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const Function &a = dynamic_cast<const Function&>( e );
    str << a;
    return str;
  }
  catch( std::bad_cast ) {}

#if 0
  std::cerr << "failed to cast term: ";
  e.ToString( std::cerr );
  std::cerr << std::endl;
#endif

  ASSERT( false );

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::StaticAbstractTermList &e )
{
  if( e.GetChildCount() < 1 )
    return str;

  str << CHAR('(');

  for( size_t i = 0; i < e.GetChildCount(); i++ )
  {  
    if( i > 0 )
    {
      str << CHAR(',');
      str << CHAR_THIN_SPACE;
    }

#ifdef PRINT_BINARY_INFIX
    str.m_termstart = true;
#endif
    str.GetNodePointer()->Push(i);
    str << *(e.GetTerm(i));
    str.GetNodePointer()->Pop();
  }

  str << CHAR(')');

  return str;
}


DataStream& operator << ( DataStream &str, const ProofLib::StaticTermList &e )
{
  if( e.GetChildCount() < 1 )
    return str;

  str << CHAR('(');

  for( size_t i = 0; i < e.GetChildCount(); i++ )
  {  
    if( i > 0 )
    {
      str << CHAR(',');
      str << CHAR_THIN_SPACE;
    }

#ifdef PRINT_BINARY_INFIX
    str.m_termstart = true;
#endif
    str.GetNodePointer()->Push(i);
    str << *(e.GetTerm(i));
    str.GetNodePointer()->Pop();
  }

  str << CHAR(')');

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::DefinedSetFormula &e )
{
  str << *(e.getDefinedSet());
  str << dynamic_cast<const ProofLib::StaticTermList&>( e );

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::DefinedSet &e )
{
#ifdef PRINT_BINARY_INFIX
#if 0
  if( e.GetArity() == 2 &&
      ( e.GetSymbol()->GetName()[0] == '=' ||
        e.GetSymbol()->GetName()[0] == '<' ||
        e.GetSymbol()->GetName()[0] == '>' ||
        e.GetSymbol()->GetName()[0] == '!' ||
        e.GetSymbol()->GetName()[0] == CHAR_ISIN ||
        e.GetSymbol()->GetName()[0] == CHAR_NOTIN ||
        e.GetSymbol()->GetName()[0] == CHAR_NI ||
        e.GetSymbol()->GetName()[0] == CHAR_CAP ||
        e.GetSymbol()->GetName()[0] == CHAR_CUP ||
        e.GetSymbol()->GetName()[0] == CHAR_LE ||
        e.GetSymbol()->GetName()[0] == CHAR_GE ||
        e.GetSymbol()->GetName()[0] == CHAR_SUB ||
        e.GetSymbol()->GetName()[0] == CHAR_SUP ||
        e.GetSymbol()->GetName()[0] == CHAR_NSUB ||
        e.GetSymbol()->GetName()[0] == CHAR_SUBE ||
        e.GetSymbol()->GetName()[0] == CHAR_SUPE ) )
#endif
  if( e.GetSymbol()->IsInfix() )
  {
//    str << "(" << CHAR_THIN_SPACE;

    str.m_termstart = true;

    str.GetNodePointer()->Push(0);
    str << *(e.GetTerm(0));
    str.GetNodePointer()->Pop();

    str << CHAR_THIN_SPACE;

    str.SetItalic();
    str << e.GetName();
    str.Reset();
    str.SetItalic( false );

    str << CHAR_THIN_SPACE;

    str.m_termstart = true;

    str.GetNodePointer()->Push(1);
    str << *(e.GetTerm(1));
    str.GetNodePointer()->Pop();

//    str << CHAR_THIN_SPACE << ")";
  }
  else
  {
    str.m_termstart = true;
#endif

  str.m_formulastart = false;
  str.SetItalic();
  str << e.GetName();
  str.Reset();
  str.SetItalic( false );

  str << dynamic_cast<const ProofLib::StaticAbstractTermList&>( e );

#ifdef PRINT_BINARY_INFIX
  }
#endif

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::AtomFormula &e )
{
#ifdef PRINT_BINARY_INFIX
#if 0
  if( e.GetArity() == 2 &&
      ( e.GetSymbol()->GetName()[0] == '=' ||
        e.GetSymbol()->GetName()[0] == '<' ||
        e.GetSymbol()->GetName()[0] == '>' ||
        e.GetSymbol()->GetName()[0] == '!' ||
        e.GetSymbol()->GetName()[0] == CHAR_ISIN ||
        e.GetSymbol()->GetName()[0] == CHAR_NOTIN ||
        e.GetSymbol()->GetName()[0] == CHAR_NI ||
        e.GetSymbol()->GetName()[0] == CHAR_CAP ||
        e.GetSymbol()->GetName()[0] == CHAR_CUP ||
        e.GetSymbol()->GetName()[0] == CHAR_LE ||
        e.GetSymbol()->GetName()[0] == CHAR_GE ||
        e.GetSymbol()->GetName()[0] == CHAR_SUB ||
        e.GetSymbol()->GetName()[0] == CHAR_SUP ||
        e.GetSymbol()->GetName()[0] == CHAR_NSUB ||
        e.GetSymbol()->GetName()[0] == CHAR_SUBE ||
        e.GetSymbol()->GetName()[0] == CHAR_SUPE ) )
#endif
  if( e.GetSymbol()->IsInfix() )
  {
//    str << "(" << CHAR_THIN_SPACE;

    str.m_termstart = true;

    str.GetNodePointer()->Push(0);
    str << *(e.GetTerm(0));
    str.GetNodePointer()->Pop();

    str << CHAR_THIN_SPACE;

    str.SetItalic();
    str << e.GetName();
    str.Reset();
    str.SetItalic( false );

    str << CHAR_THIN_SPACE;

    str.m_termstart = true;

    str.GetNodePointer()->Push(1);
    str << *(e.GetTerm(1));
    str.GetNodePointer()->Pop();

//    str << CHAR_THIN_SPACE << ")";
  }
  else
  {
    str.m_termstart = true;
#endif

  str.m_formulastart = false;
  str.SetItalic();
  str << e.GetName();
  str.Reset();
  str.SetItalic( false );

  str << dynamic_cast<const ProofLib::StaticAbstractTermList&>( e );

#ifdef PRINT_BINARY_INFIX
  }
#endif

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::BinaryConFormula &e )
{
  bool parenties = !str.m_formulastart;

  if( parenties )
    str << CHAR('(');

  str.m_formulastart = false;

  str.GetNodePointer()->Push(0);
  str << *(e.GetLeft());
  str.GetNodePointer()->Pop();

  str << CHAR_THIN_SPACE;
  str << e.GetOperatorSymbol()->GetDisplayChar();
  str.Reset();
  str << CHAR_THIN_SPACE;

  str.GetNodePointer()->Push(1);
  str << *(e.GetRight());
  str.GetNodePointer()->Pop();

  if( parenties )
    str << CHAR(')');

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::QuantorFormula &e )
{
  str.m_formulastart = false;
  str << CHAR('(');
  str << e.GetOperatorSymbol()->GetDisplayChar();
  str.Reset();
  str.GetNodePointer()->Push(0);
  str << *(e.GetVariable());
  str.GetNodePointer()->Pop();
  str << CHAR(')');
  str.GetNodePointer()->Push(1);
  str << *(e.GetFormula());
  str.GetNodePointer()->Pop();

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::SecondOrderQuantorFormula &e )
{
  str.m_formulastart = false;
  str << CHAR('(');
  str << e.GetOperatorSymbol()->GetDisplayChar();
  str.Reset();
  str.GetNodePointer()->Push(0);
  str << *(e.GetSecondOrderVariable());
  str.GetNodePointer()->Pop();
  str << CHAR(')');
  str.GetNodePointer()->Push(1);
  str << *(e.GetFormula());
  str.GetNodePointer()->Pop();

  return str;
}


DataStream& operator << ( DataStream &str, const ProofLib::UnaryConFormula &e )
{
  using namespace ProofLib;

  str.m_formulastart = false;

  str << e.GetOperatorSymbol()->GetDisplayChar();
  str.Reset();

  str.GetNodePointer()->Push(0);
  str << *(e.GetFormula());
  str.GetNodePointer()->Pop();

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::FormulaVariable &e )
{
  str.SetCal();
  str << e.GetName();
  str.Reset();
  str.SetCal( false );

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::Formula &e )
{
  using namespace ProofLib;

  try
  {
    const AtomFormula &a = dynamic_cast<const AtomFormula&>( e );
    str << a;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const BinaryConFormula &b = dynamic_cast<const BinaryConFormula&>( e );
    str << b;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const UnaryConFormula &u = dynamic_cast<const UnaryConFormula&>( e );
    str << u;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const QuantorFormula &u = dynamic_cast<const QuantorFormula&>( e );
    str << u;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const SecondOrderQuantorFormula &u = dynamic_cast<const SecondOrderQuantorFormula&>( e );
    str << u;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const DefinedSetFormula &u = dynamic_cast<const DefinedSetFormula&>( e );
    str << u;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const FormulaVariable &v = dynamic_cast<const FormulaVariable&>( e );
    str << v;
    return str;
  }
  catch( std::bad_cast ) {}

  std::cerr << "could not cast this (" << &e << "): ";
  e.ToString( std::cerr );

  ASSERT( false );

  return str;
}

DataStream& operator << ( DataStream &str,
                          const ProofLib::FormulaListVariable &e )
{
  //str.SetItalic();
  str << e.GetName();
  str.Reset();
  //str.SetItalic( false );
  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::SElementList &e )
{
  for( size_t i = 0; i < e.GetChildCount(); i++ )
  {
    if( i > 0 )
    {
      str << CHAR(',');
      str << CHAR_THIN_SPACE;
    }

    ProofLib::SElement *x = e.GetSElement( i );

    ProofLib::SFormula *f = dynamic_cast<ProofLib::SFormula*>( x );

    if( f != NULL )
    {
      str.m_formulastart = true;
      str.GetNodePointer()->Push(i);
      str.GetNodePointer()->Push(0);
      str << *(f->GetFormula());
      str.GetNodePointer()->Pop();
      str.GetNodePointer()->Pop();
      continue;
    }

    ProofLib::SFormulaListVariable *fv =
      dynamic_cast<ProofLib::SFormulaListVariable*>( x );

    if( fv != NULL )
    {
      str.GetNodePointer()->Push(i);
      str.GetNodePointer()->Push(0);
      str << *(fv->GetFormulaListVariable());
      str.GetNodePointer()->Pop();
      str.GetNodePointer()->Pop();
      continue;
    }

    ASSERT( false );
  }

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::Sequent &e )
{
  ProofLib::SElementList *l = e.GetLeft();
  ProofLib::SElementList *r = e.GetRight();

  if( l != NULL && l->GetChildCount() > 0 )
  {
    str.GetNodePointer()->Push(0);
    str << *l;
    str.GetNodePointer()->Pop();
    str << CHAR_SPACE;
  }

  str << CHAR_SEQUENT;

  if( r != NULL && r->GetChildCount() > 0 )
  {
    str << CHAR_SPACE;
    str.GetNodePointer()->Push(1);
    str << *r;
    str.GetNodePointer()->Pop();
  }

  return str;
}


DataStream& operator << ( DataStream &str, const ProofLib::LambdaTerm &e )
{
  str << CHAR_LAMBDA;

  const ProofLib::TermList vars = e.getVariables();

  for ( size_t i = 0; i < vars.GetChildCount(); i++ )
  {
    ProofLib::Variable *v = dynamic_cast<ProofLib::Variable*>( vars.GetTerm( i ) );
    ASSERT ( v != NULL );
    str << *v;
  }

  str << ".";

  str << *(e.getFormula());

  return str;
}

DataStream& operator << ( DataStream &str, const ProofLib::ProofLink &e )
{
  str << e.GetLinkedProofName();
  str.Reset();

  return str;
}

#if 0
DataStream& operator << ( DataStream &str, const ProofLib::FormulaList &e )
{
  const_iterator i;

  for( i = begin(); i != end(); i++ )
  {  
    if( i != begin() )
    {
      str << ",";
    }

    str.m_formulastart = true;
    (*i)->WriteStream( str );
  }
}

DataStream& operator << ( DataStream &str, const ProofLib::Function &e )
{
  WriteStream( str, 1 );
}

DataStream& operator << ( DataStream &str, const ProofLib::Function &e )
{
  if( GetArity() == 1 )
  {
    Term *t = front();
  
    Function *fn = dynamic_cast<Function*>( t );
  
    if( fn != NULL && GetSymbol() == fn->GetSymbol() )
    {
      fn->WriteStream( str, rek+1 );

      return;
    }
#if 0
    Function *f = this;

    while( true )
    {
      Term *t = f->front();
  
      Function *fn = dynamic_cast<Function*>( t );
  
      if( fn == NULL || GetSymbol() != fn->GetSymbol() )
        break;

      f = fn;
      rek++;
  cout << "huhu" << endl;
    }
#endif
  }

  str.SetItalic();
  str << GetSymbol()->m_name;
  str.SetItalic( false );

  if( rek != 1 )
  {
    str.SetSuperscript();
    str << CHAR_THIN_SPACE;
    str << rek;
    str.SetSuperscript( false );
  }

  TermList::WriteStream( str );
}
#endif

DataStream& operator << ( DataStream &str, const ProofLib::DataNode &e )
{
  using namespace ProofLib;

  try
  {
    const Term &a = dynamic_cast<const Term&>( e );
    str << a;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const Formula &a = dynamic_cast<const Formula&>( e );
    str << a;
    return str;
  }
  catch( std::bad_cast ) {}

  try
  {
    const Sequent &a = dynamic_cast<const Sequent&>( e );
    str << a;
    return str;
  }
  catch( std::bad_cast ) {}

  ASSERT( false );

  return str;
}

