////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/datastring.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#ifndef _DATASTRING_H_
#define _DATASTRING_H_

#include <vector>

#include "graphic/drawtypes.h"

#include "util/myassert.h"
#include "chars/latexchars.h"
#include "libproof.h"

////////////////////////////////////////////////////////////////////////////////

#define PRINT_BINARY_INFIX

////////////////////////////////////////////////////////////////////////////////

#define DCHAR_LATEX_BEGIN         CHAR_BACKSLASH
#define DCHAR_SUPERSCRIPT         CHAR('^')
#define DCHAR_SUBSCRIPT           CHAR('_') 
#define DCHAR_LEFT_BRAKET         CHAR('{') 
#define DCHAR_RIGHT_BRAKET        CHAR('}') 

////////////////////////////////////////////////////////////////////////////////

class DataString;

////////////////////////////////////////////////////////////////////////////////

class DataChar
{
public:
  UCHAR            m_char;
  bool             m_italic;
  bool             m_cal;
  DataString      *m_sup;
  DataString      *m_sub;

// metrics set by draw
  SUNIT            m_start;
  SUNIT            m_end;
  SUNIT            m_y;

  ProofLib::NodePointer m_np;

public:
  DataChar();

  bool HasSameStyle( const DataChar &other ) const
  {
    return ( m_cal == other.m_cal && m_italic == other.m_italic );
  }

  bool HasSubStrings() const { return ( m_sup != NULL || m_sub != NULL ); }

  bool isLineFeed() const { return m_char == CHAR_LINEFEED; }

#ifdef DEBUG
  void Dump( ostream &out ) const;
#endif
};

////////////////////////////////////////////////////////////////////////////////

class DataString
{
protected:
  std::vector<DataChar>  m_buffer;

private:
  DataString( const DataString &other ) {}

public:
  DataString();

  DataString( size_t len );

  inline bool IsEmpty() const { return m_buffer.empty(); }
  inline size_t GetLength() const { return m_buffer.size(); }

  inline void Clear() { m_buffer.clear(); }

  DataChar& operator [] ( size_t i );
  const DataChar& operator [] ( size_t i ) const;

  void operator += ( const DataChar &c );
  
#ifdef DEBUG
  void Dump( ostream &out ) const;
#endif
};

////////////////////////////////////////////////////////////////////////////////

class DataStream
{
protected:
  DataString            &m_dest;

  DataChar               m_char;

  ProofLib::NodePointer  m_np;

//  std::string            m_utf8_helper;
  USTRING                m_latex_helper;

  typedef enum {
    Normal,
    ReadLatexSymbol,
    StartSuperScript,
    StartSubScript,
    InSequence,
    EndOfSequence,
    WaitForNormal
  }
  StreamState;

  StreamState            m_state;

  DataStream            *m_next_stream;

public:
  bool                   m_formulastart; // helper for writing formulas
#ifdef PRINT_BINARY_INFIX
  bool                   m_termstart;
#endif


private:
  //DataStream( const DataStream &other );

protected:
  void AddChar( UCHAR c );
  StreamState GetState() const { return m_state; }

public:
  DataStream( DataString &dest );
  ~DataStream();

  inline void SetItalic( bool on = true );
  inline void SetCal( bool on = true );
  inline DataStream* GetSuperscript();
  inline DataStream* GetSubscript();

  inline void Reset();

  inline ProofLib::NodePointer* GetNodePointer() { return &m_np; }

  DataStream& operator << ( UCHAR c );
  DataStream& operator << ( char c );

  DataStream& operator << ( const char *str );
  DataStream& operator << ( const std::string &str );
  DataStream& operator << ( const USTRING &str );
  DataStream& operator << ( const UCHAR *str );

  DataStream& operator << ( int i );
};

////////////////////////////////////////////////////////////////////////////////

DataStream& operator << ( DataStream &str, const ProofLib::DataNode &e );

DataStream& operator << ( DataStream &str, const ProofLib::Variable &e );
DataStream& operator << ( DataStream &str, const ProofLib::Constant &e );
DataStream& operator << ( DataStream &str, const ProofLib::Function &e );
DataStream& operator << ( DataStream &str, const ProofLib::Term &e );
DataStream& operator << ( DataStream &str, const ProofLib::DefinedSet &e );
DataStream& operator << ( DataStream &str, const ProofLib::StaticTermList &e );
DataStream& operator << ( DataStream &str, const ProofLib::StaticAbstractTermList &e );
DataStream& operator << ( DataStream &str, const ProofLib::AtomFormula &e );
DataStream& operator << ( DataStream &str, const ProofLib::DefinedSetFormula &e );
DataStream& operator << ( DataStream &str, const ProofLib::BinaryConFormula &e );
DataStream& operator << ( DataStream &str, const ProofLib::QuantorFormula &e );
DataStream& operator << ( DataStream &str, const ProofLib::SecondOrderQuantorFormula &e );
DataStream& operator << ( DataStream &str, const ProofLib::UnaryConFormula &e );
DataStream& operator << ( DataStream &str, const ProofLib::FormulaVariable &e );
DataStream& operator << ( DataStream &str, const ProofLib::Formula &e );
DataStream& operator << ( DataStream &str, const ProofLib::FormulaListVariable &e );
DataStream& operator << ( DataStream &str, const ProofLib::SElementList &e );
DataStream& operator << ( DataStream &str, const ProofLib::Sequent &e );
DataStream& operator << ( DataStream &str, const ProofLib::ProofLink &e );
DataStream& operator << ( DataStream &str, const ProofLib::LambdaTerm &e );

////////////////////////////////////////////////////////////////////////////////

#endif
