////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/drawtypes.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <algorithm>
#include <string.h>

using namespace std;

#include "util/myassert.h"
#include "drawtypes.h"
#include "graphictypes.h"

////////////////////////////////////////////////////////////////////////////////

DrawPoint::DrawPoint( const DrawPoint &o )
{
  m_x = o.m_x;
  m_y = o.m_y;
}

DrawPoint::DrawPoint( const GraphicPoint &o )
{
  m_x = o.m_x;
  m_y = o.m_y;
}

DrawPoint& DrawPoint::TransformBy( const DrawTransMatrix &m )
{
  SUNIT m_vec[3] = { m_x, m_y, 1 };
  SUNIT *c = NULL;

  for( int i = 0; i < 2; i++ )
  {
    if( !i )
      c = &m_x;
    else
      c = &m_y;

    *c = 0;

    for( int j = 0; j < 3; j++ )
    {
      *c += m_vec[j] * m.m_matrix[i][j];
    }
  }

  return *this;
}

DrawPoint& DrawPoint::Round()
{
  m_x = round( m_x );
  m_y = round( m_y );

  return *this;
}

#ifdef DEBUG
void DrawPoint::Dump( ostream &out ) const
{
  out << "(" << m_x << "," << m_y << ")";
}
#endif

////////////////////////////////////////////////////////////////////////////////

DrawRect::DrawRect( const DrawRect &o )
{
  m_leftbottom = o.m_leftbottom;
  m_righttop = o.m_righttop;
}

DrawRect::DrawRect( const GraphicRect &o )
{
  m_leftbottom = o.m_origin;
  m_righttop = o.m_origin + o.m_size;
}

DrawRect::DrawRect( const DrawPoint &x, const DrawPoint &y )
{
  DrawPoint a = x;
  DrawPoint b = y;

  if( b < a )
    swap( a, b );

  m_leftbottom = a;
  m_righttop = b;
}

DrawPoint DrawRect::MidPoint() const
{
  return m_leftbottom + ( m_righttop / 2 );
}

GraphicRect DrawRect::GetBiggerIntRect() const
{
  GraphicRect rect;

  rect.m_origin.m_x = GUNIT( floor( m_leftbottom.m_x ) );
  rect.m_origin.m_y = GUNIT( floor( m_leftbottom.m_y ) );

  rect.m_size.m_x = GUNIT( ceil( m_righttop.m_x ) ) - rect.m_origin.m_x;
  rect.m_size.m_y = GUNIT( ceil( m_righttop.m_y ) ) - rect.m_origin.m_y;

  return rect;
}

DrawRect& DrawRect::TransformBy( const DrawTransMatrix &m )
{
  m_leftbottom.TransformBy( m );
  m_righttop.TransformBy( m );

  SUNIT tx = min( m_leftbottom.m_x, m_righttop.m_x );
  SUNIT ty = min( m_leftbottom.m_y, m_righttop.m_y );
  m_righttop.m_x = max( m_leftbottom.m_x, m_righttop.m_x );
  m_righttop.m_y = max( m_leftbottom.m_y, m_righttop.m_y );
  m_leftbottom.m_x = tx;
  m_leftbottom.m_y = ty;

  return *this;
}
DrawRect& DrawRect::Capture( const DrawRect& o )
{
  if( o.IsZero() )
    return *this;

  if( IsZero() )
  {
    (*this) = o;
    return *this;
  }

  m_leftbottom.m_x = min( m_leftbottom.m_x, o.m_leftbottom.m_x );
  m_leftbottom.m_y = min( m_leftbottom.m_y, o.m_leftbottom.m_y );
  m_righttop.m_x = max( m_righttop.m_x, o.m_righttop.m_x );
  m_righttop.m_y = max( m_righttop.m_y, o.m_righttop.m_y );

  return *this;
}

DrawRect& DrawRect::Round()
{
  m_leftbottom.Round();
  m_righttop.Round();

  return *this;
}

DrawRect& DrawRect::Shrink( SUNIT amount )
{
  ASSERT( amount > 0.0 );

  if( GetWidth() <= amount )
  {
    m_leftbottom.m_x = m_leftbottom.m_x + GetWidth() / 2;
    m_righttop.m_x   = m_leftbottom.m_x;
  }
  else
  {
    m_leftbottom.m_x += amount;
    m_righttop.m_x   -= amount;
  }

  if( GetHeight() <= amount )
  {
    m_leftbottom.m_y = m_leftbottom.m_y + GetHeight() / 2;
    m_righttop.m_y   = m_leftbottom.m_y;
  }
  else
  {
    m_leftbottom.m_y += amount;
    m_righttop.m_y   -= amount;
  }

  return *this;
}

bool DrawRect::Contains( const DrawPoint &pt ) const
{
  if( pt.m_x < m_leftbottom.m_x )
    return false;
  if( pt.m_x > m_righttop.m_x )
    return false;
  if( pt.m_y < m_leftbottom.m_y )
    return false;
  if( pt.m_y > m_righttop.m_y )
    return false;

  return true;
}
bool DrawRect::Overlap( const DrawRect &o ) const
{
  if( o.m_righttop.m_x < m_leftbottom.m_x )
    return false;

  if( o.m_leftbottom.m_x > m_righttop.m_x )
    return false;

  if( o.m_righttop.m_y < m_leftbottom.m_y )
    return false;

  if( o.m_leftbottom.m_y > m_righttop.m_y )
    return false;

  return true;
}

DrawRect DrawRect::Intersection( const DrawRect &o ) const
{
  DrawRect r;

  r.m_leftbottom.m_x = max( m_leftbottom.m_x, o.m_leftbottom.m_x );
  r.m_leftbottom.m_y = max( m_leftbottom.m_y, o.m_leftbottom.m_y );
  r.m_righttop.m_x = min( m_righttop.m_x, o.m_righttop.m_x );
  r.m_righttop.m_y = min( m_righttop.m_y, o.m_righttop.m_y );

  if( r.m_leftbottom.m_x > r.m_righttop.m_x ||
      r.m_leftbottom.m_y > r.m_righttop.m_y )
    return DrawRect();

  return r;
}

#ifdef DEBUG
void DrawRect::Dump( ostream &out ) const
{
  m_leftbottom.Dump( out );
  out << "x";
  m_righttop.Dump( out );
}
#endif

DrawTransMatrix::DrawTransMatrix()
{
  *this = Identity();
}

DrawTransMatrix::DrawTransMatrix( const DrawTransMatrix &o )
{
  memcpy( m_matrix, o.m_matrix, sizeof( m_matrix ) );
}

DrawTransMatrix::DrawTransMatrix( const SUNIT matrix[3][3] )
{
  memcpy( m_matrix, matrix, sizeof( m_matrix ) );
}

DrawTransMatrix DrawTransMatrix::Multiply( const DrawTransMatrix &o ) const
{
  DrawTransMatrix m;

  for( int i = 0; i < 3; i++ )
    for( int j = 0; j < 3; j++ )
    {
      m.m_matrix[i][j] = 0;

      for( int k = 0; k < 3; k++ )
        m.m_matrix[i][j] += m_matrix[i][k] * o.m_matrix[k][j];
    }

  return m;
}

DrawTransMatrix& DrawTransMatrix::ScaleTo( SUNIT factor )
{
  DrawTransMatrix m = Scale( factor );

  *this = Multiply( m );

  return *this;
}

DrawTransMatrix DrawTransMatrix::Scale( SUNIT factor )
{
  SUNIT matrix[3][3] = { { factor, 0, 0 }, { 0, factor, 0 }, { 0, 0, 1 } };
  
  DrawTransMatrix m( matrix );

  return m;
}

DrawTransMatrix& DrawTransMatrix::ScaleToAt( SUNIT factor, const DrawPoint &pt )
{
  DrawTransMatrix m = ScaleAtPoint( factor, pt );

  *this = Multiply( m );

  return *this;
}

DrawTransMatrix DrawTransMatrix::ScaleAtPoint( SUNIT factor, const DrawPoint &pt )
{
  const SUNIT matrix[3][3] = {  { factor, 0, (1-factor) * pt.m_x },
                                { 0, factor, (1-factor) * pt.m_y },
                                { 0, 0, 1 } };
  
  DrawTransMatrix m( matrix );

  return m;
}

DrawTransMatrix& DrawTransMatrix::TranslateTo( const DrawPoint &pt )
{
  DrawTransMatrix m = Translate( pt );

  *this = Multiply( m );

  return *this;
}

DrawTransMatrix DrawTransMatrix::Translate( const DrawPoint &pt )
{
  const SUNIT matrix[3][3] =
    { { 1, 0, pt.m_x }, { 0, 1, pt.m_y }, { 0, 0, 1 } };
  
  DrawTransMatrix m( matrix );

  return m;
}

DrawTransMatrix DrawTransMatrix::HorizontalFlip()
{
  SUNIT matrix[3][3] = { { 1, 0, 0 }, { 0, -1, 0 }, { 0, 0, 1 } };

  DrawTransMatrix m( matrix );

  return m;
}

DrawTransMatrix DrawTransMatrix::VerticalFlip()
{
  SUNIT matrix[3][3] = { { -1, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } };

  DrawTransMatrix m( matrix );

  return m;
}

DrawTransMatrix DrawTransMatrix::Inverse() const
{
  DrawTransMatrix inv;

  int ch_[5] = { 2, 0, 1, 2, 0 };
  int *ch = &ch_[1];

  SUNIT det = Determinat();

  ASSERT( det != 0.0 );

  for( int i = 0; i < 3; i++ )
    for( int j = 0; j < 3; j++ )
    {
      inv.m_matrix[i][j] =
          m_matrix[ch[j+1]][ch[i+1]] * m_matrix[ch[j-1]][ch[i-1]]
        - m_matrix[ch[j+1]][ch[i-1]] * m_matrix[ch[j-1]][ch[i+1]];
      inv.m_matrix[i][j] /= det;
    }
#if 0
  inv.m_matrix[0][0] =
    m_matrix[1][1] * m_matrix[2][2]
  - m_matrix[1][2] * m_matrix[2][1]

  inv.m_matrix[0][1] =
    m_matrix[0][2] * m_matrix[2][1]
  - m_matrix[0][1] * m_matrix[2][2]

  {{-b3 c2 + b2 c3, a3 c2 - a2 c3, -
    a3 b2 + a2 b3}, {b3 c1 - b1 c3, -a3 c1 + a1 c3, a3 b1 - a1 b3}, {-
  b2 c1 + b1 c2, a2 c1 - a1 c2, -a2 b1 + a1 b2}}
#endif

  return inv;
}

DrawTransMatrix DrawTransMatrix::Identity()
{
  SUNIT matrix[3][3] = { { 1, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } };

  return DrawTransMatrix( matrix );
}

SUNIT DrawTransMatrix::Determinat() const
{
  int ch[3] = { 0, 1, 2 };

  SUNIT det = 0;
  
  for( int i = 0; i < 6; i++ )
  {
#if 0
    if( i % 2 == 0 )
      cout << "+ ";
    else
      cout << "- ";
cout << ch[0] << ", " << ch[1] << ", " << ch[2] << endl;
#endif

    SUNIT temp = 1;

    for( int j = 0; j < 3; j++ )
    {
      temp *= m_matrix[j][ch[j]];
    }
    
    int t;
    if( i % 2 == 0 )
    {
      det += temp;
      t = ch[1];
      ch[1] = ch[2];
      ch[2] = t;
    }
    else
    {
      det -= temp;
      t = ch[1];
      ch[1] = ch[0];
      ch[0] = t;
    }
  }

  return det;
#if 0
  +
  matrix[0][0]
  matrix[1][1]
  matrix[2][2]

  -
  matrix[0][0]
  matrix[1][2]
  matrix[2][1]

  -
  matrix[0][1]
  matrix[1][0]
  matrix[2][2]

  +
  matrix[0][1]
  matrix[1][2]
  matrix[2][0]

  +
  matrix[0][2]
  matrix[1][0]
  matrix[2][1]

  -
  matrix[0][2]
  matrix[1][1]
  matrix[2][0]

  (-a3 b2 c1 + a2 b3 c1 + a3 b1 c2 - a1 b3 c2 - a2 b1 c3 + a1 b2 c3 )
#endif
}

#ifdef DEBUG
void DrawTransMatrix::Dump( ostream &out ) const
{
  for( int i = 0; i < 3; i++ )
  {
    for( int j = 0; j < 3; j++ )
      out << m_matrix[i][j] << "  ";

    out << endl;
  }
}
#endif
