////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/drawtypes.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#ifndef _DRAWTYPES_H_
#define _DRAWTYPES_H_

#include <math.h>
#ifdef DEBUG
#include <iostream>

using namespace std;
#endif

typedef double SUNIT;
//typedef int SUNIT;
typedef long SCOLOR;
#define SCOLOR_TRANSPERENT        0xff000000U
#define SCOLOR_RED                0x00ff0000U
#define SCOLOR_GREEN              0x0000ff00U
#define SCOLOR_BLUE               0x000000ffU
#define SCOLOR_WHITE              SCOLOR_RED | SCOLOR_GREEN | SCOLOR_BLUE
#define SCOLOR_BLACK              0x00000000U

class GraphicPoint;
class GraphicRect;

class DrawTransMatrix;

class DrawPoint
{
public:
  SUNIT       m_x;
  SUNIT       m_y;

public:
  DrawPoint() { m_x = 0; m_y = 0; };
  DrawPoint( const DrawPoint &o );
  DrawPoint( SUNIT x, SUNIT y ) { m_x = x; m_y = y; };
  DrawPoint( const GraphicPoint &o );

  inline long LongX() const { return lround( m_x ); }
  inline long LongY() const { return lround( m_y ); }

  inline bool operator== ( const DrawPoint &o ) const
    { return (m_x == o.m_x && m_y == o.m_y); }
  inline bool operator!= ( const DrawPoint &o ) const
    { return !( (*this) == o ); }
  inline bool operator< ( const DrawPoint &o ) const
    { return (m_x < o.m_x || ( m_x == o.m_x && m_y < o.m_y) ); }

  inline DrawPoint operator- () const
    { return DrawPoint( -m_x, -m_y ); }

  inline DrawPoint operator+ ( const DrawPoint &o ) const
    { DrawPoint pt; pt.m_x = m_x + o.m_x; pt.m_y = m_y + o.m_y; return pt; }
  inline DrawPoint operator- ( const DrawPoint &o ) const
    { DrawPoint pt; pt.m_x = m_x - o.m_x; pt.m_y = m_y - o.m_y; return pt; }
  inline void operator+= ( const DrawPoint &o )
    { m_x += o.m_x; m_y += o.m_y; }
  inline void operator-= ( const DrawPoint &o )
    { m_x -= o.m_x; m_y -= o.m_y; }

  inline DrawPoint operator+ ( SUNIT s ) const
    { return DrawPoint( m_x + s, m_y + s ); }
  inline DrawPoint operator- ( SUNIT s ) const
    { return DrawPoint( m_x - s, m_y - s ); }
  inline DrawPoint operator* ( SUNIT s ) const
    { return DrawPoint( m_x * s, m_y * s ); }
  inline DrawPoint operator/ ( SUNIT s ) const
    { return DrawPoint( m_x / s, m_y / s ); }
  inline void operator+= ( SUNIT s )
    { m_x += s; m_y += s; }
  inline void operator-= ( SUNIT s )
    { m_x -= s; m_y -= s; }
  inline void operator*= ( SUNIT s )
    { m_x *= s; m_y *= s; }
  inline void operator/= ( SUNIT s )
    { m_x /= s; m_y /= s; }

  DrawPoint& TransformBy( const DrawTransMatrix &m );
  DrawPoint& Round();

  bool IsZero() const { return m_x == 0 && m_y == 0; };

  inline static DrawPoint Zero() { return DrawPoint(); }
#ifdef DEBUG
  void Dump( ostream &out ) const;
#endif
};

#define abs( x ) ( x < 0 ? x * -1 : x )
class DrawRect
{
public:

  DrawPoint   m_leftbottom;
  DrawPoint   m_righttop;

public:
  DrawRect() {};
  DrawRect( const DrawRect &o );
  DrawRect( const DrawPoint &a, const DrawPoint &b );
  DrawRect( const GraphicRect &rect );


  DrawRect& TransformBy( const DrawTransMatrix &m );
  DrawRect& Capture( const DrawRect& o );
  DrawRect& Round();
  DrawRect& Shrink( SUNIT amount );

  SUNIT GetWidth() const { return abs(m_righttop.m_x - m_leftbottom.m_x); };
  SUNIT GetHeight() const { return abs(m_righttop.m_y - m_leftbottom.m_y); };

  DrawPoint MidPoint() const;
  DrawRect Intersection( const DrawRect &o ) const;

  GraphicRect GetBiggerIntRect() const;

  bool IsZero() const { return m_righttop.IsZero() && m_leftbottom.IsZero(); };
  bool Contains( const DrawPoint &pt ) const;
  bool Overlap( const DrawRect &o ) const;


#ifdef DEBUG
  void Dump( ostream &out ) const;
#endif
};

class DrawTransMatrix
{
public:
  SUNIT       m_matrix[3][3];

  DrawTransMatrix();
  DrawTransMatrix( const DrawTransMatrix &o );
  DrawTransMatrix( const SUNIT matrix[3][3] );

  DrawTransMatrix Multiply( const DrawTransMatrix &o ) const;
  DrawTransMatrix Inverse() const;

  DrawTransMatrix& TranslateTo( const DrawPoint &pt );
  DrawTransMatrix& ScaleTo( SUNIT factor );
  DrawTransMatrix& ScaleToAt( SUNIT factor, const DrawPoint &pt );

  SUNIT Determinat() const;


public:
  static DrawTransMatrix Identity();
  static DrawTransMatrix ScaleAtPoint( SUNIT factor, const DrawPoint &pt );
  static DrawTransMatrix Scale( SUNIT factor );
  static DrawTransMatrix Translate( const DrawPoint &pt );
  static DrawTransMatrix HorizontalFlip();
  static DrawTransMatrix VerticalFlip();

#ifdef DEBUG
  void Dump( ostream &out ) const;
#endif
};

#endif
