////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/event.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#ifndef _EVENT_H_
#define _EVENT_H_

#include <vector>

#ifdef SIREALLYNEEDED
#include "util/threads.h"
#endif

#include "graphictypes.h"
#include "drawtypes.h"
#include "acceldb.h"
#include "buttondb.h"

class GraphicSystem;
class GraphicInputListener;
class GraphicMenu;
class GraphicDialog;

class DrawItem;
class DrawSequentItem;
class DrawProofLinkItem;

class Event
{
public:
  virtual ~Event() {};
  virtual void Call( GraphicInputListener *lsnr ) = 0;
  virtual Event *Clone() = 0;
};

class ResizeEvent : public Event
{
public:
  ResizeEvent() {};
  ResizeEvent( ResizeEvent &o ) { m_size = o.m_size; };

  virtual void Call( GraphicInputListener *lsnr );
  virtual Event *Clone()
  {
    ResizeEvent *e( this );
    return e;
  }

  GraphicSize            m_size;
};

class RepaintEvent : public Event
{
public:
  RepaintEvent() {};
  RepaintEvent( RepaintEvent &o ) { m_area = o.m_area; };

  virtual void Call( GraphicInputListener *lsnr );
  virtual Event *Clone()
  {
    RepaintEvent *e( this );
    return e;
  }

  DrawRect               m_area;
};

class KeyInputEvent : public Event
{
public:
  unsigned int m_char;
  unsigned int m_mods;

public:
  KeyInputEvent() {};
  KeyInputEvent( KeyInputEvent &o )
  {
    m_char = o.m_char;
    m_mods = o.m_mods;
  }

  virtual void Call( GraphicInputListener *lsnr );
  virtual Event *Clone()
  {
    KeyInputEvent *e( this );
    return e;
  }

  virtual void SetChar( unsigned int charp ) { m_char = charp; };
  virtual void SetMods( unsigned int mods )  { m_mods = mods; };
};

class PointEvent : public Event
{
public:
  PointEvent() {};
  PointEvent( PointEvent &o ) {};

  DrawPoint m_pt;


  virtual void SetPoint( DrawPoint pt ) { m_pt = pt; };
  virtual Event *GetEventTransformedBy( DrawTransMatrix &mat )
  {
    PointEvent *e = (PointEvent *) Clone();
    e->m_pt.TransformBy( mat );

    return e;
  }
};

class MouseMoveEvent : public PointEvent
{
public:
  MouseMoveEvent() : PointEvent() { m_dragging = false; };
  virtual void Call( GraphicInputListener *lsnr );
  virtual Event *Clone()
  {
    MouseMoveEvent *e = new MouseMoveEvent( *this );
    e->m_pt = m_pt;
    return e;
  }

  virtual void SetDragging( bool on ) { m_dragging = on; };
  virtual bool IsDragging() { return m_dragging; };

protected:
  bool                   m_dragging;
};

class DragDropEvent : public MouseMoveEvent
{
public:
  virtual void Call( GraphicInputListener *lsnr );
};

class ButtonClickEvent : public PointEvent
{
public:
  ButtonClickEvent() : PointEvent() { m_mods = m_button = 0; }
  virtual void Call( GraphicInputListener *lsnr );
  virtual Event *Clone()
  {
    ButtonClickEvent *e = new ButtonClickEvent( *this );
    e->m_pt = m_pt;
    return e;
  }

  virtual void SetButton( int button ) { m_button = button; }
  virtual int GetButton() const { return m_button; }

  virtual void SetMods( int mods ) { m_mods = mods; }
  virtual int GetMods() const { return m_mods; }

private:
  int m_mods;
  int m_button;
};

////////////////////////////////////////////////////////////////////////////////

class GraphicInputDispatcher;
class GraphicInputListener;

typedef std::vector<GraphicInputListener*> GraphicInputListenerQueue;

class GraphicInputListener
{
public:
  GraphicInputListener() {}
  virtual ~GraphicInputListener() {}

  virtual void AttachToDispatcher( GraphicInputDispatcher *disp );
  virtual void DetachFromDispatcher( GraphicInputDispatcher *disp );

  virtual void OnEvent( ResizeEvent *e ) {};
  virtual void OnEvent( RepaintEvent *e ) {};
  virtual void OnEvent( MouseMoveEvent *e ) {};
  virtual void OnEvent( ButtonClickEvent *e ) {};
  virtual void OnEvent( KeyInputEvent *e ) {};
  virtual void OnEvent( DragDropEvent *e ) {};
};

class GraphicInputDispatcher : public GraphicInputListener
{
public:
  virtual ~GraphicInputDispatcher() {}

  virtual void Attach( GraphicInputListener *lsnr );
  virtual void Detach( GraphicInputListener *lsnr );

  virtual void OnEvent( ResizeEvent *e )  { Fire(e); }
  virtual void OnEvent( RepaintEvent *e ) { Fire(e); }
  virtual void OnEvent( MouseMoveEvent *e ) { Fire(e); }
  virtual void OnEvent( ButtonClickEvent *e ) { Fire(e); }
  virtual void OnEvent( KeyInputEvent *e ) { Fire(e); }
  virtual void OnEvent( DragDropEvent *e ) { Fire(e); }

  virtual void Fire( Event *e );

  GraphicInputListenerQueue m_listener;
};

class GraphicMainInputDispatcher
  : public GraphicInputDispatcher //, public TThread, public TMutex
{
  GraphicSystem         *m_sys;

  AccelDB                m_acceldb;
  ButtonDB               m_buttondb;

  // special event handling (for selecting sequent prompts etc.)
  bool                   m_wait_for_sevent;
  bool                   m_abort_select_item;
  DrawItem              *m_selected_item;

public:
  GraphicMainInputDispatcher( GraphicSystem *sys );

  AccelDB *GetAccelDb() { return &m_acceldb; }
  ButtonDB *GetButtonDb() { return &m_buttondb; }

  // popup menu
  virtual void PopupMenu( GraphicMenu *popup, size_t button );

  // selecting prompts for proof items
  virtual DrawSequentItem* SelectSequentItem();
  virtual DrawSequentItem* SelectSElement();
  virtual DrawSequentItem* SelectSFormulaListVar();
  virtual DrawSequentItem* SelectFormula();
  virtual DrawSequentItem* SelectFormulaVar();
  //virtual DrawSequentItem* SelectSElement( size_t *index );
  virtual DrawProofLinkItem* SelectProofLinkItem();

protected:
  void BeginSelectingItem();
  void EndSelectingItem();
  bool HaveToAbortSelectItem() const { return m_abort_select_item; }
public:
  bool IsSelectingItem() const { return m_wait_for_sevent; }
  void SetSelectedItem( DrawItem *item ) { m_selected_item = item; }
  DrawItem* GetSelectedItem() const { return m_selected_item; }
  bool HasSelectedItem() const { return m_selected_item != NULL; }
  void AbortSelectItem();

public:
  virtual void AbortDialog() = 0;
  virtual void DoModal( GraphicDialog *dlg ) = 0;

  virtual void CallCommand( CommandCall *cc );

  virtual void OnEvent( KeyInputEvent *e );
  virtual void Fire( Event *e );

  virtual void IndicateProcessorWorking( bool is_working = true ) = 0;

  virtual int Start() { Init(); return Run(); }
  virtual void Quit() = 0;
  virtual void Init() = 0;
  virtual int Run() = 0;
  virtual void ProcessEvents() = 0;

  void Lock() {}
  void Unlock() {}

protected:
  virtual void RunTillItemSelected() = 0;

  virtual DrawItem* SelectItem( const char *prompt, DrawPoint *pt = NULL );
};

#endif

