////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/exportpdf.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include "exportpdf.h"

#ifdef HAVE_PDFLIB
#include "pdflib.h"

#ifdef DEBUG
#include <iostream>
#endif

extern std::string GetPathForFont( const char* fone_filename );

////////////////////////////////////////////////////////////////////////////////

class PdfGraphicContext : public GraphicContext
{
  PDF *m_pdf;
  FontManager* m_font_mgr;

  int m_fonts[4];

public:
  PdfGraphicContext( PDF *pdf, FontManager* font_mgr );

  virtual void Rect( DrawRect rect ) {}
  virtual void FillRect( DrawRect rect ) {}
  virtual void Point( DrawPoint pt ) {}
  virtual void Cross( DrawPoint pt ) {}
  virtual void Line( DrawPoint from, DrawPoint to );
  virtual void Text( DrawRect rect, const char *str ) {}
  virtual void Text( DrawRect rect, DataString &str,
                     ProofLib::NodePointer *np = NULL );
  virtual void Bitmap( DrawPoint pt, DrawBitmap &bm ) {}
  virtual void Flush() {}
  virtual void Clear( const GraphicRect &screen_rect ) {}

private:
  double TextHelper(
    const DrawPoint &origin,
    double adjust_scale,
    FUNIT height,
    DataString &str,
    ProofLib::NodePointer *np = NULL );
};

////////////////////////////////////////////////////////////////////////////////

class ExportPdfImpl : public ExportPdf
{
  PDF *p;
  char *m_sourcename;

protected:
  ExportPdfImpl()
  {
    p = NULL;
    m_sourcename = NULL;
  }

  friend class ExportPdf;

public:
  virtual void Open( const char *sourcename )
  {
    m_sourcename = strdup( sourcename );

    p = PDF_new();

    std::string helper = "CMR10=" + GetPathForFont( "cmr10.afm" );
    PDF_set_parameter( p, "FontAFM", helper.c_str() );
    helper = "CMR10=" + GetPathForFont( "cmr10.pfb" );
    PDF_set_parameter( p, "FontOutline", helper.c_str() );

    helper = "CMMI10=" + GetPathForFont( "cmmi10.afm" );
    PDF_set_parameter( p, "FontAFM", helper.c_str() );
    helper = "CMMI10=" + GetPathForFont( "cmmi10.pfb" );
    PDF_set_parameter( p, "FontOutline", helper.c_str() );

    helper = "CMSY10=" + GetPathForFont( "cmsy10.afm" );
    PDF_set_parameter( p, "FontAFM", helper.c_str() );
    helper = "CMSY10=" + GetPathForFont( "cmsy10.pfb" );
    PDF_set_parameter( p, "FontOutline", helper.c_str() );

    if( p == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Couldn't create PDFlib object (out of memory)!";
      free( m_sourcename );
      m_sourcename = NULL;
      throw e;
    }

    PDF_TRY(p)
    {
      /* open new PDF file */
      if( PDF_open_file( p, m_sourcename ) == -1 )
      {
        GenericException *e = new GenericException();
        e->Stream() << "Cannot create PDF file ";
        e->Stream() << m_sourcename;
        e->Stream() << ".";
        PDF_delete(p); 
        p = NULL;
        free( m_sourcename );
        m_sourcename = NULL;
        throw e;
      }

      PDF_set_info( p, "Creator", "SEPLAN" );
      PDF_set_info( p, "Title", "Proof" );
    }
    PDF_CATCH(p)
    {
      GenericException *e = new GenericException();
      e->Stream() << "PDFlib error: " << PDF_get_errnum(p) << " ";
      e->Stream() << PDF_get_apiname(p) << ": " << PDF_get_errmsg(p);
      PDF_delete(p); 
      p = NULL;
      free( m_sourcename );
      m_sourcename = NULL;
      throw e;
    }
  }

  virtual void Close()
  {
    ASSERT( p != NULL );

    PDF_TRY(p)
    {
      PDF_close(p);
    }
    PDF_CATCH(p)
    {
      GenericException *e = new GenericException();
      e->Stream() << "PDFlib error: " << PDF_get_errnum(p) << " ";
      e->Stream() << PDF_get_apiname(p) << ": " << PDF_get_errmsg(p);
      PDF_delete(p); 
      p = NULL;
      free( m_sourcename );
      m_sourcename = NULL;
      throw e;
    }

    PDF_delete(p); 
    
    if( m_sourcename != NULL )
    {
      free( m_sourcename );
      m_sourcename = NULL;
    }
  }

  virtual void Write( const ProofLib::ProofDatabase* proofdb )
  {
    ASSERT( false );
  }

  virtual void Write( ProofSheet *sheet )
  {
    ASSERT( p != NULL );

    PDF_TRY(p)
    {
      DrawRect rect = sheet->GetBoundingRect();
  
      PDF_begin_page( p, int( rect.GetWidth() ), int( rect.GetHeight() ) );
//    PDF_set_parameter( p, "textformat", "utf8" );

      PdfGraphicContext gc( p, sheet->GetFontManager() );

      gc.SetClippingRect( GraphicRect(rect) );

      sheet->Draw( gc );

      PDF_end_page( p );
    }
    PDF_CATCH(p)
    {
      GenericException *e = new GenericException();
      e->Stream() << "PDFlib error: " << PDF_get_errnum(p) << " ";
      e->Stream() << PDF_get_apiname(p) << ": " << PDF_get_errmsg(p);
      PDF_delete(p); 
      p = NULL;
      free( m_sourcename );
      m_sourcename = NULL;
      throw e;
    }
  }
};

////////////////////////////////////////////////////////////////////////////////

ExportPdf::ExportPdf()
{
}

ExportPdf* ExportPdf::New()
{
  return new ExportPdfImpl;
}

////////////////////////////////////////////////////////////////////////////////

PdfGraphicContext::PdfGraphicContext( PDF *pdf, FontManager* font_mgr )
{
  ASSERT( pdf != NULL );
  ASSERT( font_mgr != NULL );

  m_pdf = pdf;
  m_font_mgr = font_mgr;

  m_fonts[0] = PDF_load_font( m_pdf, "CMR10", 0, "builtin", "embedding");
  m_fonts[1] = PDF_load_font( m_pdf, "CMMI10", 0, "builtin", "embedding");
  m_fonts[2] = PDF_load_font( m_pdf, "CMSY10", 0, "builtin", "embedding");
  m_fonts[3] = m_fonts[2];
}

void PdfGraphicContext::Line( DrawPoint from, DrawPoint to )
{
  from.TransformBy( m_trans );
  to.TransformBy( m_trans );

#if 0
  std::cout << "from "; from.Dump( std::cout );
  std::cout << std::endl;
  std::cout << "to "; to.Dump( std::cout );
  std::cout << std::endl;
#endif

  PDF_moveto( m_pdf, from.m_x, from.m_y );
  PDF_lineto( m_pdf, to.m_x, to.m_y );
  PDF_stroke( m_pdf );
}

double PdfGraphicContext::TextHelper(
  const DrawPoint &origin,
  double adjust_scale,
  FUNIT height,
  DataString &str,
  ProofLib::NodePointer *np /* = NULL */ )
{
  DrawPoint off = origin;

  FontBitmapDatabase &db = m_font_mgr->GetBitmapDatabase();

  double total_advance = 0;

  //DrawRect clip = GetClippingRect();

  for( size_t i = 0; i < str.GetLength(); i++ )
  {
    FontBitmap *bm = db.GetBitmapOf( str[i], height );

    off.m_x += bm->m_preadj;

    str[i].m_start = off.m_x;


    int font_num;
    long code = m_font_mgr->GetCharCode( str[i], font_num );

    PDF_setfont( m_pdf, m_fonts[font_num], height );
    PDF_set_text_pos( m_pdf, off.m_x, off.m_y + 7 );

    char out[2] = { char(code), '\0' };
    PDF_show( m_pdf, out );

    off.m_x = origin.m_x + 0;

    if( ( i < str.GetLength() - 1 ) &&
        ( !str[i].HasSameStyle( str[i+1] ) || str[i].HasSubStrings() ) )
    {
      total_advance += bm->m_full_advance * adjust_scale;
    }
    else
    {
      total_advance += bm->m_advance * adjust_scale;
    }

    double sup_advance = 0;
    double sub_advance = 0;

    if( str[i].m_sup != NULL )
    {
      DrawPoint neworigin = origin;
      FUNIT newheight = FontManager::GetSuperscriptHeight( height );
      neworigin.m_x += lround(total_advance);
      neworigin.m_y += height - newheight;

      sup_advance =
        TextHelper( neworigin, adjust_scale, newheight, *(str[i].m_sup), np );
    }
    if( str[i].m_sub != NULL )
    {
      DrawPoint neworigin = origin;
      FUNIT newheight = FontManager::GetSuperscriptHeight( height );
      neworigin.m_x += lround(total_advance);
      neworigin.m_y  += (3 * height) / 6 - newheight;

      sub_advance =
        TextHelper( neworigin, adjust_scale, newheight, *(str[i].m_sub), np );
    }

    total_advance += std::max( sup_advance, sub_advance );

    off.m_x += lround(total_advance);
  }

  return total_advance;
}

void PdfGraphicContext::Text( DrawRect rect, DataString &str,
                              ProofLib::NodePointer *np )
{
  rect.TransformBy( m_trans );

  FUNIT height = (FUNIT) rect.GetHeight();

  double adjust_scale = rect.GetHeight() / double(height);

  TextHelper( rect.m_leftbottom, adjust_scale, height, str, np );
}

#endif
