////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/freetype/freetype.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <algorithm>
#include <cmath> // for lround on readhat

#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H

#ifdef DEBUG
#include <iostream>
#include <stdio.h>

using namespace std;
#endif 

#include "util/myassert.h"
#include "chars/type1chars.h"
#include "freetype.h"

////////////////////////////////////////////////////////////////////////////////

#define FT_GROUND_FACTOR            56
//#define FT_FONT_HEIGHT_FACTOR       0.8
#define FT_FONT_HEIGHT_CALC         128

#define font_filename               "cmr10.pfb"
#define font_filename_it            "cmmi10.pfb"
#define font_filename_sym           "cmsy10.pfb"
#define font_filename_cal           "cmsy10.pfb"

#define font_kern_filename          "cmr10.afm"
#define font_kern_filename_it       "cmmi10.afm"
#define font_kern_filename_sym      "cmsy10.afm"
#define font_kern_filename_cal      "cmsy10.afm"

////////////////////////////////////////////////////////////////////////////////

FontManager* FontManager::NewFontManager( GraphicSystem *sys )
{
  return new FtFontManager( sys );
}

////////////////////////////////////////////////////////////////////////////////

FtFace::FtFace()
{
  m_face = 0;
  m_unicode_map = NULL;
}

FtFace::~FtFace()
{
  FT_Error error = FT_Done_Face( m_face );

  ASSERT( !error );

  if( m_unicode_map != NULL)
    delete m_unicode_map;
}

void FtFace::LoadFace( FtFontManager *m, const char *filename )
{
  FT_Error error;
  error = FT_New_Face( m->GetLibrary(), filename, 0, &m_face );

  if( error )
  {
    /// \todo Exception
    fprintf( stderr, "Error FT_New_Face ('%s'): 0x%04x\n", filename, error );
    ASSERT( false );
  }

  m_filename = filename;

  error = FT_Select_Charmap( m_face, ft_encoding_adobe_custom );

  if( error )
  {
    error = FT_Select_Charmap( m_face, ft_encoding_unicode );

    if( error )
    {
      /// \todo Exception
      fprintf( stderr, "Error FT_Select_Charmap for '%s': 0x%04x\n",
                filename, error );
      ASSERT( false );
    }
  }
  else
  {
    FT_UInt gindex;

    if( m_unicode_map != NULL)
      delete m_unicode_map;

    m_unicode_map = new UnicodeMap;

    // fill unicode map
    for( FT_ULong charcode = FT_Get_First_Char( m_face, &gindex );
         gindex != 0;
         charcode = FT_Get_Next_Char( m_face, charcode, &gindex ) )
    {
      char buffer[32];

      error = FT_Get_Glyph_Name( m_face, gindex, buffer, sizeof(buffer) );

      if( error )
      {
#ifdef DEBUG
        fprintf( stderr, "Glyph name for 0x%40x not found (in '%s').\n",
                 charcode, filename );
#endif
        continue;
      }

      UCHAR c = Type1Convertor.GetUnicode( buffer );

      if( c == '?' )
      {
#ifdef DEBUG
        fprintf( stderr, "Glyph name '%s' not found (in '%s').\n",
                 buffer, filename );
#endif
        continue;
      }

      (*m_unicode_map)[c] = charcode;
    }
  }
}

void FtFace::LoadKern( const char *filename )
{
  FT_Error error = FT_Attach_File( m_face, filename );
  ASSERT( !error );
}

FT_ULong FtFace::GetCharCode( UCHAR c ) const
{
  if( m_unicode_map == NULL )
    return c;

  UnicodeMap::const_iterator i = m_unicode_map->find( c );

  if( i == m_unicode_map->end() )
  {
#ifdef DEBUG
#if 0
    fprintf( stderr, "font '%s' doesn't contain unicode 0x%04x\n", 
             m_filename.c_str(), c );
#endif
#endif
    return CHAR_VOID;
  }

//fprintf( stderr, "FOUND 0x%04x for unicode 0x%04x\n", i->second, c );
  return i->second;
}

////////////////////////////////////////////////////////////////////////////////

FtFontManager::FtFontManager( GraphicSystem *sys )
  : FontManager( sys )
{
  FT_Error error;

  error = FT_Init_FreeType( &m_library );      /* initialize library */

  ASSERT( !error );

  m_faces[FACE_ROMAN].LoadFace( this, GetPathForFont( font_filename ).c_str() );
  m_faces[FACE_ITALIC].LoadFace( this, GetPathForFont( font_filename_it ).c_str() );
  m_faces[FACE_SYMBOL].LoadFace( this, GetPathForFont( font_filename_sym ).c_str() );
  m_faces[FACE_CALIGRAPH].LoadFace( this, GetPathForFont( font_filename_cal ).c_str() );
#if 0
  m_faces[FACE_ROMAN].LoadKern( font_kern_filename );
  m_faces[FACE_ITALIC].LoadKern( font_kern_filename_it );
  m_faces[FACE_SYMBOL].LoadKern( font_kern_filename_sym );
  m_faces[FACE_CALIGRAPH].LoadKern( font_kern_filename_cal );
#endif
}

FT_ULong FtFontManager::GetCharCode( const DataChar &c, FtFace **face )
{
  ASSERT( face != NULL );

  UCHAR ucode = c.m_char;

  if( ucode == CHAR_THIN_SPACE )
    ucode = 0x20;

  if( c.m_cal )
    *face = &m_faces[FACE_CALIGRAPH];
  else if( c.m_italic && !(ucode >= '0' && ucode <= '9') )
    *face = &m_faces[FACE_ITALIC];
  else
    *face = &m_faces[FACE_ROMAN];

  FT_ULong charcode = (*face)->GetCharCode( ucode );
  
  size_t i = 0;
  while( charcode == CHAR_VOID )
  {
    if( i > FACE_LAST )
      break;

    *face = &m_faces[i];
    charcode = m_faces[i++].GetCharCode( ucode );
  }

  if( charcode == CHAR_VOID )
  {
    charcode = '?';
    *face = &m_faces[0];
  }

  return charcode;
}

long FtFontManager::GetCharCode( const DataChar &c, int &font_num )
{
  UCHAR ucode = c.m_char;

  if( ucode == CHAR_THIN_SPACE )
    ucode = 0x20;

  if( c.m_cal )
    font_num = FACE_CALIGRAPH;
  else if( c.m_italic && !(ucode >= '0' && ucode <= '9') )
    font_num = FACE_ITALIC;
  else
    font_num = FACE_ROMAN;

  FT_ULong charcode = (m_faces[font_num]).GetCharCode( ucode );
  
  size_t i = 0;
  while( charcode == CHAR_VOID )
  {
    if( i > FACE_LAST )
      break;

    font_num = i;
    charcode = m_faces[i++].GetCharCode( ucode );
  }

  if( charcode == CHAR_VOID )
  {
    charcode = '?';
    font_num = 0;
  }

  return charcode;
}

void FtFontManager::MakeMetrics( const DataChar &c, CharMetrics &metrics )
{
  FtFace *face = NULL;

  FT_ULong charcode = GetCharCode( c, &face );

  ASSERT( face != NULL );
  
  FT_Error error;

  // set character size
#if 1
  error = FT_Set_Char_Size( face->GetFace(), 0,
                            FT_FONT_HEIGHT_CALC * FT_GROUND_FACTOR,
                            GetHorDPI(), GetVerDPI() );
#else
  error = FT_Set_Pixel_Sizes( face->GetFace(), 0, FT_FONT_HEIGHT_CALC );
#endif

  ASSERT( !error );
  
  // load glyph metrics
  error = FT_Load_Char( face->GetFace(), charcode, FT_LOAD_DEFAULT );

  ASSERT( !error );
  
  // get metrics
  FT_Glyph_Metrics ft_metrics = face->GetFace()->glyph->metrics;

  metrics.m_width       = double(ft_metrics.width) / (1<<6);
  metrics.m_height      = double(ft_metrics.height) / (1<<6);
  metrics.m_advance     = double(face->GetFace()->glyph->linearHoriAdvance)
                          / (1<<16);

  metrics.m_width       /= double(FT_FONT_HEIGHT_CALC);
  metrics.m_height      /= double(FT_FONT_HEIGHT_CALC);
  metrics.m_advance     /= double(FT_FONT_HEIGHT_CALC);

  // extra for thin space
  if( c.m_char == CHAR_THIN_SPACE )
  {
    metrics.m_width /= 2;
    metrics.m_advance /= 2;
  }

  ASSERT( metrics.m_advance != 0.0 );
}

FontBitmap *FtFontManager::MakeBitmap( const DataChar &c_, FUNIT height )
{
  FtFace *face = NULL;

  FT_ULong charcode = GetCharCode( c_, &face );

  ASSERT( face != NULL );
  
  FT_Error error;

  bool bthinspace = false;

  if( c_.m_char == CHAR_THIN_SPACE )
    bthinspace = true;

  DataChar c = c_;

  if( FT_Get_Char_Index( face->GetFace(), charcode ) == 0 )
  {
#ifdef DEBUG
    cerr << "Font " << face->GetFace()->family_name;
    if( face->GetFace()->style_name != NULL )
      cerr << " " << face->GetFace()->style_name;
    cerr << " does not contain charcode " << charcode;
    if( (charcode == charcode & 0xff) && isprint( (char) charcode ) )
      cerr << " '" << (char) charcode << "'";
    cerr << endl;
#endif

    charcode = '?';
    face = &(m_faces[FACE_ROMAN]);
  }

#if 1
  const FUNIT std_height = 1024;
  double scale = double(height) / double(std_height);

  /* set character size */
  error = FT_Set_Char_Size( face->GetFace(), 0, std_height * FT_GROUND_FACTOR,
                            GetHorDPI(), GetVerDPI() );

  ASSERT( !error );
  
  FT_Matrix transmat;
  transmat.xx = (FT_Fixed) ( scale * 0x10000L );
  transmat.xy = 0;
  transmat.yx = 0;
  transmat.yy = transmat.xx;

  ASSERT( transmat.yy != 0 );

  FT_Set_Transform( face->GetFace(), &transmat, NULL );
#else
  double scale = 1.0;

  /* set character size */
#if 1
  error = FT_Set_Char_Size( face->GetFace(), 0, height * FT_GROUND_FACTOR,
                            GetHorDPI(), GetVerDPI() );
#else
  error = FT_Set_Pixel_Sizes( face->GetFace(), 0, height );
#endif
    
  ASSERT( !error );
  
#endif

  error = FT_Load_Char( face->GetFace(), charcode, FT_LOAD_RENDER );
    
  if( error )
  {
    fprintf( stderr, "Error FT_Load_Char: 0x%04x\n", error );
    ASSERT( false );
  }

  FT_GlyphSlot slot = face->GetFace()->glyph;
  
#if 1
  double advance =
    double(face->GetFace()->glyph->linearHoriAdvance) / (1 << 16);
  advance *= scale;
#else
  double advance = double( slot->advance.x ) / (1 << 6);
#endif

  FUNIT preadj = slot->bitmap_left;

  if( bthinspace )
  {
    advance /= 2;
    preadj /= 2;
  }

  FUNIT width = (FUNIT) (slot->bitmap).width;
  FUNIT bheight = (FUNIT) (slot->bitmap).rows;
  GraphicSize size( width, bheight );

  FontBitmap *bm = m_sys->NewFontBitmap( size );
  bm->SetAllPixel( SCOLOR_WHITE );

  DrawOnBitmap( bm, &slot->bitmap, GraphicPoint() );

  bm->m_horz_offset = 3*height/4 - slot->bitmap_top;
  bm->m_preadj = preadj;

  const CharMetrics &metrics = GetBitmapDatabase().GetMetricsOf( c );

#if 0
  bm->m_advance = advance;
#else
  bm->m_advance = metrics.m_advance * height;
#endif

  if( metrics.m_width > metrics.m_advance )
    bm->m_full_advance = metrics.m_width * height;
  else
    bm->m_full_advance = bm->m_advance;

  return bm;
}

void FtFontManager::DrawOnBitmap( FontBitmap *bm,
                                  FT_Bitmap* bitmap,
                                  GraphicPoint off )
{
  int i,j,p,q;
  SUNIT width = bm->GetWidth();
  SUNIT height = bm->GetHeight();

  for ( i = off.m_x, p = 0; p < bitmap->width; i++, p++ )
  {
    for ( j = off.m_y, q = 0; q < bitmap->rows; j++, q++ )
    {
      if ( i < 0 || j < 0 ||
           i >= width || j >= height )
        continue;

      long col = bitmap->buffer[q * bitmap->width + p];
//      long col = 0xff - bitmap->buffer[q * bitmap->width + p];
//      col *= 0x010101;
//printf( "0x%02x: (0x%02x) 0x%06x\n", bitmap->buffer[q * bitmap->width + p], 0xff - bitmap->buffer[q * bitmap->width + p], col );

      bm->SetPixel( GraphicPoint( i, j ), col );
    }
  }
}

