////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/graphiccontext.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <algorithm>

#include "graphiccontext.h"
#include "proofsheet.h"
#include "command/execcontext.h"

#ifdef DEBUG
#include <ostream>

using namespace std;
#endif

GraphicPoint::GraphicPoint( const DrawPoint &pt )
{
  m_x = pt.LongX();
  m_y = pt.LongY();
}

GraphicRect::GraphicRect( const DrawRect &rect )
{
  m_origin = rect.m_leftbottom;
  m_size = GraphicPoint( GraphicPoint( rect.m_righttop ) - m_origin );
}
 
////////////////////////////////////////////////////////////////////////////////

void GraphicSystem::SetTopMenu( GraphicWindow *wnd )
{
  ASSERT( wnd != NULL );

  wnd->ClearTopMenu();

  for( size_t i = 0; i < m_top_menu.size(); i++ )
    wnd->AppendToMenuBar( m_top_menu[i] );
}

////////////////////////////////////////////////////////////////////////////////

void GraphicContext::SetClippingRect( const GraphicRect& rect )
{
  m_cliprect = rect;
}

////////////////////////////////////////////////////////////////////////////////

GraphicWindow::GraphicWindow( GraphicSystem &sys )
{
  m_sys = &sys;
  AttachToDispatcher( sys.GetMainInputDispatcher() );
}

void GraphicWindow::SetView( GraphicView* view )
{
  if( m_view != NULL )
    Detach( m_view );

  GraphicViewContainer::SetView( view );
  view->SetContainer( this );

  Attach( m_view );
}

////////////////////////////////////////////////////////////////////////////////

GraphicView::GraphicView( GraphicSystem *sys )
{
  m_sys = sys;
  ASSERT( m_sys != NULL );
  m_container = NULL;
}

////////////////////////////////////////////////////////////////////////////////

GraphicProofView::GraphicProofView( GraphicSystem *sys )
  : GraphicView( sys )
{
  m_zoom_factor = 1;
  m_sheet = NULL;
  m_image =
    GetGraphicSystem()->NewDrawBitmap( GetGraphicSystem()->GetMaxScreenSize() );
  ASSERT( m_image != NULL );
  m_highlight_func = NULL;
  m_last_zoom_factor = 0;
}

/// returns proof width in screen dim
GUNIT GraphicProofView::GetProofWidth() const
{
  if( m_sheet == NULL )
    return 0;

  DrawRect rect = m_sheet->GetBoundingRect();
  return lround(rect.GetWidth() * m_zoom_factor);
}

/// returns proof height in screen dim
GUNIT GraphicProofView::GetProofHeight() const
{
  if( m_sheet == NULL )
    return 0;

  DrawRect rect = m_sheet->GetBoundingRect();
  return lround(rect.GetHeight() * m_zoom_factor);
}

void GraphicProofView::SetSize( const GraphicSize &size )
{
  if( size == m_size )
    return;

  GraphicView::SetSize( size );

  if( m_image != NULL && !m_image->GetSize().Overlap( size ) )
  {
    delete m_image;
      m_image = NULL;
  }
  
  if( m_image == NULL )
    m_image = GetGraphicSystem()->NewDrawBitmap( size );
}

void GraphicProofView::SetViewPoint( const DrawPoint &vp )
{
  if( vp == m_view_pt )
    return;

  m_view_pt = vp;

  CalcTransMatrix();

  CalcScrollFieldSize();

  FireViewOriginChanged();

  UpdateScreen();
}

void GraphicProofView::ResizeViewPortEvent( const GraphicSize &size )
{
  if( size == m_size )
    return;

  SetSize( size );
  
  CalcScrollFieldSize();

  FireViewOriginChanged();

  UpdateScreen();
#if 0
  //m_view_pt += m_viewoffset_pt;
  CalcViewOffset();
  //m_view_pt -= m_viewoffset_pt;

  DrawPoint pt = m_view_pt + m_viewoffset_pt;
  pt *= m_zoom_factor;
#if 0
std::cout << "SetSize: pt is ";
pt.Dump( std::cout );
std::cout << std::endl;
#endif
  DrawPoint pt_bkp = pt;
  GUNIT scroll_width = GetProofWidth() - GetWidth();
  GUNIT scroll_height = GetProofHeight() - GetHeight();
  if( scroll_width < 0 ) scroll_width = 0;
  if( scroll_height < 0 ) scroll_height = 0;
#if 0
std::cout << "SetSize: scroll_width is " << scroll_width << std::endl;
std::cout << "SetSize: scroll_height is " << scroll_height << std::endl;
std::cout << "SetSize: GetProofWidth is " << GetProofWidth() << std::endl;
std::cout << "SetSize: GetProofHeight is " << GetProofHeight() << std::endl;
std::cout << "SetSize: GetWidth is " << GetWidth() << std::endl;
std::cout << "SetSize: GetHeight is " << GetHeight() << std::endl;
#endif

  if( pt.m_x < 0.0 )
    pt.m_x = 0.0;
  if( pt.m_y < 0.0 )
    pt.m_y = 0.0;
  if( pt.m_x > scroll_width )
    pt.m_x = SUNIT(scroll_width);
  if( pt.m_y > scroll_height )
    pt.m_y = SUNIT(scroll_height);

#if 0
std::cout << "SetSize: pt is ";
pt.Dump( std::cout );
std::cout << std::endl;
#endif
  if( pt != pt_bkp )
  {
#if 0
std::cout << "SetSize: pt is ";
std::cout << std::endl;
#endif
    m_view_pt = pt;
    m_view_pt /= m_zoom_factor;
    m_view_pt -= m_viewoffset_pt;
  }
  
  CalcTransMatrix();

  //Draw();
  UpdateScreen();
#endif
}

void GraphicProofView::ZoomAll()
{
  SUNIT zoom_factor = 1;

  DrawRect rect = m_sheet->GetBoundingRect();
  SUNIT ps_width  = ceil(rect.GetWidth());
  SUNIT ps_height = ceil(rect.GetHeight());

  GraphicSize size = GetViewportMaxSize();

  if( size.GetWidth() < ps_width ||
      size.GetHeight() < ps_height )
  {
    SUNIT wd = (SUNIT) ps_width - (SUNIT) size.GetWidth();
    SUNIT hd = (SUNIT) ps_height - (SUNIT) size.GetHeight();

    if( wd > hd )
      zoom_factor = (SUNIT) size.GetWidth() / (SUNIT) ps_width;
    else
      zoom_factor = (SUNIT) size.GetHeight() / (SUNIT) ps_height;
  }

  SetZoomMidpoint( rect.MidPoint(), zoom_factor );
#if 0
  CalcViewOffset();

  m_view_pt = -m_viewoffset_pt;
#ifdef DEBUG
std::cout << "ZoomAll: m_view_pt is ";
m_view_pt.Dump( std::cout );
std::cout << std::endl;
#endif

  CalcTransMatrix();

  //Draw();

  CalcScrollFieldSize();

  FireViewOriginChanged();

  UpdateScreen();
#endif
}

void GraphicProofView::ZoomFactor( SUNIT factor )
{
  SetZoomMidpoint( GetZoomMidpoint(), factor );
}

// 0,0 will be left lower coner (input is viewport coords)
void GraphicProofView::SetViewOrigin( const DrawPoint &pt )
{
#if 0
#ifdef DEBUG
std::cout << "m_view_pt was " << m_view_pt.m_x << "," << m_view_pt.m_y << std::endl;
#endif

  m_view_pt = pt;
  m_view_pt /= m_zoom_factor;
  m_view_pt -= m_viewoffset_pt;
  
#if 0
  m_view_pt.m_x = -pt.m_x;
  m_view_pt.m_y = -pt.m_y;
#endif

#ifdef DEBUG
std::cout << "m_view_pt is " << m_view_pt.m_x << "," << m_view_pt.m_y << std::endl;
#endif

  CalcTransMatrix();

  //Draw();

  FireViewOriginChanged();

  UpdateScreen();
#endif
}

void GraphicProofView::SetZoomMidpoint( const DrawPoint &mid_pt,
                                        double zoom_factor )
{
#ifdef DEBUG
std::cout << "SetZoomMidpoint: zoom_factor is " << zoom_factor;
std::cout << std::endl;
std::cout << "SetZoomMidpoint: mid_pt is "; mid_pt.Dump( std::cout );
std::cout << std::endl;
#endif

  bool zoom_factor_changed = false;

  if( zoom_factor == 0.0 )
    zoom_factor = m_zoom_factor;


  if( m_zoom_factor != zoom_factor )
  {
    m_zoom_factor = zoom_factor;

    zoom_factor_changed = true;
  }

  // vector from midpoint to left upper corner of viewport in proofsheet coords
  DrawPoint vec_a( SUNIT(m_size.m_x) / 2,  - SUNIT(m_size.m_y) / 2 );
  vec_a /= m_zoom_factor;

  // point of upper left corner of the proof sheet
  DrawPoint proofsheet_lefttop_pt( m_sheet->GetBoundingRect().m_leftbottom );
  proofsheet_lefttop_pt.m_y += m_sheet->GetBoundingRect().GetHeight();
  
  // vector from midpoint to left upper corner of proof sheet in
  // proofsheet coords
  DrawPoint vec_b( proofsheet_lefttop_pt - mid_pt );
#ifdef DEBUG
std::cout << "SetZoomMidpoint: vec_a is "; vec_a.Dump( std::cout );
std::cout << std::endl;
std::cout << "SetZoomMidpoint: proofsheet_lefttop_pt is ";
proofsheet_lefttop_pt.Dump( std::cout );
std::cout << std::endl;
std::cout << "SetZoomMidpoint: vec_b is "; vec_b.Dump( std::cout );
std::cout << std::endl;
#endif

  m_view_pt = vec_a + vec_b;
#ifdef DEBUG
std::cout << "SetZoomMidpoint: m_view_pt is "; m_view_pt.Dump( std::cout );
std::cout << std::endl;
#endif

  CalcTransMatrix();

  CalcScrollFieldSize();

  FireViewOriginChanged();

  UpdateScreen();
#if 0

  DrawPoint screen_mid_pt;

  screen_mid_pt.m_x = m_size.m_x;
  screen_mid_pt.m_y = m_size.m_y;

  screen_mid_pt /= SUNIT(2);

#ifdef DEBUG
std::cout << "SetZoomMidpoint: m_view_pt is ";
m_view_pt.Dump( std::cout );
std::cout << std::endl;
#endif

  m_view_pt += (pt - ViewportToProofSheet( screen_mid_pt ));

#ifdef DEBUG
std::cout << "SetZoomMidpoint: m_view_pt is ";
m_view_pt.Dump( std::cout );
std::cout << std::endl;
#endif

  CalcTransMatrix();

#ifdef DEBUG
std::cout << "SetZoomMidpoint: Zoom Midpoint is ";
ViewportToProofSheet( screen_mid_pt ).Dump( std::cout );
std::cout << std::endl;
#endif

  //Draw();

  if( zoom_factor_changed )
  {
    CalcScrollFieldSize();
  }

  FireViewOriginChanged();

  UpdateScreen();
#endif
}

void GraphicProofView::SetProofSheet( ProofSheet *ps )
{
  ASSERT( m_sheet == NULL );
  m_sheet = ps;
  m_sheet->AttachNotificant( this );
  ResizeProofSheetEvent();
}

void GraphicProofView::ResizeProofSheetEvent()
{
#ifdef DEBUG
cout << "ResizeProofSheetEvent GetProofWidth GetProofHeight: ";
cout << GetProofWidth() << " x " << GetProofHeight() << endl;
#endif
#if 0
  resizeContents( GetProofWidth(), GetProofHeight() );

  QSize qsize = viewportSize( contentsWidth(), contentsHeight() );

  GraphicSize size( qsize.width(),
                    qsize.height() );

#ifdef DEBUG
cout << "ResizeProofSheetEvent calling SetSize with: ";
cout << size.GetWidth() << " x " << size.GetHeight() << endl;
#endif
  SetSize( size );
#endif
  CalcScrollFieldSize();
  FireViewOriginChanged();
  RepaintBuffer();
  UpdateScreen();
}

void GraphicProofView::RepaintBuffer()
{
  m_last_viewport_rect = GraphicRect( 0, 0,
                                      m_size.GetWidth(), m_size.GetHeight() );
  RepaintBuffer( m_last_viewport_rect );
  m_last_zoom_factor = m_zoom_factor;
  m_last_view_pt = m_view_pt; // + m_viewoffset_pt;
}

void GraphicProofView::RepaintBuffer( const GraphicRect &cliprect )
{
#if 0
std::cout << "Drawing area: ";
cliprect.Dump( std::cout );
std::cout << std::endl;
#endif
  GraphicContext *gc = CreateGraphicContext();

  gc->SetClippingRect( cliprect );

  gc->Clear( cliprect );

#ifdef DEBUG
std::cout << "Drawing area: ";
cliprect.Dump( std::cout );
std::cout << std::endl;
#endif
  if( m_sheet != NULL )
    m_sheet->Draw( *gc );
#ifdef DEBUG
std::cout << "Drawing area done." << std::endl;
#endif

  FreeGraphicContext( gc );
}

void GraphicProofView::RepaintItem( DrawItem *item )
{
  ASSERT( item != NULL );

  DrawRect clip_ = item->GetBoundingRect();
  clip_.TransformBy( *GetTransMatrix() );

  GraphicRect clip = clip_.GetBiggerIntRect();

  RepaintBuffer( clip );
}

GraphicRect GraphicProofView::RequestBuffer( const GraphicRect& viewport_rect )
{
  GraphicRect rect = viewport_rect;
  rect.m_origin.m_x = 0;
  rect.m_origin.m_y = 0;

  if( m_last_zoom_factor != m_zoom_factor )
  {
    RepaintBuffer();
    return rect;
  }

#ifdef DEBUG
std::cout << "RequestBuffer: m_last_view_pt is ";
m_last_view_pt.Dump( std::cout );
std::cout << std::endl;
std::cout << "RequestBuffer: m_view_pt is ";
m_view_pt.Dump( std::cout );
std::cout << std::endl;
#endif

  DrawPoint a = ProofSheetToViewport( m_last_view_pt );
  DrawPoint b = ProofSheetToViewport( m_view_pt );

  DrawPoint move = b-a;

#ifdef DEBUG
std::cout << "RequestBuffer: move is ";
move.Dump( std::cout );
std::cout << std::endl;
#endif

#define MOVE_LIMIT 0.01
  if( move.m_x - round( move.m_x ) > MOVE_LIMIT ||
      move.m_y - round( move.m_y ) > MOVE_LIMIT )
  {
#ifdef DEBUG
std::cout << "DOING FULL REPAINT (move not int)";
std::cout << std::endl;
#endif
    RepaintBuffer();
    return rect;
  }

  if( move.LongX() >= m_size.GetWidth() ||
      move.LongY() >= m_size.GetHeight() )
  {
#ifdef DEBUG
std::cout << "DOING FULL REPAINT (move bigger than vp)";
std::cout << std::endl;
#endif
    RepaintBuffer();
    return rect;
  }

#if 0
std::cout << "RequestBuffer: viewport_rect is ";
viewport_rect.Dump( std::cout );
std::cout << std::endl;
std::cout << "RequestBuffer: m_last_viewport_rect is ";
m_last_viewport_rect.Dump( std::cout );
std::cout << std::endl;
#endif

  if( !move.IsZero() );
    m_image->MoveContent( move );

  GraphicPoint pt_help( move.LongX() + m_last_viewport_rect.GetWidth(),
                        move.LongY() + m_last_viewport_rect.GetHeight() );
#ifdef DEBUG
std::cout << "RequestBuffer: pt_help is ";
pt_help.Dump( std::cout );
std::cout << std::endl;
#endif

  if( move.LongX() > 0 )
  {
    RepaintBuffer( GraphicRect( 0, 0, move.LongX(),  m_size.GetHeight() ) );
  }
  if( move.LongY() > 0 )
  {
    RepaintBuffer( GraphicRect( 0, 0,  m_size.GetWidth(), move.LongY() ) );
  }
  if( pt_help.m_x < viewport_rect.GetWidth() )
  {
#if 0
std::cout << "RequestBuffer: call will have ";
GraphicRect r_( pt_help.m_x, 0,
                viewport_rect.GetWidth() - pt_help.m_x, m_size.GetHeight() );
r_.Dump( std::cout );
std::cout << std::endl;
#endif
    RepaintBuffer( GraphicRect( pt_help.m_x, 0,
                                viewport_rect.GetWidth() - pt_help.m_x,
                                m_size.GetHeight() ) );
  }
  if( pt_help.m_y < viewport_rect.GetHeight() )
  {
    RepaintBuffer( GraphicRect( 0, pt_help.m_y, m_size.GetWidth(),
                                viewport_rect.GetHeight() - pt_help.m_y ) );
  }

  m_last_zoom_factor = m_zoom_factor;
  m_last_viewport_rect = viewport_rect;
  m_last_view_pt = m_view_pt;
  //m_last_view_pt = m_view_pt + m_viewoffset_pt;
  return rect;
  
}

void GraphicProofView::CalcViewOffset()
{
  return;

  GraphicSize size = GetViewPortSize();
  
  SUNIT vp_width = SUNIT( size.GetWidth() ) / m_zoom_factor;
  SUNIT vp_height = SUNIT( size.GetHeight() ) / m_zoom_factor;

  DrawRect rect = m_sheet->GetBoundingRect();

  SUNIT ps_width  = rect.GetWidth();
  SUNIT ps_height = rect.GetHeight();

#ifdef DEBUG
std::cout << "CalcViewOffset: size = " << vp_width << "," << vp_height
          << std::endl;
std::cout << "CalcViewOffset: rect = ";
rect.Dump( std::cout );
std::cout << std::endl;
#endif

  if( vp_width > ps_width )
  {
    m_viewoffset_pt.m_x = (vp_width - ps_width) / 2;
    m_viewoffset_pt.m_x -= rect.m_leftbottom.m_x;
  }
  else
    m_viewoffset_pt.m_x = 0;

  if( vp_height > ps_height )
  {
    m_viewoffset_pt.m_y = (vp_height - ps_height) / 2;
    m_viewoffset_pt.m_y -= rect.m_leftbottom.m_y;
  }
  else
    m_viewoffset_pt.m_y = 0;
  
#if 0
  GraphicSize size = GetViewPortSize();

  DrawRect rect = m_sheet->GetBoundingRect();
  rect.TransformBy( DrawTransMatrix::Scale( m_zoom_factor ) );

  SUNIT ps_width  = rect.GetWidth();
  SUNIT ps_height = rect.GetHeight();

#ifdef DEBUG
std::cout << "CalcViewOffset: size = " << size.GetWidth() << "," << size.GetHeight() << std::endl;
std::cout << "CalcViewOffset: rect = ";
rect.Dump( std::cout );
std::cout << std::endl;
#endif

  if( size.GetWidth() > ps_width )
  {
    m_viewoffset_pt.m_x = (size.GetWidth() - ps_width) / 2;
    m_viewoffset_pt.m_x -= rect.m_leftbottom.m_x;
  }
  else
    m_viewoffset_pt.m_x = 0;

  if( size.GetHeight() > ps_height )
  {
    m_viewoffset_pt.m_y = (size.GetHeight() - ps_height) / 2;
    m_viewoffset_pt.m_y -= rect.m_leftbottom.m_y;
  }
  else
    m_viewoffset_pt.m_y = 0;
#endif

#ifdef DEBUG
std::cout << "CalcViewOffset: m_viewoffset_pt is ";
m_viewoffset_pt.Dump( std::cout );
std::cout << std::endl;
#endif
}

DrawPoint GraphicProofView::GetZoomMidpoint() const
{
  DrawPoint pt;

  pt.m_x = m_size.m_x;
  pt.m_y = m_size.m_y;

  pt /= SUNIT(2);

#ifdef DEBUG
std::cout << "GetZoomMidpoint: pt (on screen) is ";
pt.Dump( std::cout );
std::cout << std::endl;
#endif

  pt = ViewportToProofSheet( pt );

#ifdef DEBUG
std::cout << "GetZoomMidpoint: pt is ";
pt.Dump( std::cout );
std::cout << std::endl;
#endif

  return pt;
}

DrawPoint GraphicProofView::ViewportToProofSheet( const DrawPoint &pt_ ) const
{
  DrawPoint pt = pt_;

  pt.TransformBy( m_trans.Inverse() );

  return pt;
}

DrawPoint GraphicProofView::ProofSheetToViewport( const DrawPoint &pt_ ) const
{
  DrawPoint pt = pt_;

  pt.TransformBy( m_trans );

  return pt;
}

void GraphicProofView::FireViewOriginChanged()
{
  // calculate the rect of sheet relative to this vp (in screen coords)
  DrawRect org_sheet_rect = GetProofSheet()->GetBoundingRect();
  org_sheet_rect.TransformBy( m_trans );
  GraphicRect sheet_rect = org_sheet_rect.GetBiggerIntRect();

  ViewOriginChangedEvent( GraphicSize( std::max( 0, -sheet_rect.GetX() ),
                                       std::max( 0, -sheet_rect.GetY() ) ) );
}

void GraphicProofView::ChangeProofEvent( ProofLib::Proof *proof_new,
                                         ProofLib::Proof *proof_old )
{
  m_last_zoom_factor = -1; // force redraw
  ZoomAll();
}

void GraphicProofView::OnEvent( RepaintEvent *e )
{
  ASSERT( false );
  //Repaint( e->m_area );
}

void GraphicProofView::OnEvent( ButtonClickEvent *e )
{
  DrawTransMatrix m = GetTransMatrix()->Inverse();

  Event *n = e->GetEventTransformedBy( m );
  e = dynamic_cast<ButtonClickEvent*>( n );

#ifdef DEBUG
  cout << "click at ";
  e->m_pt.Dump( cout );
  cout << " with button: " << e->GetButton() << " and mods: " << e->GetMods();
  cout << endl;
#endif

  GraphicMainInputDispatcher *dis =
    GetGraphicSystem()->GetMainInputDispatcher();

  CommandCall *cmdcall = NULL;

  if( !dis->IsSelectingItem() )
  {
    cmdcall = dis->GetButtonDb()->LookupCommand( e->GetButton(), e->GetMods() );

    if( cmdcall == NULL )
      return;
  }

  DrawItem *item = GetProofSheet()->GetItemAtPoint( e->m_pt );

  if( dis->IsSelectingItem() )
  {
    if( item == NULL )
      return;

    dis->SetSelectedItem( item );
  }
  else
  {
    CommandCallArg *arg = cmdcall->NewArgument( item, &(e->m_pt) );

    GetContext()->GetCommandProcessor()->Process( cmdcall->Apply( arg ) );
  }
}

void GraphicProofView::OnEvent( MouseMoveEvent *e_ )
{
  DrawTransMatrix m = GetTransMatrix()->Inverse();

  Event *n = e_->GetEventTransformedBy( m );
  MouseMoveEvent *e = dynamic_cast<MouseMoveEvent*>( n );
#if 0
  cout << "MOUSE EVENT: at ";
  e_->m_pt.Dump( cout );
  cout << " sheet coords:";
  e->m_pt.Dump( cout );
  cout << endl;
#endif

  static DrawItem *last_item = NULL;

  if( last_item != NULL && !GetProofSheet()->Contains( last_item ) )
    last_item = NULL;

  DrawItem *item = GetProofSheet()->GetItemAtPoint( e->m_pt );

  DrawSequentItem *sitem = dynamic_cast<DrawSequentItem*>( item );
  DrawSequentItem *slast_item = dynamic_cast<DrawSequentItem*>( last_item );

  bool highlight_changed = false;

  if( slast_item != NULL && sitem != slast_item )
    slast_item->ClearHighlightPart();

  if( sitem != NULL )
  {
    try
    {
      DataChar c = sitem->GetCharAt( e_->m_pt );

//c.Dump( std::cout );
//std::cout << std::endl;
//c.m_np.Dump( std::cout );
//std::cout << std::endl;

      if( sitem != slast_item ||
          sitem->GetHighlightPart() != c.m_np )
      {
        highlight_changed = true;
        sitem->SetHighlightPart( c.m_np );

#if 0
        ProofLib::DataNode *n = c.m_np.GetDataNode( sitem->GetSequent() );

        ASSERT( n != NULL );

        DataString str;
        str << *n;
        str.Dump( std::cout );
        std::cout << std::endl;
#endif
      }
    }
#ifdef DEBUG
    catch( Exception *e )
    {
      e->Print( std::cerr );
      std::cerr << std::endl;
    }
#endif
    catch( ... )
    {
#ifdef DEBUG
      std::cerr << "Caught unknown exception in " << __FILE__ << ":" << __LINE__
                << std::endl;
#endif
    }
  }

  if( m_highlight_func != NULL && item != NULL )
  {
    if( !m_highlight_func( item ) )
    {
      item = NULL;

      if( sitem != NULL )
      {
        sitem->ClearHighlightPart();
        sitem = NULL;
      }
    }
  }


  if( item != last_item || highlight_changed )
  {
#ifdef DEBUG
  cout << "DRAW!" << endl;
#endif

    if( item != NULL )
    {
      item->SetHighlight();
      RepaintItem( item );
    }
    if(    last_item != NULL
        && item != last_item
        && GetProofSheet()->Contains( last_item ) )
    {
      last_item->SetHighlight( false );
      RepaintItem( last_item );
    }

#if 0
    GraphicContext *gc = CreateGraphicContext();

    if( item != NULL )
    {
      item->SetHighlight();
      item->Draw( *gc );
    }
    if(    last_item != NULL
        && item != last_item
        && GetProofSheet()->Contains( last_item ) )
    {
      last_item->SetHighlight( false );
      last_item->Draw( *gc );
    }

    FreeGraphicContext( gc );
#endif
    UpdateScreen();

    last_item = item;
  }

  delete n;
}

void GraphicProofView::OnEvent( DragDropEvent *e )
{
#if 0
  static DrawPoint start;

#if 0
  cout << "drag drop event!!! at ";
  e->m_pt.Dump( cout );
  cout << endl;
#endif

  if( e->IsDragging() ) // begining
  {
    start = e->m_pt;
  }
  else
  {
    GraphicPoint pt;

    pt.m_x = GUNIT (e->m_pt.m_x - start.m_x);
    pt.m_y = GUNIT (e->m_pt.m_y - start.m_y);

    Move( pt );
  }
#endif
}

