////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/graphiccontext.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#ifndef _GRAPHICCONTEXT_H_
#define _GRAPHICCONTEXT_H_

#include "libproof.h"
#ifdef SIREALLYNEEDED
#include "util/threads.h"
#endif
#include "datastring.h"

#include "graphictypes.h"
#include "drawtypes.h"
#include "event.h"
#include "proofsheetnotificant.h"

#ifdef DEBUG
#include <ostream>

using namespace std;
#endif

class ProofSheet;
class DrawBitmap;
class DrawItem;
class FontBitmap;
class FontManager;

class GraphicViewContainer;
class GraphicSystem;
class GraphicWindow;
class GraphicMenu;
class GraphicMenuItem;
class GraphicTopMenuItem;
class GraphicDialog;
class GraphicInputDialog;
class GraphicFileDialog;

class GraphicSystem
{
public:
  GraphicSystem() { m_active_proof_wnd = NULL; }
  virtual ~GraphicSystem() {}

  virtual GraphicWindow* NewProofWindow( ProofSheet *ps ) = 0;
  virtual DrawBitmap* NewDrawBitmap( const GraphicSize &size ) = 0;
  virtual FontBitmap* NewFontBitmap( const GraphicSize &size ) = 0;
  virtual GraphicMenu* NewMenu() = 0;
  virtual GraphicMenuItem* NewMenuItem( const USTRING &name ) = 0;
  virtual GraphicTopMenuItem* NewTopMenuItem( const USTRING &name ) = 0;
  virtual void ClearTopMenu() = 0;
  virtual void SetTopMenu( GraphicWindow * wnd );
  virtual GraphicInputDialog* NewInputDialog( const USTRING &name ) = 0;
  virtual GraphicFileDialog* NewFileDialog( const USTRING &name,
		                            bool for_open ) = 0;

  virtual GraphicWindow* GetActiveProofWindow() { return m_active_proof_wnd; }
  virtual GraphicMainInputDispatcher *GetMainInputDispatcher() = 0;
  virtual FontManager *GetFontManager() = 0;

  virtual GraphicSize GetMaxScreenSize() const = 0;

protected:
  GraphicWindow         *m_active_proof_wnd;
  std::vector<GraphicTopMenuItem*>  m_top_menu;
};

class GraphicContext
{
  SCOLOR                 m_fg_color;
  SCOLOR                 m_bg_color;
  SCOLOR                 m_h_fg_color;
  GraphicRect            m_cliprect;

public: // dirty
  DrawTransMatrix        m_trans;

public:
  GraphicContext()
  {
    m_fg_color = GetDefaultForegroundColor();
    m_bg_color = GetDefaultBackgroundColor();
    m_h_fg_color = GetDefaultHighlightForegroundColor();
  }

  virtual const DrawTransMatrix& GetTransMatrix() const { return m_trans; };
  virtual const GraphicRect& GetClippingRect() const { return m_cliprect; };
  virtual void SetClippingRect( const GraphicRect& rect );

  virtual void Rect( DrawRect rect ) = 0;
  virtual void FillRect( DrawRect rect ) = 0;
  virtual void Point( DrawPoint pt ) = 0;
  virtual void Cross( DrawPoint pt ) = 0;
  virtual void Line( DrawPoint from, DrawPoint to ) = 0;
  virtual void Text( DrawRect rect, const char *str ) = 0;
  virtual void Text( DrawRect rect, DataString &str,
                     ProofLib::NodePointer *np = NULL ) = 0;
  virtual void Bitmap( DrawPoint pt, DrawBitmap &bm ) = 0;
  virtual void Flush() = 0;
  virtual void Clear( const GraphicRect &screen_rect ) = 0;

  virtual SCOLOR GetForegroundColor() { return m_fg_color; }
  virtual void SetForegroundColor( SCOLOR c ) { m_fg_color = c; }
  virtual SCOLOR GetBackgroundColor() { return m_bg_color; }
  virtual void SetBackgroundColor( SCOLOR c ) { m_bg_color = c; }
  virtual SCOLOR GetHighlightForegroundColor() { return m_h_fg_color; }
  virtual void SetHighlightForegroundColor( SCOLOR c ) { m_h_fg_color = c; }

#ifdef DEBUG
  virtual SCOLOR GetDefaultForegroundColor() { return 0xccff33; }
  virtual SCOLOR GetDefaultBackgroundColor() { return 0x3366ff; }
  virtual SCOLOR GetDefaultHighlightForegroundColor() { return 0x00ff00; }
#else
  virtual SCOLOR GetDefaultForegroundColor() { return 0x000000; }
  virtual SCOLOR GetDefaultBackgroundColor() { return 0xffffff; }
  virtual SCOLOR GetDefaultHighlightForegroundColor() { return 0x0000ff; }
#endif
};

class GraphicContextCreator
{
public:
  virtual GraphicContext* CreateGraphicContext() = 0;
  virtual void FreeGraphicContext( GraphicContext* gc ) = 0;
};

class GraphicView : public GraphicInputDispatcher, public GraphicContextCreator
{
protected:
  GraphicViewContainer  *m_container;
  GraphicSystem         *m_sys;
  GraphicSize            m_size;
  DrawTransMatrix        m_trans;

public:
  GraphicView( GraphicSystem *sys );
  virtual ~GraphicView() {};

  virtual GraphicSystem* GetGraphicSystem() const { return m_sys; }
//  virtual DrawTransMatrix GetSystemTransMatrix() = 0;
//  virtual void SetToDefaultTrans() { m_trans = GetSystemTransMatrix(); }
  virtual DrawTransMatrix *GetTransMatrix() { return &m_trans; };

  virtual void SetContainer( GraphicViewContainer* container )
    { m_container = container; }
  virtual GraphicViewContainer* GetContainer() const { return m_container; };

  virtual void SetSize( const GraphicSize &size ) { m_size = size; }
  virtual GraphicSize GetSize() const { return m_size; }
  virtual GUNIT GetWidth() const { return m_size.GetWidth(); }
  virtual GUNIT GetHeight() const { return m_size.GetHeight(); }
  virtual GraphicSize GetViewPortSize() const = 0;

  virtual void UpdateScreen() = 0;
  virtual void UpdateScreen( const GraphicRect &rect ) = 0;

  virtual void ZoomAll() = 0;
  virtual void ZoomFactor( SUNIT factor ) = 0;
  virtual double GetZoomFactor() const = 0;
  virtual DrawPoint GetZoomMidpoint() const = 0;
  virtual void SetZoomMidpoint( const DrawPoint &pt, double zoom_factor ) = 0;
};

class GraphicProofView : public GraphicView, public ProofSheetNotificant
{
  ProofSheet            *m_sheet;

  DrawPoint              m_view_pt;
  DrawPoint              m_viewoffset_pt;

  DrawBitmap            *m_image;

  SUNIT                  m_zoom_factor;

  // only use in RequestBuffer
  GraphicRect            m_last_viewport_rect;
  DrawPoint              m_last_view_pt;
  SUNIT                  m_last_zoom_factor;

public:
  GraphicProofView( GraphicSystem *sys );

  virtual void SetSize( const GraphicSize &size );

  virtual ProofSheet* GetProofSheet() const { return m_sheet; };
  virtual void SetProofSheet( ProofSheet *ps );

  GUNIT GetProofWidth() const;
  GUNIT GetProofHeight() const;

  /// get view point in sheet coords
  const DrawPoint& GetViewPoint() const { return m_view_pt; }
  void SetViewPoint( const DrawPoint &vp );

  virtual GraphicSize GetViewportMinSize() const = 0;
  virtual GraphicSize GetViewportMaxSize() const = 0;

  DrawPoint ViewportToProofSheet( const DrawPoint &pt ) const;
  DrawPoint ProofSheetToViewport( const DrawPoint &pt ) const;

  void SetViewOrigin( const DrawPoint &pt );

  virtual void ZoomAll();
  virtual void ZoomFactor( SUNIT factor );
  virtual double GetZoomFactor() const { return m_zoom_factor; }
  virtual DrawPoint GetZoomMidpoint() const;
  virtual void SetZoomMidpoint( const DrawPoint &pt, double zoom_factor );

  virtual void OnEvent( RepaintEvent *e );
  virtual void OnEvent( MouseMoveEvent *e );
  virtual void OnEvent( DragDropEvent *e );
  virtual void OnEvent( ButtonClickEvent *e );

  // for highlighting
  typedef bool (*THighlightCheckFunction) (DrawItem*);
  THighlightCheckFunction m_highlight_func;

protected:
  virtual void ChangeProofEvent( ProofLib::Proof *proof_new,
                                 ProofLib::Proof *proof_old );
  virtual void ResizeProofSheetEvent();
  virtual void ResizeViewPortEvent( const GraphicSize &size );
  virtual void ViewOriginChangedEvent( const GraphicPoint &pt ) = 0;

  void FireViewOriginChanged();

  GraphicRect RequestBuffer( const GraphicRect& viewport_rect );
  DrawBitmap* GetBitmap() { return m_image; }

private:
  void RepaintBuffer();
  void RepaintBuffer( const GraphicRect &cliprect );
  void RepaintItem( DrawItem *item );

  virtual void CalcTransMatrix() = 0;
  void CalcViewOffset();
  virtual void CalcScrollFieldSize() = 0;
};

class GraphicViewContainer
{
public:
  virtual void SetView( GraphicView* view ) { m_view = view; };
  virtual GraphicView *GetView() const { return m_view; };

protected:
  GraphicView           *m_view;
};

class GraphicWindow : public GraphicInputDispatcher, public GraphicViewContainer
{
public:
  GraphicWindow( GraphicSystem &sys );

  virtual GraphicSize GetClientSize() = 0;
  virtual GraphicSystem* GetGraphicSystem() const { return m_sys; }

  virtual void SetView( GraphicView* view );

  virtual void AppendToMenuBar( GraphicTopMenuItem* menuitem ) = 0;
  virtual void ClearTopMenu() = 0;

  virtual void ShowStatusMessage( const char *msg ) = 0;
  virtual void ClearStatusMessage() = 0;

protected:
  GraphicSystem         *m_sys;
};

////////////////////////////////////////////////////////////////////////////////

class GraphicMenu
{
public:
  virtual void Append( GraphicMenuItem *item ) = 0;
  virtual void AppendSeparator() = 0;
  virtual void Popup( size_t button ) = 0;
};

class GraphicMenuItem
{
public:
  virtual void SetSubMenu( GraphicMenu *menu ) = 0;
  virtual void SetCmd( CommandCall *cc ) = 0;
  virtual void SetAccel( int modifier, int key ) = 0;
};

class GraphicTopMenuItem
{
public:
  virtual void SetCmd( CommandCall *cc ) = 0;
};

////////////////////////////////////////////////////////////////////////////////

class GraphicDialog
{
public:
  virtual ~GraphicDialog() {}

  virtual void Do() = 0;

  virtual void SetModal( bool modal ) = 0;
  virtual bool IsModal() const = 0;

  virtual void SetTitle( const USTRING &s ) = 0;
  virtual const UCHAR* GetTitle() const = 0;
};

////////////////////////////////////////////////////////////////////////////////

class GraphicInputDialog : public GraphicDialog
{
public:
  virtual void SetInput( const USTRING &s ) = 0;
  virtual const char* GetInput() const = 0;
};

////////////////////////////////////////////////////////////////////////////////

class GraphicFileDialog : public GraphicDialog
{
public:
  virtual void SetFile( const char* s ) = 0;
  virtual const char* GetFile() const = 0;
};

////////////////////////////////////////////////////////////////////////////////

#endif
