////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/proofsheet.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include "util/myassert.h"
#include "proofsheet.h"

#ifdef DEBUG
#include <iostream>
#endif


////////////////////////////////////////////////////////////////////////////////

DrawItem::DrawItem( DrawItemDatabase *db )
  : m_highlighted( false )
{
  ASSERT( db != NULL );
  m_db = db;
  m_db->Insert( this );
}

void DrawItem::SetHighlight( bool on /* = true */ )
{
  m_highlighted = on; 
}

////////////////////////////////////////////////////////////////////////////////

DrawLineItem::DrawLineItem( DrawItemDatabase *db )
  : DrawItem( db )
{
}

SUNIT DrawLineItem::GetWidth() const
{
  return abs( m_to.m_x - GetOrigin().m_x );
}

SUNIT DrawLineItem::GetHeight() const
{
  return abs( m_to.m_y - GetOrigin().m_y );
}

DrawRect DrawLineItem::GetBoundingRect() const
{
  DrawRect r;
  r.m_leftbottom = GetOrigin();
  r.m_righttop = m_to;

  return r;
}

void DrawLineItem::SetEndPoint( DrawPoint &pt )
{
  m_to = pt;
}

void DrawLineItem::Draw( GraphicContext &gc )
{
  SCOLOR col = gc.GetForegroundColor();

  if( IsHighlighted() )
    gc.SetForegroundColor(
      gc.GetHighlightForegroundColor() );

  gc.Line( GetOrigin(), m_to );

  gc.SetForegroundColor( col );
}

////////////////////////////////////////////////////////////////////////////////

DrawTextItem::DrawTextItem( DrawItemDatabase *db )
  : DrawItem( db )
{
  m_widthcache = 0;
  m_height = 0;
}

SUNIT DrawTextItem::GetWidth() const
{
  if( m_widthcache != 0 )
    return m_widthcache;

  FontManager* fm = (GetDatabase()->GetProofSheet())->GetFontManager();
  ASSERT( fm != NULL );

  m_widthcache = (fm->GetBitmapDatabase()).
    GetWidth( m_text, (FUNIT) GetHeight() );

  return m_widthcache;
}

SUNIT DrawTextItem::GetHeight() const
{
  return m_height;
}

void DrawTextItem::SetHeight( SUNIT height )
{
  if( m_height == height )
    return;

  m_height = height;
  m_widthcache = 0;
}

DrawRect DrawTextItem::GetBoundingRect() const
{
  DrawRect rect;
  rect.m_leftbottom = GetOrigin();
  rect.m_righttop = rect.m_leftbottom;
  rect.m_righttop.m_x += GetWidth();
  rect.m_righttop.m_y += GetHeight();

  return rect;
}

void DrawTextItem::Draw( GraphicContext &gc )
{
  SCOLOR col = gc.GetForegroundColor();

  if( IsHighlighted() )
    gc.SetForegroundColor(
      gc.GetHighlightForegroundColor() );

  if( GetHeight() == 0 )
    return;

  DrawRect text = GetBoundingRect();

  DrawRect t = text;
  t.TransformBy( gc.m_trans );

  gc.Text( text, m_text );

  gc.SetForegroundColor( col );
}

void DrawTextItem::SetText( const USTRING &str )
{
  DataStream stream( m_text );
  stream << str;
  m_widthcache = 0;
}

////////////////////////////////////////////////////////////////////////////////

DrawSequentLineItem::DrawSequentLineItem( DrawItemDatabase *db, DrawSequentItem *parent )
  : DrawLineItem( db )
{
  m_parent = parent;
}

////////////////////////////////////////////////////////////////////////////////

DrawSequentTextItem::DrawSequentTextItem( DrawItemDatabase *db, DrawSequentItem *parent )
  : DrawTextItem( db )
{
  m_parent = parent;
}

////////////////////////////////////////////////////////////////////////////////

DrawLambdaTermTextItem::DrawLambdaTermTextItem( DrawItemDatabase *db, DrawSequentItem *parent )
  : DrawTextItem( db )
{
  m_parent = parent;
}

void DrawLambdaTermTextItem::SetLambdaTerm( const ProofLib::LambdaTerm *lt )
{
  DataStream stream( m_text );
  stream << *lt;
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

DrawSequentItem::DrawSequentItem( DrawItemDatabase *db, DrawSequentItem *parent, bool formula_by_line ) :
  DrawItem( db ), m_formula_by_line( formula_by_line ), m_line_num( 1 )
{
  m_parent = parent;
  m_widthcache = 0;
  m_totalruleswidthcache = 0;
  m_ruleswidthcache = 0;

  m_sequent = NULL;
  m_ruletextitem = NULL;
  m_rulelineitem = NULL;
  m_lambdatextitem = NULL;
}

DrawSequentItem::~DrawSequentItem()
{
  if( m_sequent != NULL )
    m_sequent->DetachSequentEventClient( this );

  std::vector<DataString*>::iterator i;
  for ( i = m_lines.begin(); i != m_lines.end(); i++ )
  {
    delete *i;
  }
}


SUNIT DrawSequentItem::GetHeight() const
{
  return 40 * m_line_num;
}


void DrawSequentItem::SetFromRuleItem( ProofLib::RuleItem *i )
{
  ProofLib::Rule *r;
  ProofLib::Sequent *s;
  ProofLib::ProofLink *pl;

  if( (r = dynamic_cast<ProofLib::Rule*>( i )) != NULL )
  {
    // attach sequent itself
    m_sequent = r->GetConclusion();

    // add rule text
    if( !r->GetName().empty() )
    {
      m_ruletextitem = new DrawSequentTextItem( GetDatabase(), this );
      m_ruletextitem->SetText( r->GetName() );
      m_ruletextitem->SetHeight( GetRuleTextHeight() );
    }
 
    const ProofLib::LambdaTerm *lt = r->GetLambdaTerm();
    if ( lt != NULL )
    {
      m_lambdatextitem = new DrawLambdaTermTextItem( GetDatabase(), this );
      m_lambdatextitem->SetLambdaTerm( lt );
      m_lambdatextitem->SetHeight( GetRuleTextHeight() );
    }

    m_rulelineitem = new DrawSequentLineItem( GetDatabase(), this );

    // add above sequents recursivly
    for( size_t i = 0; i < r->GetRuleItemCount(); i++ )
    {
      DrawSequentItem *next = GetDatabase()->NewSequent( this, m_formula_by_line );
      m_followups.push_back( next );
      next->SetFromRuleItem( r->GetRuleItem( i ) );
    }
  }
  else if( (s = dynamic_cast<ProofLib::Sequent*>( i )) != NULL )
  {
    m_sequent = s;
  }
  else if( (pl = dynamic_cast<ProofLib::ProofLink*>( i )) != NULL )
  {
    m_sequent = pl->GetSequent();

    m_rulelineitem = new DrawSequentLineItem( GetDatabase(), this );

    DrawToProofLinkItem *next = GetDatabase()->NewToProofLink( this );
    m_followups.push_back( next );
    next->SetFromProofLink( pl );
  }
  else
  {
    ASSERT( false );
    return;
  }

  m_sequent->AttachSequentEventClient( this );

  ASSERT( m_sequent != NULL );
  DataStream stream( m_datastring );

  if ( !m_formula_by_line )
    stream << *m_sequent;
  else
  { // rearrange formulas line-by-line
    ProofLib::SElementList *l = m_sequent->GetLeft();
    ProofLib::SElementList *r = m_sequent->GetRight();

    if( l != NULL && l->GetChildCount() > 0 )
    {
      stream.GetNodePointer()->Push(0);
      for( size_t i = 0; i < l->GetChildCount(); i++ )
      {
	if( i > 0 )
	{
	  stream << CHAR(',');
          stream << CHAR_LINEFEED;
	  m_line_num++;
	}

	ProofLib::SElement *x = l->GetSElement( i );

	ProofLib::SFormula *f = dynamic_cast<ProofLib::SFormula*>( x );

	if( f != NULL )
	{
	  stream.m_formulastart = true;
	  stream.GetNodePointer()->Push(i);
	  stream.GetNodePointer()->Push(0);
	  stream << *(f->GetFormula());
	  stream.GetNodePointer()->Pop();
	  stream.GetNodePointer()->Pop();
	  continue;
	}

	ProofLib::SFormulaListVariable *fv =
	  dynamic_cast<ProofLib::SFormulaListVariable*>( x );

	if( fv != NULL )
	{
	  stream.GetNodePointer()->Push(i);
	  stream.GetNodePointer()->Push(0);
	  stream << *(fv->GetFormulaListVariable());
	  stream.GetNodePointer()->Pop();
	  stream.GetNodePointer()->Pop();
	  continue;
	}

	ASSERT( false );
      }

      stream.GetNodePointer()->Pop();
      stream << CHAR_LINEFEED;
      m_line_num++;
    }

    stream << CHAR_SEQUENT;

    if( r != NULL && r->GetChildCount() > 0 )
    {
      stream << CHAR_LINEFEED;
      m_line_num++;
      stream.GetNodePointer()->Push(1);
      for( size_t i = 0; i < r->GetChildCount(); i++ )
      {
	if( i > 0 )
	{
	  stream << CHAR(',');
          stream << CHAR_LINEFEED;
          m_line_num++;
	}

	ProofLib::SElement *x = r->GetSElement( i );

	ProofLib::SFormula *f = dynamic_cast<ProofLib::SFormula*>( x );

	if( f != NULL )
	{
	  stream.m_formulastart = true;
	  stream.GetNodePointer()->Push(i);
	  stream.GetNodePointer()->Push(0);
	  stream << *(f->GetFormula());
	  stream.GetNodePointer()->Pop();
	  stream.GetNodePointer()->Pop();
	  continue;
	}

	ProofLib::SFormulaListVariable *fv =
	  dynamic_cast<ProofLib::SFormulaListVariable*>( x );

	if( fv != NULL )
	{
	  stream.GetNodePointer()->Push(i);
	  stream.GetNodePointer()->Push(0);
	  stream << *(fv->GetFormulaListVariable());
	  stream.GetNodePointer()->Pop();
	  stream.GetNodePointer()->Pop();
	  continue;
	}

	ASSERT( false );
      }
      stream.GetNodePointer()->Pop();
    }
  }

#if 0
  cout << GetWidth() << ", ";
  cout << GetRuleWidth() << ": ";
  m_datastring.Dump( cout );
#endif
}

void DrawSequentItem::OnEvent( MouseMoveEvent *e )
{
#ifdef DEBUG
  cout << "oleole" << endl;
#endif
}

SUNIT DrawSequentItem::GetWidth() const
{
  return GetTotalRuleWidth();
}

SUNIT DrawSequentItem::GetRuleWidth() const
{
  return std::max( GetRulesWidth(), GetTextWidth() );
}

SUNIT DrawSequentItem::GetTotalRuleWidth() const
{
  return std::max( GetTotalRulesWidth(),
    GetRuleWidth() + GetRuleTextGap() + GetRuleTextWidth() );
}

SUNIT DrawSequentItem::GetRulesWidth() const
{
  if( m_followups.size() == 0 )
    return 0;

  if( m_ruleswidthcache != 0 )
  {
#if 0
m_datastring.Dump( std::cout );
std::cout << ": GetRulesWidth from cache." << std::endl;
#endif
    return m_ruleswidthcache;
  }

#if 0
m_datastring.Dump( std::cout );
std::cout << ": GetRulesWidth CALC." << std::endl;
#endif

  CalcRulesWidthCache();

  return m_ruleswidthcache;
}

SUNIT DrawSequentItem::GetTotalRulesWidth() const
{
  if( m_followups.size() == 0 )
    return 0;

  if( m_totalruleswidthcache != 0 )
  {
#if 0
m_datastring.Dump( std::cout );
std::cout << ": GetTotalRulesWidth from cache." << std::endl;
#endif
    return m_totalruleswidthcache;
  }

#if 0
m_datastring.Dump( std::cout );
std::cout << ": GetTotalRulesWidth CALC." << std::endl;
#endif

  CalcRulesWidthCache();

  return m_totalruleswidthcache;
}


std::vector<DataString*> DrawSequentItem::GetStringByLines() const
{
  if ( !m_lines.empty() )
    return m_lines;

  DataString *line = new DataString();

  for ( size_t i = 0; i < m_datastring.GetLength(); i++ )
  {
    DataChar c = m_datastring[i];
    if ( c.isLineFeed() )
    {
      m_lines.push_back(line);
      line = new DataString();
    }
    else
    {
      (*line) += c;
    }
  }
  m_lines.push_back(line);

  return m_lines;
}

void DrawSequentItem::CalcRulesWidthCache() const
{
  if( m_followups.size() == 0 )
    return;

  SUNIT rules_width = 0;
  SUNIT total_rules_width = 0;

  DrawSequentItem::TSequentList::const_iterator k,l;

  for( k = m_followups.begin(); k != m_followups.end(); k++ )
  {
    // calculate rules_width
    if( k != m_followups.begin() )
      rules_width += GetRuleGap();

    l = k;
  
    if( ++l == m_followups.end() )
      rules_width += (*k)->GetRuleWidth(); // ignore rule text for last rule
    else
      rules_width += (*k)->GetWidth();

    // calculate total_rules_width
    if( k != m_followups.begin() )
      total_rules_width += GetRuleGap();

    total_rules_width += (*k)->GetWidth();
  }

  m_ruleswidthcache = rules_width;
  m_totalruleswidthcache = total_rules_width;
}

SUNIT DrawSequentItem::GetTextWidth() const
{
  if( m_widthcache != 0 )
  {
#if 0
m_datastring.Dump( std::cout );
std::cout << ": GetTextWidth from cache." << std::endl;
#endif
    return m_widthcache;
  }

#if 0
m_datastring.Dump( std::cout );
std::cout << ": GetTextWidth CALC." << std::endl;
#endif

  FontManager* fm = (GetDatabase()->GetProofSheet())->GetFontManager();
  ASSERT( fm != NULL );

  if ( !m_formula_by_line )
  {
    m_widthcache = (fm->GetBitmapDatabase()).
      GetWidth( m_datastring, (FUNIT) GetTextHeight() );
  }
  else
  {
    std::vector<DataString*> lines = GetStringByLines();
    std::vector<DataString*>::iterator i;
    m_widthcache = 0;
    for ( i = lines.begin(); i != lines.end(); i++ )
    {
      SUNIT width = (fm->GetBitmapDatabase()).
            GetWidth( **i, (FUNIT) GetTextHeight() );
      if ( width > m_widthcache )
      {
        m_widthcache = width;
      }
    }
  }

  return m_widthcache;
}

SUNIT DrawSequentItem::GetRuleTextWidth() const
{
  if( m_ruletextitem == NULL )
    return 0;

  if( m_lambdatextitem == NULL )
  {
    return m_ruletextitem->GetWidth();
  }
  else
  {
    return m_ruletextitem->GetWidth() + m_lambdatextitem->GetWidth();
  }
}

SUNIT DrawSequentItem::GetRuleHeight() const
{
  ASSERT( false );
  return 30;
}

DrawRect DrawSequentItem::GetBoundingRect() const
{
  DrawRect r;

  r.m_leftbottom = GetTextOrigin();
  r.m_righttop = r.m_leftbottom;

  //std::cerr << "DrawSequentItem::GetBoundingRect, before, r = ";
  //r.Dump( std::cerr );
  //std::cerr << std::endl;

  r.m_righttop.m_y += GetTextHeight() * m_line_num;
  r.m_righttop.m_x += GetTextWidth();

  //std::cerr << "DrawSequentItem::GetBoundingRect, after, r = ";
  //r.Dump( std::cerr );
  //std::cerr << std::endl;

  return r;
}

void DrawSequentItem::Layout()
{
#ifdef DEBUG
#if 0
  cout << GetWidth() << ", ";
  //cout << GetRuleWidth() << ": ";
  m_datastring.Dump( cout );
#endif
#endif

  DrawPoint pt = GetOrigin();

  pt.m_y += GetHeight();

  // std::cerr << "height in layout: " << GetHeight() << endl;

  SUNIT rulewidth = GetRuleWidth();
  SUNIT ruleswidth = GetRulesWidth();

  if( rulewidth > ruleswidth )
    pt.m_x += ( rulewidth - ruleswidth ) / 2;

  DrawSequentItem::TSequentList::const_iterator k;
  for( k = m_followups.begin(); k != m_followups.end(); k++ )
  {
    if( k != m_followups.begin() )
      pt.m_x += GetRuleGap();

    SUNIT width = (*k)->GetWidth();
    
    (*k)->SetOrigin( pt );
    (*k)->Layout();

    pt.m_x += width;
  }

  // place rule line
  if( m_rulelineitem )
  {
    ASSERT( !m_followups.empty() );
    DrawSequentItem *l = *(m_followups.begin());
    DrawSequentItem *r = *(m_followups.rbegin());

    ASSERT( l != NULL && r != NULL );

    SUNIT lx = l->GetTextOrigin().m_x;
    SUNIT rx = r->GetTextOrigin().m_x + r->GetTextWidth();

    DrawPoint pt1, pt2;

    pt1 = GetOrigin();
    pt1.m_y += GetLineY();
    pt1.m_x = GetTextOrigin().m_x;
    pt2 = pt1;
    pt2.m_x += GetTextWidth();

    pt1.m_x = std::min( lx, pt1.m_x );
    pt2.m_x = std::max( rx, pt2.m_x );
    
    m_rulelineitem->SetOrigin( pt1 );
    m_rulelineitem->SetEndPoint( pt2 );

    // place rule text
    if( m_ruletextitem != NULL )
    {
      pt2.m_x += GetRuleTextGap();
      pt2.m_y -= GetRuleTextHeight() / 2;
      m_ruletextitem->SetOrigin( pt2 );
    }

    // place lambda term
    if( m_lambdatextitem != NULL )
    {
      pt2.m_x += m_ruletextitem->GetWidth() + GetRuleTextGap();
      m_lambdatextitem->SetOrigin( pt2 );
    }
  }
}

DrawPoint DrawSequentItem::GetTextOrigin() const
{
  DrawPoint pt = GetOrigin();

  SUNIT textadj = GetRuleWidth() - GetTextWidth();
  textadj /= 2;
  pt.m_x += textadj;
  pt.m_y += GetTextY();

  return pt;
}

void DrawSequentItem::Draw( GraphicContext &gc )
{
#if 0
  fprintf( stdout, "sequent at %f, %f.\n", GetOrigin().m_x, GetOrigin().m_y );
#endif
#if 0
  gc.SetColor( SCOLOR_RED );
  gc.Cross( GetOrigin() );
  gc.SetColor( SCOLOR_BLACK );
#endif

  DrawRect text = GetBoundingRect();

#if 0
  gc.SetColor( SCOLOR_RED );
  gc.Rect( text );
  gc.SetColor( SCOLOR_BLACK );
#endif

  //DrawRect box = GetBoundingRect();
  //DrawRect line;
  //DrawRect line = box;
  DrawRect t = text;

#if 0
  DrawRect line2;
  line2.m_leftbottom = GetOrigin();
  line2.m_righttop = line2.m_leftbottom;
  line2.m_righttop.m_x += GetTotalRuleWidth();
  gc.Line( line2, SCOLOR_RED );
#endif

#if 0
  text.m_leftbottom = GetTextOrigin();
  text.m_righttop = text.m_leftbottom;
  text.m_righttop.m_x += GetTextWidth();
  text.m_righttop.m_y += GetTextHeight();
#endif

  t = text;

#if 0
  std::cerr << "DrawSequentItem::Draw, t = ";
  t.Dump( std::cerr );
  std::cerr << std::endl;
#endif

  t.TransformBy( gc.m_trans );

#if 0
  std::cerr << "after transform t = ";
  t.Dump( std::cerr );
  std::cerr << std::endl;
#endif

  if( IsHighlighted() )
    gc.Text( text, m_datastring, &m_highlight_part );
  else
    gc.Text( text, m_datastring );

#if 0
  if( !m_followups.empty() )
  {
#if 0
    DrawSequentItem *l = *(m_followups.begin());
    DrawSequentItem *r = *(m_followups.rbegin());

    SUNIT lx = l->GetTextOrigin().m_x;
    SUNIT rx = r->GetTextOrigin().m_x + r->GetTextWidth();

    line = text;
    line.m_leftbottom.m_y = line.m_righttop.m_y = GetOrigin().m_y + GetLineY();
    line.m_leftbottom.m_x = min( lx, line.m_leftbottom.m_x );
    line.m_righttop.m_x   = max( rx, line.m_righttop.m_x );


    t = line;
    t.TransformBy( gc.m_trans );

    if( gc.m_cliprect.Overlap( t ) ) // TODO: clipping one lever higer
      gc.Line( line );
#endif

#if 0
    DrawRect rule;
    rule.m_leftbottom = line.m_righttop;
    rule.m_leftbottom.m_x += GetRuleTextGap();
    rule.m_leftbottom.m_y -= GetRuleTextHeight() / 2;
    rule.m_righttop = rule.m_leftbottom;
    rule.m_righttop.m_x += GetRuleTextWidth();
    rule.m_righttop.m_y += GetRuleTextHeight();

    t = rule;
    t.TransformBy( gc.m_trans );

    if( gc.m_cliprect.Overlap( t ) ) // TODO: clipping one lever higer
      gc.Text( rule, m_datastring_rulename );
#endif
  }
#endif
}

DataChar DrawSequentItem::GetCharAt( const DrawPoint &pt ) const
{
#if 0
pt.Dump( cout );
GetTextOrigin().Dump(cout);
#endif
  std::cerr << "DrawSequentItem::GetCharAt" << std::endl;

  // calculate on which line pt lies
  SUNIT start_y = m_datastring[0].m_y;
  size_t line = 0;

  for ( line = 0; start_y + line * GetTextHeight() < pt.m_y; line++ );
  line--;

  // cerr << "point lies in line " << line << endl;

  size_t i;
  size_t linefeeds = 0;

  for( i = 0; i < m_datastring.GetLength() - 1; i++ )
  {
#if 0
cout << i << ": " << m_datastring[i].m_start << ", "; 
cout << endl;
cout << "if( " << m_datastring[i].m_start << " > "
     << pt.m_x 
     // << " - " <<  GetTextOrigin().m_x
     // << " ( = " << pt.m_x - GetTextOrigin().m_x << " )";
     << " )";
cout << endl;
#endif
    if ( m_datastring[i].isLineFeed() )
    {
      linefeeds++;
      continue;
    }

    if ( line != linefeeds )
      continue;
    
    if( m_datastring[i].m_start <= pt.m_x &&
        m_datastring[i].m_end >= pt.m_x ) // - GetTextOrigin().m_x )
      break;
  }

//  ASSERT( i > 0 );
#if 0
  m_datastring[i-1].Dump( cout );
  cout << endl;
#endif
//  std::cerr << "i: " << i << endl;

  return m_datastring[i];
}

size_t DrawSequentItem::GetSElementIndexAt( const DrawPoint &pt ) const
{
  ASSERT( false );
  return 0;
#if 0
  size_t f_no = 0;
  size_t i;
#if 0
  DrawPoint o = GetTextOrigin();

cerr << "pt = ";
pt.Dump( cerr );
cerr << endl;
cerr << "o = ";
o.Dump( cerr );
cerr << endl;
#endif

  for( i = 0; i < m_datastring.GetLength(); i++ )
  {
//cerr << m_datastring[i].m_start << ">" << pt.m_x << "-" << o.m_x << endl;
    if( m_datastring[i].m_start > pt.m_x )
      break;

    if( m_datastring[i].m_seq_sep )
      f_no++;
  }

  if( m_datastring[i-1].m_seq_sep )
    return 0; // index 0 is no formula

  return f_no + 1;
#endif
}

ProofLib::SElement* DrawSequentItem::GetSElement( size_t index ) const
{
  if( index == 0 )
    return NULL; // index 0 is no formula

  index--;

  if( index >= m_sequent->GetLeft()->GetChildCount() )
  {
    return m_sequent->GetRight()->GetSElement(
      index - m_sequent->GetLeft()->GetChildCount() );
  }

  return m_sequent->GetLeft()->GetSElement( index );
}

ProofLib::SElement* DrawSequentItem::GetSElementAt( const DrawPoint &pt ) const
{
  return GetSElement( GetSElementIndexAt( pt ) );
}


void DrawSequentItem::SetHighlightPart( const ProofLib::NodePointer &np )
{
  m_highlight_part = np;
}

void DrawSequentItem::ClearHighlightPart()
{
  m_highlight_part.Clear();
}

void DrawSequentItem::SequentUpdateEvent( const ProofLib::Sequent *sequent )
{
  if( sequent != m_sequent )
  {
#ifdef DEBUG
std::cout << "SequentUpdateEvent: WRONG SEQUENT: " << (void*) sequent;
std::cout << " should " << (void*) m_sequent << std::endl;
#endif
    return;
  }

  m_highlight_part.Clear();

  m_datastring.Clear();

  DataStream stream( m_datastring );
  stream << *m_sequent;

  InvalidTextSize();
}

void DrawSequentItem::InvalidTextSize()
{
  m_widthcache = 0;

  if( m_parent != NULL )
    m_parent->InvalidFromSuccessor();
}

void DrawSequentItem::InvalidFromSuccessor()
{
  if( m_totalruleswidthcache == 0 && m_ruleswidthcache == 0 )
    return;

  m_totalruleswidthcache = 0;
  m_ruleswidthcache = 0;

  if( m_parent != NULL )
    m_parent->InvalidFromSuccessor();
}

#if 0
DrawSequentItem* DrawSequentItem::SequentWithMarkedFormula( DrawItem* item )
{
  if( item == NULL )
    return NULL;

  DrawSequentItem *i = dynamic_cast<DrawSequentItem*>( item );

  if( i == NULL )
    return NULL;

  ProofLib::NodePointer np = i->GetHighlightPart();
  ProofLib::Sequent *s = i->GetSequent();

  while( !np.IsEmpty() &&
         dynamic_cast<ProofLib::Formula*>( np.GetDataNode( s ) ) == NULL )
  {
    np.Pop();
  } 

  if( dynamic_cast<ProofLib::Formula*>( np.GetDataNode( s ) ) == NULL )
    // no formula selected
    return NULL;

  i->SetHighlightPart( np );

  return i;
}

DrawSequentItem* DrawSequentItem::SequentWithMarkedSElement( DrawItem* item )
{
  if( item == NULL )
    return NULL;

  DrawSequentItem *i = dynamic_cast<DrawSequentItem*>( item );

  if( i == NULL )
    return NULL;

  ProofLib::NodePointer np = i->GetHighlightPart();
  ProofLib::Sequent *s = i->GetSequent();

  while( !np.IsEmpty() &&
         dynamic_cast<ProofLib::SElement*>( np.GetDataNode( s ) ) == NULL )
  {
    np.Pop();
  } 

  if( dynamic_cast<ProofLib::SElement*>( np.GetDataNode( s ) ) == NULL )
    // no formula selected
    return NULL;

  i->SetHighlightPart( np );

  return i;
}

DrawSequentItem* DrawSequentItem::SequentWithMarkedSFormulaListVar( DrawItem* item )
{
  if( item == NULL )
    return NULL;

  DrawSequentItem *i = dynamic_cast<DrawSequentItem*>( item );

  if( i == NULL )
    return NULL;

  ProofLib::NodePointer np = i->GetHighlightPart();
  ProofLib::Sequent *s = i->GetSequent();

  while( !np.IsEmpty() &&
         dynamic_cast<ProofLib::SFormulaListVariable*>( np.GetDataNode( s ) )
            == NULL )
  {
    np.Pop();
  } 

  if( dynamic_cast<ProofLib::SFormulaListVariable*>( np.GetDataNode( s ) )
      == NULL )
    // no formula selected
    return NULL;

  i->SetHighlightPart( np );

  return i;
}

DrawSequentItem* DrawSequentItem::SequentWithMarkedFormulaVar( DrawItem* item )
{
  if( item == NULL )
    return NULL;

  DrawSequentItem *i = dynamic_cast<DrawSequentItem*>( item );

  if( i == NULL )
    return NULL;

  ProofLib::NodePointer np = i->GetHighlightPart();
  ProofLib::Sequent *s = i->GetSequent();

#if 0
  while( !np.IsEmpty() &&
         dynamic_cast<ProofLib::Formula*>( np.GetDataNode( s ) ) == NULL )
  {
    np.Pop();
  } 
#endif

  if( dynamic_cast<ProofLib::FormulaVariable*>( np.GetDataNode( s ) ) == NULL )
    // no formula selected
    return NULL;

  i->SetHighlightPart( np );

  return i;
}

DrawSequentItem* DrawSequentItem::SequentWithMarkedSequent( DrawItem* item )
{
  if( item == NULL )
    return NULL;

  DrawSequentItem *i = dynamic_cast<DrawSequentItem*>( item );

  if( i == NULL )
    return NULL;

  i->SetHighlightPart( ProofLib::NodePointer() );

  return i;
}
#endif

////////////////////////////////////////////////////////////////////////////////

DrawToProofLinkItem::DrawToProofLinkItem( DrawItemDatabase *db, DrawSequentItem *parent )
  : DrawProofLinkItem( db, parent )
{
  m_pl = NULL;
}

void DrawToProofLinkItem::SetFromProofLink( ProofLib::ProofLink *pl )
{
  m_pl = pl;
  ASSERT( m_pl != NULL );

  DataStream stream( m_datastring );
  stream << *m_pl;

#if 0
  cout << GetWidth() << ", ";
  cout << GetRuleWidth() << ": ";
  m_datastring.Dump( cout );
#endif
}

////////////////////////////////////////////////////////////////////////////////

DrawFromProofLinkItem::DrawFromProofLinkItem( DrawItemDatabase *db )
  : DrawProofLinkItem( db, NULL )
{
  m_from = NULL;
}

void DrawFromProofLinkItem::SetFromProof( ProofLib::Proof *p )
{
  m_from = p->GetParent();
  ASSERT( m_from != NULL );

  DataStream stream( m_datastring );
  stream << GetProofName();

  m_rulelineitem = new DrawSequentLineItem( GetDatabase(), this );

  DrawSequentItem *root = GetDatabase()->NewSequent( this );
  m_followups.push_back( root );
  root->SetFromRuleItem( p->GetEndRule() );
}

////////////////////////////////////////////////////////////////////////////////

DrawItemDatabase::DrawItemDatabase( ProofSheet *sheet )
{
  ASSERT( sheet != NULL );

  m_proofsheet = sheet;
}

DrawItemDatabase::~DrawItemDatabase()
{
  for( iterator i = begin(); i != end(); i++ )
    delete *i;

  clear();
}

void DrawItemDatabase::Layout()
{
  for( iterator i = begin(); i != end(); i++ )
    (*i)->Layout();
}

void DrawItemDatabase::Draw( GraphicContext &gc )
{
  DrawRect clip = gc.GetClippingRect();

  for( iterator i = begin(); i != end(); i++ )
  {
    DrawRect bb = (*i)->GetBoundingRect();
    bb.TransformBy( gc.m_trans );

    if( !clip.Overlap( bb ) )
      continue;

    (*i)->Draw( gc );
  }
}

void DrawItemDatabase::LoadProof( ProofLib::Proof *p, bool sequents_by_line )
{
  if( p->GetEndRule() == NULL )
    return;

  if( p->GetParent() != NULL )
  {
    DrawFromProofLinkItem *root = NewFromProofLink();
    m_proofsheet->SetRoot( root );
    root->SetFromProof( p );
  }
  else
  {
    DrawSequentItem *root = NewSequent( NULL, sequents_by_line );
    m_proofsheet->SetRoot( root );
    root->SetFromRuleItem( p->GetEndRule() );
  }
}

DrawSequentItem* DrawItemDatabase::NewSequent( DrawSequentItem *parent, bool formula_by_line )
{
  DrawSequentItem *i = new DrawSequentItem( this, parent, formula_by_line );
//printf( "size: %ld\n", size() );
  return i;
}
    
DrawToProofLinkItem* DrawItemDatabase::NewToProofLink( DrawSequentItem *parent )
{
  DrawToProofLinkItem *i = new DrawToProofLinkItem( this, parent );
//printf( "size: %ld\n", size() );
  return i;
}
    
DrawFromProofLinkItem* DrawItemDatabase::NewFromProofLink()
{
  DrawFromProofLinkItem *i = new DrawFromProofLinkItem( this );
//printf( "size: %ld\n", size() );
  return i;
}
    
void DrawItemDatabase::Insert( DrawItem *i )
{
  push_back( i );
  Attach( i );
}

bool DrawItemDatabase::Contains( const DrawItem *item ) const
{
  for( const_iterator i = begin(); i != end(); i++ )
    if( (*i) == item )
      return true;

  return false;
}

DrawRect DrawItemDatabase::GetBoundingRect() const
{
  DrawRect r;

  for( const_iterator i = begin(); i != end(); i++ )
    r.Capture( (*i)->GetBoundingRect() );

  //std::cout << "DrawItemDatabase::GetBoundingRect: ";
  //r.Dump( cout ); cout << endl;

  return r;
}

DrawItem *DrawItemDatabase::GetItemAtPoint( DrawPoint &pt ) const
{
#if 0
  std::cerr << "GetItemAtPoint: ";
  pt.Dump( std::cerr );
  std::cerr << std::endl;
#endif

  for( const_iterator i = begin(); i != end(); i++ )
    if( (*i)->GetBoundingRect().Contains( pt ) )
      return *i;

  return NULL;
}

////////////////////////////////////////////////////////////////////////////////

ProofSheet::ProofSheet()
{
  m_proofdb   = NULL;
  m_proof     = NULL;
  m_items     = NULL;
  m_rootseq   = NULL;
  m_fontmgr   = NULL;
}

ProofSheet::ProofSheet( ProofLib::ProofDatabase *db )
{
  m_proofdb   = db;
  m_proof     = NULL;
  m_items     = NULL;
  m_rootseq   = NULL;
  m_fontmgr   = NULL;
}

void ProofSheet::AttachNotificant( ProofSheetNotificant *p )
{
  m_notify_list.insert( p );
}

void ProofSheet::DetachNotificant( ProofSheetNotificant *p )
{
  ProofSheetNotificantList::iterator i = m_notify_list.find( p );

  if( i == m_notify_list.end() )
    return;

  m_notify_list.erase( i );
}


void ProofSheet::Clear()
{
  delete m_items;
  m_items = NULL;

  ProofLib::Proof *proof_old = m_proof;

  m_proof = NULL;

  m_notify_list.FireNewProof( m_proof, proof_old );
}

void ProofSheet::SetFontManager( FontManager *f )
{
  m_fontmgr = f;
}

void ProofSheet::SetRoot( DrawSequentItem *s )
{
  m_rootseq = s;
}

ProofLib::Proof* ProofSheet::SelectSequentList( const USTRING &name )
{
  ProofLib::Proof *proof_old = m_proof;

  if( m_proof != NULL )
    Clear();

  if( m_proofdb->GetSequentListCount() == 0 )
    return NULL;

  ProofLib::SequentList* l = m_proofdb->GetSequentList( name );

  if( l == NULL )
    return NULL;

  if ( !l->empty() ) // convert list into "proof" so that it can be displayed using existing code
  {
    m_proof = m_proofdb->NewAnonymousProof();
    ASSERT( m_proof != NULL );
    ProofLib::RuleItem *r = (*l)[0];
    for ( size_t i = 1; i < l->size(); i++ )
    {
      ProofLib::Rule *n_r = new ProofLib::Rule();
      n_r->AddLeft( r );
      n_r->SetConclusion( (*l)[i] );
      r = n_r;
    }
    m_proof->SetEndRule( r );
  }

#ifdef DEBUG
std::cout << "START FILLING DRAW ITEM DATABASE..." << std::endl;
#endif
  InitDatabase( true );
#ifdef DEBUG
std::cout << "DRAW ITEM DATABASE FILLED" << std::endl;
#endif
  Layout();
#ifdef DEBUG
std::cout << "DRAW ITEMS LAYOUTED" << std::endl;
#endif
  m_notify_list.FireNewProof( m_proof, proof_old );

  return m_proof;

return NULL;
}

ProofLib::Proof* ProofSheet::SelectProof( const USTRING &name )
{
  ProofLib::Proof *proof_old = m_proof;

  if( m_proof != NULL )
    Clear();

  if( m_proofdb->GetCount() == 0 )
    return NULL;

  if( name.empty() )
    m_proof = m_proofdb->GetDefaultProof();
  else
  {
    ProofLib::Proof* p = m_proofdb->GetProof( name );

    if( p == NULL )
      return NULL;

    m_proof = p;
  }

#ifdef DEBUG
std::cout << "START FILLING DRAW ITEM DATABASE..." << std::endl;
#endif
  InitDatabase();
#ifdef DEBUG
std::cout << "DRAW ITEM DATABASE FILLED" << std::endl;
#endif
  Layout();
#ifdef DEBUG
std::cout << "DRAW ITEMS LAYOUTED" << std::endl;
#endif
  m_notify_list.FireNewProof( m_proof, proof_old );

  return m_proof;
}

void ProofSheet::InitDatabase( bool sequents_by_line )
{
  m_items = new DrawItemDatabase( this );

  m_items->LoadProof( m_proof, sequents_by_line );

  Attach( m_items );
}

void ProofSheet::Layout()
{
  if( m_rootseq == NULL )
    return;

  m_rootseq->Layout();
  m_notify_list.FireResize();
}

void ProofSheet::Draw( GraphicContext &gc )
{
  if( m_items == NULL )
    return;

  m_items->Draw( gc );
}

/*
SUNIT ProofSheet::GetWidth()
{
  return m_rootseq->GetRuleWidth();
}

SUNIT ProofSheet::GetHeight()
{
  return m_rootseq->GetRuleHeight();
}
*/

DrawRect ProofSheet::GetBoundingRect() const
{
  DrawRect r;

  if( m_items == NULL )
    return r;

  return m_items->GetBoundingRect();
}

DrawItem* ProofSheet::GetItemAtPoint( DrawPoint &pt ) const
{
  if( m_items == NULL )
    return NULL;

  return m_items->GetItemAtPoint( pt );
}

bool ProofSheet::Contains( const DrawItem *i ) const
{
  if( m_items == NULL )
    return false;

  return m_items->Contains( i );
}
