////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/proofsheet.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#ifndef _PROOFSHEET_H_
#define _PROOFSHEET_H_

#include <vector>

#include "libproof.h"
#include "graphiccontext.h"

#include "fontmanager.h"

class ProofSheet;
class ProofSheetNotificant;
class DrawItemDatabase;
class DrawSequentItem;

////////////////////////////////////////////////////////////////////////////////

class DrawItem : public GraphicInputListener
{
private:
  DrawItemDatabase      *m_db;
  DrawPoint              m_origin;
  bool                   m_highlighted;

public:
  DrawItem( DrawItemDatabase* db );
  virtual ~DrawItem() {};

  template<class T> static bool IsA( DrawItem *i );

  virtual SUNIT GetWidth() const = 0;
  virtual SUNIT GetHeight() const = 0;

  virtual const DrawPoint& GetOrigin() const { return m_origin; };
  virtual void SetOrigin( DrawPoint& pt )
    { //printf( "origin: %f,%f\n", pt.m_x, pt.m_y),
      m_origin = pt; };
  virtual DrawRect GetBoundingRect() const = 0;

  virtual void Draw( GraphicContext &gc ) = 0;
  virtual void Layout() = 0;
  
  DrawItemDatabase *GetDatabase() const { return m_db; };

  void SetHighlight( bool on = true );
  bool IsHighlighted() const { return m_highlighted; }
};


template<class T> bool DrawItem::IsA( DrawItem *i )
{
  if( i == NULL )
    return false;

  return ( dynamic_cast<T*>( i ) != NULL );
}


////////////////////////////////////////////////////////////////////////////////

class DrawLineItem : public DrawItem
{
public:
  DrawLineItem( DrawItemDatabase *db );

  virtual SUNIT GetWidth() const;
  virtual SUNIT GetHeight() const;

  virtual DrawRect GetBoundingRect() const;

  virtual void Draw( GraphicContext &gc );
  virtual void Layout() {};

public:
  virtual void SetEndPoint( DrawPoint &pt );

protected:
  DrawPoint              m_to;
};

////////////////////////////////////////////////////////////////////////////////

class DrawTextItem : public DrawItem
{
public:
  DrawTextItem( DrawItemDatabase *db );

  virtual SUNIT GetWidth() const;
  virtual SUNIT GetHeight() const;

  virtual DrawRect GetBoundingRect() const;

  virtual void Draw( GraphicContext &gc );
  virtual void Layout() {};

public:
  virtual void SetText( const USTRING &str );
  virtual void SetHeight( SUNIT height );

protected:
  DataString             m_text;
  SUNIT                  m_height;

protected:
  mutable SUNIT          m_widthcache;
};

////////////////////////////////////////////////////////////////////////////////

class DrawSequentLineItem : public DrawLineItem
{
public:
  DrawSequentLineItem( DrawItemDatabase *db, DrawSequentItem *parent );

protected:
  DrawSequentItem       *m_parent;
};

////////////////////////////////////////////////////////////////////////////////

class DrawSequentTextItem : public DrawTextItem
{
public:
  DrawSequentTextItem( DrawItemDatabase *db, DrawSequentItem *parent );

protected:
  DrawSequentItem       *m_parent;
};

////////////////////////////////////////////////////////////////////////////////

class DrawLambdaTermTextItem : public DrawTextItem
{
public:
  DrawLambdaTermTextItem( DrawItemDatabase *db, DrawSequentItem *parent );

  void SetLambdaTerm( const ProofLib::LambdaTerm *lt );

protected:
  DrawSequentItem      *m_parent;
};

////////////////////////////////////////////////////////////////////////////////

/// \todo better from text item than from draw item??
class DrawSequentItem : public DrawItem,
                        protected ProofLib::SequentEventClient
{
public:
  typedef std::vector<DrawSequentItem*> TSequentList;

protected:
  DrawSequentTextItem    *m_ruletextitem;
  DrawSequentLineItem    *m_rulelineitem;
  DrawLambdaTermTextItem *m_lambdatextitem;

  ProofLib::Sequent      *m_sequent;
  DataString              m_datastring;
  ProofLib::NodePointer   m_highlight_part;

  DrawSequentItem        *m_parent;
  TSequentList            m_followups;

  bool			   m_formula_by_line; // draw formulas "line-by-line"
  unsigned int		   m_line_num;        // how many lines
  mutable std::vector<DataString*> m_lines;

  // caches
  mutable SUNIT           m_widthcache;
  mutable SUNIT           m_totalruleswidthcache;
  mutable SUNIT           m_ruleswidthcache;

public:
  DrawSequentItem( DrawItemDatabase *db, DrawSequentItem *parent, bool formula_by_line = false );
  virtual ~DrawSequentItem();

  void SetFromRuleItem( ProofLib::RuleItem *i );

protected:
  virtual SUNIT GetRuleHeight() const;

  virtual SUNIT GetTotalRulesWidth() const;
  virtual SUNIT GetRulesWidth() const;
  virtual SUNIT GetTotalRuleWidth() const;
  virtual SUNIT GetRuleWidth() const;
  virtual SUNIT GetRuleGap() const { return 30; };

  virtual DrawPoint GetTextOrigin() const;
  virtual SUNIT GetTextWidth() const;
  //virtual SUNIT GetTextHeight() const { return 30 * m_line_num + 10 * ( m_line_num - 1 ) ; }
  virtual SUNIT GetTextHeight() const { return 30; }

  virtual SUNIT GetTextY() const { return 2; };

  virtual SUNIT GetRuleTextWidth() const;
  virtual SUNIT GetRuleTextHeight()  const{ return 25; };
  virtual SUNIT GetRuleTextGap() const
    { if( m_ruletextitem == NULL ) return 0; else return 5; };

  virtual SUNIT GetLineY() const { return 40 * m_line_num; };

  void CalcRulesWidthCache() const;

  void InvalidTextSize();
  void InvalidFromSuccessor();

  std::vector<DataString*> GetStringByLines() const;

public:
  virtual SUNIT GetWidth() const;
  virtual SUNIT GetHeight() const;
  virtual DrawRect GetBoundingRect() const;

  virtual void Draw( GraphicContext &gc );
  virtual void Layout();

  virtual void OnEvent( MouseMoveEvent *e ); // ?

  ProofLib::Sequent* GetSequent() const { return m_sequent; }

  void SetHighlightPart( const ProofLib::NodePointer &np );
  const ProofLib::NodePointer& GetHighlightPart() const
    { return m_highlight_part; }
  void ClearHighlightPart();

  DataChar GetCharAt( const DrawPoint &pt ) const;
  size_t GetSElementIndexAt( const DrawPoint &pt ) const;
  ProofLib::SElement *GetSElement( size_t index ) const;
  ProofLib::SElement *GetSElementAt( const DrawPoint &pt ) const;

protected:
  virtual void SequentUpdateEvent( const ProofLib::Sequent *sequent );

public:
#if 0
  static DrawSequentItem* SequentWithMarkedFormula( DrawItem* item );
  static DrawSequentItem* SequentWithMarkedFormulaVar( DrawItem* item );
  static DrawSequentItem* SequentWithMarkedSFormulaListVar( DrawItem* item );
  static DrawSequentItem* SequentWithMarkedSElement( DrawItem* item );
  static DrawSequentItem* SequentWithMarkedSequent( DrawItem* item );

  static bool IsSequentWithMarkedFormula( DrawItem* item )
    { return (SequentWithMarkedFormula( item ) != NULL); }
  static bool IsSequentWithMarkedFormulaVar( DrawItem* item )
    { return (SequentWithMarkedFormulaVar( item ) != NULL); }
  static bool IsSequentWithMarkedSFormulaListVar( DrawItem* item )
    { return (SequentWithMarkedSFormulaListVar( item ) != NULL); }
  static bool IsSequentWithMarkedSElement( DrawItem* item )
    { return (SequentWithMarkedSElement( item ) != NULL); }
  static bool IsSequentWithMarkedSequent( DrawItem* item )
    { return (SequentWithMarkedSequent( item ) != NULL); }
#endif
  template<class T>
    static DrawSequentItem* SequentWithMarkedType( DrawItem* item );
  template<class T>
    static bool IsSequentWithMarkedType( DrawItem* item );
};

template<class T>
bool DrawSequentItem::IsSequentWithMarkedType( DrawItem* item )
{
  return ( SequentWithMarkedType<T>( item ) != NULL );
}

template<class T>
DrawSequentItem* DrawSequentItem::SequentWithMarkedType( DrawItem* item )
{
  if( item == NULL )
    return NULL;

  DrawSequentItem *i = dynamic_cast<DrawSequentItem*>( item );

  if( i == NULL )
    return NULL;

  ProofLib::NodePointer np = i->GetHighlightPart();
  ProofLib::Sequent *s = i->GetSequent();

  while( !np.IsEmpty() &&
         dynamic_cast<T*>(np.GetDataNode(s)) == NULL )
  {
    np.Pop();
  } 

  if( dynamic_cast<T*>(np.GetDataNode(s)) == NULL )
    return NULL;

  i->SetHighlightPart( np );

  return i;
}

////////////////////////////////////////////////////////////////////////////////

class DrawProofLinkItem : public DrawSequentItem
{
public:
  DrawProofLinkItem( DrawItemDatabase *db, DrawSequentItem *parent )
    : DrawSequentItem( db, parent ) {}
};

////////////////////////////////////////////////////////////////////////////////

class DrawToProofLinkItem : public DrawProofLinkItem
{
protected:
  ProofLib::ProofLink      *m_pl;

public:
  DrawToProofLinkItem( DrawItemDatabase *db, DrawSequentItem *parent );
  void SetFromProofLink( ProofLib::ProofLink *i );

  const USTRING &GetProofName() const { return m_pl->GetLinkedProofName(); }
  ProofLib::ProofLink* GetProofLink() const { return m_pl; }
};

////////////////////////////////////////////////////////////////////////////////

class DrawFromProofLinkItem : public DrawProofLinkItem
{
protected:
  ProofLib::Proof          *m_from;

public:
  DrawFromProofLinkItem( DrawItemDatabase *db );
  void SetFromProof( ProofLib::Proof *p );

  const USTRING &GetProofName() const { return m_from->GetName(); }
};

////////////////////////////////////////////////////////////////////////////////

class DrawItemDatabase : std::vector<DrawItem*>, public GraphicInputDispatcher
{
protected:
  ProofSheet            *m_proofsheet;

public:
  DrawItemDatabase( ProofSheet *sheet );
  ~DrawItemDatabase();

  void LoadProof( ProofLib::Proof *p, bool sequents_by_line = false );
  inline ProofSheet* GetProofSheet() const { return m_proofsheet; };

  DrawSequentItem *NewSequent( DrawSequentItem *parent, bool formula_by_line = false );
  DrawToProofLinkItem *NewToProofLink( DrawSequentItem *parent );
  DrawFromProofLinkItem *NewFromProofLink();
  void Insert( DrawItem *i );
  bool Contains( const DrawItem *i ) const;

  void Draw( GraphicContext &gc );
  void Layout();
  DrawRect GetBoundingRect() const;

  DrawItem *GetItemAtPoint( DrawPoint &pt ) const;
};

////////////////////////////////////////////////////////////////////////////////

class ProofSheet : public GraphicInputDispatcher
{
  ProofLib::ProofDatabase   *m_proofdb;
  ProofLib::Proof           *m_proof;
  DrawItemDatabase          *m_items;
  DrawSequentItem           *m_rootseq;
  FontManager               *m_fontmgr;

  ProofSheetNotificantList m_notify_list;

public:
  ProofSheet();
  ProofSheet( ProofLib::ProofDatabase *db );
// ProofSheet( ProofDatabase *db, const char *name );

  void AttachNotificant( ProofSheetNotificant *p );
  void DetachNotificant( ProofSheetNotificant *p );

public:
  void Clear();
  ProofLib::Proof* SelectProof( const USTRING &name );
  ProofLib::Proof* SelectSequentList( const USTRING &name );
  ProofLib::Proof* GetActiveProof() const { return m_proof; }

  void SetRoot( DrawSequentItem *s );

  void Layout();
  void Draw( GraphicContext &gc );

  DrawRect GetBoundingRect() const;

  ProofLib::ProofDatabase* GetProofDatabase() const { return m_proofdb; }
  FontManager* GetFontManager() const { return m_fontmgr; }
  void SetFontManager( FontManager *f );

  DrawItem* GetItemAtPoint( DrawPoint &pt ) const;
  bool Contains( const DrawItem *i ) const;

  void SetProofDatabase( ProofLib::ProofDatabase *db ) { m_proofdb = db; }

protected:
  void InitDatabase( bool sequents_by_line = false );
};

#endif
