////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/qt3/qt3impl.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <stack>

#include "qt3impl.h"

#include "command/execcontext.h"

// for dev
#ifdef DEBUG
#include <iostream>
using namespace std;
#endif

#define QT_WINDOW_LAYOUT_MARGIN         6

#define ALPHA_WEIGHT( a )  int(double(0xff) * (a) * (a))

/******************************************************************************/

void QtGraphicMenuItemSep::AddEnrtyToMenu( QPopupMenu *menu )
{
  ASSERT( menu != NULL );

  menu->insertSeparator();
}

////////////////////////////////////////////////////////////////////////////////

QtGraphicMenu::~QtGraphicMenu()
{
  while( !m_items.empty() )
  {
    delete m_items.back();
    m_items.pop_back();
  }
}

void QtGraphicMenu::Popup( size_t button )
{
  ASSERT( this != NULL );
  
  QPopupMenu *popup = new QPopupMenu;

  GenerateMenu( popup );

  popup->exec( QCursor::pos() );

  delete popup;
}

void QtGraphicMenu::GenerateMenu( QPopupMenu *menu )
{
  ASSERT( menu != NULL );

  std::vector<QtGraphicMenuBase*>::const_iterator i;

  for( i = m_items.begin(); i != m_items.end(); i++ )
    (*i)->AddEnrtyToMenu( menu );
}

void QtGraphicMenu::Append( GraphicMenuItem *item )
{
  ASSERT( item != NULL );

  QtGraphicMenuItem *i = dynamic_cast<QtGraphicMenuItem*>( item );
  ASSERT( i != NULL );

  m_items.push_back( i );
}

void QtGraphicMenu::AppendSeparator()
{
  ASSERT( this != NULL );
  m_items.push_back( new QtGraphicMenuItemSep );
}

////////////////////////////////////////////////////////////////////////////////

QtGraphicMenuItem::QtGraphicMenuItem( const char* utf8_lbl )
{
  ASSERT( utf8_lbl != NULL );
  m_cmd_call = NULL;
  m_label = utf8_lbl;
  m_popupMenu = NULL;
}

QtGraphicMenuItem::~QtGraphicMenuItem()
{
  if( m_popupMenu != NULL )
    delete m_popupMenu;
}

void QtGraphicMenuItem::SetSubMenu( GraphicMenu *menu )
{
  ASSERT( this != NULL );
  ASSERT( menu != NULL );

  QtGraphicMenu *m = dynamic_cast<QtGraphicMenu*>( menu );
  ASSERT( m != NULL );
  
  m_popupMenu = m;
}

void QtGraphicMenuItem::SetCmd( CommandCall *cc )
{
  m_cmd_call = cc;
}

void QtGraphicMenuItem::SetAccel( int modifier, int key )
{
  if( ACCEL_MODIFIER_SHIFT & modifier )
    key |= Qt::SHIFT;
  if( ACCEL_MODIFIER_CTRL & modifier )
    key |= Qt::CTRL;
  if( ACCEL_MODIFIER_ALT & modifier )
    key |= Qt::ALT;
  if( ACCEL_MODIFIER_META & modifier )
    key |= Qt::META;

  m_accel = QKeySequence( key );

#ifdef DEBUG
  QString str = m_accel;
  cout << "ACCEL " << str << endl;
#endif
}

void QtGraphicMenuItem::AddEnrtyToMenu( QPopupMenu *menu )
{
  ASSERT( menu != NULL );

  if( m_popupMenu != NULL )
  {
    QPopupMenu *popup = new QPopupMenu;
    m_popupMenu->GenerateMenu( popup );
    menu->insertItem( QString::fromUtf8(m_label.c_str()), popup );
  }
  else
  {
    int id = menu->insertItem( QString::fromUtf8(m_label.c_str()), this,
                               SLOT( SlotAction() ) );
    menu->setAccel( m_accel, id );
  }
}

void QtGraphicMenuItem::SlotAction()
{
#ifdef DEBUG
  cout << "CLICK MENU ITEM" << endl;
#endif

  if( m_cmd_call == NULL )
    return;

  GetContext()->GetGraphicSystem()->GetMainInputDispatcher()->
    CallCommand( m_cmd_call->Apply() );
}

////////////////////////////////////////////////////////////////////////////////

QtGraphicTopMenuItem::QtGraphicTopMenuItem( const char *utf8_lbl )
{
  m_cmd_call = NULL;
  m_label = utf8_lbl;

  connect( this, SIGNAL( aboutToShow() ), this, SLOT( SlotSetupItems() ) );
}

void QtGraphicTopMenuItem::SetCmd( CommandCall *cc )
{
  m_cmd_call = cc;
}

void QtGraphicTopMenuItem::SlotSetupItems()
{
#ifdef DEBUG
  cout << "TOP MENU SELECT" << endl;
#endif

  ASSERT( this != NULL );

  if( m_cmd_call == NULL )
    return;

  //m_cmd_call->SetWaitForResult( true );
  GetContext()->GetGraphicSystem()->GetMainInputDispatcher()->
    CallCommand( m_cmd_call );

  clear();

  QtGraphicMenu *m =
    dynamic_cast<QtGraphicMenu*>( m_cmd_call->GetResult() );
  
  if( m == NULL )
    return;

  m->GenerateMenu( this ); 
}

#if 0
  virtual void on_select()
  {
    set_submenu( *m );
    m->show_all_children();

    if( oldsub != NULL )
      delete oldsub;

    Qt::Item::on_select();
  }
#endif

/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////

QtGraphicSystem::QtGraphicSystem( int argc, char **argv )
{
  m_fontmgr = FontManager::NewFontManager( this );
  m_inputdispatcher = new QtGraphicMainInputDispatcher( this, argc, argv );
}

QtGraphicSystem::~QtGraphicSystem()
{
  delete m_fontmgr;
  delete m_inputdispatcher;
}

GraphicMainInputDispatcher* QtGraphicSystem::GetMainInputDispatcher()
{
  return m_inputdispatcher;
}

GraphicWindow* QtGraphicSystem::NewProofWindow( ProofSheet *ps )
{
  QtGraphicWindow *wnd = NULL;
  QtGraphicProofView *view = NULL;

  try
  {
#if 0
    SUNIT display_width = gdk_screen_width();
    SUNIT display_height = gdk_screen_height();
  
    DrawRect r;
  
    r.m_righttop.m_x = display_width;
    r.m_righttop.m_y = display_height / 2;
#endif
    DrawRect r;
  
    r.m_righttop.m_x = 600;
    r.m_righttop.m_y = 600;
  
/**/
  
    wnd = new QtGraphicWindow( this, r );
  
    view = new QtGraphicProofView(
            //wnd,
            wnd->GetClientWidget(),
            (QtGraphicSystem*) wnd->GetGraphicSystem() );
  
    wnd->SetView( view );

//    view->SetSize( wnd->GetClientSize() );
    //view->SetToDefaultTrans();
    view->SetProofSheet( ps );
  
    view->show();
  
    SetTopMenu( wnd );

    m_active_proof_wnd = wnd;
  }
  catch( ... )
  {
    if( view != NULL )
      delete view;
    if( wnd != NULL )
      delete wnd;

    throw;
  }

  return wnd;
}

DrawBitmap* QtGraphicSystem::NewDrawBitmap( const GraphicSize &size )
{
  return new QtDrawBitmap( size, this );
}

FontBitmap* QtGraphicSystem::NewFontBitmap( const GraphicSize &size )
{
  return new QtFontBitmap( size, this );
}

GraphicMenu* QtGraphicSystem::NewMenu()
{
  return new QtGraphicMenu();
}

GraphicMenuItem* QtGraphicSystem::NewMenuItem( const USTRING &name )
{
  std::string utf8;
  UTF8FromUString( name, utf8 );
  return new QtGraphicMenuItem( utf8.c_str() );
}

GraphicTopMenuItem* QtGraphicSystem::NewTopMenuItem( const USTRING &name )
{
  std::string utf8;
  UTF8FromUString( name, utf8 );
  QtGraphicTopMenuItem *m = new QtGraphicTopMenuItem( utf8.c_str() );

  m_top_menu.push_back( m );
//std::cout << "Top menues count: " << m_top_menu.size() << std::endl;

  return m;
}

void QtGraphicSystem::ClearTopMenu()
{
  m_top_menu.clear();
#if 0
  while( !m_top_menu.empty() )
  {
    GraphicTopMenuItem *m = m_top_menu.back();
    delete m;
    m_top_menu.pop_back();
  }
#endif
}

GraphicInputDialog* QtGraphicSystem::NewInputDialog( const USTRING &name )
{
  QtGraphicInputDialog *dlg = new QtGraphicInputDialog( this );

  dlg->SetTitle( name );

  return dlg;
}

GraphicFileDialog* QtGraphicSystem::NewFileDialog( const USTRING &name,
		                                   bool for_open )
{
  QtGraphicFileDialog *dlg = new QtGraphicFileDialog( this, for_open );

  dlg->SetTitle( name );

  return dlg;
}

QColor QtGraphicSystem::QtColorFromSCOLOR( SCOLOR c )
{
  int red   = c / 0x10000;
  int green = (c / 0x100) % 0x100;
  int blue  = c % 0x100;

  return QColor( red, green, blue );
}

GraphicSize QtGraphicSystem::GetMaxScreenSize() const
{
  QDesktopWidget *desk = QApplication::desktop();

  return GraphicSize( desk->width(), desk->height() );
}

////////////////////////////////////////////////////////////////////////////////

void QConsoleComboBox::keyReleaseEvent( QKeyEvent *e )
{
  QComboBox::keyReleaseEvent( e );

  if( e->key() == Qt::Key_Return )
    emit executeCommand();
}

////////////////////////////////////////////////////////////////////////////////

QtConsole::QtConsole( CommandProcessor *cp, QWidget *parent, const char *name )
  : QDockWindow( QDockWindow::InDock, parent, name )
{
  ASSERT( cp != NULL );
  m_cp = cp;

  setResizeEnabled( true );
  setFixedExtentHeight( 100 );

  QFrame *frame = new QFrame( this );

  m_layout = new QVBoxLayout( frame );//, QT_WINDOW_LAYOUT_MARGIN );

  m_log = new QTextEdit( frame, "log" );
  m_log->setReadOnly( true );
  m_log->setTextFormat( Qt::LogText );
  m_layout->addWidget( m_log );

  m_edit = new QConsoleComboBox( frame, "editor" );
  m_edit->setInsertionPolicy( QComboBox::AtTop );
  m_edit->setAutoCompletion( true );
  connect( m_edit, SIGNAL( textChanged( const QString& ) ),
           this, SLOT( SlotEditTextChanged( const QString& ) ) );
  connect( m_edit, SIGNAL( executeCommand() ), this, SLOT( SlotExecute() ) );
  m_layout->addWidget( m_edit );

#if 0
  m_buttonLayout = new QHBoxLayout;

  m_execbutton = new QPushButton( "Execute", frame );
  connect( m_execbutton, SIGNAL( clicked() ), this, SLOT( SlotExecute() ) );
  m_execbutton->setEnabled( false );
  m_execbutton->setDefault( true );
  m_buttonLayout->addWidget( m_execbutton );

  m_layout->addLayout( m_buttonLayout );
#endif

  m_edit->setFocus();
  setCaption( "Console" );

  setWidget( frame );
  frame->show();
  QMainWindow *wnd = dynamic_cast<QMainWindow*>( parent );
  wnd->addDockWindow( this, DockBottom);
}

void QtConsole::Show()
{
  show();
}

void QtConsole::Hide()
{
  hide();
}

QString QtConsole::PrepareMessage( const char *msg )
{
  QString str = QString::fromUtf8( msg );
  str.replace( QChar('&'), "&amp;" );
  str.replace( QChar('<'), "&lt;" );
  str.replace( QChar('>'),"&gt;");

  return str;
}

void QtConsole::ShowPreparedMessage( const QString &str )
{
#ifdef DEBUG
  std::cout << "CONSOLE: " << str << std::endl;
#endif
  m_log->append( str );
}

void QtConsole::ShowMessage( const char *msg )
{
  QString str = PrepareMessage( msg );

  ShowPreparedMessage( str );
}

void QtConsole::ShowError( const char *msg )
{
  QString str = "<font color=red>";
  str += PrepareMessage( msg );
  str += "</font>";

  ShowPreparedMessage( str );
}

void QtConsole::ShowPrompt( const char *msg )
{
  GetContext()->GetProofWindow()->ShowStatusMessage( msg );

  QString str = "<font color=green>";
  str += PrepareMessage( msg );
  str += ":</font>";

  ShowPreparedMessage( str );
}

void QtConsole::ClearPrompt()
{
  GetContext()->GetProofWindow()->ClearStatusMessage();
}

void QtConsole::SlotExecute()
{
  QString cmd = m_edit->currentText();
  if( cmd.isEmpty() )
    return;

  ShowPreparedMessage( cmd );
  m_edit->clearEdit();

  m_cp->Process( cmd, true );
}

void QtConsole::SlotEditTextChanged( const QString &str )
{
//  m_execbutton->setEnabled( !str.isEmpty() );
}

////////////////////////////////////////////////////////////////////////////////

QtGraphicMainInputDispatcher::
QtGraphicMainInputDispatcher( GraphicSystem *sys, int argc, char** argv )
  : QApplication( argc, argv ), GraphicMainInputDispatcher( sys )
{
  m_is_working.Set( false );
  m_progressbar_timer = NULL;
  m_processor_working_bar = NULL;

  connect( this, SIGNAL( lastWindowClosed() ),
           this, SLOT( SlotQuit() ) );
}

void QtGraphicMainInputDispatcher::Init()
{
  //m_modal_dlg = NULL;
  //m_processor_working_bar = NULL;
  m_is_working.Set( false );
}

#if 0
gboolean timeout_func( gpointer data )
{
//  cout << "TIMEOUT" << endl;
  QtGraphicMainInputDispatcher *dis = (QtGraphicMainInputDispatcher*) data;

  dis->TimeOutCall();

  return true;
}

gboolean timeout_func_dlg( gpointer data )
{
  //std::cout << "DLG TIMEOUT" << std::endl;

  QtGraphicMainInputDispatcher *dis = (QtGraphicMainInputDispatcher*) data;

  dis->TimeOutCallDlg();

  return true;
}
#endif

void QtGraphicMainInputDispatcher::TimeOutCallDlg()
{
#if 0
  Unlock();
  if( m_modal_dlg == NULL || HaveToQuit() )
    Qt::Main::quit();
  Lock();
#endif
}

void QtGraphicMainInputDispatcher::TimeOutCall()
{
#if 0
  if( HaveToQuit() )
    Exit();

  if( m_modal_dlg != NULL )
  {
//std::cout << "DIALOG MODAL" << std::endl;
    m_modal_dlg->GetDialog()->set_modal( true );
    m_modal_dlg->GetDialog()->show_all();
    guint th = gtk_timeout_add( 100, &timeout_func_dlg, this );
    Qt::Main::run();
    gtk_timeout_remove( th );
    m_modal_dlg = NULL;
    m_cond_modal_done.Broadcast();
  }

  if( m_processor_working_bar != NULL && m_is_working.Get() )
  {
    m_processor_working_bar->pulse();
  }

  Unlock();
/*
  sleep(1);
*/
  Lock();
#endif
}

void QtGraphicMainInputDispatcher::DoModal( GraphicDialog *dlg )
{
#if 0
  QtGraphicDialog *gdlg = dynamic_cast<QtGraphicDialog*>( dlg );
  ASSERT( gdlg != NULL );

  Lock();
  ASSERT( m_modal_dlg == NULL );
  m_modal_dlg = gdlg;
  m_cond_modal_done.WaitOnMutex( *this );
  Unlock();
#endif
/*
  gdlg->GetDialog()->set_modal( true );
  gdlg->GetDialog()->show_all();
  Qt::Main::run();
  Unlock();
*/
}

void QtGraphicMainInputDispatcher::AbortDialog()
{
#if 0
  if( m_modal_dlg == NULL )
    return;

  Lock();
  m_modal_dlg = NULL;
  Unlock();
#endif
}

void QtGraphicMainInputDispatcher::RunTillItemSelected()
{
#define RUN_TIMEOUT 500
  while( IsSelectingItem() && !HasSelectedItem() &&
         !HaveToAbortSelectItem() )
  {
//eventLoop()->processEvents( QEventLoop::AllEvents | QEventLoop::WaitForMore,1000 );
    processEvents();
    usleep( RUN_TIMEOUT );
  }
#undef RUN_TIMEOUT
}

int QtGraphicMainInputDispatcher::Run()
{
#if 0
  gtk_timeout_add( 100, &timeout_func, this );
#endif

#if 1
//  Lock();
  exec();
//  Unlock();
#else
  while(1)
  {
    Lock();
    while( Qt::Main::events_pending() )
      Qt::Main::iteration( false );
    sleep(1);
    Unlock();
  }
#endif

  return 0;
}

void QtGraphicMainInputDispatcher::ProcessEvents()
{
  processEvents();
}
void QtGraphicMainInputDispatcher::Quit()
{
  exit();
}

void
QtGraphicMainInputDispatcher::SetProcessorProgressbar( QProgressBar *bar )
{
  m_processor_working_bar = bar;
  m_processor_working_bar->setProgress( -1, -1 );

  if( m_progressbar_timer != NULL )
    delete m_progressbar_timer;

  m_progressbar_timer = new QTimer( this );
  connect( m_progressbar_timer, SIGNAL( timeout() ),
           this, SLOT( SlotTriggerProgressBar() ) );
}

void QtGraphicMainInputDispatcher::
IndicateProcessorWorking( bool is_working /*= true*/ )
{
  if( m_is_working.Get() == is_working )
    return;

  m_is_working.Set( is_working );

  if( is_working )
  {
    m_processor_working_bar->setProgress( 0, 0 );
    m_progressbar_timer->start( 1000 );
 //   qApp->processEvents( 500 );
  }
  else
  {
    m_progressbar_timer->stop();
    m_processor_working_bar->setProgress( -1, -1 );
  }
}

void QtGraphicMainInputDispatcher::SlotTriggerProgressBar()
{
  static int i = 0;

#ifdef DEBUG
std::cout << "TIMER" << std::endl;
#endif
  m_processor_working_bar->setProgress( ++i );
}

////////////////////////////////////////////////////////////////////////////////

QtDrawBitmap::QtDrawBitmap( GraphicSize size, QtGraphicSystem *sys )
  : QPixmap( size.GetWidth(), size.GetHeight() )
{
  m_sys = sys;
}

GraphicContext* QtDrawBitmap::CreateGraphicContext()
{
  return new QtGraphicContext( new QPainter( this ), m_sys );
}

void QtDrawBitmap::FreeGraphicContext( GraphicContext* gc_ )
{
  QtGraphicContext *gc = dynamic_cast<QtGraphicContext*>( gc_ );
  ASSERT( gc != NULL );
  gc->QtEnd();
  delete gc;
}

void QtDrawBitmap::SetPixel( GraphicPoint pt, SCOLOR c )
{
  ASSERT( false );
  QPainter *p = new QPainter( this );

  p->setPen( QtGraphicSystem::QtColorFromSCOLOR( c ) );
  p->drawPoint( pt.m_x, pt.m_y );

  delete p;
}

SCOLOR QtDrawBitmap::GetPixel( GraphicPoint pt ) const
{
  // TODO
  ASSERT( false );

  return SCOLOR_BLACK;
}

GUNIT QtDrawBitmap::GetWidth() const
{
  return width();
}

GUNIT QtDrawBitmap::GetHeight() const
{
  return height();
}

GraphicSize QtDrawBitmap::GetSize() const
{
  return GraphicSize( width(), height() );
}

void QtDrawBitmap::SetAllPixel( SCOLOR c )
{
  fill( QtGraphicSystem::QtColorFromSCOLOR( c ) );
}

void QtDrawBitmap::MoveContent( const GraphicPoint &v )
{
  QPixmap tmp( *this );
  tmp.detach();

  int x_my = 0;
  int y_my = 0;
  int x_tmp = 0;
  int y_tmp = 0;

  if( v.m_x < 0 )
    x_tmp = -v.m_x;
  else
    x_my = v.m_x;

  if( v.m_y < 0 )
    y_tmp = -v.m_y;
  else
    y_my = v.m_y;

  copyBlt( this, x_my, y_my,
           &tmp, x_tmp, y_tmp, width(), height() );
}

////////////////////////////////////////////////////////////////////////////////

#if 0
QtFontBitmap::QtFontBitmap( GraphicSize size, QtGraphicSystem *sys )
  : QtDrawBitmap( size, sys )
{
}
#endif

#define FONT_IMAGE_DEPTH 8

QtFontBitmap::QtFontBitmap( GraphicSize size, QtGraphicSystem *sys )
  : QImage( size.GetWidth(), size.GetHeight(), FONT_IMAGE_DEPTH, 0x100 )
            //QImage::systemBitOrder() )
{
  setAlphaBuffer( true );
  m_lastColor = SCOLOR_BLACK;
}

void QtFontBitmap::SetPixel( GraphicPoint pt, SCOLOR c )
{
  ASSERT( c >= 0 );
  ASSERT( c <= 0xff );

  setPixel( pt.m_x, pt.m_y, c );
  setColor( c, qRgba( 0, 0, 0, c ) );
}

SCOLOR QtFontBitmap::GetPixel( GraphicPoint pt ) const
{
  // TODO
  ASSERT( false );

  return SCOLOR_BLACK;
}

GUNIT QtFontBitmap::GetWidth() const
{
  return width();
}

GUNIT QtFontBitmap::GetHeight() const
{
  return height();
}

GraphicSize QtFontBitmap::GetSize() const
{
  return GraphicSize( width(), height() );
}

void QtFontBitmap::SetAllPixel( SCOLOR c )
{
  fill( c );
}

void QtFontBitmap::SetColor( SCOLOR c )
{
  if( m_lastColor == c )
    return;

  m_lastColor = c;

  for( int i = 0; i < numColors(); i++ )
  {
    int alpha = qAlpha( color( i ) );

    QColor qc = QtGraphicSystem::QtColorFromSCOLOR( c );

    setColor( i, qRgba( qc.red(), qc.green(), qc.blue(), alpha ) );
  }
}

void QtFontBitmap::MoveContent( const GraphicPoint &v )
{
  ASSERT( false );
}

#if 0
QtFontBitmap::QtFontBitmap( GraphicSize size, QtGraphicSystem *sys )
{
  if( size.GetWidth() <= 0 || size.GetHeight() <= 0 )
    return;

  m_pixbuf = Gdk::Pixbuf::create( Gdk::COLORSPACE_RGB, true, 8,
                                  size.GetWidth(), size.GetHeight() );

  ASSERT( !m_pixbuf.is_null() );
#if 0
  guint8 *pixels = m_pixbuf->get_pixels();

  for( size_t i = 0; i < (size_t) m_pixbuf->get_width(); i++ )
    for( size_t j = 0; j < (size_t) m_pixbuf->get_height(); j++ )
  {
    size_t index = j * m_pixbuf->get_rowstride() +
                   i * m_pixbuf->get_n_channels();

    if( i % 2 == 0 )
      pixels[index] = 0x00;
    else
      pixels[index] = 0xff;

    pixels[index+1] = 0x00;
    pixels[index+2] = 0x00;
  }
#endif
}

QtFontBitmap::~QtFontBitmap()
{
}

void QtFontBitmap::SetColor( SCOLOR c )
{
  if( m_pixbuf.is_null() )
    return;

  guint8 *pixels = m_pixbuf->get_pixels();

  for( size_t i = 0; i < (size_t) m_pixbuf->get_width(); i++ )
    for( size_t j = 0; j < (size_t) m_pixbuf->get_height(); j++ )
  {
    size_t index = j * m_pixbuf->get_rowstride() +
                   i * m_pixbuf->get_n_channels();

    pixels[index]   = c / 0x10000;
    pixels[index+1] = (c / 0x100) % 0x100;
    pixels[index+2] = c % 0x100;
  }
}

void QtFontBitmap::SetPixel( GraphicPoint pt, SCOLOR col )
{
  ASSERT( !m_pixbuf.is_null() );

  ASSERT( pt.m_x <= m_pixbuf->get_width() );
  ASSERT( pt.m_y <= m_pixbuf->get_height() );

  guint8 *pixels = m_pixbuf->get_pixels();

  size_t index = pt.m_y * m_pixbuf->get_rowstride() +
                 pt.m_x * m_pixbuf->get_n_channels();

#if 0
  pixels[index]   = col / 0x10000;
  pixels[index+1] = (col / 0x100) % 0x100;
  pixels[index+2] = col % 0x100;
#endif
  pixels[index]   = 0xff;
  pixels[index+1] = 0xff;
  pixels[index+2] = 0xff;
  pixels[index+3] = (~col) & 0xff;
}

SCOLOR QtFontBitmap::GetPixel( GraphicPoint pt ) const
{
#ifdef DEBUG
cerr << "QtFontBitmap::GetPixel not implemented " << __FILE__ <<  ":" << __LINE__ << endl;
#endif
  // TODO
  return SCOLOR_BLACK;
}

GUNIT QtFontBitmap::GetWidth() const
{
  if( m_pixbuf.is_null() )
    return 0;

  return m_pixbuf->get_width();
}

GUNIT QtFontBitmap::GetHeight() const
{
  if( m_pixbuf.is_null() )
    return 0;

  return m_pixbuf->get_height();
}

void QtFontBitmap::SetAllPixel( SCOLOR c )
{
  if( m_pixbuf.is_null() )
    return;

  m_pixbuf->fill( 0xffffff00 );
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

QtGraphicContextWidget::QtGraphicContextWidget(
  Glib::RefPtr<Qt::Widget> w, QtGraphicSystem* sys )
    : QtGraphicContext( sys )
{
  ASSERT( !w.is_null() );
  m_wg = w;
}

Glib::RefPtr<Gdk::GC> QtGraphicContextWidget::GetQtGC()
{
  return m_wg->get_style()->get_black_gc();
}

Glib::RefPtr<Gdk::Drawable> QtGraphicContextWidget::GetQtDrawable()
{
  return m_wg->get_window();
}

////////////////////////////////////////////////////////////////////////////////

QtGraphicContextLayout::QtGraphicContextLayout(
  Glib::RefPtr<Qt::Layout> l, QtGraphicSystem* sys )
    : QtGraphicContext( sys )
{
  ASSERT( !l.is_null() );
  m_lo = l;
}

Glib::RefPtr<Gdk::GC> QtGraphicContextLayout::GetQtGC()
{
  return m_lo->get_style()->get_black_gc();
}

Glib::RefPtr<Gdk::Drawable> QtGraphicContextLayout::GetQtDrawable()
{
  return m_lo->get_bin_window();
}
#endif

////////////////////////////////////////////////////////////////////////////////

#if 0
QtGraphicContextPixmap::QtGraphicContextPixmap(
  QPainter* p, QtGraphicSystem* sys )
    : QtGraphicContext( sys )
{
  ASSERT( p != NULL );
  m_painter = p;
}
#endif

////////////////////////////////////////////////////////////////////////////////

QtGraphicContext::QtGraphicContext( QPainter* p, QtGraphicSystem* sys )
    : GraphicContext()
{
  ASSERT( sys != NULL );
  m_sys = sys;

  ASSERT( p != NULL );
  m_painter = p;
}

void QtGraphicContext::SetClippingRect( const GraphicRect& rect )
{
  GraphicContext::SetClippingRect( rect );

  m_painter->setClipRect( QRect( rect.GetX(), rect.GetY(),
                                 rect.GetWidth(), rect.GetHeight() ) );
}


SCOLOR GetAlphaColor( SCOLOR bg, SCOLOR fg, double alpha )
{
  if( alpha < 0.0 ) alpha = 0.0;
  if( alpha > 1.0 ) alpha = 1.0;

  int r_bg = (bg >> 16) % 0x100;
  int g_bg = (bg >> 8) % 0x100;
  int b_bg = bg % 0x100;

  int r_fg = (fg >> 16) % 0x100;
  int g_fg = (fg >> 8) % 0x100;
  int b_fg = fg % 0x100;

  int r = r_bg + int( double(r_fg-r_bg) * alpha );
  int g = g_bg + int( double(g_fg-g_bg) * alpha );
  int b = b_bg + int( double(b_fg-b_bg) * alpha );
  
  SCOLOR c = (r << 16) + (g << 8) + b;

#if 0
fprintf( stdout, "bg color: 0x%06x\n", bg );
fprintf( stdout, "fg color: 0x%06x\n", fg );
fprintf( stdout, "alpha is %f aa color: 0x%06x\n", alpha, c );
#endif

  return c;
}

void QtGraphicContext::Rect( DrawRect rect )
{
  if( m_painter == NULL )
    return;

  rect.TransformBy( m_trans );

  rect.Round();

  m_painter->save();
  m_painter->
    setPen( QtGraphicSystem::QtColorFromSCOLOR( GetForegroundColor() ) );
  m_painter->drawRect( (GUNIT) rect.m_leftbottom.m_x,
                       (GUNIT) rect.m_leftbottom.m_y,
                       (GUNIT) rect.GetWidth(),
                       (GUNIT) rect.GetHeight() );
  m_painter->restore();
}

void QtGraphicContext::FillRect( DrawRect rect )
{
#define DO_ANTIALIASE
  if( m_painter == NULL )
    return;

  rect.TransformBy( m_trans );

#if 0
  std::cout << "rect (dbl) "; rect.Dump( std::cout ); std::cout << std::endl;
  std::cout << "clip rect (int) "; GetClippingRect().Dump( std::cout );
  std::cout << std::endl;
#endif
  DrawRect img_rect_ = rect.Intersection( GetClippingRect() );
#ifdef DO_ANTIALIASE
  DrawRect img_rect_org = img_rect_;
#endif
#if 0
  std::cout << "intersec rect (dbl) "; img_rect_.Dump( std::cout );
  std::cout << std::endl;
#endif

  img_rect_.m_leftbottom.m_x = floor( img_rect_.m_leftbottom.m_x );
  img_rect_.m_leftbottom.m_y = floor( img_rect_.m_leftbottom.m_y );
  img_rect_.m_righttop.m_x = ceil( img_rect_.m_righttop.m_x );
  img_rect_.m_righttop.m_y = ceil( img_rect_.m_righttop.m_y );

  GraphicRect img_rect( img_rect_ );

  if( img_rect.IsNull() )
    return;

#if 0
  std::cout << "clipped rect (int) "; img_rect.Dump( std::cout );
  std::cout << std::endl;
#endif

  QColor fg = QtGraphicSystem::QtColorFromSCOLOR( GetForegroundColor() );

  QImage img( img_rect.GetWidth(), img_rect.GetHeight(), 8, 0x100 );

#ifdef DO_ANTIALIASE
  img.setAlphaBuffer( true );

  img.setColor( 0, qRgba( fg.red(), fg.green(), fg.blue(), 0xff ) );
#else
  img.setColor( 0, fg );
#endif

  img.fill( 0 );

#ifdef DO_ANTIALIASE
  DrawPoint pt1_diff = img_rect_.m_leftbottom + DrawPoint( 1, 1 );
  pt1_diff -= img_rect_org.m_leftbottom;

  DrawPoint pt2_diff = img_rect_org.m_righttop;
  pt2_diff -= img_rect_.m_righttop - DrawPoint( 1, 1 );

  if( img_rect.GetWidth() == 1 )
  {
    pt1_diff.m_x = abs( pt1_diff.m_x - pt2_diff.m_x );
    pt2_diff.m_x = 1;
  }

  if( img_rect.GetHeight() == 1 )
  {
    pt1_diff.m_y = abs( pt1_diff.m_y - pt2_diff.m_y );
    pt2_diff.m_y = 1;

    if( img_rect.GetWidth() == 1 )
    {
      pt1_diff.m_x = std::max( pt1_diff.m_x, pt1_diff.m_y );
      pt1_diff.m_y = 1;
    }
  }

#if 0
  std::cout << "pt1_diff "; pt1_diff.Dump( std::cout ); std::cout << std::endl;
  std::cout << "pt2_diff "; pt2_diff.Dump( std::cout ); std::cout << std::endl;
#endif

  int col_num = 0;

#define HAVE_TO_ALAIS( a ) ( ((a) >= 0.0) && ((a) < 1.0) )
  // upper line
  if( HAVE_TO_ALAIS(pt2_diff.m_y) )
  {
    img.setColor( ++col_num, qRgba( fg.red(), fg.green(), fg.blue(),
                            ALPHA_WEIGHT( pt2_diff.m_y ) ) );

    for( int i = 0; i < img_rect.GetWidth(); i++ )
      img.setPixel( i, 0, col_num );
  }
  // lower line
  if( HAVE_TO_ALAIS(pt1_diff.m_y) )
  {
    img.setColor( ++col_num, qRgba( fg.red(), fg.green(), fg.blue(),
                            ALPHA_WEIGHT( pt1_diff.m_y ) ) );

    for( int i = 0; i < img_rect.GetWidth(); i++ )
      img.setPixel( i, img_rect.GetHeight() - 1, col_num );
  }
  // left line
  if( HAVE_TO_ALAIS(pt1_diff.m_x) )
  {
    img.setColor( ++col_num, qRgba( fg.red(), fg.green(), fg.blue(),
                            ALPHA_WEIGHT( pt1_diff.m_x ) ) );

    for( int i = 0; i < img_rect.GetHeight(); i++ )
      img.setPixel( 0, i, col_num );

    // upper left corner
    if( HAVE_TO_ALAIS(pt1_diff.m_x) )
    {
      img.setColor( ++col_num, qRgba( fg.red(), fg.green(), fg.blue(),
                    ALPHA_WEIGHT( pt1_diff.m_x * pt2_diff.m_y ) ) );

      img.setPixel( 0, 0, col_num );
    }
    // lower left corner
    if( HAVE_TO_ALAIS(pt2_diff.m_x) )
    {
      img.setColor( ++col_num, qRgba( fg.red(), fg.green(), fg.blue(),
                    ALPHA_WEIGHT( pt1_diff.m_x * pt1_diff.m_y ) ) );

      img.setPixel( 0, img_rect.GetHeight() - 1, col_num );
    }
  }
  // right line
  if( HAVE_TO_ALAIS(pt2_diff.m_x) )
  {
    img.setColor( ++col_num, qRgba( fg.red(), fg.green(), fg.blue(),
                            ALPHA_WEIGHT( pt2_diff.m_x ) ) );

    for( int i = 0; i < img_rect.GetHeight(); i++ )
      img.setPixel( img_rect.GetWidth() - 1, i, col_num );

    // upper right corner
    if( HAVE_TO_ALAIS(pt1_diff.m_x) )
    {
      img.setColor( ++col_num, qRgba( fg.red(), fg.green(), fg.blue(),
                    ALPHA_WEIGHT( pt2_diff.m_x * pt2_diff.m_y ) ) );

      img.setPixel( img_rect.GetWidth() - 1, 0, col_num );
    }
    // lower right corner
    if( HAVE_TO_ALAIS(pt2_diff.m_x) )
    {
      img.setColor( ++col_num, qRgba( fg.red(), fg.green(), fg.blue(),
                    ALPHA_WEIGHT( pt2_diff.m_x * pt1_diff.m_y ) ) );

      img.setPixel( img_rect.GetWidth() - 1,
                    img_rect.GetHeight() - 1, col_num );
    }
  }
#undef HAVE_TO_ALAIS
#endif

  m_painter->drawImage( QPoint( img_rect.GetX(), img_rect.GetY() ), img );

#if 0
  rect.Round();

  m_painter->fillRect( (GUNIT) rect.m_leftbottom.m_x,
                       (GUNIT) rect.m_leftbottom.m_y,
                       (GUNIT) rect.GetWidth(),
                       (GUNIT) rect.GetHeight(),
             QtGraphicSystem::QtColorFromSCOLOR( GetForegroundColor() ) );
#endif
}

void QtGraphicContext::Point( DrawPoint pt )
{
#if 0
  GetQtGC()->set_foreground(
    QtGraphicSystem::GdkColorFromSCOLOR(
      GetForegroundColor(),
      GetQtDrawable()->get_colormap() )
    );

#endif
}

void QtGraphicContext::Cross( DrawPoint pt )
{
#if 0
  GetQtGC()->set_foreground(
    QtGraphicSystem::GdkColorFromSCOLOR(
      GetForegroundColor(),
      GetQtDrawable()->get_colormap() )
    );

  pt.TransformBy( m_trans );

  int i=1;
  int o=4;

  GetQtDrawable()->draw_line( GetQtGC(),
    (int) pt.m_x-i, (int) pt.m_y,
    (int) pt.m_x-o, (int) pt.m_y );
  GetQtDrawable()->draw_line( GetQtGC(),
    (int) pt.m_x+i, (int) pt.m_y,
    (int) pt.m_x+o, (int) pt.m_y );
  GetQtDrawable()->draw_line( GetQtGC(),
    (int) pt.m_x, (int) pt.m_y-i,
    (int) pt.m_x, (int) pt.m_y-o );
  GetQtDrawable()->draw_line( GetQtGC(),
    (int) pt.m_x, (int) pt.m_y+i,
    (int) pt.m_x, (int) pt.m_y+o );
#endif
}

void QtGraphicContext::Line( DrawPoint from, DrawPoint to )
{
  if( m_painter == NULL )
    return;

  DrawRect rect( from, to );
  rect.TransformBy( m_trans );
  rect.Intersection( GetClippingRect() );

  from = rect.m_leftbottom;
  to = rect.m_righttop;

#if 0
  std::cout << "from "; from.Dump( std::cout );
  std::cout << std::endl;
  std::cout << "to "; to.Dump( std::cout );
  std::cout << std::endl;
#endif

  double up_diff = from.m_x - floor(from.m_x);
  
  GraphicRect img_rect;

  if( up_diff > 0.5 )
    img_rect.m_origin.m_x = GUNIT( floor( from.m_x ) );
  else
    img_rect.m_origin.m_x = GUNIT( floor( from.m_x ) ) - 1;
  img_rect.m_origin.m_y = GUNIT( floor( from.m_y ) );
  img_rect.m_size.m_x = GUNIT( ceil( to.m_x - from.m_x ) );
  img_rect.m_size.m_y = 2;

  if( img_rect.IsNull() )
    return;

#if 0
  std::cout << "img_rect "; img_rect.Dump( std::cout );
  std::cout << std::endl;
#endif
  QColor fg = QtGraphicSystem::QtColorFromSCOLOR( GetForegroundColor() );

  QImage img( img_rect.GetWidth(), img_rect.GetHeight(), 8, 0x100 );

#ifdef DO_ANTIALIASE
  img.setAlphaBuffer( true );

  img.setColor( 0, qRgba( fg.red(), fg.green(), fg.blue(), 0xff ) );
#else
  img.setColor( 0, fg );
#endif

  img.fill( 0 );

#ifdef DO_ANTIALIASE
  int col_num = 0;

  if( up_diff == 0.5 )
  {
    img.setColor( ++col_num, qRgba( fg.red(), fg.green(), fg.blue(), 0x0 ) );

    for( int i = 0; i < img_rect.GetWidth(); i++ )
      img.setPixel( i, 1, col_num );
  }
  else
  {
    img.setColor( ++col_num, qRgba( fg.red(), fg.green(), fg.blue(),
                                    ALPHA_WEIGHT( up_diff ) ) );

    for( int i = 0; i < img_rect.GetWidth(); i++ )
      img.setPixel( i, 0, col_num );

    img.setColor( ++col_num, qRgba( fg.red(), fg.green(), fg.blue(),
                                    ALPHA_WEIGHT( 1.0 - up_diff ) ) );

    for( int i = 0; i < img_rect.GetWidth(); i++ )
      img.setPixel( i, 1, col_num );
  }
#endif

  m_painter->drawImage( QPoint( img_rect.GetX(), img_rect.GetY() ), img );
#if 0
  from.Round();
  to.Round();

  m_painter->save();
  m_painter->
    setPen( QtGraphicSystem::QtColorFromSCOLOR( GetForegroundColor() ) );
  m_painter->drawLine( (GUNIT) from.m_x, (GUNIT) from.m_y,
                       (GUNIT) to.m_x, (GUNIT) to.m_y );
  m_painter->restore();
#endif
}

void QtGraphicContext::Text( DrawRect rect, const char *str )
{
  ASSERT( false );
}

void QtGraphicContext::Text(
  DrawRect rect,
  DataString &str,
  ProofLib::NodePointer *np /* = NULL */ )
{
  DrawRect box = rect;

  rect.TransformBy( m_trans );

  DrawRect clip = GetClippingRect();
  if( !clip.Overlap( rect ) )
    return;

  size_t num_lines = 1;

  for( size_t i = 0; i < str.GetLength(); i++ )
  {
    if ( str[i].isLineFeed() )
    {
      num_lines++;
    }
  }

#define TEXT_HEIGHT_LIMIT 8
  if( rect.GetHeight() < TEXT_HEIGHT_LIMIT * num_lines )
  {
    SCOLOR col;
    col = GetForegroundColor();

    if( np != NULL )
    {
      np->Clear();
      SetForegroundColor( GetHighlightForegroundColor() );
    }
    box.Shrink( box.GetHeight() * 0.15 );
    FillRect( box );
    SetForegroundColor( col );
    return;
  }
  
  m_painter->fillRect( (GUNIT) rect.m_leftbottom.m_x,
                       (GUNIT) rect.m_leftbottom.m_y,
                       (GUNIT) rect.GetWidth(),
                       (GUNIT) rect.GetHeight(),
             QtGraphicSystem::QtColorFromSCOLOR( GetBackgroundColor() ) );

  FUNIT height = (FUNIT) rect.GetHeight();

  height = ceil(double(height) / double(num_lines));

  //double adjust_scale = rect.GetHeight() / double(height);
  double adjust_scale = 1;

  TextHelper( rect.m_leftbottom, adjust_scale, height, str, np );
}

double QtGraphicContext::TextHelper(
  const DrawPoint &origin,
  double adjust_scale,
  FUNIT height,
  DataString &str,
  ProofLib::NodePointer *np /* = NULL */ )
{
  DrawPoint off = origin, adjusted_origin = origin;

  FontBitmapDatabase& db = m_sys->GetFontManager()->GetBitmapDatabase();

  double total_advance = 0;

  DrawRect clip = GetClippingRect();

  for( size_t i = 0; i < str.GetLength(); i++ )
  {

    if ( str[i].isLineFeed() )
    {
      adjusted_origin.m_y += height;
      adjusted_origin.m_x = origin.m_x;
      off = adjusted_origin;
      total_advance = 0;
      continue;
    }

    if( off.m_x > clip.m_righttop.m_x )
      continue;

    FontBitmap *bm = db.GetBitmapOf( str[i], height );

    off.m_x += bm->m_preadj;

    str[i].m_start = off.m_x; // / m_trans.m_matrix[0][0]; // aware scale
    str[i].m_y = off.m_y;

    DrawRect c_bb;
    c_bb.m_leftbottom = off;
    c_bb.m_righttop.m_x = off.m_x + bm->GetWidth() ;
    c_bb.m_righttop.m_y = off.m_y + bm->GetHeight();

    if( clip.Overlap( c_bb ) )
    {
      bool reset_color = false;
      SCOLOR col = SCOLOR_WHITE;

      if( np != NULL && *np <= str[i].m_np )
      {
        col = GetForegroundColor();
        SetForegroundColor( GetHighlightForegroundColor() );
        reset_color = true;
      }

      Bitmap( DrawPoint( off.m_x, off.m_y + bm->m_horz_offset ), *bm );

      if( reset_color )
        SetForegroundColor( col );
    }

    off.m_x = adjusted_origin.m_x + 0;

    if( ( i < str.GetLength() - 1 ) &&
        ( !str[i].HasSameStyle( str[i+1] ) || str[i].HasSubStrings() ) )
    {
      total_advance += bm->m_full_advance * adjust_scale;
    }
    else
    {
      total_advance += bm->m_advance * adjust_scale;
    }

    double sup_advance = 0;
    double sub_advance = 0;

    if( str[i].m_sup != NULL )
    {
      DrawPoint neworigin = adjusted_origin;
      FUNIT newheight = FontManager::GetSuperscriptHeight( height );
      neworigin.m_x += lround(total_advance);
      neworigin.m_y  += (3 * height) / 6 - newheight;

      sup_advance =
        TextHelper( neworigin, adjust_scale, newheight, *(str[i].m_sup), np );
    }
    if( str[i].m_sub != NULL )
    {
      DrawPoint neworigin = adjusted_origin;
      FUNIT newheight = FontManager::GetSuperscriptHeight( height );
      neworigin.m_x += lround(total_advance);
      neworigin.m_y += height - newheight;

      sub_advance =
        TextHelper( neworigin, adjust_scale, newheight, *(str[i].m_sub), np );
    }

    total_advance += std::max( sup_advance, sub_advance );

    off.m_x += lround(total_advance);
    str[i].m_end = off.m_x;
  }

  return total_advance;
  //return lround(total_advance);
}

void QtGraphicContext::Bitmap( DrawPoint pt, FontBitmap &bm )
{
  //QtGraphicContext::Bitmap( pt, (DrawBitmap&) bm );
  // NO TRANSFROM :-(
  GUNIT x = (GUNIT) pt.m_x;
  GUNIT y = (GUNIT) pt.m_y;

  QtFontBitmap *qtbm = dynamic_cast<QtFontBitmap*>( &bm );
  ASSERT( qtbm != NULL );

  qtbm->SetColor( GetForegroundColor() );

  QPixmap pixm( *qtbm );

  bitBlt( m_painter->device(), QPoint( x, y ), &pixm,
          QRect( 0, 0, qtbm->GetWidth(), qtbm->GetHeight() ), Qt::CopyROP );
}

void QtGraphicContext::Bitmap( DrawPoint pt, DrawBitmap &bm )
{
  // NO TRANSFROM :-(
  GUNIT x = (GUNIT) pt.m_x;
  GUNIT y = (GUNIT) pt.m_y;

  QtDrawBitmap *qtbm = dynamic_cast<QtDrawBitmap*>( &bm );
  ASSERT( qtbm != NULL );

  bitBlt( m_painter->device(), QPoint( x, y ), qtbm,
          QRect( 0, 0, qtbm->GetWidth(), qtbm->GetHeight() ), Qt::CopyROP );
}

void QtGraphicContext::Flush()
{
  m_painter->flush();
}

void QtGraphicContext::Clear( const GraphicRect &rect )
{
  m_painter->fillRect( rect.GetX(), rect.GetY(),
                       rect.GetWidth(), rect.GetHeight(),
                       QtGraphicSystem::
                         QtColorFromSCOLOR( GetBackgroundColor() ) );
}

////////////////////////////////////////////////////////////////////////////////

#if 0
class QtGraphicWindow::_WinHelper : public Qt::Window
{
public:
  _WinHelper( QtGraphicWindow *maincl ) { m_maincl = maincl; }


  Qt::VBox m_Box;
  Qt::MenuBar m_MenuBar;
  Qt::Menu m_Menu_File;
  Qt::Statusbar m_StatusBar;
  Qt::ProgressBar m_ProgressBar;

  void AddView( Qt::Widget& widget )
  {
//    m_MenuBar.items().push_back( Qt::Menu_Helpers::MenuElem("_File", m_Menu_File) );
  
    m_StatusBar.pack_end( m_ProgressBar, Qt::PACK_SHRINK );
    m_Box.pack_start( m_MenuBar, Qt::PACK_SHRINK );
    m_Box.add( widget );
    m_Box.pack_end( m_StatusBar, Qt::PACK_SHRINK );

    add( m_Box );

    show_all_children();

    m_ProgressBar.set_fraction( 1.0 );
  }

  void AppendToMenuBar( GraphicTopMenuItem* menuitem )
  {
    ASSERT( menuitem != NULL );

    QtGraphicTopMenuItem *i = dynamic_cast<QtGraphicTopMenuItem*>( menuitem );
    ASSERT( i != NULL );

    m_MenuBar.append( *i );
  }

  void ClearTopMenu()
  {
    m_MenuBar.items().clear();
  }

private:
  QtGraphicWindow      *m_maincl;
};
#endif

QtGraphicWindow::QtGraphicWindow( QtGraphicSystem *sys, DrawRect rect )
  : GraphicWindow( *sys )
{
  setCaption( "ProofTool" );
  resize( (GUNIT) rect.GetWidth(), (GUNIT) rect.GetHeight() );

  m_menuBar = menuBar();

#if 0
  m_progressBar = new QProgressBar( -1, statusBar() );
  m_progressBar->setMaximumWidth( 100 );
  statusBar()->addWidget( m_progressBar, 0, true );

  QtGraphicMainInputDispatcher* mid =
    dynamic_cast<QtGraphicMainInputDispatcher*>(sys->GetMainInputDispatcher());

  ASSERT( mid != NULL );

  mid->SetProcessorProgressbar( m_progressBar );
#endif

#if 0
  m_scroll_view = new QScrollView( this );
  m_scroll_view->setResizePolicy( QScrollView::Manual );
  //m_scroll_view->setStaticBackground( true );
  setCentralWidget( m_scroll_view );
#endif

  show();
}

QtGraphicWindow::~QtGraphicWindow()
{
}

GraphicSize QtGraphicWindow::GetClientSize()
{
  GraphicSize size_( size().width(),
                     size().height() );
#if 0
  QSize qsize = m_scroll_view->viewportSize(
                m_scroll_view->contentsWidth(),
                m_scroll_view->contentsHeight() );

  GraphicSize size( qsize.width(),
                    qsize.height() );
#endif

  return size_;
}

void QtGraphicWindow::SetProofSize( GUNIT w, GUNIT h )
{
//  m_scroll_view->resizeContents( w, h );
}

void QtGraphicWindow::SetView( GraphicView *view_ )
{
  GraphicWindow::SetView( view_ );

  ASSERT( view_ != NULL );

  QtGraphicProofView *view = dynamic_cast<QtGraphicProofView*>( view_ );

  if( view == NULL )
    ASSERT( false );

  setCentralWidget( view );
  view->setFocus();
}

void QtGraphicWindow::OnEvent( ResizeEvent *e )
{
  // TODO
}

void QtGraphicWindow::AppendToMenuBar( GraphicTopMenuItem* menuitem )
{
  ASSERT( menuitem != NULL );

  QtGraphicTopMenuItem *i = dynamic_cast<QtGraphicTopMenuItem*>( menuitem );
  ASSERT( i != NULL );

  int idx = m_menuBar->insertItem( i->GetLabel(), i );

  m_inserted_menus.push_back( idx );

  i->SlotSetupItems();
}

void QtGraphicWindow::ClearTopMenu()
{
  while( !m_inserted_menus.empty() )
  {
    m_menuBar->removeItem( m_inserted_menus.back() );
    m_inserted_menus.pop_back();
  }
}

void QtGraphicWindow::ShowStatusMessage( const char *msg )
{
  statusBar()->message( msg );
}

void QtGraphicWindow::ClearStatusMessage()
{
  statusBar()->clear();
}

////////////////////////////////////////////////////////////////////////////////


QtGraphicProofView::QtGraphicProofView( QWidget *parent, QtGraphicSystem *sys )
  : QScrollView( parent ), //, NULL, WNoAutoErase | WStaticContents ),
    GraphicProofView( sys ),
    m_view_frame( viewport(), this )
{
  m_ignore_ViewOriginChanged = false;
  m_ignore_SlotReposition = false;

  setFocusPolicy( QWidget::StrongFocus );

  setResizePolicy( QScrollView::Manual );
  setStaticBackground( true );

  setVScrollBarMode( QScrollView::AlwaysOn );
  setHScrollBarMode( QScrollView::AlwaysOn );

  connect( this, SIGNAL( contentsMoving ( int, int ) ),
           this, SLOT( SlotReposition( int, int ) ) );
}

QtGraphicProofView::~QtGraphicProofView()
{
}

GraphicSize QtGraphicProofView::GetViewportMinSize() const
{
  QSize min_vp_size = viewportSize( m_size.GetWidth() * 2,
                                    m_size.GetHeight() * 2 );

  return GraphicSize( min_vp_size.width(), min_vp_size.height() );
}

GraphicSize QtGraphicProofView::GetViewportMaxSize() const
{
  QSize max_vp_size = viewportSize( 0, 0 );

  return GraphicSize( max_vp_size.width(), max_vp_size.height() );
}

GraphicContext* QtGraphicProofView::CreateGraphicContext()
{
  GraphicContext *gc = GetBitmap()->CreateGraphicContext();

  gc->m_trans = m_trans;

  gc->SetClippingRect( 
            GraphicRect( 0, 0, m_size.GetWidth(), m_size.GetHeight() ) );

  return gc;
}

void QtGraphicProofView::FreeGraphicContext( GraphicContext* gc )
{
  GetBitmap()->FreeGraphicContext( gc );
}

void QtGraphicProofView::CalcScrollFieldSize()
{
  // calculate the rect of sheet relative to this vp (in screen coords)
  DrawRect org_sheet_rect = GetProofSheet()->GetBoundingRect();
  
  org_sheet_rect.TransformBy( m_trans );
  GraphicRect sheet_rect = org_sheet_rect.GetBiggerIntRect();

  org_sheet_rect = GetProofSheet()->GetBoundingRect();
  
#ifdef DEBUG
std::cout << "CalcScrollFieldSize org_sheet_rect: ";
org_sheet_rect.Dump( std::cout );
std::cout << std::endl;
std::cout << "CalcScrollFieldSize sheet_rect: ";
sheet_rect.Dump( std::cout );
std::cout << std::endl;
#endif

#if 0
  m_view_pt_at_scroll_origin = DrawPoint::Zero();

  if( org_sheet_rect.m_leftbottom.m_x > 0.0 )
    m_view_pt_at_scroll_origin.m_x = org_sheet_rect.m_leftbottom.m_x;
  if( org_sheet_rect.m_leftbottom.m_y > 0.0 )
    m_view_pt_at_scroll_origin.m_y = org_sheet_rect.m_leftbottom.m_y;
  
  m_view_pt_at_scroll_origin = org_sheet_rect.m_leftbottom;
  m_view_pt_at_scroll_origin /= GetZoomFactor();

  m_view_pt_at_scroll_origin =
     ViewportToProofSheet( DrawPoint( left_extra, top_extra ) );
  m_view_pt_at_scroll_origin -= DrawPoint( org_sheet_rect.m_leftbottom.m_x,
                                           org_sheet_rect.m_righttop.m_y );
#endif


  GUNIT left_extra = std::max( 0, -sheet_rect.GetX() );
  GUNIT top_extra = std::max( 0, -sheet_rect.GetY() );

  m_view_pt_at_scroll_origin = DrawPoint::Zero();

  if( sheet_rect.GetX() > 0 )
    m_view_pt_at_scroll_origin.m_x = GetViewPoint().m_x;
  if( sheet_rect.GetY() > 0 )
    m_view_pt_at_scroll_origin.m_y = GetViewPoint().m_y;

#ifdef DEBUG
std::cout << "CalcScrollFieldSize M_VIEW_PT_AT_SCROLL_ORIGIN: ";
m_view_pt_at_scroll_origin.Dump( std::cout );
std::cout << std::endl;
#endif

  // calculate the min vp possible by enable scrollbars
  QSize min_vp_size = viewportSize( m_size.GetWidth() * 2,
                                    m_size.GetHeight() * 2 );
#ifdef DEBUG
std::cout << "CalcScrollFieldSize (min_vp) width: " << min_vp_size.width();
std::cout << " height: " << min_vp_size.height() << std::endl;
#endif

  // calc new contents size
  GUNIT contents_width = left_extra +
                         std::max( min_vp_size.width(),
                              sheet_rect.GetX() + sheet_rect.GetWidth() );
  GUNIT contents_height = top_extra +
                          std::max( min_vp_size.height(),
                               sheet_rect.GetY() + sheet_rect.GetHeight() );

  if( contents_width != contentsWidth() || contents_height != contentsHeight() )
  {
#ifdef DEBUG
  std::cout << "CalcScrollFieldSize contents_width: " << contents_width;
  std::cout << " contents_height: " << contents_height;
  std::cout << std::endl;
#endif

#if 0
    QSize qsize = viewportSize( contents_width, contents_height );
  
    if( qsize.width() != m_size.GetWidth() ||
        qsize.height() != m_size.GetHeight() )
    {
#ifdef DEBUG
std::cout << "SIZE CHANGE width: " << qsize.width();
std::cout << " height: " << qsize.height() << std::endl;
#endif
      // calc new contents size if possible
      if( left_extra > 0 &&
          min_vp_size.width() > sheet_rect.GetX() + sheet_rect.GetWidth() )
        contents_width = left_extra +
                         max( qsize.width(),
                              sheet_rect.GetX() + sheet_rect.GetWidth() );

      if( top_extra > 0 &&
          min_vp_size.height() > sheet_rect.GetY() + sheet_rect.GetHeight() )
        contents_height = top_extra +
                          max( qsize.height(),
                               sheet_rect.GetY() + sheet_rect.GetHeight() );
#ifdef DEBUG
std::cout << "CalcScrollFieldSize correct contents_width: " << contents_width;
std::cout << " contents_height: " << contents_height;
std::cout << std::endl;
#endif

      if( viewportSize( contents_width, contents_height ) != qsize );
      {
        qsize = viewportSize( contents_width, contents_height );
#ifdef DEBUG
std::cout << "ALERT:  SIZE CHANGE width: " << qsize.width();
std::cout << " height: " << qsize.height() << std::endl;
#endif
      }
  

      SetSize( GraphicSize( qsize.width(), qsize.height() ) );
    }
#endif
    
    resizeContents( contents_width, contents_height );
  }
}

void QtGraphicProofView::CalcTransMatrix()
{
  // reset
  m_trans = DrawTransMatrix::Identity();

  // move to sheet offset
//  m_trans.TranslateTo( DrawPoint( 100, 200 ) );

  // window coords y -> goes from upper to lower, sheet low to upper
  m_trans = m_trans.Multiply( DrawTransMatrix::HorizontalFlip() );

  // zoom
  m_trans.ScaleTo( GetZoomFactor() );

  // move down proof height
  DrawRect rect_proof;

  if( GetProofSheet() != NULL )
    rect_proof= GetProofSheet()->GetBoundingRect();

  m_trans.TranslateTo( DrawPoint( 0, -rect_proof.GetHeight() ) );
  m_trans.TranslateTo( rect_proof.m_leftbottom * -1 );

  m_trans.TranslateTo( GetViewPoint() );
}

GraphicSize QtGraphicProofView::GetViewPortSize() const
{
  return GetSize();
}

void QtGraphicProofView::UpdateScreen()
{
#ifdef DEBUG
std::cout << "UpdateScreen()" << std::endl;
#endif
  m_view_frame.update();
}

void QtGraphicProofView::UpdateScreen( const GraphicRect &rect )
{
#ifdef DEBUG
std::cout << "UpdateScreen( const GraphicRect &rect )" << std::endl;
#endif
  m_view_frame.update();
}

void QtGraphicProofView::SlotReposition( int x, int y )
{
  if( m_ignore_SlotReposition )
    return;

#ifdef DEBUG
std::cout << "SlotReposition " << x << "," << y << std::endl;
#endif

  DrawPoint view_pt( m_view_pt_at_scroll_origin );

  view_pt -= DrawPoint( SUNIT(x) / GetZoomFactor(),
                        -SUNIT(y) / GetZoomFactor() );
  
  if( x == m_last_scroll.m_x )
    view_pt.m_x = GetViewPoint().m_x;
  if( y == m_last_scroll.m_y )
    view_pt.m_y = GetViewPoint().m_y;

  m_last_scroll = GraphicPoint( x, y );

#ifdef DEBUG
std::cout << "SlotReposition view_pt: "; view_pt.Dump( std::cout );
std::cout << std::endl;
#endif

  m_ignore_ViewOriginChanged = true;
  SetViewPoint( view_pt );
  m_ignore_ViewOriginChanged = false;
  
#if 0
  // calculate the rect of sheet relative to this vp (in screen coords)
  DrawRect org_sheet_rect = GetProofSheet()->GetBoundingRect();
  org_sheet_rect.TransformBy( m_trans );
  GraphicRect sheet_rect = org_sheet_rect.GetBiggerIntRect();

  GUNIT left_extra = max( 0, -sheet_rect.GetX() );
  GUNIT top_extra = max( 0, -sheet_rect.GetY() );

  x -= left_extra;
  y -= top_extra;

  DrawPoint move( SUNIT(x) / GetZoomFactor(), SUNIT(y) / GetZoomFactor() );

#if 0
  m_ignore_ViewOriginChanged = true;
  SetZoomMidpoint( GetZoomMidpoint() + move, GetZoomFactor() );
  m_ignore_ViewOriginChanged = false;
#endif

  // vector from midpoint to left upper corner of viewport in proofsheet coords
  DrawPoint vec_a( SUNIT(m_size.m_x) / 2,  - SUNIT(m_size.m_y) / 2 );
  vec_a /= m_zoom_factor;

  // point of upper left corner of the proof sheet
  DrawPoint proofsheet_lefttop_pt( 0, m_sheet->GetBoundingRect().GetHeight() );
  
  // vector from midpoint to left upper corner of proof sheet in
  // proofsheet coords
  DrawPoint vec_b( proofsheet_lefttop_pt - mid_pt );
#ifdef DEBUG
std::cout << "SetZoomMidpoint: vec_a is "; vec_a.Dump( std::cout );
std::cout << std::endl;
std::cout << "SetZoomMidpoint: proofsheet_lefttop_pt is ";
proofsheet_lefttop_pt.Dump( std::cout );
std::cout << std::endl;
std::cout << "SetZoomMidpoint: vec_b is "; vec_b.Dump( std::cout );
std::cout << std::endl;
#endif

  m_view_pt = vec_a + vec_b;
#ifdef DEBUG
std::cout << "SetZoomMidpoint: m_view_pt is "; m_view_pt.Dump( std::cout );
std::cout << std::endl;
#endif

  CalcTransMatrix();

  CalcScrollFieldSize();

  FireViewOriginChanged();

  UpdateScreen();
#endif
}

void QtGraphicProofView::ViewOriginChangedEvent( const GraphicPoint &pt )
{
  if( m_ignore_ViewOriginChanged )
    return;

#ifdef DEBUG
std::cout << "VIEWORIGINCHANGED " << pt.m_x << "," << pt.m_y << std::endl;
#endif

  m_ignore_SlotReposition = true;
  setContentsPos( pt.m_x, pt.m_y );
  m_ignore_SlotReposition = false;

  m_last_scroll = pt;
}

void QtGraphicProofView::viewportResizeEvent( QResizeEvent* e )
{
#ifdef DEBUG
cout << "resizeEvent ";
cout << e->size().width() << " x " << e->size().height() << endl;
#endif
#if 0
  ResizeViewPortEvent( GraphicSize( e->size().width(), e->size().height() ) );
#endif

  m_view_frame.resize( e->size() );

  QSize qsize = viewportSize( contentsWidth(), contentsHeight() );

  GraphicSize size( qsize.width(),
                    qsize.height() );

#ifdef DEBUG
cout << "viewportResizeEvent: CLIENT SIZE ";
cout << size.GetWidth() << " x " << size.GetHeight() << endl;
#endif
  ResizeViewPortEvent( size );
}

void QtGraphicProofView::viewportPaintEvent( QPaintEvent* e )
{
}

void QtGraphicProofView::QtGraphicProofViewFrame::paintEvent( QPaintEvent* e )
{
#ifdef DEBUG
cout << "paintEvent (rect ignored) ";
cout << 
  e->rect().x() << "," <<
  e->rect().y() << " " <<
  e->rect().width() << "x" <<
  e->rect().height() <<endl;
#endif
#if 0
  GraphicRect rect( e->rect().x(), 
                    e->rect().y(), 
                    e->rect().width(), 
                    e->rect().height() );
#endif

  QtDrawBitmap *bm = dynamic_cast<QtDrawBitmap*>( m_view->GetBitmap() );
  ASSERT( bm != NULL );

  GraphicRect rect( 0, 0,
                    m_view->GetSize().GetWidth(),
                    m_view->GetSize().GetHeight() );

  GraphicRect vp_rect = rect;
  vp_rect.m_origin.m_x = m_view->contentsX();
  vp_rect.m_origin.m_y = m_view->contentsY();
  GraphicRect r = m_view->RequestBuffer( vp_rect );

  bitBlt( this, QPoint( rect.GetX(), rect.GetY() ),
          bm, QRect( r.GetX(), r.GetY(), r.GetWidth(), r.GetHeight() ),
          Qt::CopyROP );
}

void QtGraphicProofView::keyPressEvent( QKeyEvent* e )
{
  KeyInputEvent *ev = new KeyInputEvent;
  ev->SetChar( e->key() );
  ev->SetMods( e->state() );

  ev->Call( GetGraphicSystem()->GetMainInputDispatcher() );
}

void QtGraphicProofView::contentsMousePressEvent( QMouseEvent* e )
{
#if 0
std::cout << "BUTTON PRESS! " << std::endl;
#endif
  //int x = e->x(), y = e->y(); 
  int x,y; 
  contentsToViewport( e->x(), e->y(), x, y );

  ButtonClickEvent *ev = new ButtonClickEvent();
  ev->SetPoint( DrawPoint( x, y ) );
  ev->SetMods( e->state() );
  ev->SetButton( e->button() );

  ev->Call( GetGraphicSystem()->GetMainInputDispatcher() );
}

void QtGraphicProofView::contextMenuEvent( QContextMenuEvent *e )
{
  int x = e->x(), y = e->y(); 
  //int x,y; 
  //contentsToViewport( e->x(), e->y(), x, y );

  ButtonClickEvent *ev = new ButtonClickEvent();
  ev->SetPoint( DrawPoint( x, y ) );
//  ev->SetMods( e->state() );
  ev->SetButton( BUTTONDB_CONTEXT_MENU );

  ev->Call( GetGraphicSystem()->GetMainInputDispatcher() );
}

void QtGraphicProofView::contentsMouseMoveEvent( QMouseEvent *e )
{
}
        
void QtGraphicProofView::
QtGraphicProofViewFrame::mouseMoveEvent( QMouseEvent *e )
{
#if 0
std::cout << "MOTION " << e->pos().x() << "," << e->pos().y() << std::endl;
#endif
  QPoint pt = e->pos();
  //QPoint pt = m_view->contentsToViewport( e->pos() );


  MouseMoveEvent *ev = new MouseMoveEvent();
  ev->SetPoint( DrawPoint( pt.x(), pt.y() ) );
  ev->SetDragging( false );

  ev->Call( m_view->GetGraphicSystem()->GetMainInputDispatcher() );
}

////////////////////////////////////////////////////////////////////////////////

QtGraphicInputDialog::QtGraphicInputDialog( QtGraphicSystem *sys )
  : QtGraphicDialog( sys )
{
  resize( 200, 100 );

  m_layout = new QVBoxLayout( this, QT_WINDOW_LAYOUT_MARGIN );

  m_edit = new QLineEdit( this, "edit" );
  m_layout->addWidget( m_edit );

  m_buttonLayout = new QHBoxLayout;

  m_ok = new QPushButton( "Ok", this );
  connect( m_ok, SIGNAL( clicked() ), this, SLOT( accept() ) );
  m_ok->setDefault( true );
  m_buttonLayout->addWidget( m_ok );
  m_cancel = new QPushButton( "Cancel", this );
  connect( m_cancel, SIGNAL( clicked() ), this, SLOT( reject() ) );
  m_buttonLayout->addWidget( m_cancel );

  m_layout->addLayout( m_buttonLayout );
}

QtGraphicInputDialog::~QtGraphicInputDialog()
{
}

void QtGraphicInputDialog::Do()
{
  exec();

  if( result() == QDialog::Accepted )
    m_return = m_edit->text().utf8();
  else
    m_return.empty();
}

void QtGraphicInputDialog::SetModal( bool modal )
{
  setModal( true );
}

bool QtGraphicInputDialog::IsModal() const
{
  return isModal(); 
}


void QtGraphicInputDialog::SetTitle( const USTRING &s )
{
  std::string utf8;
  UTF8FromUString( s, utf8 );
  setCaption( utf8 );
}

const UCHAR* QtGraphicInputDialog::GetTitle() const
{
  return NULL;
}

void QtGraphicInputDialog::SetInput( const USTRING& s )
{
  std::string str;
  UTF8FromUString( s, str );
  m_edit->setText( str );
}

const char* QtGraphicInputDialog::GetInput() const
{
  if( result() != QDialog::Accepted )
    return NULL;

  return m_return.c_str();
}

////////////////////////////////////////////////////////////////////////////////

QtGraphicFileDialog::QtGraphicFileDialog( QtGraphicSystem *sys, bool for_open )
  : QtGraphicDialog( sys )
{
  if( !for_open )
    setMode( QFileDialog::AnyFile );
}

QtGraphicFileDialog::~QtGraphicFileDialog()
{
}

void QtGraphicFileDialog::Do()
{
  exec();

  if( result() == QDialog::Accepted )
    m_return = (const char *) selectedFile().utf8();
  else
    m_return.empty();
}

void QtGraphicFileDialog::SetModal( bool modal )
{
  setModal( modal );
}

bool QtGraphicFileDialog::IsModal() const
{
  return isModal(); 
}


void QtGraphicFileDialog::SetTitle( const USTRING &s )
{
  std::string utf8;
  UTF8FromUString( s, utf8 );
  setCaption( utf8 );
}

const UCHAR* QtGraphicFileDialog::GetTitle() const
{
  return NULL;
}

void QtGraphicFileDialog::SetFile( const char *s )
{
  //m_idHelper->set_filename( s );
}

const char* QtGraphicFileDialog::GetFile() const
{
  if( result() != QDialog::Accepted )
    return NULL;

  return m_return.c_str();
}

