////////////////////////////////////////////////////////////////////////////////
//
// ProofTool -------------------------------------------------------------------
//
// File: ./graphic/qt3/qt3impl.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#ifndef _QT3IMPL_H_
#define _QT3IMPL_H_

#include "graphic/graphiccontext.h"
#include "graphic/proofsheet.h"
#include "graphic/bitmap.h"
#include "graphic/fontmanager.h"

#include <qapplication.h>
#include <qmainwindow.h>
#include <qtextedit.h>
#include <qlineedit.h>
#include <qsplitter.h>
#include <qvbox.h>
#include <qpushbutton.h>
#include <qpainter.h>
#include <qdialog.h>
#include <qfiledialog.h>
#include <qlayout.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qcursor.h>
#include <qimage.h>
#include <qstatusbar.h>
#include <qprogressbar.h>
#include <qtimer.h>
#include <qscrollview.h>
#include <qdockwindow.h>
#include <qcombobox.h>


class QtGraphicMainInputDispatcher;


////////////////////////////////////////////////////////////////////////////////

class QtGraphicSystem : public GraphicSystem
{
protected:
  FontManager                   *m_fontmgr;
  QtGraphicMainInputDispatcher  *m_inputdispatcher;

public:
  QtGraphicSystem( int argc, char** argv );
  virtual ~QtGraphicSystem();

  virtual GraphicWindow* NewProofWindow( ProofSheet *ps );
  virtual DrawBitmap* NewDrawBitmap( const GraphicSize &size );
  virtual FontBitmap* NewFontBitmap( const GraphicSize &size );
  virtual GraphicMenu* NewMenu();
  virtual GraphicMenuItem* NewMenuItem( const USTRING &name );
  virtual GraphicTopMenuItem* NewTopMenuItem( const USTRING &name );
  virtual void ClearTopMenu();
  virtual GraphicInputDialog* NewInputDialog( const USTRING &name );
  virtual GraphicFileDialog* NewFileDialog( const USTRING &name,
		                            bool for_open );

  //virtual QApplication* GetQtApp() { return m_qtapp; };

  virtual FontManager* GetFontManager() { return m_fontmgr; };
  inline GraphicMainInputDispatcher* GetMainInputDispatcher();

  virtual GraphicSize GetMaxScreenSize() const;

  static QColor QtColorFromSCOLOR( SCOLOR c );
};


////////////////////////////////////////////////////////////////////////////////

class QConsoleComboBox : public QComboBox
{
  Q_OBJECT

public:
  QConsoleComboBox( QWidget* parent = NULL, const char *name = NULL )
    : QComboBox( true, parent, name ) {}

protected:
  virtual void keyReleaseEvent( QKeyEvent *e );

signals:
  void executeCommand();
};


class QtConsole : public QDockWindow, public CommandConsole
{
  Q_OBJECT

  QVBoxLayout           *m_layout;
  QHBoxLayout           *m_buttonLayout;

  QTextEdit             *m_log;
  QConsoleComboBox      *m_edit;
  QPushButton           *m_execbutton;

  CommandProcessor      *m_cp;

public:
  QtConsole( CommandProcessor *cp, QWidget *parent = 0, const char *name = 0 );

  void Show();
  void Hide();

  virtual void ShowMessage( const char *msg );
  virtual void ShowError( const char *msg );
  virtual void ShowPrompt( const char *msg );
  virtual void ClearPrompt();

private:
  QString PrepareMessage( const char *msg );
  void ShowPreparedMessage( const QString &str );

protected slots:
  void SlotExecute();
  void SlotEditTextChanged( const QString &str );
};

////////////////////////////////////////////////////////////////////////////////

class QtGraphicMainInputDispatcher : public QApplication,
                                     public GraphicMainInputDispatcher
{
  Q_OBJECT

  TSyncVar<bool>         m_is_working;
  QProgressBar          *m_processor_working_bar;

  QTimer                *m_progressbar_timer;

public:
  QtGraphicMainInputDispatcher( GraphicSystem *sys, int argc, char** argv );

  virtual void Init();
  virtual int Run();
  virtual void Quit();
  virtual void ProcessEvents();

  virtual void DoModal( GraphicDialog *dlg );
  virtual void AbortDialog();

  virtual void IndicateProcessorWorking( bool is_working = true );

  void TimeOutCall();
  void TimeOutCallDlg();

  void SetProcessorProgressbar( QProgressBar *bar );

protected:
  virtual void RunTillItemSelected();

protected slots:
  void SlotTriggerProgressBar();
  void SlotQuit() { Quit(); }
};

////////////////////////////////////////////////////////////////////////////////

class QtGraphicContext : public GraphicContext
{
  QtGraphicSystem       *m_sys;

protected:
  QPainter              *m_painter;

public:
  QtGraphicContext( QPainter *p, QtGraphicSystem *sys );
  virtual ~QtGraphicContext() {};

  virtual void SetClippingRect( const GraphicRect& rect );

  virtual void Rect( DrawRect rect );
  virtual void FillRect( DrawRect rect );
  virtual void Point( DrawPoint pt );
  virtual void Cross( DrawPoint pt );
  virtual void Line( DrawPoint from, DrawPoint to );
  virtual void Text( DrawRect rect, const char *str );
  virtual void Text( DrawRect rect, DataString &str,
                     ProofLib::NodePointer *np = NULL );
  virtual void Bitmap( DrawPoint pt, DrawBitmap &bm );
  virtual void Bitmap( DrawPoint pt, FontBitmap &bm );
  virtual void Flush();
  virtual void Clear( const GraphicRect &screen_rect );

  void QtEnd() { m_painter->end(); }
protected:
  double TextHelper( const DrawPoint &origin, double adjust_scale,
                    FUNIT height,
                    DataString &str, ProofLib::NodePointer *np );
};

#if 0
////////////////////////////////////////////////////////////////////////////////

class QtGraphicContextWidget : public QtGraphicContext
{
  Glib::RefPtr<Qt::Widget> m_wg;

public:
  QtGraphicContextWidget( Glib::RefPtr<Qt::Widget> w, QtGraphicSystem* sys );

  virtual Glib::RefPtr<Gdk::GC> GetQtGC();
  virtual Glib::RefPtr<Gdk::Drawable> GetQtDrawable();
};

////////////////////////////////////////////////////////////////////////////////

class QtGraphicContextLayout : public QtGraphicContext
{
  Glib::RefPtr<Qt::Layout> m_lo;

public:
  QtGraphicContextLayout( Glib::RefPtr<Qt::Layout> l, QtGraphicSystem* sys );

  virtual Glib::RefPtr<Gdk::GC> GetQtGC();
  virtual Glib::RefPtr<Gdk::Drawable> GetQtDrawable();
};

////////////////////////////////////////////////////////////////////////////////

class QtGraphicContextPixmap : public QtGraphicContext
{
//  Glib::RefPtr<Gdk::Pixmap> m_pi;
//  Glib::RefPtr<Gdk::GC> m_gc;

public:
  QtGraphicContextPixmap( QPainter *p, QtGraphicSystem *sys );

//  virtual Glib::RefPtr<Gdk::GC> GetQtGC();
//  virtual Glib::RefPtr<Gdk::Drawable> GetQtDrawable();
};
#endif

////////////////////////////////////////////////////////////////////////////////

class QtGraphicWindow : public QMainWindow, public GraphicWindow
{
  Q_OBJECT

  QMenuBar             *m_menuBar;
  QProgressBar         *m_progressBar;

  std::vector<int>      m_inserted_menus;
protected:
  QtGraphicSystem      *m_sys;

public:
  QtGraphicWindow( QtGraphicSystem *sys, DrawRect rect );
  virtual ~QtGraphicWindow();

  virtual void SetView( GraphicView* view );

  QtGraphicSystem* GetQtGraphicSystem() { return m_sys; };
  virtual GraphicSize GetClientSize();

  virtual void OnEvent( ResizeEvent *e );

  virtual void AppendToMenuBar( GraphicTopMenuItem* menuitem );
  virtual void ClearTopMenu();

  virtual void ShowStatusMessage( const char *msg );
  virtual void ClearStatusMessage();

  void SetProofSize( GUNIT w, GUNIT h );

  QWidget* GetClientWidget() { return this; }
};

////////////////////////////////////////////////////////////////////////////////

class QtGraphicProofView : public QScrollView, public GraphicProofView
{
  Q_OBJECT

  bool                  m_ignore_ViewOriginChanged;
  bool                  m_ignore_SlotReposition;

  DrawPoint             m_view_pt_at_scroll_origin;

  GraphicPoint          m_last_scroll;

  class QtGraphicProofViewFrame : public QFrame
  {
    QtGraphicProofView *m_view;

  public:
    QtGraphicProofViewFrame( QWidget *parent, QtGraphicProofView *v )
      : QFrame( parent, 0, WNoAutoErase ) //| WStaticContents )
    {
      m_view = v;

      setMouseTracking( true );
    }
  protected:
    virtual void paintEvent( QPaintEvent* e );
    virtual void mouseMoveEvent( QMouseEvent *e );
  };
  
  QtGraphicProofViewFrame   m_view_frame;

public:
  QtGraphicProofView( QWidget *parent, QtGraphicSystem *sys );
  virtual ~QtGraphicProofView();

  virtual GraphicSize GetViewportMinSize() const;
  virtual GraphicSize GetViewportMaxSize() const;

  virtual void UpdateScreen();
  virtual void UpdateScreen( const GraphicRect &rect );

  virtual GraphicSize GetViewPortSize() const;

  virtual GraphicContext* CreateGraphicContext();
  virtual void FreeGraphicContext( GraphicContext* gc );

protected:
  virtual void ViewOriginChangedEvent( const GraphicPoint &pt );

protected:
  // from QScrolllbar
  virtual void viewportPaintEvent( QPaintEvent *e );
  virtual void viewportResizeEvent( QResizeEvent *e );
  virtual void keyPressEvent( QKeyEvent *e );
  virtual void contentsMousePressEvent( QMouseEvent *e );
  virtual void contextMenuEvent( QContextMenuEvent *e );
  virtual void contentsMouseMoveEvent( QMouseEvent *e );

protected slots:
  void SlotReposition( int x, int y );

private:
  virtual void CalcTransMatrix();
  virtual void CalcScrollFieldSize();
};

////////////////////////////////////////////////////////////////////////////////

class QtDrawBitmap : public DrawBitmap, public QPixmap
{
  QtGraphicSystem         *m_sys;

protected:
  QtDrawBitmap( GraphicSize size, QtGraphicSystem *sys );
  friend class QtGraphicSystem;

public:
  virtual ~QtDrawBitmap() {};

  virtual void SetPixel( GraphicPoint pt, SCOLOR col ); //, int mode = XOR );
  virtual SCOLOR GetPixel( GraphicPoint pt ) const;

  inline virtual GUNIT GetWidth() const; //{ return m_xi->width; };
  inline virtual GUNIT GetHeight() const; //{ return m_xi->height; };
  virtual GraphicSize GetSize() const;

  virtual void SetAllPixel( SCOLOR c );
  virtual void MoveContent( const GraphicPoint &v );
  
  virtual GraphicContext* CreateGraphicContext();
  virtual void FreeGraphicContext( GraphicContext* gc );
};

#if 0
class QtFontBitmap : public QtDrawBitmap
{
protected:
  QtFontBitmap( GraphicSize size, QtGraphicSystem *sys );
  friend class QtGraphicSystem;
};
#endif

////////////////////////////////////////////////////////////////////////////////

class QtFontBitmap : public FontBitmap, public QImage
{
  SCOLOR                   m_lastColor;

protected:
  QtFontBitmap( GraphicSize size, QtGraphicSystem *sys );
  friend class QtGraphicSystem;

public:
  virtual ~QtFontBitmap() {};

  virtual void SetPixel( GraphicPoint pt, SCOLOR col ); //, int mode = XOR );
  virtual SCOLOR GetPixel( GraphicPoint pt ) const;

  inline virtual GUNIT GetWidth() const; //{ return m_xi->width; };
  inline virtual GUNIT GetHeight() const; //{ return m_xi->height; };
  virtual GraphicSize GetSize() const;
  virtual void MoveContent( const GraphicPoint &v );

  virtual void SetAllPixel( SCOLOR c );
  virtual void SetColor( SCOLOR c );

  virtual GraphicContext* CreateGraphicContext() { return NULL; }
  virtual void FreeGraphicContext( GraphicContext* gc ) {}
};

////////////////////////////////////////////////////////////////////////////////

class QtGraphicDialog
{
public:
  QtGraphicDialog( QtGraphicSystem *sys )
  {
    m_sys = sys;
    ASSERT( m_sys != NULL );
  }

  QtGraphicSystem* GetGraphicSystem() const { return m_sys; }

private:
  QtGraphicSystem      *m_sys;
};

////////////////////////////////////////////////////////////////////////////////

class QtGraphicInputDialog : public GraphicInputDialog, public QtGraphicDialog,
  public QDialog     
{
  QVBoxLayout           *m_layout;
  QHBoxLayout           *m_buttonLayout;

  QLineEdit             *m_edit;
  QPushButton           *m_ok;
  QPushButton           *m_cancel;

  std::string            m_return;

public:
  QtGraphicInputDialog( QtGraphicSystem *sys );
  virtual ~QtGraphicInputDialog();

  virtual void Do();

  virtual void SetModal( bool modal );
  virtual bool IsModal() const;

  virtual void SetTitle( const USTRING &s );
  virtual const UCHAR* GetTitle() const;

  virtual void SetInput( const USTRING &s );
  virtual const char* GetInput() const;
};

////////////////////////////////////////////////////////////////////////////////

class QtGraphicFileDialog : public GraphicFileDialog, public QtGraphicDialog,
  public QFileDialog     
{
  std::string            m_return;

public:
  QtGraphicFileDialog( QtGraphicSystem *sys, bool for_open );
  virtual ~QtGraphicFileDialog();

  virtual void Do();

  virtual void SetModal( bool modal );
  virtual bool IsModal() const;

  virtual void SetTitle( const USTRING &s );
  virtual const UCHAR* GetTitle() const;

  virtual void SetFile( const char *s );
  virtual const char* GetFile() const;
};

////////////////////////////////////////////////////////////////////////////////

class QtGraphicMenuBase
{
public:
  virtual void AddEnrtyToMenu( QPopupMenu *menu ) = 0;
};

class QtGraphicMenuItemSep : public QtGraphicMenuBase
{
public:
  void AddEnrtyToMenu( QPopupMenu *menu );
};

class QtGraphicMenu : public GraphicMenu
{
  std::vector<QtGraphicMenuBase*>   m_items;

public:
  virtual ~QtGraphicMenu();

  virtual void Popup( size_t button );
  void GenerateMenu( QPopupMenu *menu );

  virtual void Append( GraphicMenuItem *item );
  virtual void AppendSeparator();
};

class QtGraphicMenuItem : public QObject,
                          public GraphicMenuItem, public QtGraphicMenuBase
{
  Q_OBJECT

  std::string            m_label;
  CommandCall           *m_cmd_call;
  QtGraphicMenu         *m_popupMenu;
  QKeySequence           m_accel;

public:
  QtGraphicMenuItem( const char* utf8_lbl );
  virtual ~QtGraphicMenuItem();

  virtual void SetSubMenu( GraphicMenu *menu );
  virtual void SetCmd( CommandCall *cc );
  virtual void SetAccel( int modifier, int key );

  void AddEnrtyToMenu( QPopupMenu *menu );

protected slots:
  virtual void SlotAction();
};

class QtGraphicTopMenuItem : public QPopupMenu, public GraphicTopMenuItem
{
  Q_OBJECT

  CommandCall           *m_cmd_call;
  std::string            m_label;

public:
  QtGraphicTopMenuItem( const char *utf8_lbl );

  virtual void SetCmd( CommandCall *cc );

  const char* GetLabel() const { return m_label.c_str(); }

public slots:
  void SlotSetupItems();
};

#endif

