////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: export/exportxml.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <stack>
#include "util/hash.h"
#include "util/myassert.h"
#include "chars/latexchars.h"
#include "libproof.h"
#include "recursive.h"

#include <typeinfo>
#include "exportlatex.h"

using namespace std;
using namespace ProofLib;

////////////////////////////////////////////////////////////////////////////////

class ExportLaTeXImpl : public ExportLaTeX
{
  bool m_export_flat;
  bool m_export_cascade;
  bool m_first_term;
  bool m_first_formula;

  stack<Proof*> m_todo;
  stlhash::hash_set< const Symbol* > m_done;


public:
  ExportLaTeXImpl( bool export_flat )
  {
    m_export_flat = export_flat;
    m_first_formula = false;
    m_first_term = false;
    m_export_cascade = false;
  }

public:
  virtual bool IsFlatExport() const { return m_export_flat; }
  virtual bool IsCascadeExport() const { return m_export_cascade; }

  virtual void AddCascade( Proof *p )
  {
    m_todo.push( p );
  }

  virtual bool IsFirstFormula()
  {
    bool f = m_first_formula;
    m_first_formula = false;
    return f;
  }
  virtual void FirstFormula() { m_first_formula = true; }

  virtual bool IsFirstTerm()
  {
    bool f = m_first_term;
    m_first_term = false;
    return f;
  }
  virtual void FirstTerm() { m_first_term = true; }

  virtual void Write( const ProofLib::ProofDatabase* proofdb )
  {
    ASSERT( proofdb != NULL );

    bool export_cascade = m_export_cascade;
    m_export_cascade = true;
    
    WriteHeader();

    const ProofLib::Proof* proof = proofdb->GetDefaultProof();
    WriteProofCascade( proof, proofdb );

    for( size_t i = 0; i < proofdb->GetCount(); i++ )
    {
      proof = proofdb->GetProof(i);

      if( m_done.find( proof->GetSymbol() ) != m_done.end() )
        continue;

      WriteProofCascade( proof, proofdb );
    }

    WriteFooter();

    m_export_cascade = export_cascade;
#if 0

//    m_pdb = proofdb;

    GetStream() << "\\documentclass{article}" << endl;
    GetStream() << "\\usepackage{proof}" << endl;
    GetStream() << endl;

    GetStream() << "\\newcommand{\\SEQ}{\\vdash}" << endl;
    GetStream() << "\\newcommand{\\VAR}[1]{#1}" << endl;
    GetStream() << "\\newcommand{\\CONST}[1]{\\mathrm{#1}}" << endl;
    GetStream() << "\\newcommand{\\FUNC}[1]{#1}" << endl;
    GetStream() << "\\newcommand{\\PRED}[1]{#1}" << endl;
    GetStream() << "\\newcommand{\\FORMULAVAR}[1]{\\mathcal{#1}}" << endl;
    GetStream() << "\\newcommand{\\FORMULALISTVAR}[1]{#1}" << endl;
    GetStream() << "\\newcommand{\\PROOFNAME}[1]{#1}" << endl;
    GetStream() << "\\newcommand{\\QUANT}[3]{(#1 #2) #3}" << endl;
    GetStream() << "\\newcommand{\\PROOF}[2]{$\\PROOFNAME{#1} :=$ $$ #2 $$}"
                << endl;
    GetStream() << endl;

    GetStream() << "\\begin{document}" << endl;
    GetStream() << endl;

    if( IsFlatExport() )
      (*this) << *( proofdb->GetDefaultProof() );
    else
      (*this) << *proofdb;

    GetStream() << endl;
    GetStream() << "\\end{document}" << endl;
#endif
  }

  virtual void Write( const ProofLib::Proof* proof,
                      const ProofLib::ProofDatabase* proofdb )
  {
    WriteHeader();

    WriteProofCascade( proof, proofdb );

    WriteFooter();
  }

  virtual void Write( const ProofLib::SequentList* list,
                      const ProofLib::ProofDatabase* proofdb )
  {
    WriteHeader();

    WriteSequentListCascade( list, proofdb );

    WriteFooter();
  }


  void WriteHeader()
  {
    GetStream() << "\\documentclass{article}" << endl;
    GetStream() << "\\usepackage{proof}" << endl;
    GetStream() << endl;

    GetStream() << "\\newcommand{\\SEQ}{\\vdash}" << endl;
    GetStream() << "\\newcommand{\\VAR}[1]{#1}" << endl;
    GetStream() << "\\newcommand{\\CONST}[1]{\\mathbf{#1}}" << endl;
    GetStream() << "\\newcommand{\\FUNC}[1]{#1}" << endl;
    GetStream() << "\\newcommand{\\PRED}[1]{#1}" << endl;
    GetStream() << "\\newcommand{\\FORMULAVAR}[1]{\\mathcal{#1}}" << endl;
    GetStream() << "\\newcommand{\\FORMULALISTVAR}[1]{#1}" << endl;
    GetStream() << "\\newcommand{\\PROOFNAME}[1]{#1}" << endl;
    GetStream() << "\\newcommand{\\QUANT}[3]{(#1 #2) #3}" << endl;
    GetStream() << "\\newcommand{\\PROOF}[2]{$\\PROOFNAME{#1} :=$ $$ #2 $$}"
                << endl;
    GetStream() << "\\newcommand{\\SEQUENTLIST}[2]{$\\PROOFNAME{#1} :=$ $$ #2 $$}"
                << endl;
    GetStream() << endl;

    GetStream() << "\\begin{document}" << endl;
    GetStream() << endl;
  }

  void WriteProofCascade( const ProofLib::Proof* proof,
                          const ProofLib::ProofDatabase* proofdb )
  {
    if( proof != NULL )
      (*this) << *proof;

    m_done.insert( proof->GetSymbol() );

    while( !m_todo.empty() )
    {
      GetStream() << endl;
      GetStream() << endl;
      Proof *p = m_todo.top();
      m_todo.pop();
      (*this) << *p;
      m_done.insert( p->GetSymbol() );
    }
  }

  void WriteSequentListCascade( const ProofLib::SequentList* list,
                          const ProofLib::ProofDatabase* proofdb )
  {
    if( list != NULL )
      (*this) << *list;

    m_done.insert( list->GetSymbol() );

#if 0 // FIXME: implement todo list for sequentlists?
    while( !m_todo.empty() )
    {
      GetStream() << endl;
      GetStream() << endl;
      SequentList *p = m_todo.top();
      m_todo.pop();
      (*this) << *p;
      m_done.insert( p->GetSymbol() );
    }
#endif
  }


  void WriteFooter()
  {
    GetStream() << endl;
    GetStream() << "\\end{document}" << endl;
  }
};

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX* ExportLaTeX::New( bool export_flat )
{
  return new ExportLaTeXImpl( export_flat );
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const Variable& v )
{
  IsFirstTerm();

  string latex;
  LatexConvertor.GetLatex( v.GetName(), latex );

  GetStream() << "\\VAR{" << latex << "}";

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const Constant& c )
{
  IsFirstTerm();

  string latex;
  LatexConvertor.GetLatex( c.GetName(), latex );

  GetStream() << "\\CONST{" << latex << "}";

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const Function& f )
{
  bool first = IsFirstTerm();

  if( f.GetSymbol()->IsInfix() )
  {
    ASSERT( f.GetChildCount() == 2 );

    if( !first )
      GetStream() << "( ";

    (*this) << *(f.GetTerm(0));
    GetStream() << " ";
  }

  bool print_args = true;

  try
  {
    const RecursiveFunction &f_r = dynamic_cast<const RecursiveFunction&>( f );

    if( f_r.GetChildCount() < 1 )
    {
      GetStream() << "\\CONST{";
      print_args = false;
    }
    else
      GetStream() << "\\FUNC{";

    WriteRecursiveName( f.GetName(), f_r );

    GetStream() << "}";
  }
  catch( bad_cast )
  {
    string latex;
    LatexConvertor.GetLatex( f.GetName(), latex );

    size_t iter = 1;

    if( f.GetArity() == 1 )
    {
      const Symbol* s = f.GetSymbol();
      const Function *f_1 = &f;
      const Function *f_0 = f_1;

      while( ( f_1 = f_1->GetTerm(0)->CastTo<Function>() ) != NULL &&
             s == f_1->GetSymbol() )
      {
        iter++;
        f_0 = f_1;
      }

      if( iter > 1 )
      {
        GetStream() << "\\FUNC{{" << latex << "}^{" << iter << "}}";
        (*this) << dynamic_cast<const StaticAbstractTermList&>( *f_0 );
        return (*this);
      }
    }

    GetStream() << "\\FUNC{" << latex << "}";
  }

  if( f.GetSymbol()->IsInfix() )
  {
    GetStream() << " ";
    (*this) << *(f.GetTerm(1));

    if( !first )
      GetStream() << " )";
  }
  else if( print_args )
    (*this) << dynamic_cast<const StaticAbstractTermList&>( f );

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::WriteRecursiveName( const USTRING& name,
                                              const RecursiveObject& o )
{
  RecursiveExpr* e = o.GetRecursiveExpr();

  ASSERT( e->GetVariableName() == NULL );

  string latex;
  LatexConvertor.GetLatex( name, latex );

  GetStream() << latex << "[" << e->GetNum() << "]";
  
  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const AbstractTerm& t )
{
  try
  {
    const Term &a = dynamic_cast<const Term&>( t );
    (*this) << a;
    return (*this);
  }
  catch( bad_cast ) {}

  // TODO: implement rest

  ASSERT( false );
  
  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const Term& t )
{
  try
  {
    const Variable &a = dynamic_cast<const Variable&>( t );
    (*this) << a;
    return (*this);
  }
  catch( bad_cast ) {}
  
  try
  {
    const Constant &a = dynamic_cast<const Constant&>( t );
    (*this) << a;
    return (*this);
  }
  catch( bad_cast ) {}
  
  try
  {
    const Function &a = dynamic_cast<const Function&>( t );
    (*this) << a;
    return (*this);
  }
  catch( bad_cast ) {}

  ASSERT( false );
  
  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const StaticTermList& l )
{
  GetStream() << "( ";

  for( size_t i = 0; i < l.GetChildCount(); i++ )
  {
    if( i != 0 )
      GetStream() << ", ";

    FirstTerm();
    (*this) << *(l.GetTerm(i));
  }

  GetStream() << " )";

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const StaticAbstractTermList& l )
{
  GetStream() << "( ";

  for( size_t i = 0; i < l.GetChildCount(); i++ )
  {
    if( i != 0 )
      GetStream() << ", ";

    FirstTerm();
    (*this) << *(l.GetTerm(i));
  }

  GetStream() << " )";

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const AtomFormula& f )
{
  //bool first = IsFirstFormula();

  if( f.GetSymbol()->IsInfix() )
  {
    ASSERT( f.GetChildCount() == 2 );

#if 0
    if( !first )
      GetStream() << "( ";
#endif

    FirstTerm();
    (*this) << *(f.GetTerm(0));
    GetStream() << " ";
  }

  GetStream() << "\\PRED{";

  try
  {
    const RecursivePredicate &f_r =
      dynamic_cast<const RecursivePredicate&>( f );

    WriteRecursiveName( f.GetName(), f_r );
  }
  catch( bad_cast )
  {
    string latex;
    LatexConvertor.GetLatex( f.GetName(), latex );

    GetStream() << latex;
  }

  GetStream() << "}";

  if( f.GetSymbol()->IsInfix() )
  {
    GetStream() << " ";
    FirstTerm();
    (*this) << *(f.GetTerm(1));

#if 0
    if( !first )
      GetStream() << " )";
#endif
  }
  else if( f.GetChildCount() > 0 )
    (*this) << dynamic_cast<const StaticAbstractTermList&>( f );

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const QuantorFormula& f )
{
  IsFirstFormula();

  const char *latex =
    LatexConvertor.GetLatex( f.GetOperatorSymbol()->GetDisplayChar() );

  GetStream() << "\\QUANT{\\" << latex << "}{";

  (*this) << *(f.GetVariable());

  GetStream() << "}{";

  (*this) << *(f.GetFormula());

  GetStream() << "}";

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const UnaryConFormula& f )
{
  IsFirstFormula();

  const char *latex =
    LatexConvertor.GetLatex( f.GetOperatorSymbol()->GetDisplayChar() );

  GetStream() << "\\" << latex << " ";

  (*this) << *(f.GetFormula());

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const BinaryConFormula& f )
{
  bool first = IsFirstFormula();

  if( !first )
    GetStream() << "( ";

  (*this) << *(f.GetLeft());

  const char *latex =
    LatexConvertor.GetLatex( f.GetOperatorSymbol()->GetDisplayChar() );

  GetStream() << " \\" << latex << " ";

  (*this) << *(f.GetRight());

  if( !first )
    GetStream() << " )";

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const FormulaVariable& f )
{
  string latex;
  LatexConvertor.GetLatex( f.GetName(), latex );

  GetStream() << "\\FORMULAVAR{" << latex << "}";

//  AddFormulaVar( &f );
  
  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const Formula& f )
{
  try
  {
    const AtomFormula &a = dynamic_cast<const AtomFormula&>( f );
    (*this) << a;
    return (*this);
  }
  catch( bad_cast ) {}

  try
  {
    const QuantorFormula &a = dynamic_cast<const QuantorFormula&>( f );
    (*this) << a;
    return (*this);
  }
  catch( bad_cast ) {}

  try
  {
    const UnaryConFormula &a = dynamic_cast<const UnaryConFormula&>( f );
    (*this) << a;
    return (*this);
  }
  catch( bad_cast ) {}

  try
  {
    const BinaryConFormula &a = dynamic_cast<const BinaryConFormula&>( f );
    (*this) << a;
    return (*this);
  }
  catch( bad_cast ) {}

  try
  {
    const FormulaVariable &a = dynamic_cast<const FormulaVariable&>( f );
    (*this) << a;
    return (*this);
  }
  catch( bad_cast ) {}

  ASSERT( false );

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const FormulaList& l )
{
  if( &l != NULL )
    for( size_t i = 0; i < l.GetChildCount(); i++ )
    {
      if( i != 0 )
        GetStream() << ", ";

      (*this) << *(l.GetElement(i));
    }

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const FormulaListVariable& l )
{
  string latex;
  LatexConvertor.GetLatex( l.GetName(), latex );

  GetStream() << "\\FORMULALISTVAR{" << latex << "}";

//  AddFormulaListVar( &l );

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const FormulaListElement& e )
{
  try
  {
    const Formula &a = dynamic_cast<const Formula&>( e );
    (*this) << a;
    return (*this);
  }
  catch( bad_cast ) {}

  try
  {
    const FormulaListVariable &a =
      dynamic_cast<const FormulaListVariable&>( e );
    (*this) << a;
    return (*this);
  }
  catch( bad_cast ) {}

  ASSERT( false );
  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const SElement& e )
{
  try
  {
    const SFormula &a = dynamic_cast<const SFormula&>( e );
    FirstFormula();
    (*this) << *(a.GetFormula());
    return (*this);
  }
  catch( bad_cast ) {}

  try
  {
    const SFormulaListVariable &a =
      dynamic_cast<const SFormulaListVariable&>( e );
    (*this) << *(a.GetFormulaListVariable());
    return (*this);
  }
  catch( bad_cast ) {}

  ASSERT( false );
  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const SElementList& l )
{
  if( &l != NULL )
    for( size_t i = 0; i < l.GetChildCount(); i++ )
    {
      if( i != 0 )
        GetStream() << ", ";

      (*this) << *(l.GetSElement(i));
    }

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::write ( const SElementList& l, bool formulas_by_line )
{
  if( &l != NULL )
  {
    for( size_t i = 0; i < l.GetChildCount(); i++ )
    {
      if( i != 0 )
        GetStream() << ",\\\\" << endl;

      (*this) << *(l.GetSElement(i));
    }

    if ( l.GetChildCount() > 0)
      GetStream() << "\\\\";
  }

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const Sequent& s )
{
  (*this) << *(s.GetLeft());

  GetStream() << " \\SEQ ";

  (*this) << *(s.GetRight());

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const Rule& r )
{
  if( r.GetRuleItemCount() == 0 )
  {
    (*this) << *(r.GetConclusion());
    return (*this);
  }

  string latex;
  LatexConvertor.GetLatex( r.GetName(), latex );

  GetStream() << "\\infer[" << latex << "]" << endl << "{";

  (*this) << *(r.GetConclusion());

  GetStream() << "}" << endl << "{";

  for( size_t i = 0; i < r.GetRuleItemCount(); i++ )
  {
    if( i != 0 )
      GetStream() << " & ";

    (*this) << *(r.GetRuleItem(i));
  }

  GetStream() << "}" << endl;

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const ProofLink& p )
{
  if( IsFlatExport() )
  {
    (*this) << *(p.GetLinkedProof()->GetEndRule());
  }
  else
  {
    string latex;
    LatexConvertor.GetLatex( p.GetLinkedProofName(), latex );

    GetStream() << "\\PROOFNAME{" << latex << "}";
    
    if( IsCascadeExport() )
      AddCascade( p.GetLinkedProof() );
  }

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const RuleItem& ruleitem )
{
  try
  {
    const Sequent &s = dynamic_cast<const Sequent&>( ruleitem );
    (*this) << s;
    return (*this);
  }
  catch( bad_cast ) { }

  try
  {
    const Rule &r = dynamic_cast<const Rule&>( ruleitem );
    (*this) << r;
    return (*this);
  }
  catch( bad_cast ) { }

  try
  {
    const ProofLink &pl = dynamic_cast<const ProofLink&>( ruleitem );
    (*this) << pl;
    return (*this);
  }
  catch( bad_cast ) { }

  ASSERT( false ); // unknow type of ruleitem

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const Proof& p )
{
  string latex;
  LatexConvertor.GetLatex( p.GetName(), latex );

  GetStream() << "\\PROOF{" << latex << "}{";

  (*this) << *(p.GetEndRule());

  GetStream() << "}";

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::write ( const ProofLib::Sequent& s, bool formulas_by_line )
{
  write( *(s.GetLeft()), formulas_by_line );

  GetStream() << endl <<"\\SEQ \\\\" << endl;

  write( *(s.GetRight()), formulas_by_line );

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////
ExportLaTeX& ExportLaTeX::operator << ( const SequentList& l )
{
  string latex;
  LatexConvertor.GetLatex( l.GetName(), latex );

  GetStream() << "\\SEQUENTLIST{" << latex << "}{\\begin{array}{l}" << endl;

  SequentList::const_iterator i = l.begin();

  if ( i != l.end() )
  {
    write( **i, true );
    i++;
  }

  for ( ; i != l.end(); i++ )
  {
    GetStream() << "\\hline" << endl;
    write( **i, true );
  }

  GetStream() << "\\end{array}}";

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////

ExportLaTeX& ExportLaTeX::operator << ( const ProofDatabase& pdb )
{
  for( size_t i = 0; i < pdb.GetCount(); i++ )
  {
    Proof *p = pdb.GetProof( i );

    (*this) <<  *p;
  }

//  WriteVariables();

  return (*this);
}

////////////////////////////////////////////////////////////////////////////////
