////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: export/savedbxml.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <stack>

#include "util/myassert.h"
#include "util/hash.h"
#include "util/xmlwrite.h"
#include "util/proofdbdtd.h"
#include "libproof.h"

#include "savedbxml.h"

using namespace std;
using namespace ProofLib;

////////////////////////////////////////////////////////////////////////////////

class SaveDbXMLImpl : public SaveDbXML
{
  XmlWriter             *m_xml_writer;
  int                    m_compression;

  const ProofDatabase   *m_pdb;

public:
  typedef stlhash::hash_set< const DbNode* > VarsSet;
  typedef stlhash::hash_set< const DbNode* > ListVarsSet;
  typedef stack< const FormulaVariable* > VarsStack;
  typedef stack< const FormulaListVariable* > ListVarsStack;

  VarsSet                m_vars;
  ListVarsSet            m_list_vars;

  VarsStack              m_vars_todo_stack;
  ListVarsStack          m_list_vars_todo_stack;

  void AddFormulaVar( const FormulaVariable *v )
  {
    if( m_vars.find( v ) != m_vars.end() )
      return;

    m_vars.insert( v );
    m_vars_todo_stack.push( v );
  }

  void AddFormulaListVar( const FormulaListVariable *v )
  {
    if( m_list_vars.find( v ) != m_list_vars.end() )
      return;

    m_list_vars.insert( v );
    m_list_vars_todo_stack.push( v );
  }

  void WriteVariables()
  {
    GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_VARIABLE_DEFINITIONS );

    WriteFormulaListVars();
    WriteFormulaVars();

    GetXmlWriter()->CloseElement();
  }

  void WriteFormulaVars()
  {
    while( !m_vars_todo_stack.empty() )
    {
      const FormulaVariable* f = m_vars_todo_stack.top();

      m_vars_todo_stack.pop();

      GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_FORMULAVAR_DEF );

      GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL,
                                      f->GetName() );

      (*this) << *(f->GetFormula());

      GetXmlWriter()->CloseElement();
    }
  }

  void WriteFormulaListVars()
  {
    while( !m_list_vars_todo_stack.empty() )
    {
      const FormulaListVariable* l = m_list_vars_todo_stack.top();

      m_list_vars_todo_stack.pop();

      GetXmlWriter()->OpenElement( PROOFDBDTD_NODE_FORMULALISTVAR_DEF );

      GetXmlWriter()->WriteAttribute( PROOFDBDTD_ATTR_SYMBOL,
                                      l->GetName() );

      (*this) << (const FormulaList&) *l;

      GetXmlWriter()->CloseElement();
    }
  }

protected:
  SaveDbXMLImpl()
  {
    m_xml_writer = NULL;
    m_compression = 0;
    m_pdb = NULL;
  }

  friend class SaveDbXML;

public:
  XmlWriter* GetXmlWriter() const
  {
    return m_xml_writer;
  }

  const ProofDatabase* GetProofDatabase() const
  {
    return m_pdb;
  }

  virtual void Write( const ProofLib::ProofDatabase* proofdb )
  {
    ASSERT( proofdb != NULL );
    ASSERT( m_xml_writer != NULL );

    m_pdb = proofdb;

    m_xml_writer->StartDocument( PROOFDBDTD_VERSION, PROOFDBDTD_ENCODING );

    m_xml_writer->WriteDTD( PROOFDBDTD_NODE_PROOFDATABASE,
                            PROOFDBDTD_DTD_PROOFDATABASE_PUBLIC,
                            PROOFDBDTD_DTD_PROOFDATABASE_SYSTEM,
                            PROOFDBDTD_DTD_PROOFDATABASE_SUBSET );

    (*this) << *proofdb;

    m_xml_writer->EndDocument();
  }

  virtual void Write( const ProofLib::Proof* proof,
                      const ProofLib::ProofDatabase* proofdb )
  {
    ASSERT( false ); // not implemented
  }

  virtual void Open( const char *sourcename )
  {
    ASSERT( m_xml_writer == NULL );

    m_xml_writer = XmlWriter::New();

    m_xml_writer->SetCompression( m_compression );
    m_xml_writer->Open( sourcename );
  }

  virtual void Close()
  {
    ASSERT( m_xml_writer != NULL );

    m_xml_writer->Close();
    delete m_xml_writer;
    m_xml_writer = NULL;
  }
};

////////////////////////////////////////////////////////////////////////////////

SaveDbXML::SaveDbXML()
{
}

SaveDbXML* SaveDbXML::New()
{
  return new SaveDbXMLImpl();
}

////////////////////////////////////////////////////////////////////////////////
