////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: import/importotter.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <vector>
#include <wchar.h>

#include "util/myassert.h"
#include "util/proofexceptions.h"
#include "libproof.h"
#include "chars/latexchars.h"
#include "importotter.h"

#ifdef DEBUG
//#define PRINT_DEBUG_INFO
#endif

////////////////////////////////////////////////////////////////////////////////
// otter proof object tokens

#define OTTER_MAX_TOKEN_LEN             32

#define OTTER_PO_BEGIN                  ";; BEGINNING OF PROOF OBJECT"

#define OTTER_INFER_UNDEFINED           "undefined"
#define OTTER_INFER_INPUT               "input"
#define OTTER_INFER_EQ_AXIOM            "eq-axiom"
#define OTTER_INFER_INSTANTIATE         "instantiate"
#define OTTER_INFER_PROPOSITIONAL       "propositional"
#define OTTER_INFER_RESOLVE             "resolve"
#define OTTER_INFER_PARAMOD             "paramod"
#define OTTER_INFER_FLIP                "flip"
#define OTTER_INFER_UNKNOWN             "unknown"

#define OTTER_OR                        "or"
#define OTTER_NOT                       "not"
#define OTTER_NIL                       "NIL"
#define OTTER_BOX                       "false"

#define OTTER_OPEN_BRACKET              '('
#define OTTER_CLOSE_BRACKET             ')'
#define OTTER_DOT                       '.'

#define OTTER_VAR_START                 'v'

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////
class OtterException;
  class UnknownInferenceException;
  class UnexpectedCharException;
  class UnexpectedTokenException;

class Substitution;
class SubstitutionList;

class OtterInfer;
  class OtterInferUndefined;
  class OtterInferInput;
  class OtterInferEqAxiom;
  class OtterInferInstantiate;
  class OtterInferPropositional;
  class OtterInferResolve;
  class OtterInferParamod;
  class OtterInferFlip;
  class OtterInferUnknown;

class OtterRule;
class OtterProofReader;

class ImportOtterProofImpl;


ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    Term** t );
ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    TermList& l );
ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    Formula** f );

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    Substitution& s );
ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    SubstitutionList& l );

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterInferUndefined& i );
ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterInferInput& i );
ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterInferEqAxiom& i );
ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterInferInstantiate& i );
ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterInferPropositional& i );
ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterInferResolve& i );
ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterInferParamod& i );
ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterInferFlip& i );
ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterInferUnknown& i );

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterInfer** i );

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterRule& r );

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in,
                                    OtterProofReader& opr );

////////////////////////////////////////////////////////////////////////////////
//
// This file contains classes only needed to import a otter proof object.
//
// Example of an otter output containing a proof object:
//
// ----- Otter 3.3, August 2003 -----
//
// [...]
//
// set(build_proof_object_2).
//    dependent: set(order_history).
//
// [...]
//
// ;; BEGINNING OF PROOF OBJECT
// (
// (1 (input) (or (not (P (i v0 v1))) (or (not (P v0)) (P v1))) (1))
// (2 (input) (not (P (i (i (i (a) (b)) (c)) (i (b) (c))))) (2))
// (3 (input) (P (i (i v0 v1) (i (i v1 v2) (i v0 v2)))) (3))
// (4 (input) (P (i (i (n v0) v0) v0)) (4))
// (5 (input) (P (i v0 (i (n v0) v1))) (5))
// (6 (instantiate 1 ((v0 . (i v64 v65))(v1 . (i (i v65 v66) (i v64 v66))))) (or (not (P (i (i v64 v65) (i (i v65 v66) (i v64 v66))))) (or (not (P (i v64 v65))) (P (i (i v65 v66) (i v64 v66))))) NIL)
// (7 (instantiate 3 ((v0 . v64)(v1 . v65)(v2 . v66))) (P (i (i v64 v65) (i (i v65 v66) (i v64 v66)))) NIL)
// (8 (resolve 6 (1) 7 ()) (or (not (P (i v64 v65))) (P (i (i v65 v66) (i v64 v66)))) NIL)
//
// [...]
//
// (159 (resolve 2 () 158 ()) false (1424))
// )
// ;; END OF PROOF OBJECT
//
// [...]
//
// end of example.
//
//  The proof object is a lisp expression enveloped by two lines
// (";; BEGINNING OF PROOF OBJECT" and ";; END OF PROOF OBJECT"). The expression
// is a list containing inferences. Each inference is a list containing four
// elements:
//   1. id of the inference, a number counting the inferences.
//   2. the inference function.
//   3. the conclusion of the inference.
//   4. unkown and not used.
//
//  ImportOtterProofImpl is the reader classes wich read the chars from the
// stream, finding the proof object and let read the stream input operator
// function the proof object by building vector containing individual objects
// for each inference type. After reading the proof object the reader class will
// build the ProofLib proof tree out of these objects read.
//
//
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
//
// 1. Exception throw only by otter import
//
////////////////////////////////////////////////////////////////////////////////

class OtterException : public Exception
{
  size_t                 m_lineno;
  size_t                 m_charno;

protected:
  OtterException( size_t lineno, size_t charno )
  {
    m_lineno = lineno;
    m_charno = charno;
  }

public:
  virtual void Print( std::ostream &out ) const
  {
    out << "Error";
    if( m_lineno > 0 )
    {
      out << " at line " << m_lineno << " pos " << m_charno;
    }
    out << ": ";
  }
};

////////////////////////////////////////////////////////////////////////////////

class UnknownInferenceException : public OtterException
{
public:
  UnknownInferenceException( size_t lineno, size_t charno, const char* infer_name )
    : OtterException( lineno, charno )
  {
  }

  virtual void Print( std::ostream &out ) const
  {
    OtterException::Print( out );
    out << "UnknownInferenceException";
  }
};

////////////////////////////////////////////////////////////////////////////////

class UnexpectedCharException : public OtterException
{
  char                   m_got;
  char                   m_expected;

public:
  UnexpectedCharException( size_t lineno, size_t charno,
                           char got, char expected = '\0' )
    : OtterException( lineno, charno )
  {
    m_got = got;
    m_expected = expected;
  }

  virtual void Print( std::ostream &out ) const
  {
    OtterException::Print( out );
    out << "UnexpectedCharException: " << m_got;
    if( m_expected != '\0' )
      out << " expected: " << m_expected;
  }
};

////////////////////////////////////////////////////////////////////////////////

class UnexpectedTokenException : public OtterException
{
  std::string            m_got;
  std::string            m_expected;

public:
  UnexpectedTokenException( size_t lineno, size_t charno, std::string got, std::string expected )
    : OtterException( lineno, charno )
  {
    m_got = got;
    m_expected = expected;
  }

  virtual void Print( std::ostream &out ) const
  {
    OtterException::Print( out );
    out << "UnexpectedTokenException: " << m_got << " expected: " << m_expected;
  }
};


////////////////////////////////////////////////////////////////////////////////
//
// 2. Simple (otter-) Substition and Substitionlist
//
////////////////////////////////////////////////////////////////////////////////

class Substitution
{
public:
  Variable    *m_var;
  Term        *m_term;
};

class SubstitutionList : public std::vector<Substitution>
{
};


////////////////////////////////////////////////////////////////////////////////
//
// 3. otter proof object interferences classes (declaration only)
//
////////////////////////////////////////////////////////////////////////////////

/// This class is the base class of all other inferences class
class OtterInfer
{
public:
  /// polymorph
  virtual ~OtterInfer() {}

  /// virtual function to create a RuleItem form the inference.
  virtual RuleItem *MakeRuleItem( OtterRule *rule,
                                            OtterProofReader* reader,
                                            ProofDatabase *pdb );
private:
  static void MakeSequentHelper( Formula *f,
                          SElementList *l,
                          SElementList *r )
  {
    using namespace ProofLib;
  
    UnaryConFormula *uf = dynamic_cast<UnaryConFormula*>( f );
    
    if( uf != NULL && uf->GetOperatorSymbol()->IsNegJunctor() )
    {
      l->AddRight( new SFormula( uf->GetFormula() ) );
  
      return;
    }
  
    BinaryConFormula *bf = dynamic_cast<BinaryConFormula*>( f );
  
    if( bf != NULL && bf->GetOperatorSymbol()->IsOrJunctor() )
    {
      MakeSequentHelper( bf->GetLeft(), l, r );
      MakeSequentHelper( bf->GetRight(), l, r );
  
      return;
    }
  
    r->AddRight( new SFormula( f ) );
  }
  
protected:
  static Sequent* MakeSequent( Formula *f )
  {
    using namespace ProofLib;
  
  #if 0
    SElementList *r = new SElementList;
    r->AddRight( new SFormula( f ) );
    return new Sequent( new SElementList, r );
  #endif
    SElementList *l = new SElementList;
    SElementList *r = new SElementList;
  
    MakeSequentHelper( f, l, r );
  
    return new Sequent( l, r );
  }
};

////////////////////////////////////////////////////////////////////////////////

/// exists but not implemented (could not generate it with otter example files)
class OtterInferUndefined : public OtterInfer
{
};

////////////////////////////////////////////////////////////////////////////////

class OtterInferInput : public OtterInfer
{
public:
  virtual RuleItem *MakeRuleItem( OtterRule *rule,
                                            OtterProofReader* reader,
                                            ProofDatabase *pdb );
};

////////////////////////////////////////////////////////////////////////////////

/// exists but not implemented (could not generate it with otter example files)
class OtterInferEqAxiom : public OtterInfer
{
};

////////////////////////////////////////////////////////////////////////////////

class OtterInferInstantiate : public OtterInfer
{
public:
  size_t                 m_id;
  SubstitutionList       m_subst;

  virtual RuleItem *MakeRuleItem( OtterRule *rule,
                                            OtterProofReader* reader,
                                            ProofDatabase *pdb );
};

////////////////////////////////////////////////////////////////////////////////

class OtterInferPropositional : public OtterInfer
{
public:
  size_t                 m_id;

  virtual RuleItem *MakeRuleItem( OtterRule *rule,
                                            OtterProofReader* reader,
                                            ProofDatabase *pdb );
};

////////////////////////////////////////////////////////////////////////////////

class OtterInferResolve : public OtterInfer
{
public:
  size_t                 m_id_l;
  size_t                 m_id_r;

  std::vector<size_t>    m_subf_l;
  std::vector<size_t>    m_subf_r;

  virtual RuleItem *MakeRuleItem( OtterRule *rule,
                                            OtterProofReader* reader,
                                            ProofDatabase *pdb );
};

////////////////////////////////////////////////////////////////////////////////

class OtterInferParamod : public OtterInfer
{
public:
  size_t                 m_id_l;
  size_t                 m_id_r;

  std::vector<size_t>    m_subf_l;
  std::vector<size_t>    m_subf_r;

  virtual RuleItem *MakeRuleItem( OtterRule *rule,
                                            OtterProofReader* reader,
                                            ProofDatabase *pdb );
};

////////////////////////////////////////////////////////////////////////////////

class OtterInferFlip : public OtterInfer
{
public:
  size_t                 m_id;

  virtual RuleItem *MakeRuleItem( OtterRule *rule,
                                            OtterProofReader* reader,
                                            ProofDatabase *pdb );
};

////////////////////////////////////////////////////////////////////////////////

/// exists but not implemented (could not generate it with otter example files)
class OtterInferUnknown : public OtterInfer
{
};


////////////////////////////////////////////////////////////////////////////////
//
// 4. otter inferences container
//
////////////////////////////////////////////////////////////////////////////////

class OtterRule
{
public:
  size_t                 m_id;
  OtterInfer            *m_infer;
  Formula     *m_clause;
  bool                   m_done;

  OtterRule()
  {
    m_id = 0;
    m_infer = NULL;
    m_clause = NULL;
    m_done = false;
  }

  RuleItem *MakeRuleItem( OtterProofReader* r,
                                    ProofDatabase *pdb )
  {
    return m_infer->MakeRuleItem( this, r, pdb );
  }
};

////////////////////////////////////////////////////////////////////////////////

class OtterProofReader : public std::vector<OtterRule>
{
public:
  Proof* GenerateProof( ProofDatabase *pdb, int num )
  {
    ASSERT( pdb != NULL );

    wchar_t str[16];
    swprintf( str, 16, L"otter_%d", num );

    const Symbol *symbol =
      pdb->GetSymbolTable()->GetOrAdd( str, 0, SYMBOL_TYPE_PROOF );

    Proof *p = pdb->NewProof( symbol );

    if( p == NULL )
      throw new ProofExistsException( str );

    p->SetEndRule( (*this)[size()-1].MakeRuleItem( this, pdb ) );

    return p;
  }
};


////////////////////////////////////////////////////////////////////////////////
//
// 4. otter proof object interferences classes (implementation of MakeRuleItem)
//
////////////////////////////////////////////////////////////////////////////////

RuleItem* OtterInfer::MakeRuleItem(
  OtterRule *rule,
  OtterProofReader* reader,
  ProofDatabase *pdb )
{
  ASSERT( false );
  return NULL;
}

////////////////////////////////////////////////////////////////////////////////

RuleItem* OtterInferInput::MakeRuleItem(
  OtterRule *rule,
  OtterProofReader* reader,
  ProofDatabase *pdb )
{
  return MakeSequent( rule->m_clause );
}

////////////////////////////////////////////////////////////////////////////////

RuleItem* OtterInferResolve::MakeRuleItem(
  OtterRule *rule,
  OtterProofReader* reader,
  ProofDatabase *pdb )
{
  using namespace ProofLib;

  Sequent *s = MakeSequent( rule->m_clause );

  Rule *ru = new Rule;
  USTRING name;
  LatexConvertor.GetUnicode( "Res", name );
  ru->SetName( name );
  ru->SetConclusion( s );

  ru->AddRight( (*reader)[m_id_l-1].MakeRuleItem( reader, pdb ) );
  ru->AddRight( (*reader)[m_id_r-1].MakeRuleItem( reader, pdb ) );

  return ru;
}

////////////////////////////////////////////////////////////////////////////////

RuleItem* OtterInferParamod::MakeRuleItem(
  OtterRule *rule,
  OtterProofReader* reader,
  ProofDatabase *pdb )
{
  using namespace ProofLib;

  Sequent *s = MakeSequent( rule->m_clause );

  Rule *ru = new Rule;
  USTRING name;
  LatexConvertor.GetUnicode( "Paramod", name );
  ru->SetName( name );
  ru->SetConclusion( s );

  ru->AddRight( (*reader)[m_id_l-1].MakeRuleItem( reader, pdb ) );
  ru->AddRight( (*reader)[m_id_r-1].MakeRuleItem( reader, pdb ) );

  return ru;
}

////////////////////////////////////////////////////////////////////////////////

RuleItem* OtterInferInstantiate::MakeRuleItem(
  OtterRule *rule,
  OtterProofReader* reader,
  ProofDatabase *pdb )
{
  using namespace ProofLib;

  Sequent *s = MakeSequent( rule->m_clause );

  Rule *ru = new Rule;
  USTRING name;
  LatexConvertor.GetUnicode( "Inst", name );
  ru->SetName( name );
  ru->SetConclusion( s );

  ru->AddRight( (*reader)[m_id-1].MakeRuleItem( reader, pdb ) );

  return ru;
}

////////////////////////////////////////////////////////////////////////////////

RuleItem* OtterInferFlip::MakeRuleItem(
  OtterRule *rule,
  OtterProofReader* reader,
  ProofDatabase *pdb )
{
  using namespace ProofLib;

  Sequent *s = MakeSequent( rule->m_clause );

  Rule *ru = new Rule;
  USTRING name;
  LatexConvertor.GetUnicode( "Flip", name );
  ru->SetName( name );
  ru->SetConclusion( s );

  ru->AddRight( (*reader)[m_id-1].MakeRuleItem( reader, pdb ) );

  return ru;
}

////////////////////////////////////////////////////////////////////////////////

RuleItem* OtterInferPropositional::MakeRuleItem(
  OtterRule *rule,
  OtterProofReader* reader,
  ProofDatabase *pdb )
{
  using namespace ProofLib;

  Sequent *s = MakeSequent( rule->m_clause );

  Rule *ru = new Rule;
  USTRING name;
  LatexConvertor.GetUnicode( "Prop", name );
  ru->SetName( name );
  ru->SetConclusion( s );

  ru->AddRight( (*reader)[m_id-1].MakeRuleItem( reader, pdb ) );

  return ru;
}

////////////////////////////////////////////////////////////////////////////////
//
// 5. otter proof object reader class
//
////////////////////////////////////////////////////////////////////////////////

class ImportOtterProofImpl : public ImportOtterProof
{
  friend class ImportOtterProof;

  size_t                 m_lineno;
  size_t                 m_charno;

public:
  ImportOtterProofImpl( ProofDatabase* pdb )
    : ImportOtterProof( pdb )
  {
    m_lineno = 1;
    m_charno = 1;
  }

  size_t GetLineNo() const { return m_lineno; }
  size_t GetCharNo() const { return m_charno; }

  Proof* Read()
  {
    int p_num = 1;
    Proof* p1 = NULL;

    OtterProofReader reader;
  
    m_lineno = 1;
    m_charno = 1;

    size_t len = strlen( OTTER_PO_BEGIN );
  
    char *cmp = new char[len+1];
  
    while( (*m_in) )
    {
      m_in->get( cmp, len+1 );
  
      if( !strcmp( OTTER_PO_BEGIN, cmp ) )
      {
        EatWhiteSpaces();

        (*this) >> reader;
        
        Proof* p = reader.GenerateProof( m_pdb, p_num++ );

        if( p1 == NULL )
          p1 = p;
  
//        break;
      }
  
      char c;
  
      do
      {
        m_in->get( c );
      }
      while( (*m_in) && c != '\n' );
  
      m_lineno++;

      EatWhiteSpaces();
    }
  
    delete[] cmp;
  
    return p1;
  }

  void AssertChar( char c )
  {
    char r;

    is().get( r );

    if( r == '\n' )
    {
      m_lineno++;
      m_charno = 0;
    }
    else
      m_charno++;

    if( r != c )
      throw new UnexpectedCharException( m_lineno, m_charno, r, c );
  }

  void AssertString( const char *str )
  {
    size_t len = strlen( str );

    char *cmp = new char[len+1];

    ReadString( cmp, len+1 );

    if( strcmp( str, cmp ) != 0 )
      throw new UnexpectedTokenException( m_lineno, m_charno, cmp, str );

    delete cmp;
  }

  char CurrentChar()
  {
    char c;

    is().get( c );

    is().putback( c );

    return c;
  }

  bool IsWhiteChar( char c ) const
  {
    return ( c == ' ' || c == '\t' || c == '\r' || c == '\n' );
  }
  
  bool IsOtterStringChar( char c ) const
  {
    if( IsWhiteChar( c ) || c == '(' || c == ')' || c == '\'' || c == '\"' )
      return false;

    return true;
#if 0
    if( c >= 'a' && c <= 'z' )
      return true;
    if( c >= 'A' && c <= 'Z' )
      return true;
    if( c >= '0' && c <= '9' )
      return true;
    if( c == '-' || c == '+' || c == '=' || c == '*' )
      return true;

    return false;
#endif
  }

  void ReadString( char *str, size_t len )
  {
    size_t pos = 0;

    if( CurrentChar() == '\'' || CurrentChar() == '\"' )
    {
      char qtype = CurrentChar();
      char dummy;
      is().get( dummy );

      while( CurrentChar() != qtype )
      {
        if( len-- < 2 )
          break;

        is().get( str[pos++] );
        m_charno++;
      }

      is().get( dummy );
    }
    else while( IsOtterStringChar( CurrentChar() ) )
    {
      if( len-- < 2 )
        break;

      is().get( str[pos++] );
      m_charno++;
    }

    str[pos] = '\0';
  }

  void ReadInteger( size_t &i )
  {
    is() >> i;
    m_charno += is().gcount();
  }

  void EatWhiteSpaces()
  {
    while( true )
    {
      char c = CurrentChar();

      if( c != ' ' && c != '\t' && c != '\n' && c != '\r' )
        break;

      is().get( c );

      if( c == '\n' )
      {
        m_lineno++;
        m_charno = 0;
      }
      else
        m_charno++;
    }
  }

  /// throws UnknownInferenceException for reading (and streaming) functions
  /// adds line and char number.
  void ThrowUnknownInferenceException( const char* infer_name ) const
  {
    throw new UnknownInferenceException( m_lineno, m_charno, infer_name );
  }
};

////////////////////////////////////////////////////////////////////////////////

/// factory function for ImportOtterProofImpl
/// \return an ImportOtterProofImpl obj usable by the general interface
/// ImportOtterProof (user have to delete non null pointer)
ImportOtterProof* ImportOtterProof::New( ProofDatabase* pdb )
{
  return new ImportOtterProofImpl( pdb );
}

////////////////////////////////////////////////////////////////////////////////
//
// A. input stream functions
//
////////////////////////////////////////////////////////////////////////////////

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, Term** t )
{
  if( in.CurrentChar() != OTTER_OPEN_BRACKET )
  {
    in.EatWhiteSpaces();
    char token[OTTER_MAX_TOKEN_LEN];
    in.ReadString( token, OTTER_MAX_TOKEN_LEN );

#ifdef PRINT_DEBUG_INFO
    std::cout << token << '#' << std::endl;
#endif

    if( strlen( token ) < 1 )
      throw new UnexpectedCharException( in.GetLineNo(), in.GetCharNo(),
                                         in.CurrentChar() );

    if( strlen( token ) < OTTER_MAX_TOKEN_LEN - 3 )
    {
      char *p = &token[ strlen( token ) ];

      *(p+3) = '\0';
      *(p+2) = '}';
      p--;

      while( isdigit( *p ) )
      {
        *(p+2) = *p;
        p--;
      }

      *(p+1) = '_';
      *(p+2) = '{';
    }
  
    USTRING name;
    LatexConvertor.GetUnicode( token, name );

    const Symbol *symbol =
      in.GetSymbolTable()->GetOrAdd( name, 0, SYMBOL_TYPE_VARIABLE );

    *t = Variable::New( in.GetNodeDatabase(), symbol );
  }
  else
  {
    in.AssertChar( OTTER_OPEN_BRACKET );

    in.EatWhiteSpaces();
    char token[OTTER_MAX_TOKEN_LEN];
    in.ReadString( token, OTTER_MAX_TOKEN_LEN );

#ifdef PRINT_DEBUG_INFO
    std::cout << token << '#' << std::endl;
#endif
  
    in.EatWhiteSpaces();
    if( in.CurrentChar() == OTTER_CLOSE_BRACKET )
    {
      USTRING name;
      LatexConvertor.GetUnicode( token, name );

      const Symbol *symbol =
        in.GetSymbolTable()->GetOrAdd( name, 0, SYMBOL_TYPE_CONSTANT );

      *t = Constant::New( in.GetNodeDatabase(), symbol );
    }
    else
    {
      TermList l;

      in >> l;

      USTRING name;
      LatexConvertor.GetUnicode( token, name );

      const Symbol *symbol =
        in.GetSymbolTable()->GetOrAdd( name, l.GetChildCount(),
                                       SYMBOL_TYPE_FUNCTION );

      *t = Function::New( in.GetNodeDatabase(), symbol, &l );
    }

    in.EatWhiteSpaces();
    in.AssertChar( OTTER_CLOSE_BRACKET );
  }

  ASSERT( t != NULL );

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, TermList& l )
{
  while( in.CurrentChar() != OTTER_CLOSE_BRACKET )
  {
    Term *t;

    in >> &t;

    l.AddRight( t );

    in.EatWhiteSpaces();
  }

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, Formula** f )
{
  if( in.CurrentChar() != OTTER_OPEN_BRACKET )
  {
    in.AssertString( OTTER_BOX );

    TermList l;
    USTRING name;
    LatexConvertor.GetUnicode( "\\box", name );

    const Symbol *symbol =
      in.GetSymbolTable()->GetOrAdd( name, l.GetChildCount(),
                                     SYMBOL_TYPE_PREDICATE );

    *f = AtomFormula::New( in.GetNodeDatabase(), symbol, &l );

    return in;
  }

  in.AssertChar( OTTER_OPEN_BRACKET );

  in.EatWhiteSpaces();
  char token[OTTER_MAX_TOKEN_LEN];
  in.ReadString( token, OTTER_MAX_TOKEN_LEN );

#ifdef PRINT_DEBUG_INFO
  std::cout << token << '#' << std::endl;
#endif

  if( strlen( token ) < 1 )
    throw new UnexpectedCharException( in.GetLineNo(), in.GetCharNo(),
                                       in.CurrentChar() );

  if( !strcmp( token, OTTER_OR ) )
  {
    Formula *l = NULL, *r = NULL;

    in.EatWhiteSpaces();
    in >> &l;
    
    in.EatWhiteSpaces();
    in >> &r;
    
    const OperatorSymbol *opsym =
      in.GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_OR );
    ASSERT( opsym != NULL );

    *f = BinaryConFormula::New( in.GetNodeDatabase(), opsym, l, r );
  }
  else if( !strcmp( token, OTTER_NOT ) )
  {
    Formula *l = NULL;

    in.EatWhiteSpaces();
    in >> &l;
    
    const OperatorSymbol *opsym =
      in.GetSymbolTable()->GetOperatorSymbol( OP_SYMBOL_NEG );
    ASSERT( opsym != NULL );

    *f = UnaryConFormula::New( in.GetNodeDatabase(), opsym, l );
  }
  else
  {
    TermList l;

    in.EatWhiteSpaces();
    in >> l;

    USTRING name;
    LatexConvertor.GetUnicode( token, name );

    const Symbol *symbol =
      in.GetSymbolTable()->GetOrAdd( name, l.GetChildCount(),
                                     SYMBOL_TYPE_PREDICATE );

    *f = AtomFormula::New( in.GetNodeDatabase(), symbol, &l );
  }


  ASSERT( *f != NULL );

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_CLOSE_BRACKET );

  return in;
}

////////////////////////////////////////////////////////////////////////////////

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, Substitution& s ) 
{
  in.AssertChar( OTTER_OPEN_BRACKET );

  Term *t = NULL;
  Variable *v = NULL;
  
  in.EatWhiteSpaces();
  in >> &t;
  v = dynamic_cast<Variable*>( t );
  ASSERT( v != NULL );

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_DOT );

  in.EatWhiteSpaces();
  in >> &t;

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_CLOSE_BRACKET );

  s.m_var = v;
  s.m_term = t;

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, SubstitutionList& l ) 
{
  in.AssertChar( OTTER_OPEN_BRACKET );

  while( true )
  {
    in.EatWhiteSpaces();
    if( in.CurrentChar() != OTTER_OPEN_BRACKET )
      break;

    l.resize( l.size() + 1 );

    in >> l[l.size() -1];
  }

  in.AssertChar( OTTER_CLOSE_BRACKET );

  return in;
}

////////////////////////////////////////////////////////////////////////////////

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterInferUndefined& i )
{
  ASSERT( false ); // TODO: implement!

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterInferInput& i ) 
{
  // no parameters
  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterInferEqAxiom& i ) 
{
  ASSERT( false ); // TODO: implement!

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterInferInstantiate& i ) 
{
  in.EatWhiteSpaces();
  in.ReadInteger( i.m_id );

  in.EatWhiteSpaces();
  in >> i.m_subst;

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterInferPropositional& i ) 
{
  in.EatWhiteSpaces();
  in.ReadInteger( i.m_id );

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterInferResolve& i ) 
{
  in.EatWhiteSpaces();
  in.ReadInteger( i.m_id_l );

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_OPEN_BRACKET );

  in.EatWhiteSpaces();
  while( in.CurrentChar() != OTTER_CLOSE_BRACKET )
  {
    size_t x;
    in.ReadInteger( x );
    i.m_subf_l.push_back( x );
    in.EatWhiteSpaces();
  }

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_CLOSE_BRACKET );

  in.EatWhiteSpaces();
  in.ReadInteger( i.m_id_r );

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_OPEN_BRACKET );

  in.EatWhiteSpaces();
  while( in.CurrentChar() != OTTER_CLOSE_BRACKET )
  {
    size_t x;
    in.ReadInteger( x );
    i.m_subf_r.push_back( x );
    in.EatWhiteSpaces();
  }

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_CLOSE_BRACKET );

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterInferParamod& i ) 
{
  in.EatWhiteSpaces();
  in.ReadInteger( i.m_id_l );

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_OPEN_BRACKET );

  in.EatWhiteSpaces();
  while( in.CurrentChar() != OTTER_CLOSE_BRACKET )
  {
    size_t x;
    in.ReadInteger( x );
    i.m_subf_l.push_back( x );
    in.EatWhiteSpaces();
  }

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_CLOSE_BRACKET );

  in.EatWhiteSpaces();
  in.ReadInteger( i.m_id_r );

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_OPEN_BRACKET );

  in.EatWhiteSpaces();
  while( in.CurrentChar() != OTTER_CLOSE_BRACKET )
  {
    size_t x;
    in.ReadInteger( x );
    i.m_subf_r.push_back( x );
    in.EatWhiteSpaces();
  }

  in.AssertChar( OTTER_CLOSE_BRACKET );

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterInferFlip& i ) 
{
  in.EatWhiteSpaces();
  in.ReadInteger( i.m_id );

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_OPEN_BRACKET );

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_CLOSE_BRACKET );

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterInferUnknown& i ) 
{
  ASSERT( false ); // TODO: implement!

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterInfer** i )
{
  in.AssertChar( OTTER_OPEN_BRACKET );

  in.EatWhiteSpaces();
  char infer_name[32];
  in.ReadString( infer_name, 32 );

//  std::cout << infer_name << '#' << std::endl;

  if( !strcmp( infer_name, OTTER_INFER_UNDEFINED ) )
  {
    OtterInferUndefined *a = new OtterInferUndefined;
    in >> *a; 
    *i = a;
  }
  else if( !strcmp( infer_name, OTTER_INFER_INPUT ) )
  {
    OtterInferInput *a = new OtterInferInput;
    in >> *a; 
    *i = a;
  }
  else if( !strcmp( infer_name, OTTER_INFER_EQ_AXIOM ) )
  {
    OtterInferEqAxiom *a = new OtterInferEqAxiom;
    in >> *a; 
    *i = a;
  }
  else if( !strcmp( infer_name, OTTER_INFER_INSTANTIATE ) )
  {
    OtterInferInstantiate *a = new OtterInferInstantiate;
    in >> *a; 
    *i = a;
  }
  else if( !strcmp( infer_name, OTTER_INFER_PROPOSITIONAL ) )
  {
    OtterInferPropositional *a = new OtterInferPropositional;
    in >> *a; 
    *i = a;
  }
  else if( !strcmp( infer_name, OTTER_INFER_RESOLVE ) )
  {
    OtterInferResolve *a = new OtterInferResolve;
    in >> *a; 
    *i = a;
  }
  else if( !strcmp( infer_name, OTTER_INFER_PARAMOD ) )
  {
    OtterInferParamod *a = new OtterInferParamod;
    in >> *a; 
    *i = a;
  }
  else if( !strcmp( infer_name, OTTER_INFER_FLIP ) )
  {
    OtterInferFlip *a = new OtterInferFlip;
    in >> *a; 
    *i = a;
  }
  else if( !strcmp( infer_name, OTTER_INFER_UNKNOWN ) )
  {
    OtterInferUnknown *a = new OtterInferUnknown;
    in >> *a; 
    *i = a;
  }
  else
    in.ThrowUnknownInferenceException( infer_name );

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_CLOSE_BRACKET );

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterRule& r )
{
  in.AssertChar( OTTER_OPEN_BRACKET );

  in.EatWhiteSpaces();
  in.ReadInteger( r.m_id );

  in.EatWhiteSpaces();
  in >> &(r.m_infer);

  in.EatWhiteSpaces();
  in >> &(r.m_clause);

  in.EatWhiteSpaces();
  if( in.CurrentChar() != OTTER_OPEN_BRACKET )
    in.AssertString( OTTER_NIL );
  else
  {
    in.AssertChar( OTTER_OPEN_BRACKET );

    in.EatWhiteSpaces();
    size_t dummy;
    in.ReadInteger( dummy );

    in.EatWhiteSpaces();
    in.AssertChar( OTTER_CLOSE_BRACKET );
  }

  in.EatWhiteSpaces();
  in.AssertChar( OTTER_CLOSE_BRACKET );

  return in;
}

ImportOtterProofImpl& operator >> ( ImportOtterProofImpl &in, OtterProofReader& opr )
{
  in.AssertChar( OTTER_OPEN_BRACKET );
  
  do
  {
    in.EatWhiteSpaces();

    if( in.CurrentChar() != OTTER_OPEN_BRACKET )
      break;

    opr.resize( opr.size()+1 );

    in >> opr[ opr.size()-1 ];
  }
  while( true );

  in.AssertChar( OTTER_CLOSE_BRACKET );

  return in;
}

}
