
//
// Proof Library for C++ -------------------------------------------------------
//
// File: libproof.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Datastructures for Proofs
////////////////////////////////////////////////////////////////////////////////
#ifndef _LIBPROOF_H_
#define _LIBPROOF_H_

#include <vector>
#include <set>

#include "util/hash.h"
#include "util/exceptions.h"
#include "symboltable.h"
#include "operations_def.h"

#include <iostream>

namespace HLK
{
  class SymbolTable;
}


namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

class Formula;
class FormulaList;
class FormulaListVariable;
class SElementList;
class Sequent;
class Proof;
class ProofDatabase;
class NodeDatabase;
class StaticTermList;
class StaticAbstractTermList;
class LambdaTerm;
class Variable;

class Operation;
class OperationCheckEquality;

typedef std::pair< Variable*, Variable* > VarMap;
typedef std::vector< VarMap > VarMapping;


////////////////////////////////////////////////////////////////////////////////
/// Base class for _all_ nodes in the data tree
///
////////////////////////////////////////////////////////////////////////////////

class DataNode
{
private:
#ifdef DEBUG
  static size_t          m_dbg_new_count;
#endif

public:
  DataNode();
#ifdef DEBUG
  virtual ~DataNode();
#else
  virtual ~DataNode() {};
#endif

  /// applies an operation to this datanode and to its subnodes.
  DECLARE_APPLY_ABSTRACT

  /// applies an operation to a datanode and to its subnodes and cast return
  /// type to its type.
  template<class T> static T* ApplyTo( T* node, Operation &op );

  /// prints information about a node
  virtual std::ostream& ToString( std::ostream &out ) const = 0;

  /// check this node and all its subnodes for having the given pointer.
  bool ContainsPointer( const DataNode *o ) const;

  template<class T> const T* CastTo() const;
  template<class T> T* CastTo();

#ifdef DEBUG
  /// NEVER USE THIS FUNCTION for type identification or something else,
  /// it's just for debug printing!
  const char *DEBUGGetTypeName() const;
  void* operator new ( size_t size );
  void operator delete ( void *ptr );
#endif
};

template<class T> T* DataNode::ApplyTo( T* node, Operation &op )
{
  return dynamic_cast<T*>( node->Apply( op ) );
}

template<class T> T* DataNode::CastTo()
{
  return dynamic_cast<T*>( this );
}

template<class T> const T* DataNode::CastTo() const
{
  return dynamic_cast<const T*>( this );
}

////////////////////////////////////////////////////////////////////////////////
/// Additional base class for all internal nodes in the data tree
///
////////////////////////////////////////////////////////////////////////////////

class TreeInternalNode
{
public:
  virtual ~TreeInternalNode() {}
  
  virtual size_t GetChildCount() const = 0;
  virtual DataNode* GetChild( size_t index ) const = 0;
};

////////////////////////////////////////////////////////////////////////////////
/// Base class for all nodes that are stored in the pointer database.
///
////////////////////////////////////////////////////////////////////////////////

/// \todo eliminate copy constructor, copy operator
class DbNode
{
private:
  NodeDatabase          *m_node_db;
  size_t                 m_ref_count;

  friend class NodeDatabase;

protected:
  DbNode() { m_node_db = NULL; m_ref_count = 0; }

protected:
  virtual ~DbNode() {};

public:
  /// \return hash value for the db node.
  virtual size_t Hash() const = 0;

  /// \return true if n points to a node that would have the same entry in the
  ///         database (an internal node (one with children) should point
  ///         already to children that are already contained in the database;
  ///         false otherwise
  virtual bool DbEqual( const DbNode *n ) const { return DbEqual( n, VarMapping() ); }

  /// the same as DbEqual, but with a variable map to check modulo
  /// first order alpha equality
  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const = 0;

  /// this function will increase the reference count.
  void Attach() { m_ref_count++; }

  /// this function will decrease the reference count. If the reference count
  /// results to 0, the node will be deleted.
  void Detach();

  NodeDatabase* GetNodeDatabase() const { return m_node_db; }

#ifdef DEBUG
  size_t GetRefCount() const { return m_ref_count; }
#endif

protected:
  /// This function is called by DbNode classes static New function
  /// to add the class to the node database.
  template<class T> T* AddToNodeDatabase( NodeDatabase *ndb );
};

////////////////////////////////////////////////////////////////////////////////

} // namespace ProofLib


namespace stlhash
{
  template<> struct hash<const ProofLib::DbNode*> {
    size_t operator()(const ProofLib::DbNode* x) const { return x->Hash(); }
  };
  template<> struct hash<ProofLib::DbNode*> {
    size_t operator()(ProofLib::DbNode* x) const { return x->Hash(); }
  };
}

namespace std
{
  template<> struct equal_to<const ProofLib::DbNode*> {
    size_t operator()(const ProofLib::DbNode* x,const ProofLib::DbNode* y) const
      { 
	return x->DbEqual( y ); 
      }
  };
  template<> struct equal_to<ProofLib::DbNode*> {
    size_t operator()(ProofLib::DbNode* x,const ProofLib::DbNode* y) const
      { 
	return x->DbEqual( y ); 
      }  };
}

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class Expression : public DbNode
{
public:
  virtual ~Expression() {};
};

////////////////////////////////////////////////////////////////////////////////

class AbstractTerm : public Expression, public DataNode
{
public:
  virtual std::ostream& ToString( std::ostream &out ) const = 0;
};

////////////////////////////////////////////////////////////////////////////////

class SetTerm : public AbstractTerm
{

};


////////////////////////////////////////////////////////////////////////////////

class NodePointer : private std::vector<size_t>
{
public:
  void Push( size_t n );
  size_t Pop();

  void Clear() { clear(); }

  size_t GetLength() const { return size(); }
  bool IsEmpty() const { return empty(); }

  DataNode* GetDataNode( DataNode *root ) const;
  template<class T> T* GetTypedDataNode( DataNode *root ) const;

  void Dump( std::ostream &out ) const;

  size_t operator [] ( size_t index ) const;
  bool operator == ( const NodePointer &other ) const;
  bool operator != ( const NodePointer &other ) const;
  bool operator < ( const NodePointer &other ) const;
  bool operator > ( const NodePointer &other ) const;
  bool operator <= ( const NodePointer &other ) const;
  bool operator >= ( const NodePointer &other ) const;

  NodePointer operator + ( const NodePointer &other ) const;
  NodePointer operator + ( size_t index ) const;

  NodePointer GetCommonBeginning( const NodePointer &o ) const;

private:
  DataNode* GetDataNode( DataNode *root, size_t depth ) const;
};

template<class T> T* NodePointer::GetTypedDataNode( DataNode *root ) const
{
  return dynamic_cast<T*>( GetDataNode( root ) );
}

class NoSuchNodeException : public Exception
{
private:
  const NodePointer     *m_np;

public:
  NoSuchNodeException( const NodePointer *np );

  virtual void Print( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////

class NodePointers : private std::vector< NodePointer >
{
public:
  virtual ~NodePointers() {}

  virtual void Add( const NodePointer &np ) { push_back( np ); }

  virtual const NodePointer& operator [] ( size_t index ) const;

  virtual size_t GetCount() const { return size(); }
  virtual int Find( const NodePointer &np ) const;
  virtual bool Contains( const NodePointer &np ) const { return Find(np) >= 0; }

  virtual bool IsAboveAny( const NodePointer &np ) const;
};

////////////////////////////////////////////////////////////////////////////////
/// Base Class for Nodes carring Symbols
///
/// base class for atom fomulas (predicate symbol), functions (function symbol),
/// etc.
////////////////////////////////////////////////////////////////////////////////

class SymbolNode
{
private:
  /// pointer to the symbol table entry.
  const Symbol      *m_symbolentry;

public:
  /// set m_symbolentry to an exsiting or newly created ymbol table entry.
  SymbolNode( const Symbol* e );
  virtual ~SymbolNode() {}

  /// \return a pointer to the symbol table entry
  const Symbol *GetSymbol() const { return m_symbolentry; };

  const USTRING& GetName() const { return m_symbolentry->GetName(); }
  size_t GetArity() const { return m_symbolentry->GetArity(); }
  SymbolType GetType() const; // { return m_symbolentry->GetType(); }

  virtual std::ostream& ToString( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////
/// Base Class for Nodes carring Symbols that are stored in the pointer database
///
////////////////////////////////////////////////////////////////////////////////

class DbSymbolNode : public SymbolNode
{
public:
  DbSymbolNode( const Symbol *e );
  virtual ~DbSymbolNode() {}

  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const;
};


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class Term : public AbstractTerm
{
};

class AbstractTermListBase : public TreeInternalNode
{
public:
  virtual AbstractTerm* GetTerm( size_t index ) const = 0;
  virtual int FindTerm( const AbstractTerm *t ) const = 0;

  virtual bool IsEmpty() const = 0;

  virtual std::ostream& ToString( std::ostream &out ) const;
};

class AbstractTermList : public AbstractTermListBase, private std::vector<class AbstractTerm*>
{
public:
  AbstractTermList() {};
  AbstractTermList( size_t arity );
  AbstractTermList( const StaticAbstractTermList& stl );
  virtual ~AbstractTermList() {};

  void AddLeft( AbstractTerm *t );
  void AddRight( AbstractTerm *t );
  AbstractTerm* GetTerm( size_t index ) const;

  void SetTerms( AbstractTermList *other );

  virtual int FindTerm( const AbstractTerm *t ) const;

  void RemoveAt( size_t index );

  virtual bool IsEmpty() const { return GetChildCount() == 0; }

  virtual size_t Hash() const;
  // TreeInternalNode overrides
  virtual size_t GetChildCount() const;
  virtual DataNode* GetChild( size_t index ) const { return GetTerm( index ); }
};

class StaticAbstractTermList : public AbstractTermListBase
{
private:
  AbstractTerm           **m_terms;
  size_t                 m_arity;

protected:
  void SetTerms( const AbstractTermListBase *tl );

public:
  StaticAbstractTermList( size_t arity );
  StaticAbstractTermList( const AbstractTermListBase* stl );
  virtual ~StaticAbstractTermList();

  AbstractTerm* GetTerm( size_t index ) const;

  virtual int FindTerm( const AbstractTerm *t ) const;

  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const;
  virtual size_t Hash() const;

  virtual bool IsEmpty() const { return m_arity == 0; }

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return m_arity; }
  virtual DataNode* GetChild( size_t index ) const { return GetTerm( index ); }
};

class TermList : public AbstractTermList
{
public:
  TermList() {};
  TermList( const StaticTermList& stl );
  virtual ~TermList() {};

  void AddLeft( Term *t );
  void AddRight( Term *t );
  Term* GetTerm( size_t index ) const;

  void SetTerms( TermList *other );

  virtual int FindTerm( const Term *t ) const;
};

class StaticTermList : public StaticAbstractTermList
{
protected:
  void SetTerms( const TermList *tl );
  void SetTerms( const StaticTermList *tl );
public:
  StaticTermList( const TermList *terms );
  StaticTermList( const StaticTermList *terms );
  StaticTermList( size_t arity );
  virtual ~StaticTermList();

  Term* GetTerm( size_t index ) const;

  virtual int FindTerm( const Term *t ) const;
};

class SecondOrderVariable : public DbSymbolNode, public SetTerm
{
protected:
  SecondOrderVariable( const Symbol *e );
public:
  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const
    { return DbSymbolNode::DbEqual( n, var_map ); }
  virtual size_t Hash() const;

  virtual std::ostream& ToString( std::ostream &out ) const;

//  LambdaTerm* getEtaNF( NodeDatabase *ndb, HLK::SymbolTable* symtbl ) const;

  DECLARE_APPLY

public:
  static SecondOrderVariable* New( NodeDatabase *ndb, const Symbol *e );
};

class Variable : public Term, public DbSymbolNode
{
protected:
  Variable( const Symbol *e );

public:
  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const
    { 
#if 0
      // check modulo alpha equality
      std::cerr << "checking dbequal for variable: ";
      ToString(std::cerr);
      std::cerr << std::endl;
#endif
      for ( size_t i = 0; i < var_map.size(); i++ )
      {
	Variable *v = var_map[i].first;
#if 0
	std::cerr << "comparing variable: ";
	v->ToString(std::cerr);
	std::cerr << std::endl;
#endif
	if ( this == v )
	{
#if 0
	  std::cerr << "found in var map!" << std::endl;
#endif
	  return var_map[i].second->DbEqual( n, VarMapping() );
	}
      }
      return DbSymbolNode::DbEqual( n, var_map );
    }
  virtual size_t Hash() const;

  DECLARE_APPLY

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static Variable* New( NodeDatabase *ndb, const Symbol *e );
};

class Constant : public Term, public DbSymbolNode
{
protected:
  Constant( const Symbol *e );

public:
  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const
    { return DbSymbolNode::DbEqual( n, var_map ); }
  virtual size_t Hash() const;

  DECLARE_APPLY

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static Constant* New( NodeDatabase *ndb, const Symbol *e );
};

class Function : public Term, public DbSymbolNode, public StaticAbstractTermList
{
protected:
  Function( const Symbol *e, int arity );

public:
  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const;
  virtual size_t Hash() const;

  DECLARE_APPLY

  virtual Function* ReplaceTerms( AbstractTermList &terms ) const;

  virtual bool HaveEqualSymbol( const Function *o ) const;

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static Function* New( NodeDatabase *ndb, const Symbol *e,
                        AbstractTermList *terms );
};


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class FormulaListInsertable : public DataNode
{
public:
  virtual void InsertIntoFormulaList( FormulaList *list, size_t index ) = 0;
  virtual void InsertIntoSElementList( SElementList *list, size_t index ) = 0;
};

////////////////////////////////////////////////////////////////////////////////

class FormulaListElement : public FormulaListInsertable
{
};

////////////////////////////////////////////////////////////////////////////////

class Formula : public FormulaListElement, public Expression
{
public:
  virtual void InsertIntoFormulaList( FormulaList *list, size_t index );
  virtual void InsertIntoSElementList( SElementList *list, size_t index );
};

}

namespace stlhash
{
  template<> struct hash<const ProofLib::Formula*> {
    size_t operator()(const ProofLib::Formula* x) const { return x->Hash(); }
  };
  template<> struct hash<ProofLib::Formula*> {
    size_t operator()(ProofLib::Formula* x) const { return x->Hash(); }
  };
}

namespace ProofLib
{


////////////////////////////////////////////////////////////////////////////////
class AtomFormula : public Formula, public DbSymbolNode, public StaticAbstractTermList
{
private:
  SecondOrderVariable* m_var;

protected:
  AtomFormula( const Symbol *e, int arity, SecondOrderVariable* var = NULL );

public:
  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const;
  virtual size_t Hash() const;

  DECLARE_APPLY

  virtual AtomFormula* ReplaceTerms( AbstractTermList &terms ) const;

  virtual std::ostream& ToString( std::ostream &out ) const;

  SecondOrderVariable* GetSecondOrderVariable() const;

public:
  static AtomFormula* New( NodeDatabase *ndb, const Symbol *e,
                           AbstractTermList *terms, SecondOrderVariable* var = NULL );
};

////////////////////////////////////////////////////////////////////////////////

class DefinedSet : public SetTerm, public DbSymbolNode, public StaticAbstractTermList
{
protected:
  DefinedSet( const Symbol *e, int arity );
public:
  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const;
  virtual size_t Hash() const;
  DECLARE_APPLY
  virtual std::ostream& ToString( std::ostream &out ) const;
  virtual DefinedSet* ReplaceTerms( AbstractTermList &terms ) const;
  static DefinedSet* New( NodeDatabase *ndb, const Symbol *e, AbstractTermList *terms );
};
////////////////////////////////////////////////////////////////////////////////

class DefinedSetFormula : public Formula, public StaticTermList
{
private:
  DefinedSet* m_defined_set;

protected:
  DefinedSetFormula( DefinedSet *d, TermList *t );

  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const;
  virtual size_t Hash() const;

  DECLARE_APPLY

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static DefinedSetFormula* New( NodeDatabase *ndb, DefinedSet *d,
                                 TermList *t );
  // TreeInternalNode overrides
  virtual size_t GetChildCount() const;
  virtual DataNode* GetChild( size_t index ) const;

  DefinedSet* getDefinedSet() const { return m_defined_set; }

  virtual ~DefinedSetFormula() {};
};
////////////////////////////////////////////////////////////////////////////////

class OperatorFormula : public Formula, public TreeInternalNode
{
private:
  const OperatorSymbol  *m_sym;

protected:
  OperatorFormula( const OperatorSymbol *sym );

public:
  const OperatorSymbol* GetOperatorSymbol() const { return m_sym; }

  // DbNode overrides
  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const;
};

////////////////////////////////////////////////////////////////////////////////

class UnaryConFormula : public OperatorFormula
{
private:
  Formula               *m_formula;

protected:
  UnaryConFormula( const OperatorSymbol *sym, Formula *f );

public:
  DECLARE_APPLY

  virtual Formula* GetFormula() const { return m_formula; }

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return 1; }
  virtual DataNode* GetChild( size_t index ) const;

  virtual size_t Hash() const;

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static UnaryConFormula* New( NodeDatabase *ndb, const OperatorSymbol *sym,
                               Formula *f );
};

////////////////////////////////////////////////////////////////////////////////

class QuantorFormula : public OperatorFormula
{
private:
  Variable              *m_variable;
  Formula               *m_formula;

protected:
  QuantorFormula( const OperatorSymbol *sym,
                  Variable *v, Formula *f );

public:
  DECLARE_APPLY

  virtual Formula* GetFormula() const { return m_formula; }
  virtual Variable* GetVariable() const { return m_variable; }

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return 2; }
  virtual DataNode* GetChild( size_t index ) const;

  virtual size_t Hash() const;

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static QuantorFormula* New( NodeDatabase *ndb, const OperatorSymbol *sym,
                              Variable *v, Formula *f );
};

////////////////////////////////////////////////////////////////////////////////

class SecondOrderQuantorFormula : public OperatorFormula
{
private:
  SecondOrderVariable   *m_variable;
  Formula               *m_formula;

protected:
  SecondOrderQuantorFormula( const OperatorSymbol *sym,
                  SecondOrderVariable *v, Formula *f );

public:
  DECLARE_APPLY

  virtual Formula* GetFormula() const { return m_formula; }
  virtual SecondOrderVariable* GetSecondOrderVariable() const { return m_variable; }

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return 1; }
  virtual DataNode* GetChild( size_t index ) const;

  virtual size_t Hash() const;

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static SecondOrderQuantorFormula* New( NodeDatabase *ndb, const OperatorSymbol *sym,
                              SecondOrderVariable *v, Formula *f );
};


////////////////////////////////////////////////////////////////////////////////

class BinaryConFormula : public OperatorFormula
{
private:
  Formula               *m_left;
  Formula               *m_right;

protected:
  BinaryConFormula( const OperatorSymbol *sym,
                    Formula *fl, Formula *fr );

public:
  DECLARE_APPLY

  virtual Formula* GetLeft() const { return m_left; }
  virtual Formula* GetRight() const { return m_right; }

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return 2; }
  virtual DataNode* GetChild( size_t index ) const;

  virtual size_t Hash() const;

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static BinaryConFormula* New( NodeDatabase *ndb, const OperatorSymbol *sym,
                                Formula *fl, Formula *fr );
};

////////////////////////////////////////////////////////////////////////////////

class FormulaVariable : public Formula, public DbSymbolNode,
                        public TreeInternalNode
{
private:
  Formula               *m_formula;

protected:
  FormulaVariable( const Symbol *e );

  void SetFormula( Formula *f );

public:
  DECLARE_APPLY

  Formula* GetFormula() const;

  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const;
  virtual size_t Hash() const;

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return 1; }
  virtual DataNode* GetChild( size_t index ) const;

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static FormulaVariable* New( NodeDatabase *ndb,
                               const Symbol *e, Formula *f );
};

////////////////////////////////////////////////////////////////////////////////

class FormulaList : public FormulaListElement,
                    private std::vector<FormulaListElement*>,
                    public TreeInternalNode
{
public:
  void AddLeft( FormulaListElement *f );
  void AddLeft( FormulaList *o );
  void AddRight( FormulaListElement *f );
  void AddRight( FormulaList *o );

  void InsertAt( size_t index, FormulaListElement *e );
  void InsertAt( size_t index, const FormulaList *l );

  FormulaListElement* GetElement( size_t index ) const;

  DECLARE_APPLY

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return size(); }
  virtual DataNode* GetChild( size_t index ) const
    { return GetElement( index ); }

  virtual void InsertIntoFormulaList( FormulaList *list, size_t index );
  virtual void InsertIntoSElementList( SElementList *list, size_t index );

  virtual std::ostream& ToString( std::ostream &out ) const;

private:
  bool FindSubListAtIndex( const FormulaList *list, size_t &index ) const;
  bool IsSubListAtIndex( const FormulaList *list, size_t index ) const;
  void ReplaceFormulas( FormulaListVariable *fv, size_t start_index );
  void UnsubstFormulaListVar( FormulaListVariable *fv, size_t index );
};

////////////////////////////////////////////////////////////////////////////////

class FormulaListVariable : public DbNode, public DbSymbolNode,
                            public FormulaListElement,
                            public TreeInternalNode
{
  FormulaList *list;

protected:
  FormulaListVariable( const Symbol *e );
  virtual ~FormulaListVariable();

  friend class NodeDatabase;

public:
  virtual bool DbEqual( const DbNode *n, VarMapping var_map ) const;
  virtual size_t Hash() const;

  DECLARE_APPLY

  const FormulaList* GetFormulaList() const { return list; }

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return ( list == NULL ? 0 : 1 ); }
  virtual DataNode* GetChild( size_t index ) const;

  virtual void InsertIntoFormulaList( FormulaList *list, size_t index );
  virtual void InsertIntoSElementList( SElementList *list, size_t index );

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static FormulaListVariable* New( NodeDatabase *ndb,
                                   const Symbol *e, FormulaList *l );
};

///////////////////////////////////////////////////////////////////////////////

class LambdaTerm : public SetTerm, public TreeInternalNode
{
  Formula *m_form;
  TermList m_vars;
  
protected:
  LambdaTerm( const TermList &v, const Formula *f );

public:
  Formula *getFormula() const { return m_form; }
  const TermList getVariables() const { return m_vars; }

  Formula *apply( ProofDatabase* pdb, const TermList &tv ) const;

  // DbNode overrides
  virtual size_t Hash() const;
  virtual bool DbEqual(const DbNode *n, VarMapping var_map) const;

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return m_vars.GetChildCount() + 1; }
  virtual DataNode* GetChild(size_t i) const
  {
  	if ( i < m_vars.GetChildCount() )
  	{
  	  return m_vars.GetTerm(i);
  	}
  	else
  	{
  	  return m_form;
  	}
  }
  
  static LambdaTerm* New( NodeDatabase *ndb, const TermList &v, const Formula *f );
  
  virtual std::ostream& ToString( std::ostream &out ) const;
  
  DECLARE_APPLY
};
///////////////////////////////////////////////////////////////////////////////

class SElement : public TreeInternalNode,
                 public FormulaListInsertable
{
public:
  virtual SElement* Clone() const = 0;
};

///////////////////////////////////////////////////////////////////////////////

class SFormula : public SElement
{
private:
  Formula               *m_formula;

public:
  SFormula( Formula* f );
  virtual ~SFormula();

  Formula* GetFormula() const { return m_formula; }
//  void SetFormula( Formula *f );

  DECLARE_APPLY

  virtual SElement* Clone() const;

  virtual size_t GetChildCount() const { return 1; }
  virtual DataNode* GetChild( size_t index ) const;

  virtual void InsertIntoFormulaList( FormulaList *list, size_t index );
  virtual void InsertIntoSElementList( SElementList *list, size_t index );

  virtual std::ostream& ToString( std::ostream &out ) const
    { return m_formula->ToString(out); }
};

///////////////////////////////////////////////////////////////////////////////

class SFormulaListVariable : public SElement
{
private:
  FormulaListVariable   *m_formula_list_var;

public:
  SFormulaListVariable( FormulaListVariable *fv );
  virtual ~SFormulaListVariable();

  FormulaListVariable* GetFormulaListVariable() const
    { return m_formula_list_var; }

  DECLARE_APPLY

  virtual size_t GetChildCount() const { return 1; }
  virtual DataNode* GetChild( size_t index ) const;

  virtual SElement* Clone() const;

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  virtual void InsertIntoFormulaList( FormulaList *list, size_t index );
  virtual void InsertIntoSElementList( SElementList *list, size_t index );
};

///////////////////////////////////////////////////////////////////////////////

class SElementListEventClient
{
public:
  virtual ~SElementListEventClient() {}
  virtual void SElementListUpdateEvent( const SElementList *list ) = 0;
};

class SElementListNotifier
{
  std::set<SElementListEventClient*> m_clients;

public:
  virtual ~SElementListNotifier() {}

  void AttachSElementListEventClient( SElementListEventClient *client );
  void DetachSElementListEventClient( SElementListEventClient *client );

  void FireUpdateEvent() const;
};

///////////////////////////////////////////////////////////////////////////////

class SElementList : public DataNode, private std::vector<SElement*>,
                     public TreeInternalNode,
                     public SElementListNotifier
{
public:
  virtual ~SElementList();

  void AddLeft( SElement *f );
  void AddLeft( SElementList *o );
  void AddLeft( FormulaList *o );
  void AddRight( SElement *f );
  void AddRight( SElementList *o );
  void AddRight( FormulaList *o );

  void InsertAt( size_t index, SElement *e );
  void InsertAt( size_t index, SElementList *l );
  void InsertAt( size_t index, const FormulaList *l );

  SElement* GetSElement( size_t index ) const;

  void RemoveAt( size_t index );

  void ReplaceSElement( size_t index, SElement *e );
  void ExchangeSElements( size_t a, size_t b );

  DECLARE_APPLY

  SElementList* Clone() const;

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return size(); }
  virtual DataNode* GetChild( size_t index ) const
    { return GetSElement( index ); }

  bool IsSubListOf( const SElementList* o, FormulaList *miss = NULL ) const;
  bool ContainsFormula( const Formula *f ) const;
  size_t CountOfFormula( const Formula *f ) const;
  bool FindFormula( Formula *f, size_t &index ) const;

  bool FindSubListAtIndex( const FormulaList *list, size_t &index ) const;
  bool IsSubListAtIndex( const FormulaList *list, size_t index ) const;

  virtual std::ostream& ToString( std::ostream &out ) const;
#if 0
private:
  void ReplaceSFormulas( FormulaListVariable *fv, size_t start_index );
  void UnsubstSFormulaListVar( FormulaListVariable *fv, size_t index );
#endif
};

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class RuleItem : public DataNode
{
private:
  RuleItem              *m_parent;

public:
  RuleItem() { m_parent = NULL; }

  RuleItem* GetParent() const { return m_parent; }
  void SetParent( RuleItem *parent ) { m_parent = parent; }
};

///////////////////////////////////////////////////////////////////////////////

class SequentEventClient
{
public:
  virtual ~SequentEventClient() {}
  virtual void SequentUpdateEvent( const Sequent *sequent ) = 0;
};

class SequentNotifier
{
  std::set<SequentEventClient*> m_clients;

public:
  virtual ~SequentNotifier() {}

  void AttachSequentEventClient( SequentEventClient *client );
  void DetachSequentEventClient( SequentEventClient *client );

  void FireUpdateEvent() const;
};

///////////////////////////////////////////////////////////////////////////////

class Sequent :
  public RuleItem,
  public TreeInternalNode,
  public SequentNotifier,
  public SElementListEventClient
{
private:
  SElementList       *m_leftside;
  SElementList       *m_rightside;

public:
  Sequent( SElementList *l, SElementList *r );
  virtual ~Sequent();

  DECLARE_APPLY

  SElementList* GetLeft() const { return m_leftside; }
  SElementList* GetRight() const { return m_rightside; }
  void SetLeft( SElementList* l );
  void SetRight( SElementList* l );

  void AddLeft( SElement* e );
  void AddRight( SElement* e );
  void AddLeftRight( SElement* e );
  void AddRightLeft( SElement* e );

  void AddLeft( Formula *f );
  void AddRight( Formula *f );

  Sequent* Clone() const;

  virtual size_t GetChildCount() const { return 2; }
  virtual DataNode* GetChild( size_t index ) const;

  bool IsSubSequentOf( const Sequent* o, FormulaList *miss_l = NULL,
                       FormulaList *miss_r = NULL ) const;
  bool ContainsFormula( const Formula *f, bool left ) const;
  size_t CountOfFormula( const Formula *f, bool left ) const;

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  void SElementListUpdateEvent( const SElementList *list );
};

////////////////////////////////////////////////////////////////////////////////

class SequentList : public std::vector< Sequent* >, public SymbolNode
{
public:
  SequentList( const Symbol *e ) : SymbolNode( e ) {}
//  USTRING GetName() const;
};

////////////////////////////////////////////////////////////////////////////////

class ProofLink : public RuleItem
{
private:
  Proof                 *m_linkto;
  Proof                 *m_owner;

protected:
  ProofLink( Proof *owner, Proof *linkto );

  friend class Proof;

public:
  virtual ~ProofLink();

  Sequent *GetSequent();

  DECLARE_APPLY

  const USTRING& GetLinkedProofName() const;
  const USTRING& GetOwnerProofName() const;
  Proof* GetOwner() { return m_owner; }
  Proof* GetLinkedProof() const { return m_linkto; }

  void SetNewOwner( Proof *owner );

  virtual std::ostream& ToString( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////

class RuleItemList : private std::vector<RuleItem *>, public RuleItem,
                     public TreeInternalNode
{
public:
  virtual ~RuleItemList();

  void AddLeft( RuleItem *i );
  void AddRight( RuleItem *i );

  void SetRuleItemList( RuleItemList *other );
  void ReplaceRuleItem( RuleItem *torep, RuleItem *by );

  void RemoveAt( size_t index );

  //DECLARE_APPLY

  RuleItem* GetRuleItem( size_t index ) const;
  virtual size_t GetRuleItemCount() const { return size(); }

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return GetRuleItemCount(); }
  virtual DataNode* GetChild( size_t index ) const
    { return GetRuleItem( index ); }
};

////////////////////////////////////////////////////////////////////////////////

class Rule : public RuleItemList
{
private:
  Sequent               *m_conclusion;
  const LambdaTerm      *m_lambda_term;
  USTRING                m_name;
  USTRING                m_type;
  USTRING                m_param;

public:
  Rule();
  virtual ~Rule();

  void SetName( const USTRING &name );
  const USTRING& GetName() const { return m_name; }
  void SetType( const USTRING &type );
  const USTRING& GetType() const { return m_type; }
  void SetParam( const USTRING &param );
  const USTRING& GetParam() const { return m_param; }
  void SetLambdaTerm( const LambdaTerm* lambda_term );
  const LambdaTerm* GetLambdaTerm() const { return m_lambda_term; }

  void SetConclusion( Sequent *i );
  inline Sequent* GetConclusion() const { return m_conclusion; };

  DECLARE_APPLY

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return GetRuleItemCount() + 1; }
  virtual DataNode* GetChild( size_t index ) const;

  virtual std::ostream& ToString( std::ostream &out ) const;
};


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class Proof : public DataNode, public SymbolNode, public TreeInternalNode
{
private:
  RuleItem              *m_endseq;
  ProofLink             *m_parent_link;

protected:
  Proof( const Symbol *e );

  friend class ProofDatabase;
  friend ProofLink::~ProofLink();

public:
  virtual ~Proof();

  void SetEndRule( RuleItem *r );
  inline RuleItem* GetEndRule() const { return m_endseq; }

  ProofLink *NewProofLink( Proof *parent );
  Proof* GetParent() const;
  inline ProofLink* GetUpLink() const { return m_parent_link; }

  DECLARE_APPLY

  // TreeInternalNode overrides
  virtual size_t GetChildCount() const { return 1; }
  virtual DataNode* GetChild( size_t index ) const;

  virtual std::ostream& ToString( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class NodeDatabase : private stlhash::hash_set<DbNode*>
{
protected:
  /// type safe access to nodes
  template<class T> T* GetOrAdd( T *n );
  void Remove( DbNode *n );

  friend class DbNode;
#ifdef DEBUG // fot stats
  friend class ProofDatabase;
#endif
};

////////////////////////////////////////////////////////////////////////////////

template<class T> T* NodeDatabase::GetOrAdd( T *n )
{
  const_iterator i = find(n);

  if( i == end() ) // not found
  {
    insert(n);
    n->m_node_db = this;

    TreeInternalNode *tn = dynamic_cast<TreeInternalNode*>( n );

    if( tn != NULL )
    {
      for( size_t i = 0; i < tn->GetChildCount(); i++ )
      {
	DbNode *c = dynamic_cast<DbNode*>( tn->GetChild(i) );

	if( c != NULL )
	  c->Attach();
      }

      // workaround for secondordervariables
      AtomFormula *f = dynamic_cast<AtomFormula*>( tn );
      if ( f != NULL && f->GetSecondOrderVariable() != NULL )
      {
	f->GetSecondOrderVariable()->Attach();
      }
    }
#if 0
    std::cerr << "insert " << (void*) n << std::endl;
#endif
  }
  else
  {
    // FIXED?
    // FIXME: WE LEAK A LOT OF MEMORY!!!
    // FIXED?
    // delete n;
    n = (T*) *i;

#if 0
    std::cerr << "found " << (void*) n << std::endl;
#endif
  }

  return n;
}

template<class T> T* DbNode::AddToNodeDatabase( NodeDatabase *ndb )
{
  return ndb->GetOrAdd( (T*) this );
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

class ProofDatabase :
  private stlhash::hash_map< USTRING, Proof* >
{
private:
  Proof                 *m_def_proof;
  NodeDatabase          *m_node_db;
  SymbolTable           *m_symbol_table;
  bool                   m_own_node_db;
  stlhash::hash_map< USTRING, SequentList* > m_sequent_lists;

public:
  ProofDatabase( SymbolTable *symtbl );
  ProofDatabase( NodeDatabase* node_db, SymbolTable *symtbl );
  virtual ~ProofDatabase();

  NodeDatabase* GetNodeDatabase() { return m_node_db; }
  SymbolTable* GetSymbolTable() { return m_symbol_table; }

  void Clear();

  Proof* NewProof( const Symbol *sym );
  void RemoveProof( const Symbol *sym );

  // creates a new proof without name, and without adding it
  // to the official list of proofs.
  Proof* NewAnonymousProof( );

  SequentList* NewSequentList( const Symbol *sym );
  SequentList* GetSequentList( const USTRING &name ) const;
  SequentList* GetSequentList( const Symbol *sym ) const;

  Proof* GetDefaultProof() const;
  Proof* GetProof( const Symbol *sym ) const;
  Proof* GetProof( const USTRING &name ) const;

  void SplitProof( Proof *proof, Sequent *splitat, const USTRING &linkname );
  void UnsplitProof( ProofLink *plink );

#ifdef DEBUG
  void DumpStats() const;
#endif

  virtual size_t GetCount() const { return size(); }
  virtual size_t GetSequentListCount() const { return m_sequent_lists.size(); }
  virtual Proof* GetProof( size_t index ) const;
  virtual SequentList* GetSequentList( size_t index ) const;
};

////////////////////////////////////////////////////////////////////////////////

} // namespace Proof

#endif
