////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: nodepointer.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Node Pointer
////////////////////////////////////////////////////////////////////////////////
#include "util/myassert.h"
#include "libproof.h"

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

void NodePointer::Push( size_t n )
{
  push_back( n );
}

size_t NodePointer::Pop()
{
  size_t n = 0;

  if( empty() )
    return n;

  n = back();

  pop_back();

  return n;
}

void NodePointer::Dump( std::ostream &out ) const
{
  out << '(';

  for( size_t i = 0; i < size(); i++ )
  {
    out << ' ';
    out << (*this)[i];
  }

  if( size() > 0 )
    out << ' ';
  
  out << ')';
}

DataNode* NodePointer::GetDataNode( DataNode *root ) const
{
//Dump( std::cout );
//std::cout << std::endl;
  return GetDataNode( root, 0 );
}

DataNode* NodePointer::GetDataNode( DataNode *root, size_t depth ) const
{
  if( size() == depth )
    return root;

  size_t n = (*this)[depth];
//std::cout << n << std::endl;

  TreeInternalNode *tnode = dynamic_cast<TreeInternalNode*>( root );

  if( tnode == NULL )
  {
//std::cout << "no tnode" << std::endl;
    throw new NoSuchNodeException( this );
  }

  if( tnode->GetChildCount() < n )
    throw new NoSuchNodeException( this );

  return GetDataNode( tnode->GetChild( n ), depth+1 );
}

size_t NodePointer::operator [] ( size_t index ) const
{
  ASSERT( index < size() );
  return std::vector<size_t>::operator[]( index );
}

bool NodePointer::operator == ( const NodePointer &other ) const
{
  return (std::vector<size_t>) (*this) == (std::vector<size_t>) other;
}

bool NodePointer::operator != ( const NodePointer &other ) const
{
  return !( (*this) == other );
}

bool NodePointer::operator < ( const NodePointer &other ) const
{
  if( size() >= other.size() )
    return false;

  for( size_t i = 0; i < size(); i++ )
    if( (*this)[i] != other[i] )
      return false;

  return true;
}

bool NodePointer::operator > ( const NodePointer &other ) const
{
  return ( other < (*this) );
}

bool NodePointer::operator <= ( const NodePointer &other ) const
{
  if( size() == other.size() )
    return ( (*this) == other );

  return ( (*this) < other );
}

bool NodePointer::operator >= ( const NodePointer &other ) const
{
  if( size() == other.size() )
    return ( (*this) == other );

  return ( (*this) > other );
}

NodePointer NodePointer::operator + ( const NodePointer &other ) const
{
  NodePointer np( *this );

  np.insert( np.end(), other.begin(), other.end() );

  return np;
}

NodePointer NodePointer::operator + ( size_t index ) const
{
  NodePointer np( *this );

  np.Push( index );

  return np;
}

NodePointer NodePointer::GetCommonBeginning( NodePointer const &o ) const
{
  size_t min_len = std::min( GetLength(), o.GetLength() );

  NodePointer np;

  for( size_t i = 0; i < min_len; i++ )
  {
    if( (*this)[i] != o[i] )
      return np;

    np.Push( (*this)[i] );
  }

  return np;
}

////////////////////////////////////////////////////////////////////////////////

const NodePointer& NodePointers::operator [] ( size_t index ) const
{
  ASSERT( index < GetCount() );

  return std::vector<NodePointer>::operator[]( index );
}

int NodePointers::Find( const NodePointer &np ) const
{
  for( size_t i = 0; i < GetCount(); i++ )
  {
    if( (*this)[i] == np )
      return i;
  }

  return -1;
}

bool NodePointers::IsAboveAny( const NodePointer &np ) const
{
  for( size_t i = 0; i < GetCount(); i++ )
  {
#if 0
#include <iostream>
#include "../printhlk.h"
std::cerr << "TEST: ";
HLK::Print::PrintNodePointer( &((*this)[i]), std::cerr );
std::cerr << " < ";
HLK::Print::PrintNodePointer( &np, std::cerr );
std::cerr << std::endl;
#endif

    if( (*this)[i] <= np )
      return true;
  }

  return false;
}

////////////////////////////////////////////////////////////////////////////////

NoSuchNodeException::NoSuchNodeException( const NodePointer *np )
{
  ASSERT( np != NULL );
  m_np = np;
}

void NoSuchNodeException::Print( std::ostream &out ) const
{
  out << "No such node";

  if( m_np == NULL )
    return;

  out << ": ";
  m_np->Dump( out );
}

////////////////////////////////////////////////////////////////////////////////

} // namespace ProofLib

