////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: operations.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
// OPERATIONS IMPLEMENTATION
//
////////////////////////////////////////////////////////////////////////////////

#include "util/myassert.h"
#include "libproof.h"
#include "resolution.h"
#include "operations.h"

#include <iostream>
//#include "../printhlk.h"

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////
//                    //////////////////////////////////////////////////////////
// GENERAL  OPERATION //////////////////////////////////////////////////////////
//                    //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// OperationScope //////////////////////////////////////////////////////////////
OperationScope::OperationScope( const NodePointer &scope_sequent,
                                const NodePointer &scope_proof )
{
  m_scope_sequent = scope_sequent;
  m_scope_proof = scope_proof;
}

bool OperationScope::IsInSequentScope( const NodePointer &np ) const
{
  return np <= m_scope_sequent;
}

bool OperationScope::IsInProofScope( const NodePointer &np ) const
{
  return np <= m_scope_proof;
}

bool OperationScope::IsSequentScope( const NodePointer &np ) const
{
  return np == m_scope_sequent;
}

bool OperationScope::IsProofScope( const NodePointer &np ) const
{
  return np == m_scope_proof;
}

// Operation ///////////////////////////////////////////////////////////////////

void Operation::Push( size_t index )
{
  m_position.Push( index );
}

size_t Operation::Pop()
{
  ASSERT( !m_position.IsEmpty() );
  return m_position.Pop();
}

void Operation::SetScope( OperationScope *scope )
{
  m_scope = scope;
}

bool Operation::InSequentScope() const
{
  if( m_scope == NULL )
    return true;

  return m_scope->IsInSequentScope( m_position );
}

bool Operation::InProofScope() const
{
  if( m_scope == NULL )
    return true;

  return m_scope->IsInProofScope( m_position );
}

bool Operation::AtSequentScope() const
{
  if( m_scope == NULL )
    return true;

  return m_scope->IsSequentScope( m_position );
}

bool Operation::AtProofScope() const
{
  if( m_scope == NULL )
    return true;

  return m_scope->IsProofScope( m_position );
}

////////////////////////////////////////////////////////////////////////////////
//                    //////////////////////////////////////////////////////////
// STANDARD OPERATION //////////////////////////////////////////////////////////
//                    //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// ExpressionOperation ///////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD( LambdaTerm, SetTerm, ExpressionOperation, t )
{
  ASSERT( t != NULL );

  Push(0);
//  std::cerr << "recursing from lambdaterm into formula: ";
//  HLK::Print::PrintFormula( t->getFormula(), std::cerr ) ;
//  std::cerr << std::endl;
  Formula *old_f = t->getFormula();
  Formula *f = DataNode::ApplyTo( t->getFormula(), *this );
  Pop();

  if ( old_f != f )
    return LambdaTerm::New( GetNodeDatabase(), t->getVariables(), f );
  else
    return t;
}

DECLARE_APPLY_FUNCTION_HEAD( DefinedSet, SetTerm, ExpressionOperation, s )
{
  ASSERT( s != NULL );

  bool changed = false;
  AbstractTermList tl( s->GetChildCount() );
  for ( size_t i = 0; i < s->GetChildCount(); i++ )
  {
    Push(i);
    AbstractTerm *t = DataNode::ApplyTo( s->GetTerm( i ), *this );
    Pop();

    tl.AddRight( t );

    if( s->GetTerm(i) != t )
      changed = true;
  }

  if ( changed )
    s = s->ReplaceTerms( tl );
  
  return s;
}

DECLARE_APPLY_FUNCTION_HEAD( SecondOrderVariable, SetTerm, ExpressionOperation, v )
{
  return v;
}

DECLARE_APPLY_FUNCTION_HEAD( Variable, Term, ExpressionOperation, t )
{
  std::cerr << "entered ExpressionOperation" << std::endl;
  return t;
}

DECLARE_APPLY_FUNCTION_HEAD( Constant, Term, ExpressionOperation, t )
{
  std::cerr << "entered ExpressionOperation" << std::endl;
  return t;
}

DECLARE_APPLY_FUNCTION_HEAD( Function, Term, ExpressionOperation, t )
{
  std::cerr << "entered ExpressionOperation" << std::endl;
  ASSERT( t != NULL );

  AbstractTermList list;
  bool changed = false;

  for( size_t i = 0; i < t->GetChildCount(); i++ )
  {
    Push(i);
    AbstractTerm *s = DataNode::ApplyTo( t->GetTerm(i), *this );
    Pop();

    list.AddRight( s );

    if( t->GetTerm(i) != s )
      changed = true;
  }

  if( changed )
      return t->ReplaceTerms( list );

  return t;
}

DECLARE_APPLY_FUNCTION_HEAD( DefinedSetFormula, Formula, ExpressionOperation, sf )
{
  bool changed = false;
  Push(0);
  SetTerm *s = DataNode::ApplyTo( sf->getDefinedSet(), *this );
  Pop();

  if ( s != sf->getDefinedSet() )
    changed = true;


  TermList *tl = new TermList();
  for ( size_t i = 0; i < sf->GetChildCount() ; i++ )
  {
    Push(i + 1);
    Term *t = dynamic_cast<Term*>( DataNode::ApplyTo( sf->GetTerm( i ), *this ) );
    Pop();

    ASSERT( t != NULL );
    tl->AddRight( t );
    if ( sf->GetChild( i ) != t )
      changed = true;
  }

  if ( !changed )
    return sf;

  DefinedSet *ds = dynamic_cast<DefinedSet*>( s );

  if ( ds != NULL )
    return DefinedSetFormula::New( GetNodeDatabase(), ds, tl );

  // FIXME: if s is lambdaterm, apply and return result
  //        if s is secondordervariable, construct appropriate atomformula and return
  ASSERT( false );
  return NULL;
}

DECLARE_APPLY_FUNCTION_HEAD( AtomFormula, Formula, ExpressionOperation, f )
{
  AbstractTermList list;
  bool changed = false;

  // FIXME const_cast!
  SecondOrderVariable* var = const_cast<SecondOrderVariable*>( f->GetSecondOrderVariable() );
  SetTerm *new_s = NULL;

  if ( var != NULL )
  {
    // FIXME: this does not use nodepointer!
    new_s = ApplyToSecondOrderVariable( var ) ;

    if ( new_s != var )
      changed = true;
  }

  for( size_t i = 0; i < f->GetChildCount(); i++ )
  {
    Push(i);
    AbstractTerm *s = DataNode::ApplyTo( f->GetTerm(i), *this );
    Pop();

    list.AddRight( s );

    if( f->GetTerm(i) != s )
      changed = true;
  }

  if( changed )
  {
    // TODO: have to construct new formula from setterm + arguments!!!
    if ( new_s != NULL )
      var = dynamic_cast<SecondOrderVariable*>( new_s );
    return
      AtomFormula::New( GetNodeDatabase(), f->GetSymbol(), &list, var );
    ASSERT( false );
  }

  return f;
}

DECLARE_APPLY_FUNCTION_HEAD( UnaryConFormula, Formula, ExpressionOperation, f )
{
  Push(0);
  Formula *s = DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  if( s == f->GetFormula() )
    return f;

  return UnaryConFormula::New( GetNodeDatabase(), f->GetOperatorSymbol(), s );
}

DECLARE_APPLY_FUNCTION_HEAD( QuantorFormula, Formula, ExpressionOperation, f )
{
  Variable *v = f->GetVariable();

  Push(1);
  Formula *s = DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  if( s == f->GetFormula() && v == f->GetVariable() )
    return f;

  return QuantorFormula::New( GetNodeDatabase(), f->GetOperatorSymbol(), v, s );
}

DECLARE_APPLY_FUNCTION_HEAD( SecondOrderQuantorFormula, Formula, ExpressionOperation, f )
{
  Push(0);
  Formula *s = DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  if( s == f->GetFormula() )
    return f;

  return SecondOrderQuantorFormula::New( GetNodeDatabase(), f->GetOperatorSymbol(), 
      SecondOrderVariable::New( GetNodeDatabase(), f->GetSecondOrderVariable()->GetSymbol() ), s );
}

DECLARE_APPLY_FUNCTION_HEAD( BinaryConFormula, Formula, ExpressionOperation, f )
{
  Push(0);
  Formula *l = DataNode::ApplyTo( f->GetLeft(), *this );
  Pop();

  Push(1);
  Formula *r = DataNode::ApplyTo( f->GetRight(), *this );
  Pop();

  if( l == f->GetLeft() && r == f->GetRight() )
    return f;

  return
    BinaryConFormula::New( GetNodeDatabase(), f->GetOperatorSymbol(), l, r );
}

DECLARE_APPLY_FUNCTION_HEAD( FormulaVariable, Formula, ExpressionOperation, f )
{
  Push(0);
  Formula *c = DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  if( c == f->GetFormula() )
    return f;

  return c; // unsubst formula automatically
}

DECLARE_APPLY_FUNCTION_HEAD( FormulaList, FormulaList, ExpressionOperation, l )
{
  FormulaList new_list;
  bool changed = false;

  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    Push(i);
    FormulaListInsertable *n = DataNode::ApplyTo( l->GetElement(i), *this );
    Pop();
  
    if( n != l->GetElement(i) )
      changed = true;

    n->InsertIntoFormulaList( &new_list, new_list.GetChildCount() );
  }

  if( changed )
    return new FormulaList( new_list );

  return l;
#if 0
  // REPLACEMENT for a s-element by a formula list variable
  const SElementReplacement *sr =
    dynamic_cast<const SElementReplacement*>( rep );

  bool replaced = false;

  if( sr != NULL )
  {
    NodePointer np = sr->GetReplacePosition();

    if( !np.IsEmpty() )
    {
      size_t start_index = np.Pop();
      
      if( np == rep->GetCurrentPosition() && 
          IsSubListAtIndex( sr->GetReplacement(), start_index ) )
      {
        ReplaceFormulas( sr->GetReplacement(), start_index );

        replaced = true;
      }
    }
    else
    {
      size_t index = 0;

      while( FindSubListAtIndex( sr->GetReplacement(), index ) )
      {
        ReplaceFormulas( sr->GetReplacement(), index );

        index++;

        replaced = true;
      }
    }
  }

  // REPLACEMENT unfold of a formula list variable
  const FormulaListVarReplacement *fr =
    dynamic_cast<const FormulaListVarReplacement*>( rep );

  if( fr != NULL )
  {
    NodePointer np = fr->GetReplacePosition();

    if( !np.IsEmpty() )
    {
      size_t index = np.Pop();
      
      if( np == rep->GetCurrentPosition() &&
          GetElement( index )->CastTo<FormulaListVariable>()
             == fr->GetReplacement() )
      {
        UnsubstFormulaListVar( fr->GetReplacement(), index );
      }
    }
    else
    {
      for( size_t index = 0; index < GetChildCount(); index++ )
      {
        if( GetElement( index )->CastTo<FormulaListVariable>()
              == fr->GetReplacement() )
        {
          UnsubstFormulaListVar( fr->GetReplacement(), index );
          index = fr->GetReplacement()->GetChildCount() - 1;
        }
      }
    }
  }

  // REPLACEMENT for childs
  if( !replaced )
  {
    size_t num = 0;
  
    for( iterator i = begin(); i != end(); i++ )
    {
      rep->Push(num);
      FormulaListElement *e = ApplyReplacementTo( (*i), rep );
      rep->Pop();
  
      if( e != *i )
      {
        DbNode *db_e = dynamic_cast<DbNode*>( e );
        DbNode *db_i = dynamic_cast<DbNode*>( *i );

        ASSERT( db_e != NULL && db_i != NULL );

        db_e->Attach();
        *i = e;
        db_i->Detach();
      }
  
      num++;
    }
  }

  return this;
#endif
}

DECLARE_APPLY_FUNCTION_HEAD
  ( FormulaListVariable, FormulaListInsertable, ExpressionOperation, f )
{
  Push(0);
  FormulaListInsertable *n = dynamic_cast<FormulaListInsertable*>
    ( ((FormulaList*) f->GetFormulaList())->Apply( *this ) );
  Pop();

  if( n != f->GetFormulaList() )
    return n;
  
  return f;
}
#if 0
// TermOperation ///////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD( Variable, Term, TermOperation, t )
{
  return t;
}

DECLARE_APPLY_FUNCTION_HEAD( Constant, Term, TermOperation, t )
{
  return t;
}

DECLARE_APPLY_FUNCTION_HEAD( Function, Term, TermOperation, t )
{
  ASSERT( t != NULL );

  TermList list;
  bool changed = false;

  for( size_t i = 0; i < t->GetChildCount(); i++ )
  {
    Push(i);
    Term *s = DataNode::ApplyTo( t->GetTerm(i), *this );
    Pop();

    list.AddRight( s );

    if( t->GetTerm(i) != s )
      changed = true;
  }

  if( changed )
      return t->ReplaceTerms( list );

  return t;
}

// FormulaOperation ////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD( AtomFormula, Formula, FormulaOperation, f )
{
  TermOperation *t_op = dynamic_cast<TermOperation*>( this );

  if( t_op == NULL )
    return f;

  TermList list;
  bool changed = false;

  for( size_t i = 0; i < f->GetChildCount(); i++ )
  {
    Push(i);
    Term *s = DataNode::ApplyTo( f->GetTerm(i), *this );
    Pop();

    list.AddRight( s );

    if( f->GetTerm(i) != s )
      changed = true;
  }

  if( changed )
    return
      AtomFormula::New( GetNodeDatabase(), f->GetSymbol(), &list, f->GetSecondOrderVariable() );

  return f;
}

DECLARE_APPLY_FUNCTION_HEAD( UnaryConFormula, Formula, FormulaOperation, f )
{
  Push(0);
  Formula *s = DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  if( s == f->GetFormula() )
    return f;

  return UnaryConFormula::New( GetNodeDatabase(), f->GetOperatorSymbol(), s );
}

DECLARE_APPLY_FUNCTION_HEAD( QuantorFormula, Formula, FormulaOperation, f )
{
  Variable *v = f->GetVariable();

  TermOperation *t_op = dynamic_cast<TermOperation*>( this );

  if( t_op != NULL )
  {
    Push(0);
    v = DataNode::ApplyTo( f->GetVariable(), *this );
    Pop();
  }

  Push(1);
  Formula *s = DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  if( s == f->GetFormula() && v == f->GetVariable() )
    return f;

  return QuantorFormula::New( GetNodeDatabase(), f->GetOperatorSymbol(), v, s );
}

DECLARE_APPLY_FUNCTION_HEAD( SecondOrderQuantorFormula, Formula, FormulaOperation, f )
{
  Push(0);
  Formula *s = DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  if( s == f->GetFormula() )
    return f;

  return SecondOrderQuantorFormula::New( GetNodeDatabase(), f->GetOperatorSymbol(), 
      SecondOrderVariable::New( GetNodeDatabase(), f->GetSecondOrderVariable()->GetSymbol() ), s );
}

DECLARE_APPLY_FUNCTION_HEAD( BinaryConFormula, Formula, FormulaOperation, f )
{
  Push(0);
  Formula *l = DataNode::ApplyTo( f->GetLeft(), *this );
  Pop();

  Push(1);
  Formula *r = DataNode::ApplyTo( f->GetRight(), *this );
  Pop();

  if( l == f->GetLeft() && r == f->GetRight() )
    return f;

  return
    BinaryConFormula::New( GetNodeDatabase(), f->GetOperatorSymbol(), l, r );
}

DECLARE_APPLY_FUNCTION_HEAD( FormulaVariable, Formula, FormulaOperation, f )
{
  Push(0);
  Formula *c = DataNode::ApplyTo( f->GetFormula(), *this );
  Pop();

  if( c == f->GetFormula() )
    return f;

  return c; // unsubst formula automatically
}

DECLARE_APPLY_FUNCTION_HEAD( FormulaList, FormulaList, FormulaOperation, l )
{
  FormulaList new_list;
  bool changed = false;

  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    Push(i);
    FormulaListInsertable *n = DataNode::ApplyTo( l->GetElement(i), *this );
    Pop();
  
    if( n != l->GetElement(i) )
      changed = true;

    n->InsertIntoFormulaList( &new_list, new_list.GetChildCount() );
  }

  if( changed )
    return new FormulaList( new_list );

  return l;
#if 0
  // REPLACEMENT for a s-element by a formula list variable
  const SElementReplacement *sr =
    dynamic_cast<const SElementReplacement*>( rep );

  bool replaced = false;

  if( sr != NULL )
  {
    NodePointer np = sr->GetReplacePosition();

    if( !np.IsEmpty() )
    {
      size_t start_index = np.Pop();
      
      if( np == rep->GetCurrentPosition() && 
          IsSubListAtIndex( sr->GetReplacement(), start_index ) )
      {
        ReplaceFormulas( sr->GetReplacement(), start_index );

        replaced = true;
      }
    }
    else
    {
      size_t index = 0;

      while( FindSubListAtIndex( sr->GetReplacement(), index ) )
      {
        ReplaceFormulas( sr->GetReplacement(), index );

        index++;

        replaced = true;
      }
    }
  }

  // REPLACEMENT unfold of a formula list variable
  const FormulaListVarReplacement *fr =
    dynamic_cast<const FormulaListVarReplacement*>( rep );

  if( fr != NULL )
  {
    NodePointer np = fr->GetReplacePosition();

    if( !np.IsEmpty() )
    {
      size_t index = np.Pop();
      
      if( np == rep->GetCurrentPosition() &&
          GetElement( index )->CastTo<FormulaListVariable>()
             == fr->GetReplacement() )
      {
        UnsubstFormulaListVar( fr->GetReplacement(), index );
      }
    }
    else
    {
      for( size_t index = 0; index < GetChildCount(); index++ )
      {
        if( GetElement( index )->CastTo<FormulaListVariable>()
              == fr->GetReplacement() )
        {
          UnsubstFormulaListVar( fr->GetReplacement(), index );
          index = fr->GetReplacement()->GetChildCount() - 1;
        }
      }
    }
  }

  // REPLACEMENT for childs
  if( !replaced )
  {
    size_t num = 0;
  
    for( iterator i = begin(); i != end(); i++ )
    {
      rep->Push(num);
      FormulaListElement *e = ApplyReplacementTo( (*i), rep );
      rep->Pop();
  
      if( e != *i )
      {
        DbNode *db_e = dynamic_cast<DbNode*>( e );
        DbNode *db_i = dynamic_cast<DbNode*>( *i );

        ASSERT( db_e != NULL && db_i != NULL );

        db_e->Attach();
        *i = e;
        db_i->Detach();
      }
  
      num++;
    }
  }

  return this;
#endif
}

DECLARE_APPLY_FUNCTION_HEAD
  ( FormulaListVariable, FormulaListInsertable, FormulaOperation, f )
{
  Push(0);
  FormulaListInsertable *n = dynamic_cast<FormulaListInsertable*>
    ( ((FormulaList*) f->GetFormulaList())->Apply( *this ) );
  Pop();

  if( n != f->GetFormulaList() )
    return n;
  
  return f;
}
#endif
// SequentOperation ////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD( Sequent, Sequent, SequentOperation, s )
{
  if( s->GetLeft() != NULL )
  {
    Push(0);
    SElementList *l = DataNode::ApplyTo( s->GetLeft(), *this );
    Pop();

    if( l != s->GetLeft() )
    {
      ASSERT( false ); // TODO: What to do now? ;-)
    }
  }

  if( s->GetRight() != NULL )
  {
    Push(1);
    SElementList *r = DataNode::ApplyTo( s->GetRight(), *this );
    Pop();
  
    if( r != s->GetRight() )
    {
      ASSERT( false ); // TODO: What to do now? ;-)
    }
  }

  return s;
}

DECLARE_APPLY_FUNCTION_HEAD( SElementList, SElementList, SequentOperation, l )
{
  if( l == NULL )
    return l;

  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    Push(i);
    FormulaListInsertable *n = dynamic_cast<FormulaListInsertable*>
      ( l->GetSElement(i)->Apply( *this ) );
    Pop();

    if( n != l->GetSElement(i) )
    {
      //l->ReplaceSElement( i, n );
      l->RemoveAt( i );

      if( n != NULL )
        n->InsertIntoSElementList( l, i );
    }
  }

  return l;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( SFormula, FormulaListInsertable, SequentOperation, sf )
{
  ExpressionOperation *f_op = dynamic_cast<ExpressionOperation*>( this );

  if( f_op == NULL )
    return sf;

  Push(0);
  Formula *f = DataNode::ApplyTo( sf->GetFormula(), *this );
  Pop();
  
  if( f != sf->GetFormula() )
    return new SFormula( f );

  return sf;
}

DECLARE_APPLY_FUNCTION_HEAD
  ( SFormulaListVariable, FormulaListInsertable, SequentOperation, s )
{
  ExpressionOperation *f_op = dynamic_cast<ExpressionOperation*>( this );

  if( f_op == NULL )
    return s;

  Push(0);
  FormulaListInsertable *n = dynamic_cast<FormulaListInsertable*>
    ( s->GetFormulaListVariable()->Apply( *this ) );
  Pop();

  return n;
}

DECLARE_APPLY_FUNCTION_HEAD( ClauseList, ClauseList, SequentOperation, c )
{
  ASSERT( false ); // TODO
  return c;
}

DECLARE_APPLY_FUNCTION_HEAD( Clause, Clause, SequentOperation, c )
{
  Push(0);
  LiteralList *l = DataNode::ApplyTo( &(c->GetNegatives()), *this );
  Pop();

  if( l != &(c->GetNegatives()) )
  {
    ASSERT( false ); // TODO: What to do now? ;-)
  }

  Push(1);
  LiteralList *r = DataNode::ApplyTo( &(c->GetPositives()), *this );
  Pop();
  
  if( r != &(c->GetPositives()) )
  {
    ASSERT( false ); // TODO: What to do now? ;-)
  }

  return c;
}

DECLARE_APPLY_FUNCTION_HEAD( LiteralList, LiteralList, SequentOperation, l )
{
  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    Push(i);
    Literal *li = DataNode::ApplyTo( l->GetAt(i), *this );
    Pop();

    if( li != l->GetAt(i) )
    {
      l->ReplaceAt( i, li );
    }
  }

  return l;
}

DECLARE_APPLY_FUNCTION_HEAD( Literal, Literal, SequentOperation, l )
{
  ExpressionOperation *f_op = dynamic_cast<ExpressionOperation*>( this );

  if( f_op == NULL )
    return l;

  Push(0);
  AtomFormula *f = DataNode::ApplyTo( l->GetAtomFormula(), *this );
  Pop();
  
  if( f != l->GetAtomFormula() )
    return new Literal( f );

  return l;
}

////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_HEAD( Proof, Proof, ProofOperation, p )
{
  p->GetEndRule()->Apply( *this );
  return p;
}

DECLARE_APPLY_FUNCTION_HEAD( Rule, Rule, ProofOperation, r )
{
  r->GetConclusion()->Apply( *this );

  for( size_t i = 0; i < r->GetRuleItemCount(); i++ )
  {
    r->GetRuleItem(i)->Apply( *this );
  }

  return r;
}

DECLARE_APPLY_FUNCTION_HEAD( ProofLink, ProofLink, ProofOperation, p )
{
  p->GetLinkedProof()->Apply( *this );
  return p;
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//                    //////////////////////////////////////////////////////////
// COMPARE OPERATION  //////////////////////////////////////////////////////////
//                    //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

NodeDatabase* CompareOperation::GetNodeDatabase()
{
  ASSERT( false );
}

ProofDatabase* CompareOperation::GetProofDatabase()
{
  ASSERT( false );
}

// ExpressionCompareOperation //////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Variable, ExpressionCompareOperation, t )
{
  std::cerr << "comparing variable" << std::endl;
  return ( t == o );
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Constant, ExpressionCompareOperation, t )
{
  std::cerr << "comparing constant" << std::endl;
  bool ret = ( t == o );

  if (!ret)
    std::cerr << "no match: ";
  else
    std::cerr << "match: ";

//  HLK::Print::PrintTerm( t, std::cerr );
//  std::cerr << " == ";
//  HLK::Print::PrintTerm( dynamic_cast<const Term*>( o ), std::cerr );
//  std::cerr << std::endl;

  return ret;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Function, ExpressionCompareOperation, t )
{
  std::cerr << "comparing function" << std::endl;
  ASSERT( t != NULL );
  ASSERT( o != NULL );

  const Function *f_o = o->CastTo<const Function>();

  if( f_o == NULL || !t->HaveEqualSymbol( f_o ) )
    return false;

  for( size_t i = 0; i < t->GetChildCount(); i++ )
  {
    Push(i);
    bool ret = t->GetTerm(i)->ApplyCompare( *this, f_o->GetTerm(i) );
    Pop();

    if( !ret )
      return false;
  }

  return true;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( DefinedSetFormula, ExpressionCompareOperation, f )
{
  std::cerr << "comparing definedsetformula" << std::endl;
  ASSERT( f != NULL );
  ASSERT( o != NULL );

  const DefinedSetFormula *f_o = o->CastTo<const DefinedSetFormula>();

  if( f_o == NULL )
  {
    std::cerr << "f_o == NULL" << std::endl;
    return false;
  }

  if ( f->GetChildCount() != f_o->GetChildCount() )
  {
    std::cerr << "f->GetChildCount() != f_o->GetChildCount()" << std::endl;
    return false;
  }

  Push(0);
  bool ret = f->getDefinedSet()->ApplyCompare( *this, f_o->getDefinedSet() );
  Pop();

  if (!ret)
    return false;

  for ( size_t i = 0; i < f->GetChildCount(); i++ )
  {
    Push(i + 1);
    ret = f->GetTerm( i )->ApplyCompare( *this, f_o->GetTerm( i ) );
    Pop();

    if ( !ret )
      return false;
  }

  return true;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( AtomFormula, ExpressionCompareOperation, f )
{
  ASSERT( f != NULL );
  ASSERT( o != NULL );

  const AtomFormula *f_o = o->CastTo<const AtomFormula>();

  if( f_o == NULL || f->GetSymbol() != f_o->GetSymbol() )
    return false;

  for( size_t i = 0; i < f->GetChildCount(); i++ )
  {
    Push(i);
    bool ret = f->GetTerm(i)->ApplyCompare( *this, f_o->GetTerm(i) );
    Pop();

    if( !ret )
      return false;
  }

  return true;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( UnaryConFormula, ExpressionCompareOperation, f )
{
  const UnaryConFormula *f_o = o->CastTo<const UnaryConFormula>();
#if 0
  std::cerr << "comparing unary formula: ";
  HLK::Print::PrintFormula( f, std::cerr );
  std::cerr << std::endl << "with other: ";
  const Formula *f_o2 = dynamic_cast<const Formula*>( o ) ;
  HLK::Print::PrintFormula( f_o2, std::cerr );
  std::cerr  << std::endl;
#endif

  if( f_o == NULL || f->GetOperatorSymbol() != f_o->GetOperatorSymbol() )
    return false;

  Push(0);
#if 0
  std::cerr << "comparing subformula:";
  HLK::Print::PrintFormula( f_o->GetFormula(), std::cerr );
  std::cerr  << std::endl;
#endif
  bool ret = f->GetFormula()->ApplyCompare( *this, f_o->GetFormula() );
  Pop();

  return ret;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( QuantorFormula, ExpressionCompareOperation, f )
{
#if 0
  std::cerr << "comparing quantor formula" << std::endl;
#endif
  const QuantorFormula *f_o = o->CastTo<const QuantorFormula>();

  if( f_o == NULL || f->GetOperatorSymbol() != f_o->GetOperatorSymbol() )
    return false;

  bool ok = true;

  Push(0);
  ok = f->GetVariable()->ApplyCompare( *this, f_o->GetVariable() );
  Pop();

  if( !ok )
    return false;

  Push(1);
  ok = f->GetFormula()->ApplyCompare( *this, f_o->GetFormula() );
  Pop();

  return ok;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( SecondOrderQuantorFormula, ExpressionCompareOperation, f )
{
#if 0
  std::cerr << "comparing SO-quantor formula" << std::endl;
#endif
  const SecondOrderQuantorFormula *f_o = o->CastTo<const SecondOrderQuantorFormula>();

  if( f_o == NULL || f->GetOperatorSymbol() != f_o->GetOperatorSymbol() )
    return false;

  Push(0);
  bool ret = f->GetFormula()->ApplyCompare( *this, f_o->GetFormula() );
  Pop();

  return ret;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( BinaryConFormula, ExpressionCompareOperation, f )
{
#if 0
  std::cerr << "comparing binary formula" << std::endl;
#endif
  const BinaryConFormula *f_o = o->CastTo<const BinaryConFormula>();

  if( f_o == NULL || f->GetOperatorSymbol() != f_o->GetOperatorSymbol() )
    return false;

  Push(0);
  bool ok = f->GetLeft()->ApplyCompare( *this, f_o->GetLeft() );
  Pop();

  if( !ok )
    return false;

  Push(1);
  ok = f->GetRight()->ApplyCompare( *this, f_o->GetRight() );
  Pop();

  return ok;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( FormulaVariable, ExpressionCompareOperation, f )
{
  ASSERT( false ); // TODO
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( FormulaList, ExpressionCompareOperation, f )
{
  ASSERT( false ); // TODO
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( FormulaListVariable, ExpressionCompareOperation, f )
{
  ASSERT( false ); // TODO
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( LambdaTerm, ExpressionCompareOperation, t )
{
  bool ok = false;
  if ( const LambdaTerm *n_t = dynamic_cast<const LambdaTerm *> (o) )
  {
    if ( t->getVariables().GetChildCount() != n_t->getVariables().GetChildCount() )
    {
      return false;
    }

    // substitute bound variables and check for alpha-equivalence

    // FIXME: check whether variables of term2 are free in term1
    // and abort if this is the case!

    OperationVariableSubstitution op_sub( GetNodeDatabase() );
    for ( unsigned int i = 0; i < t->getVariables().GetChildCount(); i++ )
    {
      op_sub.AddSubstitution( dynamic_cast<Variable*>( t->getVariables().GetTerm( i ) ), n_t->getVariables().GetTerm( i ) );
    }
    Formula* form = DataNode::ApplyTo( t->getFormula(), op_sub );
    Push(0);
    ok = form->ApplyCompare( *this, n_t->getFormula() );
    Pop();

    //if ( form != m_form )
    //  form->Detach();
  }

  return ok;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( SecondOrderVariable, ExpressionCompareOperation, t )
{
  return t == o;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( DefinedSet, ExpressionCompareOperation, f )
{
  std::cerr << "comparing defined set" << std::endl;
  ASSERT( f != NULL );
  ASSERT( o != NULL );

  const DefinedSet *f_o = o->CastTo<const DefinedSet>();

  if( f_o == NULL || f->GetSymbol() != f_o->GetSymbol() )
    return false;
  std::cerr << "symbols match" << std::endl;

  for( size_t i = 0; i < f->GetChildCount(); i++ )
  {
    Push(i);
    bool ret = f->GetTerm(i)->ApplyCompare( *this, f_o->GetTerm(i) );
    Pop();

    if( !ret )
      return false;
  }
  std::cerr << "terms match" << std::endl;

  return true;
}

// TermCompareOperation ////////////////////////////////////////////////////////
#if 0
DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Variable, TermCompareOperation, t )
{
  return ( t == o );
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Constant, TermCompareOperation, t )
{
  return ( t == o );
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Function, TermCompareOperation, t )
{
  ASSERT( t != NULL );
  ASSERT( o != NULL );

  const Function *f_o = o->CastTo<const Function>();

  if( f_o == NULL || !t->HaveEqualSymbol( f_o ) )
    return false;

  for( size_t i = 0; i < t->GetChildCount(); i++ )
  {
    Push(i);
    bool ret = t->GetTerm(i)->ApplyCompare( *this, f_o->GetTerm(i) );
    Pop();

    if( !ret )
      return false;
  }

  return true;
}

// FormulaCompareOperation /////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( AtomFormula, FormulaCompareOperation, f )
{
  ASSERT( f != NULL );
  ASSERT( o != NULL );

  const AtomFormula *f_o = o->CastTo<const AtomFormula>();

  if( f_o == NULL || f->GetSymbol() != f_o->GetSymbol() )
    return false;

  TermCompareOperation *t_op = dynamic_cast<TermCompareOperation*>( this );

  if( t_op == NULL )
    return ( f == o );

  for( size_t i = 0; i < f->GetChildCount(); i++ )
  {
    Push(i);
    bool ret = f->GetTerm(i)->ApplyCompare( *this, f_o->GetTerm(i) );
    Pop();

    if( !ret )
      return false;
  }

  return true;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( UnaryConFormula, FormulaCompareOperation, f )
{
  const UnaryConFormula *f_o = o->CastTo<const UnaryConFormula>();

  if( f_o == NULL || f->GetOperatorSymbol() != f_o->GetOperatorSymbol() )
    return false;

  Push(0);
  bool ret = f->GetFormula()->ApplyCompare( *this, f_o->GetFormula() );
  Pop();

  return ret;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( QuantorFormula, FormulaCompareOperation, f )
{
  const QuantorFormula *f_o = o->CastTo<const QuantorFormula>();

  if( f_o == NULL || f->GetOperatorSymbol() != f_o->GetOperatorSymbol() )
    return false;

  bool ok = true;

  TermCompareOperation *t_op = dynamic_cast<TermCompareOperation*>( this );

  if( t_op != NULL )
  {
    Push(0);
    ok = f->GetVariable()->ApplyCompare( *this, f_o->GetVariable() );
    Pop();
  }
  else
    ok = ( f->GetVariable() == f_o->GetVariable() );

  if( !ok )
    return false;

  Push(1);
  ok = f->GetFormula()->ApplyCompare( *this, f_o->GetFormula() );
  Pop();

  return ok;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( SecondOrderQuantorFormula, FormulaCompareOperation, f )
{
  const SecondOrderQuantorFormula *f_o = o->CastTo<const SecondOrderQuantorFormula>();

  if( f_o == NULL || f->GetOperatorSymbol() != f_o->GetOperatorSymbol() )
    return false;

  Push(0);
  bool ret = f->GetFormula()->ApplyCompare( *this, f_o->GetFormula() );
  Pop();

  return ret;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( BinaryConFormula, FormulaCompareOperation, f )
{
  const BinaryConFormula *f_o = o->CastTo<const BinaryConFormula>();

  if( f_o == NULL || f->GetOperatorSymbol() != f_o->GetOperatorSymbol() )
    return false;

  Push(0);
  bool ok = f->GetLeft()->ApplyCompare( *this, f_o->GetLeft() );
  Pop();

  if( !ok )
    return false;

  Push(1);
  ok = f->GetRight()->ApplyCompare( *this, f_o->GetRight() );
  Pop();

  return ok;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( FormulaVariable, FormulaCompareOperation, f )
{
  ASSERT( false ); // TODO
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( FormulaList, FormulaCompareOperation, f )
{
  ASSERT( false ); // TODO
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( FormulaListVariable, FormulaCompareOperation, f )
{
  ASSERT( false ); // TODO
  return false;
}
#endif
////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Sequent, SequentCompareOperation, s )
{
  const Sequent *s_o = o->CastTo<const Sequent>();

  if( s_o == NULL )
    return false;

  bool ok = false;

  if( s->GetLeft() == NULL )
    ok = ( s_o->GetLeft() == NULL );
  else if( s_o->GetLeft() != NULL )
  {
    Push(0);
    ok = s->GetLeft()->ApplyCompare( *this, s_o->GetLeft() );
    Pop();
  }

  if( !ok )
    return false;

  ok = false; // reset

  if( s->GetRight() == NULL )
    ok = ( s_o->GetRight() == NULL );
  else if( s_o->GetRight() != NULL )
  {
    Push(1);
    ok = s->GetRight()->ApplyCompare( *this, s_o->GetRight() );
    Pop();
  }

  return ok;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( SFormula, SequentCompareOperation, s )
{
  const SFormula *s_o = o->CastTo<const SFormula>();

  if( s_o == NULL )
    return false;

  Push(0);
  bool ok = s->GetFormula()->ApplyCompare( *this, s_o->GetFormula() );
  Pop();

  return ok;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( SElementList, SequentCompareOperation, s )
{
  ASSERT( false ); // TODO
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( SFormulaListVariable, SequentCompareOperation, s )
{
  ASSERT( false ); // TODO
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
( ClauseList, SequentCompareOperation, c )
{
  ASSERT( false ); // TODO
  return false;
}
DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
( Clause, SequentCompareOperation, c )
{
  ASSERT( false ); // TODO
  return false;
}
DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
( LiteralList, SequentCompareOperation, c )
{
  ASSERT( false ); // TODO
  return false;
}
DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
( Literal, SequentCompareOperation, l )
{
  const Literal *l_o = o->CastTo<const Literal>();

  if( l_o == NULL )
    return false;

  Push(0);
  bool ok = l->GetAtomFormula()->ApplyCompare( *this, l_o->GetAtomFormula() );
  Pop();

  return ok;
}


////////////////////////////////////////////////////////////////////////////////

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Proof, ProofCompareOperation, p )
{
  ASSERT( false ); // TODO
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( Rule, ProofCompareOperation, p )
{
  ASSERT( false ); // TODO
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( RuleItemList, ProofCompareOperation, p )
{
  ASSERT( false ); // TODO
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( ProofLink, ProofCompareOperation, p )
{
  ASSERT( false ); // TODO
  return false;
}

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////

} // namespace ProofLib

