////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: operations.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#ifndef _OPERATIONS_H_
#define _OPERATIONS_H_

#include "operations_def.h"

namespace ProofLib
{

class NodeDatabase;
class ProofDatabase;

class Variable;
class Constant;
class Function;

class AtomFormula;
class UnaryConFormula;
class QuantorFormula;
class SecondOrderQuantorFormula;
class BinaryConFormula;
class FormulaVariable;
class FormulaList;
class FormulaListVariable;

class Sequent;
class SFormula;
class SElementList;
class SFormulaListVariable;
class ClauseList;
class Clause;
class LiteralList;
class Literal;

class ProofLink;
class RuleItemList;
class Rule;
class Proof;

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//                    //////////////////////////////////////////////////////////
// STANDARD OPERATION //////////////////////////////////////////////////////////
//                    //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/// this define is used to make the ApplyTo... function delcartions
#define DECLARE_APPLY_FUNCTION( CLASS, RETURN ) \
  virtual ProofLib::RETURN* ApplyTo##CLASS( ProofLib::CLASS* )

#define DECLARE_APPLY_FUNCTION_HEAD( CLASS, RETURN, OPERATION_CLASS, VAR ) \
  ProofLib::RETURN* OPERATION_CLASS::ApplyTo##CLASS( ProofLib::CLASS* VAR )

////////////////////////////////////////////////////////////////////////////////

class OperationScope
{
  NodePointer m_scope_sequent;
  NodePointer m_scope_proof;

public:
  OperationScope( const NodePointer &scope_sequent,
                  const NodePointer &scope_proof );
  virtual ~OperationScope() {}

  virtual bool IsInSequentScope( const NodePointer &np ) const;
  virtual bool IsInProofScope( const NodePointer &np ) const;
  virtual bool IsSequentScope( const NodePointer &np ) const;
  virtual bool IsProofScope( const NodePointer &np ) const;
};

////////////////////////////////////////////////////////////////////////////////

class Operation
{
  OperationScope *m_scope;
  NodePointer     m_position;

public:
  Operation() { m_scope = NULL; }
  virtual ~Operation() {}

  virtual NodeDatabase* GetNodeDatabase() = 0;
  virtual ProofDatabase* GetProofDatabase() = 0;

protected:
  /// operations for nodepointers
  virtual void Push( size_t index );
  virtual size_t Pop();
  virtual const NodePointer& GetPosition() const { return m_position; }

public:
  void SetScope( OperationScope *scope );
  const OperationScope* GetScope() const { return m_scope; }

protected:
  bool InSequentScope() const;
  bool InProofScope() const;
  bool AtSequentScope() const;
  bool AtProofScope() const;
};

////////////////////////////////////////////////////////////////////////////////

class ProofOperation : public Operation
{
public:
  virtual ~ProofOperation() {}

  DECLARE_APPLY_FUNCTION( Proof, Proof );
  DECLARE_APPLY_FUNCTION( Rule, Rule );
  DECLARE_APPLY_FUNCTION( ProofLink, ProofLink );
};

////////////////////////////////////////////////////////////////////////////////

class SequentOperation : public ProofOperation
{
public:
  virtual ~SequentOperation() {}

  DECLARE_APPLY_FUNCTION( Sequent, Sequent );
  DECLARE_APPLY_FUNCTION( SElementList, SElementList );
  DECLARE_APPLY_FUNCTION( SFormula, FormulaListInsertable );
  DECLARE_APPLY_FUNCTION( SFormulaListVariable, FormulaListInsertable );
  DECLARE_APPLY_FUNCTION( ClauseList, ClauseList );
  DECLARE_APPLY_FUNCTION( Clause, Clause );
  DECLARE_APPLY_FUNCTION( LiteralList, LiteralList );
  DECLARE_APPLY_FUNCTION( Literal, Literal );
};

////////////////////////////////////////////////////////////////////////////////
class ExpressionOperation : public SequentOperation
{
public:
  virtual ~ExpressionOperation() {}

  DECLARE_APPLY_FUNCTION( AtomFormula, Formula );
  DECLARE_APPLY_FUNCTION( UnaryConFormula, Formula );
  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
  DECLARE_APPLY_FUNCTION( SecondOrderQuantorFormula, Formula );
  DECLARE_APPLY_FUNCTION( BinaryConFormula, Formula );
  DECLARE_APPLY_FUNCTION( DefinedSetFormula, Formula );
  DECLARE_APPLY_FUNCTION( FormulaVariable, Formula );
  DECLARE_APPLY_FUNCTION( FormulaList, FormulaList );
  DECLARE_APPLY_FUNCTION( FormulaListVariable, FormulaListInsertable );

  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( Constant, Term );
  DECLARE_APPLY_FUNCTION( Function, Term );

  DECLARE_APPLY_FUNCTION( LambdaTerm, SetTerm );
  DECLARE_APPLY_FUNCTION( SecondOrderVariable, SetTerm );
  DECLARE_APPLY_FUNCTION( DefinedSet, SetTerm );
};
#if 0
////////////////////////////////////////////////////////////////////////////////
class FormulaOperation : public SequentOperation
{
public:
  virtual ~FormulaOperation() {}

  DECLARE_APPLY_FUNCTION( AtomFormula, Formula );
  DECLARE_APPLY_FUNCTION( UnaryConFormula, Formula );
  DECLARE_APPLY_FUNCTION( QuantorFormula, Formula );
  DECLARE_APPLY_FUNCTION( SecondOrderQuantorFormula, Formula );
  DECLARE_APPLY_FUNCTION( BinaryConFormula, Formula );
  DECLARE_APPLY_FUNCTION( FormulaVariable, Formula );
  DECLARE_APPLY_FUNCTION( FormulaList, FormulaList );
  DECLARE_APPLY_FUNCTION( FormulaListVariable, FormulaListInsertable );
};

////////////////////////////////////////////////////////////////////////////////

class TermOperation : public FormulaOperation
{
public:
  virtual ~TermOperation() {}

  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( Constant, Term );
  DECLARE_APPLY_FUNCTION( Function, Term );
};
#endif
////////////////////////////////////////////////////////////////////////////////
//                    //////////////////////////////////////////////////////////
// COMPARE OPERATION  //////////////////////////////////////////////////////////
//                    //////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/// this define is used to make the ApplyTo... function delcartions
#define DECLARE_APPLY_FUNCTION_FOR_COMPARE( CLASS ) \
  virtual bool ApplyTo##CLASS( const CLASS*, const DataNode* )
#define DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD( CLASS, OPERATION_CLASS, VAR ) \
  bool OPERATION_CLASS::ApplyTo##CLASS( const CLASS* VAR, const DataNode *o )

////////////////////////////////////////////////////////////////////////////////

class CompareOperation
{
public:
  virtual ~CompareOperation() {}

  virtual NodeDatabase* GetNodeDatabase();
  virtual ProofDatabase* GetProofDatabase();

  /// operations for nodepointers
  virtual void Push( size_t index ) = 0;
  virtual size_t Pop() = 0;
};

////////////////////////////////////////////////////////////////////////////////

class ProofCompareOperation : public CompareOperation
{
public:
  virtual ~ProofCompareOperation() {}

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Proof );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Rule );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( RuleItemList );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( ProofLink );
};

////////////////////////////////////////////////////////////////////////////////

class SequentCompareOperation : public ProofCompareOperation
{
public:
  virtual ~SequentCompareOperation() {}

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Sequent );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( SElementList );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( SFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( SFormulaListVariable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( ClauseList );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Clause );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( LiteralList );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Literal );
};

////////////////////////////////////////////////////////////////////////////////

class ExpressionCompareOperation : public SequentCompareOperation
{
public:
  virtual ~ExpressionCompareOperation() {}

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( AtomFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( UnaryConFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( QuantorFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( SecondOrderQuantorFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( BinaryConFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( DefinedSetFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( FormulaVariable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( FormulaList );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( FormulaListVariable );

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Variable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Constant );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Function );

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( LambdaTerm );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( SecondOrderVariable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( DefinedSet );
};

////////////////////////////////////////////////////////////////////////////////
#if 0
class FormulaCompareOperation : public SequentCompareOperation
{
public:
  virtual ~FormulaCompareOperation() {}

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( AtomFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( UnaryConFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( QuantorFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( SecondOrderQuantorFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( BinaryConFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( FormulaVariable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( FormulaList );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( FormulaListVariable );
};

////////////////////////////////////////////////////////////////////////////////

class TermCompareOperation : public FormulaCompareOperation
{
public:
  virtual ~TermCompareOperation() {}

  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Variable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Constant );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Function );
};
#endif
////////////////////////////////////////////////////////////////////////////////

}

#endif
