////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: proofdb.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Datastructures for Proofs
////////////////////////////////////////////////////////////////////////////////
#ifdef DEBUG
#include <iostream>
#include <set>
#include <stack>
#include <math.h>
#include "util/hash.h"
#endif

#include <typeinfo>
#include "util/myassert.h"
#include "libproof.h"
#include "operations.h"
#include "util/proofexceptions.h"
/*
#include "replacement.h"
#include "proofexceptions.h"
*/

namespace ProofLib
{

  
ProofDatabase::ProofDatabase( SymbolTable *symtbl )
{
  m_def_proof     = NULL;
  m_node_db       = new NodeDatabase;
  m_own_node_db   = true;
  m_symbol_table  = symtbl;
  ASSERT( m_symbol_table != NULL );
}

ProofDatabase::ProofDatabase( NodeDatabase* node_db, SymbolTable *symtbl )
{
  m_def_proof   = NULL;
  m_node_db     = node_db;
  m_own_node_db = false;
  m_symbol_table  = symtbl;
  ASSERT( m_symbol_table != NULL );
}

ProofDatabase::~ProofDatabase()
{
  Clear();

  if( m_own_node_db ) delete m_node_db;
}

#if 0
Variable* ProofDatabase::NewVariable( const USTRING &name )
{
  return m_node_db.GetOrAdd( new Variable( name ) );
}

Constant* ProofDatabase::NewConstant( const USTRING &name )
{
  return m_node_db.GetOrAdd( new Constant( name ) );
}

Function* ProofDatabase::NewFunction( const USTRING &name, TermList *terms )
{
  ASSERT( terms != NULL );

  Function *f = new Function( name, terms->GetChildCount() );

  f->SetTerms( terms );

  return m_node_db.GetOrAdd( f );
}

#if 0
LambdaTerm* ProofDatabase::NewLambdaTerm( const TermList &v, const Formula *f)
{
  ASSERT( f != NULL );
  
  LambdaTerm *t = new LambdaTerm( v, f );
  
  return m_node_db.GetOrAdd( t );	
} 
#endif

AtomFormula* ProofDatabase::NewAtomFormula( const USTRING &name, TermList *terms )
{
  AtomFormula *f;

  if( terms == NULL )
    f = new AtomFormula( name, 0 );
  else
  {
    f = new AtomFormula( name, terms->GetChildCount() );

    f->SetTerms( terms );
  }

  return m_node_db.GetOrAdd( f );
}

UnaryConFormula*
ProofDatabase::NewUnaryConFormula( const OperatorSymbol *sym, Formula *f )
{
  return m_node_db.GetOrAdd( new UnaryConFormula( sym, f ) );
}

QuantorFormula*
ProofDatabase::NewQuantorFormula( const OperatorSymbol *sym, Variable *v, Formula *f )
{
  return m_node_db.GetOrAdd( new QuantorFormula( sym, v, f ) );
}

BinaryConFormula*
ProofDatabase::NewBinaryConFormula( const OperatorSymbol *sym,
                                    Formula *fl, Formula *fr )
{
  return m_node_db.GetOrAdd( new BinaryConFormula( sym, fl, fr ) );
}

FormulaVariable*
ProofDatabase::NewFormulaVariable( const USTRING &name, Formula *f )
{
  FormulaVariable *fv = new FormulaVariable( name );

  FormulaVariable *fv_db = m_node_db.GetOrAdd( fv );

  if( fv_db != fv )
  {
    fv = NULL;

    if( f == NULL )
      return fv_db;
    
    if( fv_db->GetFormula() == NULL )
    {
      fv_db->SetFormula( f );
      return fv_db;
    }

    if( f == fv_db->GetFormula() )
      return fv_db;

    return NULL;
  }

  fv_db->SetFormula( f );

  return fv_db;
}

FormulaListVariable*
ProofDatabase::NewFormulaListVariable( const USTRING &name, FormulaList *list )
{
  FormulaListVariable *fv = new FormulaListVariable( name );

  FormulaListVariable *fv_db = m_node_db.GetOrAdd( fv );

  if( fv_db != fv )
  {
    fv = NULL;

    if( list == NULL || list->GetChildCount() == 0 )
      return fv_db;

    if( fv_db->GetChildCount() == 0 )
    {
      fv_db->AddLeft( list );

      for( size_t i = 0; i < list->GetChildCount(); i++ )
      {
        DbNode *c = dynamic_cast<DbNode*>( list->GetChild(i) );
        if( c != NULL )
          c->Attach();
      }

      return fv_db;
    }

    if( fv_db->GetChildCount() != list->GetChildCount() )
      return NULL;

    for( size_t i = 0; i < list->GetChildCount(); i++ )
    {
      if( fv_db->GetElement(i) != list->GetElement(i) )
        return NULL;
    }

    return fv_db;
  }

  fv_db->AddLeft( list );

  if( list != NULL )
    for( size_t i = 0; i < list->GetChildCount(); i++ )
    {
      DbNode *c = dynamic_cast<DbNode*>( list->GetChild(i) );
      if( c != NULL )
        c->Attach();
    }

  return fv_db;
}
#endif

void ProofDatabase::Clear()
{
  while( !empty() )
  {
    iterator i = begin();
    Proof *p = i->second;

    while( p->GetParent() != NULL )
      p = p->GetParent();

#if 0
    USTRING name;
    bool ret = UStringFromUTF8( p->GetSymbol()->GetName(), name );
    ASSERT( ret );

    i = find( name );
#endif
    i = find( p->GetSymbol()->GetName() );

    ASSERT( i != end() );

    erase( i );

    delete p->GetEndRule();
    delete p;
  }
}

SequentList* ProofDatabase::NewSequentList( const Symbol *sym )
{
#if 0
  USTRING name;
  bool ret = UStringFromUTF8( sym->GetName(), name );
  ASSERT( ret );
#endif
  USTRING name = sym->GetName();

  if( m_sequent_lists.find( name ) != m_sequent_lists.end() )
    return NULL;
  
  SequentList *l = new SequentList( sym );

  m_sequent_lists[name] = l;
#if 0
fprintf( stderr, "proof '%s' at %p\n", name, p );
#endif
  return l;
}

Proof* ProofDatabase::NewAnonymousProof()
{
  Proof *p = new Proof( NULL );

  return p;
}

Proof* ProofDatabase::NewProof( const Symbol *sym )
{
#if 0
  USTRING name;
  bool ret = UStringFromUTF8( sym->GetName(), name );
  ASSERT( ret );
#endif
  USTRING name = sym->GetName();

  if( find( name ) != end() )
    return NULL;
  
  Proof *p = new Proof( sym );

  (*this)[name] = p;
#if 0
fprintf( stderr, "proof '%s' at %p\n", name, p );
#endif
  if( m_def_proof == NULL )
    m_def_proof = p;
  return p;
}

Proof* ProofDatabase::GetDefaultProof() const
{
  return m_def_proof;
}

SequentList* ProofDatabase::GetSequentList( const Symbol *sym ) const
{
  stlhash::hash_map< USTRING, SequentList* >::const_iterator l = 
    m_sequent_lists.find( sym->GetName() );

  if( l == m_sequent_lists.end() )
    return NULL;

  return l->second;
}

Proof* ProofDatabase::GetProof( const Symbol *sym ) const
{
#if 0
  USTRING name;
  bool ret = UStringFromUTF8( sym->GetName(), name );
  ASSERT( ret );

  if( find( name ) == end() )
    return NULL;

  return (find(name))->second;
#endif
  if( find( sym->GetName() ) == end() )
    return NULL;

  return find( sym->GetName() )->second;
}

Proof* ProofDatabase::GetProof( const USTRING &name ) const
{
  if( find( name ) == end() )
    return NULL;

  return find( name )->second;
}

SequentList* ProofDatabase::GetSequentList( const USTRING &name ) const
{
  if( m_sequent_lists.find( name ) == m_sequent_lists.end() )
    return NULL;

  return m_sequent_lists.find( name )->second;
}



/// Will only unlink the proof, will not free memory.
void ProofDatabase::RemoveProof( const Symbol *sym )
{
#if 0
  USTRING name;
  bool ret = UStringFromUTF8( sym->GetName(), name );
  ASSERT( ret );

  erase( name );
#endif
  erase( sym->GetName() );
}

Proof* ProofDatabase::GetProof( size_t index ) const
{
  ASSERT( index < size() );

  const_iterator i = begin();

  while( index > 0 ) { i++; index--; }

  return i->second;
}

SequentList* ProofDatabase::GetSequentList( size_t index ) const
{
  ASSERT( index < m_sequent_lists.size() );

  stlhash::hash_map< USTRING, SequentList* >::const_iterator i = m_sequent_lists.begin();

  while( index > 0 ) { i++; index--; }

  return i->second;
}

class OpReplaceProofParent : protected ProofOperation
{
  ProofDatabase *m_pdb;
  Proof         *m_new_parent;

public:
  OpReplaceProofParent( ProofDatabase *pdb, Proof *new_parent )
  {
    m_pdb = pdb;
    m_new_parent = new_parent;
  }

  void ReplaceFromHere( RuleItem *r )
  {
    r->Apply( *this );
  }

protected:
  virtual NodeDatabase* GetNodeDatabase()  { ASSERT( false ); return NULL; }
  virtual ProofDatabase* GetProofDatabase() { ASSERT( false ); return NULL; }

  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

protected:
  DECLARE_APPLY_FUNCTION( ProofLink, ProofLink );
};

DECLARE_APPLY_FUNCTION_HEAD( ProofLink, ProofLink, OpReplaceProofParent, p )
{
  p->SetNewOwner( m_new_parent );
  return p;
}

void ProofDatabase::SplitProof( Proof *proof, Sequent *splitat,
                                const USTRING &linkname )
{
  ASSERT( proof != NULL );
  ASSERT( splitat != NULL );
  ASSERT( !linkname.empty() );

  Rule *rule = dynamic_cast<Rule*>( splitat->GetParent() );

  if( rule == NULL )
  {
    GenericException *e = new GenericException();
    e->Stream() << "First sequent in proof can't be splitted.";
    throw e;
  }

  Rule *rparent = dynamic_cast<Rule*>( rule->GetParent() );

  if( rparent == NULL ) // rule ist the first in proof => no split possible
  {
    GenericException *e = new GenericException();
    e->Stream() << "First sequent in proof can't be splitted.";
    throw e;
  }

  // create proof 'name'
  const Symbol *symbol =
    GetSymbolTable()->Add( linkname, 0, SYMBOL_TYPE_PROOF );

  if( symbol == NULL ) // proof already exists
    throw new ProofExistsException( linkname );

  Proof *p = NewProof( symbol );
  ASSERT( p != NULL );

#if 0
  p->SetParent( proof );

/*
  // clone sequent
  Sequent *seq = new Sequent( *splitat );
*/

  ProofLink* plink = new ProofLink( p );
//  plink->SetSequent( seq );
#endif
  ProofLink* plink = p->NewProofLink( proof );

  // place proof above seq in new proof
  if( rule->GetConclusion() == splitat )
  {
    rparent->ReplaceRuleItem( rule, plink );
    p->SetEndRule( rule );

    OpReplaceProofParent ppr( this, p );
    ppr.ReplaceFromHere( rule );
  }
  else // seq is a leaf in the proof tree
  {
    rule->ReplaceRuleItem( splitat, plink );
    p->SetEndRule( splitat );
  }
}

void ProofDatabase::UnsplitProof( ProofLink *plink )
{
  ASSERT( plink != NULL );

  // get rule that contains the proof link
  Rule *rule = dynamic_cast<Rule*>( plink->GetParent() );
  ASSERT( rule != NULL );

  // get the linked proof
  Proof *proof = plink->GetLinkedProof();
  ASSERT( proof != NULL );

  rule->ReplaceRuleItem( plink, proof->GetEndRule() );

  OpReplaceProofParent ppr( this, proof->GetParent()  );
  ppr.ReplaceFromHere( proof->GetEndRule() );

  delete plink;
  RemoveProof( proof->GetSymbol() );
  delete proof;
}

////////////////////////////////////////////////////////////////////////////////

#ifdef DEBUG
extern std::set<DataNode*> DEBUG_DataNode_Table;

void ProofDatabase::DumpStats() const
{
  using namespace std;
  using namespace stlhash;

  set<DataNode*> t = DEBUG_DataNode_Table;
  set<DataNode*> done;
  stack<DataNode*> todo;
  hash_map<DbNode*,size_t> db_ref;

  size_t count = 0;
  size_t dbcount = 0;
  size_t size = 0;

  for( const_iterator p = begin(); p != end(); p++ )
  {
//cout << "PROOF: " << (void *) p->second << endl;
    todo.push( (Proof*) p->second );
    done.insert( (Proof*) p->second );

    while( !todo.empty() )
    {
      DataNode *n = todo.top();
      todo.pop();
//cout << "NODE (#" << count << ") : " << (void *) n << endl;

      size += sizeof( *n );

      t.erase(n);

      DbNode *dbn = dynamic_cast<DbNode*>( n );

      if( dbn != NULL )
        dbcount++;

      TreeInternalNode *tn = dynamic_cast<TreeInternalNode*>( n );

      if( tn != NULL )
        for( size_t i = 0; i < tn->GetChildCount(); i++ )
        {
          DbNode *x = dynamic_cast<DbNode*>( tn->GetChild(i) );

          if( x != NULL )
            db_ref[x]++;
          
          if( done.find( tn->GetChild(i) ) == done.end() )
          {
            todo.push( tn->GetChild(i) );
            done.insert( tn->GetChild(i) );
          }
        }

      count++;
    }
  }

  cout << "Used DataNodes: " << count
       << " (" << dbcount << " in node db)" << endl;
  cout << "Used Bytes (without external mem like symbols): " << size << endl;
  cout << endl;

  cout << "Unreferenced Items: " << t.size() << endl;

  for( set<DataNode*>::const_iterator i = t.begin(); i != t.end(); i++ )
  {
    cout << (void *) *i;

    cout << ": " << (*i)->DEBUGGetTypeName() << endl;
  }

  if( dbcount != m_node_db->size() || dbcount != db_ref.size() )
  {
    cout << "WARNING: referenced db items (" << dbcount << ") differs to "
         << "Node db size (" << m_node_db->size() << ")" << endl;
    cout << "refed db elements (" << db_ref.size() << ")" << endl;
  }

  size_t bucket_count = m_node_db->bucket_count();
  double ideal_count_in_one_bucket = double(dbcount) / double(bucket_count);

  vector< hash_set<DbNode*> > bucket_usage(bucket_count);

  for( hash_map<DbNode*,size_t>::const_iterator i = db_ref.begin();
       i != db_ref.end();
       i++ )
  {
    DbNode *dbn = i->first;

    if( m_node_db->find(dbn) == m_node_db->end() )
    {
      cout << "DB Node " << (void *) dbn;
      cout << " '" << dynamic_cast<DataNode*>(dbn)->DEBUGGetTypeName() << "'";
      cout << " not in NodeDatabase." << endl;
    }
    else
    {
      bucket_usage[dbn->Hash() % bucket_count].insert( dbn );
    }

    if( dbn->GetRefCount() != i->second )
    {
      cout << "DB Node " << (void *) dbn;
      cout << " '" << dynamic_cast<DataNode*>(dbn)->DEBUGGetTypeName() << "'";
      cout << " with wrong ref count " << dbn->GetRefCount()
           << " should be " << i->second << "." << endl;
    }
  }

  cout << endl;
  cout << "Node database stats: " << dbcount << " elements "
       << bucket_count << " buckets ( ideal count for each bucket: "
       << ideal_count_in_one_bucket << " )" << endl;

  double worst_dist = (double(dbcount) - ideal_count_in_one_bucket) * 2;
  double dist = worst_dist;

  for( size_t i = 0; i < bucket_count; i++ )
  {
    dist -= abs( long(bucket_usage[i].size() - ideal_count_in_one_bucket) );

    if( !bucket_usage[i].empty() &&
        lround( abs( long(bucket_usage[i].size() - ideal_count_in_one_bucket) )) > 1 )
    {
      cout << "bucket " << i << ": " << bucket_usage[i].size() << " elements:";
      for( hash_set<DbNode*>::iterator j = bucket_usage[i].begin();
           j != bucket_usage[i].end(); j++ )
      {
        if( j != bucket_usage[i].begin() )
          cout << ", ";
        cout << "'" << dynamic_cast<DataNode*>(*j)->DEBUGGetTypeName() << "'";
        cout << " (" << (*j)->Hash() << ")";
      }
      cout << endl;
    };
  }

  dist /= worst_dist; 

  cout << "hash distribution quote: " << dist * 100 << "% of optimum" << endl;
}

const char* DataNode::DEBUGGetTypeName() const
{
  char *type = NULL;

#define CHECK_TYPE( a ) \
  { if( dynamic_cast<const a*>( this ) != NULL ) type = #a; }

  CHECK_TYPE( TreeInternalNode );
  CHECK_TYPE( DbNode );
  CHECK_TYPE( NodePointer );
  CHECK_TYPE( NoSuchNodeException );
  CHECK_TYPE( SymbolNode );
  CHECK_TYPE( DbSymbolNode );
  CHECK_TYPE( Term );
  CHECK_TYPE( TermList );
  CHECK_TYPE( StaticTermList );
  CHECK_TYPE( Variable );
  CHECK_TYPE( Constant );
  CHECK_TYPE( Function );
  CHECK_TYPE( FormulaListElement );
  CHECK_TYPE( Formula );
  CHECK_TYPE( AtomFormula );
  CHECK_TYPE( OperatorFormula );
  CHECK_TYPE( UnaryConFormula );
  CHECK_TYPE( QuantorFormula );
  CHECK_TYPE( BinaryConFormula );
  CHECK_TYPE( FormulaVariable );
  CHECK_TYPE( FormulaList );
  CHECK_TYPE( FormulaListVariable );
  CHECK_TYPE( SElement );
  CHECK_TYPE( SFormula );
  CHECK_TYPE( SFormulaListVariable );
  CHECK_TYPE( SElementList );
  CHECK_TYPE( RuleItem );
  CHECK_TYPE( Sequent );
  CHECK_TYPE( ProofLink );
  CHECK_TYPE( RuleItemList );
  CHECK_TYPE( Rule );
  CHECK_TYPE( Proof );
  CHECK_TYPE( NodeDatabase );
  CHECK_TYPE( ProofDatabase );
#undef CHECK_TYPE

  return type;
}
#endif

////////////////////////////////////////////////////////////////////////////////

} // namespace ProofLib

