////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: proofs.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Datastructures for Proofs
////////////////////////////////////////////////////////////////////////////////
#ifdef DEBUG
#include <iostream>
#include <set>
#include <stack>
#include "util/hash.h"
#endif

#include <typeinfo>
#include "util/myassert.h"
#include "libproof.h"
#include "operations.h"

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

ProofLink::ProofLink( Proof *owner, Proof *linkto )
{
  ASSERT( linkto != NULL );
  ASSERT( owner != NULL );
  m_linkto = linkto;
  m_owner = owner;
}

ProofLink::~ProofLink()
{
  m_linkto->m_parent_link = NULL;
}

Sequent* ProofLink::GetSequent()
{
  ASSERT( m_linkto != NULL );

  RuleItem *ri = m_linkto->GetEndRule();

  if( ri == NULL )
    return NULL;

  Sequent *s = ri->CastTo<Sequent>();

  if( s != NULL )
    return s;

  Rule *r = ri->CastTo<Rule>();

  if( r != NULL )
    return r->GetConclusion();

  return NULL;
}

const USTRING& ProofLink::GetLinkedProofName() const
{
  return m_linkto->GetName();
}

const USTRING& ProofLink::GetOwnerProofName() const
{
  return m_owner->GetName();
}

void ProofLink::SetNewOwner( Proof *owner )
{
  ASSERT( owner != NULL );
  m_owner = owner;
}

IMPLEMENT_APPLY( ProofLink, Proof )

std::ostream& ProofLink::ToString( std::ostream &out ) const
{
  out << "ProofLink" <<  std::endl;
  return out;
}

////////////////////////////////////////////////////////////////////////////////

RuleItemList::~RuleItemList()
{
  for( iterator i = begin(); i != end(); i++ )
    delete *i;
}

void RuleItemList::AddLeft( RuleItem *i )
{
  ASSERT( i != NULL );

  insert( begin(), i );
  i->SetParent( this );
}

void RuleItemList::AddRight( RuleItem *i )
{
  ASSERT( i != NULL );

  push_back( i );
  i->SetParent( this );
}

void RuleItemList::SetRuleItemList( RuleItemList *other )
{
  for( iterator i = begin(); i != end(); i = erase(i) )
    delete (*i);

  if( other == NULL )
    return;

  for( const_iterator j = other->begin(); j != other->end(); j++ )
  {
    (*j)->SetParent( this );
    push_back( *j );
  }

  other->clear();
}

void RuleItemList::ReplaceRuleItem( RuleItem *torep, RuleItem *by )
{
  for( iterator i = begin(); i != end(); i++ )
  {
    if( *i == torep )
    {
      *i = by;
      torep->SetParent( NULL );
      by->SetParent( this );
      return;
    }
  }
}

void RuleItemList::RemoveAt( size_t index )
{
  if( GetRuleItem( index )->GetParent() == this )
    GetRuleItem( index )->SetParent( NULL );

  erase( begin() + index );
}

RuleItem* RuleItemList::GetRuleItem( size_t index ) const
{
  ASSERT( index < size() );
  return (*this)[index];
}

//IMPLEMENT_APPLY( RuleItemList, Proof )

////////////////////////////////////////////////////////////////////////////////

Rule::Rule()
{
  m_conclusion = NULL;
  m_lambda_term = NULL;
}

Rule::~Rule()
{
  if( m_conclusion != NULL );
    delete m_conclusion;
}

void Rule::SetName( const USTRING &name )
{
  m_name = name;
}

void Rule::SetType( const USTRING &type )
{
  m_type = type;
}

void Rule::SetParam( const USTRING &param )
{
  m_param = param;
}

void Rule::SetLambdaTerm( const LambdaTerm *lambda_term )
{
  m_lambda_term = lambda_term;
}

void Rule::SetConclusion( Sequent *i )
{
  ASSERT( i != NULL );

  i->SetParent( this );
  m_conclusion = i;
}

DataNode* Rule::GetChild( size_t index ) const
{
  if( index == 0 )
    return GetConclusion();

  return GetRuleItem( index - 1 );
}

IMPLEMENT_APPLY( Rule, Proof )

std::ostream& Rule::ToString( std::ostream &out ) const
{
  out << "Rule" <<  std::endl;
  return out;
}

////////////////////////////////////////////////////////////////////////////////

Proof::Proof( const Symbol *sym ) : SymbolNode( sym )
{
  m_endseq = NULL;
  m_parent_link = NULL;
}

Proof::~Proof()
{
  //ASSERT( m_parent_link == NULL );
}

void Proof::SetEndRule( RuleItem* r )
{
  ASSERT( r != NULL );

  m_endseq = r;
}

Proof* Proof::GetParent() const
{
  if( m_parent_link == NULL )
    return NULL;

  return m_parent_link->GetOwner();
}

ProofLink *Proof::NewProofLink( Proof *parent )
{
  ASSERT( parent != NULL );
  ASSERT( parent != this );

  if( m_parent_link != NULL )
    return NULL;

  m_parent_link = new ProofLink( parent, this );

  return m_parent_link;
}

DataNode* Proof::GetChild( size_t index ) const
{
  ASSERT( index == 0 );
  return GetEndRule();
}

IMPLEMENT_APPLY( Proof, Proof )

std::ostream& Proof::ToString( std::ostream &out ) const
{
  out << "Proof ";
  SymbolNode::ToString(out);
  out <<  std::endl;
  return out;
}

////////////////////////////////////////////////////////////////////////////////

} // namespace ProofLib

