////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: recursive.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include "libproof/util/myassert.h"
#include "libproof/libproof.h"
#include "libproof/operations.h"
#include "recursive.h"

namespace ProofLib
{

RecursiveExpr::RecursiveExpr( const wchar_t *var,
                              char op, unsigned int num )
{
  if( var != NULL )
    m_var = var;

  ASSERT( op == '+' || op == '-' || op == '*' );
  
  m_op = op;
  m_num = num;
}

bool RecursiveExpr::DbEqual( const DbNode *n, VarMapping var_map ) const
{
  const RecursiveExpr *t = dynamic_cast<const RecursiveExpr*>(n);

  if( t == NULL )
    return false;

  if( m_var != t->m_var )
    return false;

  return ( m_op == t->m_op && m_num == t->m_num );
}

size_t RecursiveExpr::Hash() const
{
  return size_t(m_op) % 10 + size_t(m_num) * 10;
}

RecursiveExpr* RecursiveExpr::New( NodeDatabase *ndb, const wchar_t *var,
                                   char op, unsigned int num )
{
  ASSERT( ndb != NULL );

  RecursiveExpr *e = new RecursiveExpr( var, op, num );

  return e->AddToNodeDatabase<RecursiveExpr>( ndb );
}

unsigned int RecursiveExpr::GetValueAtLevel( unsigned int lvl ) const
{
  switch( m_op )
  {
  case '+':
    return lvl + m_num;
  case '-':
    if( lvl < m_num )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Recursion level substraction would result in neagtive "
                     "value (" << lvl << " - " << m_num << ")";
      throw e;
    }
    return lvl - m_num;
  case '*':
    return lvl * m_num;
  default:
    ASSERT( false );
  }

  return 0;
}

////////////////////////////////////////////////////////////////////////////////

RecursiveObject::RecursiveObject( RecursiveExpr *expr )
{
  ASSERT( expr != NULL );
  m_expr = expr;
}

////////////////////////////////////////////////////////////////////////////////

RecursiveVariable::RecursiveVariable( const Symbol *e,
                                      RecursiveExpr *expr )
  : Variable( e ), RecursiveObject( expr )
{
}

bool RecursiveVariable::DbEqual( const DbNode *n, VarMapping var_map ) const
{
  const RecursiveVariable *t = dynamic_cast<const RecursiveVariable*>(n);

  if( t == NULL )
    return false;

  return ( Variable::DbEqual( n, var_map ) &&
           GetRecursiveExpr()->DbEqual( t->GetRecursiveExpr(), var_map ) );
}


size_t RecursiveVariable::Hash() const
{
  return (Variable::Hash() << 1) ^ GetRecursiveExpr()->Hash();
}

RecursiveVariable* RecursiveVariable::New( NodeDatabase *ndb,
                                           const Symbol *e,
                                           RecursiveExpr *expr )
{
  ASSERT( ndb != NULL );
  ASSERT( e != NULL );
  ASSERT( expr != NULL );

  RecursiveVariable *t =
    new RecursiveVariable( e, expr );

  return t->AddToNodeDatabase<RecursiveVariable>( ndb );
}


////////////////////////////////////////////////////////////////////////////////

RecursiveConstant::RecursiveConstant( const Symbol *e,
                                      RecursiveExpr *expr )
  : Constant( e ), RecursiveObject( expr )
{
}

bool RecursiveConstant::DbEqual( const DbNode *n, VarMapping var_map ) const
{
  const RecursiveConstant *t = dynamic_cast<const RecursiveConstant*>(n);

  if( t == NULL )
    return false;

  return ( Constant::DbEqual( n, var_map ) &&
           GetRecursiveExpr()->DbEqual( t->GetRecursiveExpr(), var_map ) );
}


size_t RecursiveConstant::Hash() const
{
  return (Constant::Hash() << 1) ^ GetRecursiveExpr()->Hash();
}

RecursiveConstant* RecursiveConstant::New( NodeDatabase *ndb,
                                           const Symbol *e,
                                           RecursiveExpr *expr )
{
  ASSERT( ndb != NULL );
  ASSERT( e != NULL );
  ASSERT( expr != NULL );

  RecursiveConstant *t =
    new RecursiveConstant( e, expr );

  return t->AddToNodeDatabase<RecursiveConstant>( ndb );
}


////////////////////////////////////////////////////////////////////////////////

RecursiveFunction::RecursiveFunction( const Symbol *e,
                                      int arity, RecursiveExpr *expr ) :
  Function( e, arity ),
//  DbSymbol( e ), StaticTermList( arity ),
  RecursiveObject( expr )
{
}

RecursiveFunction*
RecursiveFunction::New( NodeDatabase* ndb, const Symbol *fs,
                        AbstractTermList* terms, RecursiveExpr* expr )
{
  ASSERT( ndb != NULL );
  ASSERT( fs != NULL );
  ASSERT( terms != NULL );
  ASSERT( expr != NULL );

  RecursiveFunction *f =
    new RecursiveFunction( fs, terms->GetChildCount(), expr );

  f->SetTerms( terms );

  return f->AddToNodeDatabase<RecursiveFunction>( ndb );
}

bool RecursiveFunction::DbEqual( const DbNode *n, VarMapping var_map ) const
{
  const RecursiveFunction *t = dynamic_cast<const RecursiveFunction*>(n);

  if( t == NULL )
    return false;

  return ( Function::DbEqual( n, var_map ) &&
           GetRecursiveExpr()->DbEqual( t->GetRecursiveExpr(), var_map ) );
}


size_t RecursiveFunction::Hash() const
{
  return (Function::Hash() << 1) ^ GetRecursiveExpr()->Hash();
}

RecursiveFunction* RecursiveFunction::ReplaceTerms( AbstractTermList &terms ) const
{
  return New( GetNodeDatabase(), GetSymbol(),
              &terms, GetRecursiveExpr() );
}

bool RecursiveFunction::HaveEqualSymbol( const Function *o ) const
{
  if( GetSymbol() != o->GetSymbol() )
    return false;

  const ProofLib::RecursiveFunction *b_r = NULL;

  b_r = o->CastTo<const ProofLib::RecursiveFunction>();

  if( b_r == NULL )
    return false;

  ASSERT( GetRecursiveExpr()->GetVariableName() == NULL );
  ASSERT( b_r->GetRecursiveExpr()->GetVariableName() == NULL );

  return GetRecursiveExpr()->GetValueAtLevel( 0 ) ==
         b_r->GetRecursiveExpr()->GetValueAtLevel( 0 );
}

////////////////////////////////////////////////////////////////////////////////

RecursivePredicate::RecursivePredicate( const Symbol *e, int arity,
                                        RecursiveExpr *expr  )
 : AtomFormula( e, arity ), RecursiveObject( expr )
{
}

RecursivePredicate* RecursivePredicate::New(
  NodeDatabase *ndb, const Symbol *ps,
  AbstractTermList *terms, RecursiveExpr *expr )
{
  ASSERT( ndb != NULL );
  ASSERT( ps != NULL );
  ASSERT( terms != NULL );
  ASSERT( expr != NULL );

  RecursivePredicate *p =
    new RecursivePredicate( ps, terms->GetChildCount(), expr );

  p->SetTerms( terms );

  return p->AddToNodeDatabase<RecursivePredicate>( ndb );
}

bool RecursivePredicate::DbEqual( const DbNode *n, VarMapping var_map ) const
{
  const RecursivePredicate *t = dynamic_cast<const RecursivePredicate*>(n);

  if( t == NULL )
    return false;

  return ( AtomFormula::DbEqual( n, var_map ) &&
           GetRecursiveExpr()->DbEqual( t->GetRecursiveExpr(), var_map ) );
}

size_t RecursivePredicate::Hash() const
{
  return (AtomFormula::Hash() << 1) ^ GetRecursiveExpr()->Hash();
}

RecursivePredicate* RecursivePredicate::ReplaceTerms( AbstractTermList &terms ) const
{
  return New( GetNodeDatabase(), GetSymbol(),
              &terms, GetRecursiveExpr() );
}

////////////////////////////////////////////////////////////////////////////////

}
