////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: resolution.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////

#include <iostream> // TEST
#include <resolution.h>
#include <typeinfo>

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

Literal::Literal( AtomFormula *f )
{
  m_formula = NULL;
  SetAtomFormula(f);
}

Literal::~Literal()
{
  if( m_formula != NULL )
    m_formula->Detach();
}

IMPLEMENT_APPLY( Literal, Sequent )

void Literal::SetAtomFormula( AtomFormula *f )
{
  if( m_formula == f )
    return;

  if( m_formula != NULL )
    m_formula->Detach();

  m_formula = f;
  m_formula->Attach();
}

bool Literal::operator==( const Literal &o ) const
{
#if 0
m_formula->ToString( std::cerr );
std::cerr << "(" << (void*) m_formula << ") =?= ";
o.m_formula->ToString( std::cerr );
std::cerr << "(" << (void*) o.m_formula << "): " << ( m_formula == o.m_formula ? "yes" : "no" );
std::cerr << std::endl;
#endif
  return( m_formula == o.m_formula );
}

DataNode* Literal::GetChild( size_t index ) const
{
  ASSERT( index == 0 );
  return m_formula;
}

std::ostream& Literal::ToString( std::ostream &out ) const
{
  m_formula->ToString( out );
  return out;
}

////////////////////////////////////////////////////////////////////////////////

IMPLEMENT_APPLY( LiteralList, Sequent )

void LiteralList::AddLiteral( Literal *l )
{
  for( size_t i = 0; i < size(); i++ )
  {
    if( *l == *((*this)[i]) )
      return;
  }

  push_back(l);
}

void LiteralList::ReplaceAt( size_t index, Literal *l )
{
#if 0
this->ToString(std::cerr);
std::cerr << std::endl;
std::cerr << "to replace at " << index << ": ";
l->ToString(std::cerr);
std::cerr << std::endl;
#endif
  ASSERT( index < size() );

  iterator i = begin() + index;

  delete (*i);

#if 0
  if( Contains( l ) )
  {
    erase(i);
#if 0
this->ToString(std::cerr);
std::cerr << std::endl;
#endif
    return;
  }
#endif

  erase(i); 
  iterator j = begin() + index;
  insert( j, l );
#if 0
this->ToString(std::cerr);
std::cerr << std::endl;
#endif
}

void LiteralList::BuildFromSElementList( SElementList *l )
{
  if( l == NULL ) return;

  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    SElement *e = l->GetSElement(i);

    SFormula *sf = e->CastTo<SFormula>();

    if( sf == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Expected SFormula.";
      throw e;
    }

    AtomFormula *f = sf->GetFormula()->CastTo<AtomFormula>();

    if( f == NULL )
    {
      GenericException *e = new GenericException();
      e->Stream() << "Expected AtomFormula.";
      throw e;
    }

    AddLiteral( new Literal(f) );
  }
}

std::ostream& LiteralList::ToString( std::ostream &out ) const
{
  if( IsEmpty() ) return out;

  for( const_iterator i = begin(); i != end(); i++ )
  {
    if( i != begin() )
      out << ", ";

    (*i)->ToString(out);
  }

  return out;
}

Literal* LiteralList::GetAt( size_t index )
{
  ASSERT( index < size() );
  return (*this)[index];
}

const Literal* LiteralList::GetAt( size_t index ) const
{
  ASSERT( index < size() );
  return (*this)[index];
}

bool LiteralList::Contains( const Literal *l ) const
{
  for( const_iterator i = begin(); i != end(); i++ )
  {
    const Literal *o = *i;
    if( *l == *o )
      return true;
  }

  return false;
}

bool LiteralList::HaveSameLiterals( const LiteralList &o ) const
{
  if( o.GetChildCount() != GetChildCount() )
    return false;

  std::vector<size_t> checks( GetChildCount(), GetChildCount() + 1 );

  for( size_t i = 0; i < GetChildCount(); i++ )
  {
    bool found = false;

    for( size_t j = 0; j < GetChildCount(); j++ )
    {
      if( *GetAt(i) == *(o.GetAt(j)) )
      {
        if( checks[i] != GetChildCount() + 1 )
          continue;

        checks[i] = j;
        found = true;
        break;
      }
    }

    if( !found )
      return false;
  }

  return true;
}

DataNode* LiteralList::GetChild( size_t index ) const
{
  return (DataNode*) GetAt( index );
}

////////////////////////////////////////////////////////////////////////////////

IMPLEMENT_APPLY( Clause, Sequent )

bool Clause::HaveSameLiterals( const Clause &o ) const
{
  if( ! GetNegatives().HaveSameLiterals( o.GetNegatives() ) )
    return false;

  return GetPositives().HaveSameLiterals( o.GetPositives() );
}

std::ostream& Clause::ToString( std::ostream &out ) const
{
  if( !m_negatives.IsEmpty() )
  {
    m_negatives.ToString(out);
    out << " ";
  }

  out << ":-";

  if( !m_positives.IsEmpty() )
  {
    out << " ";
    m_positives.ToString(out);
  }

  return out;
}

////////////////////////////////////////////////////////////////////////////////

GivenClause* GivenClause::BuildFromSequent( Sequent *s )
{
  if( s == NULL ) return NULL;

  GivenClause *c = new GivenClause;

  c->m_negatives.BuildFromSElementList( s->GetLeft() );
  c->m_positives.BuildFromSElementList( s->GetRight() );

  return c;
}

std::ostream& GivenClause::ToString( std::ostream &out ) const
{
  out << "given " << m_id << ": ";
  Clause::ToString(out);
  return out;
}

DataNode* Clause::GetChild( size_t index ) const
{
  ASSERT( index < 2 );
  if( index == 0 )
    return (DataNode*) &m_negatives;
  else
    return (DataNode*) &m_positives;
}

////////////////////////////////////////////////////////////////////////////////

ResolvedClause::ResolvedClause( const Clause *l, const Clause *r,
                                const OperationMGU &mgu ) :
  m_mgu(mgu)
{
  m_left_parent = l;
  m_right_parent = r;
}

std::ostream& ResolvedClause::ToString( std::ostream &out ) const
{
  out << "bin resolvent " << m_id << " of " 
      << m_left_parent->GetId() << ","
      << m_right_parent->GetId() << ": ";
  Clause::ToString(out);
  out << " with mgu ";
  m_mgu.ToString( out );
  return out;
}

std::ostream& ResolvedClause::ToStringDetailed( std::ostream &out ) const
{
  out << "use ";
  m_left_parent->ToString(out);
  out << std::endl;
  out << "use ";
  m_right_parent->ToString(out);
  out << std::endl;
  ToString(out);
  return out;
}

void ResolvedClause::Finalize()
{
  m_mgu.Finalize();
}

////////////////////////////////////////////////////////////////////////////////

ParamodClause::ParamodClause( Term *s, Term *t,
                              const Clause *l, const Clause *r,
                              const OperationMGU &mgu ) :
  m_mgu(mgu)
{
  m_left_parent = l;
  m_right_parent = r;

  m_s = s;
  m_t = t;
  m_s->Attach();
  m_t->Attach();
}

ParamodClause::~ParamodClause()
{
  m_s->Detach();
  m_t->Detach();
}

std::ostream& ParamodClause::ToString( std::ostream &out ) const
{
  out << "paramodulated " << m_id << " of " 
      << m_left_parent->GetId() << ","
      << m_right_parent->GetId() << ": ";
  Clause::ToString(out);
  out << " using ";
  m_s->ToString(out);
  out << " <- ";
  m_t->ToString(out);
  out << " with mgu ";
  m_mgu.ToString( out );
  return out;
}

std::ostream& ParamodClause::ToStringDetailed( std::ostream &out ) const
{
  out << "use ";
  m_left_parent->ToString(out);
  out << std::endl;
  out << "use ";
  m_right_parent->ToString(out);
  out << std::endl;
  ToString(out);
  return out;
}

void ParamodClause::Finalize()
{
  m_mgu.Finalize();
}

////////////////////////////////////////////////////////////////////////////////

void ClauseList::Add( const ClauseList &o )
{
  insert( begin(), o.begin(), o.end() );
}

void ClauseList::AddClause( Clause *c )
{
  push_back(c);
}

Clause* ClauseList::Pop()
{
  Clause *c = *begin();

  erase(begin());

  return c;
}

Clause* ClauseList::GetAt( size_t index )
{
  ASSERT( index < size() );
  iterator i = begin();
  while( index > 0 )
  {
    i++;
    index--;
  }
  Clause *c = *i;

  return c;
}

const Clause* ClauseList::GetAt( size_t index ) const
{
  ASSERT( index < size() );
  const_iterator i = begin();
  while( index > 0 )
  {
    i++;
    index--;
  }
  Clause *c = *i;

  return c;
}

std::ostream& ClauseList::ToString( std::ostream &out ) const
{
  for( const_iterator i = begin(); i != end(); i++ )
  {
    (*i)->ToString( out );
    out << std::endl;
  }
  return out;
}

////////////////////////////////////////////////////////////////////////////////

void BinaryGen::CreateNewClause( const Literal *l1, const Literal *l2,
                                 Literal *r1, Literal *r2,
                                 const OperationMGU &mgu,
                                 const Clause *c1, const Clause *c2,
                                 Clause *c,
                                 ClauseInserter* ins )
{
#if 0
        std::cerr << std::endl;
        c1->ToString(std::cerr);
        std::cerr << std::endl;
        c2->ToString(std::cerr);
        std::cerr << std::endl;
        mgu.ToString( std::cerr );
        std::cerr << std::endl;
#endif
  bool left = true;

  //OperationVariableSubstitution op( GetNodeDatabase(), mgu );

  while( true )
  {
    const LiteralList *a, *b;
    LiteralList *l;

    if( left )
    {
      a = &(c1->GetNegatives());
      b = &(c2->GetNegatives());
      l = &(c->GetNegatives());
    }
    else
    {
      a = &(c1->GetPositives());
      b = &(c2->GetPositives());
      l = &(c->GetPositives());
    }

    bool use_a = true;

    while( true )
    {
      const LiteralList *x;
      const Literal *xl;
      Literal *xr;

      OperationVariableSubstitution op( GetNodeDatabase() );
      mgu.GetSubstitution( use_a ? OperationMGU::Left : OperationMGU::Right,
                           op );
#if 0
      std::cerr << std::endl;
      op.ToString( std::cerr );
      std::cerr << std::endl;
      std::cerr << std::endl;
#endif

      if( use_a )
      {
        x = a;
        xl = l1;
        xr = r1;
      }
      else
      {
        x = b;
        xl = l2;
        xr = r2;
      }

      for( size_t i = 0; i < x->GetChildCount(); i++ )
      {
        if( xl != NULL && x->GetAt(i) == xl )
        {
          if( xr != NULL )
            l->AddLiteral( xr );
          continue;
        }

        AtomFormula *af = x->GetAt(i)->GetAtomFormula();

        l->AddLiteral( new Literal( DataNode::ApplyTo( af, op ) ) );
      }

      if( use_a ) use_a = false; else break;
    }

    if( left ) left = false; else break;
  }
#if 0
        c->ToString(std::cerr);
        std::cerr << std::endl;
#endif
  m_hit_count++;
  ins->AddClause( c );
}
////////////////////////////////////////////////////////////////////////////////

BinaryResolution::BinaryResolution( FreshVariableGenerator *fg )
{
  m_fg = fg;
  m_hit_count = 0;
}

void BinaryResolution::Inference( Clause *c1, Clause *c2, ClauseInserter* ins )
{
  bool left = true;

  FreshVarMap map1( m_fg, c1 );
  FreshVarMap map2( m_fg, c2, map1.GetNextVarNum() + 1 );
#if 0
  OperationNormVars op1( m_fg );
  c1 = DataNode::ApplyTo( c1, op1 );
  OperationNormVars op2( m_fg, op1.GetVarCount() + 1 );
  c2 = DataNode::ApplyTo( c2, op2 );
#endif

  while( true )
  {
    LiteralList *a, *b;

    if( left )
    {
      a = &(c1->GetNegatives());
      b = &(c2->GetPositives());
    }
    else
    {
      a = &(c1->GetPositives());
      b = &(c2->GetNegatives());
    }

    for( size_t i = 0; i < a->GetChildCount(); i++ )
    {
      for( size_t j = 0; j < b->GetChildCount(); j++ )
      {
        OperationMGU mgu( m_fg, &map1, &map2 );

#if 0
        a->GetAt(i)->ToString(std::cerr);
        std::cerr << std::endl;
        b->GetAt(j)->ToString(std::cerr);
        std::cerr << std::endl;
#endif

        //if( !mgu.Unify( l1, l2 ) )
        if( !mgu.Unify( a->GetAt(i), b->GetAt(j) ) )
          continue;

        CreateNewClause( a->GetAt(i), b->GetAt(j), NULL, NULL, mgu, c1, c2,
                         new ResolvedClause( c1, c2, mgu ), ins );
      }
    }

    if( left )
      left = false;
    else
      break;
  }
}

#if 0
void BinaryResolution::CreateNewClause( const Literal *l1, const Literal *l2,
                                        const OperationMGU &mgu,
                                        const Clause *c1, const Clause *c2,
                                        ClauseInserter* ins )
{
#if 0
        std::cerr << std::endl;
        c1->ToString(std::cerr);
        std::cerr << std::endl;
        c2->ToString(std::cerr);
        std::cerr << std::endl;
        mgu.ToString( std::cerr );
        std::cerr << std::endl;
#endif
  OperationVariableSubstitution op( m_fg->GetNodeDatabase(), mgu );

  bool left = true;
  Clause *c = new ResolvedClause( c1, c2, mgu );

  while( true )
  {
    const LiteralList *a, *b;
    LiteralList *l;

    if( left )
    {
      a = &(c1->GetNegatives());
      b = &(c2->GetNegatives());
      l = &(c->GetNegatives());
    }
    else
    {
      a = &(c1->GetPositives());
      b = &(c2->GetPositives());
      l = &(c->GetPositives());
    }

    bool use_a = true;

    while( true )
    {
      const LiteralList *x;
      const Literal *xl;

      if( use_a )
      {
        x = a;
        xl = l1;
      }
      else
      {
        x = b;
        xl = l2;
      }

      for( size_t i = 0; i < x->GetChildCount(); i++ )
      {
        if( x->GetAt(i) == xl )
          continue;

        AtomFormula *af = x->GetAt(i)->GetAtomFormula();

        l->AddLiteral( new Literal( DataNode::ApplyTo( af, op ) ) );
      }

      if( use_a ) use_a = false; else break;
    }

    if( left ) left = false; else break;
  }
#if 0
        c->ToString(std::cerr);
        std::cerr << std::endl;
#endif
  m_hit_count++;
  ins->AddClause( c );
}
#endif
                                        
std::ostream& BinaryResolution::DumpStats( std::ostream &out ) const
{
  out << "Clauses generated by binary resolution: " << m_hit_count << std::endl;
  return out;
}

////////////////////////////////////////////////////////////////////////////////

Paramodulation::Paramodulation( FreshVariableGenerator *fg )
{
  m_fg = fg;
  m_hit_count = 0;
}

void Paramodulation::Inference( Clause *c1, Clause *c2, ClauseInserter* ins )
{
#if 0
  OperationNormVars op1( m_fg );
  c1 = DataNode::ApplyTo( c1, op1 );
  OperationNormVars op2( m_fg, op1.GetVarCount() + 1 );
  c2 = DataNode::ApplyTo( c2, op2 );
#endif
  FreshVarMap map1( m_fg, c1 );
  FreshVarMap map2( m_fg, c2, map1.GetNextVarNum() + 1 );

  InferenceInternal( c1, c2, map1, map2, ins );
  InferenceInternal( c2, c1, map2, map1, ins );
}

std::ostream& Paramodulation::DumpStats( std::ostream &out ) const
{
  out << "Clauses generated by paramodulation: " << m_hit_count << std::endl;
  return out;
}

void Paramodulation::InferenceInternal( Clause *e, Clause *c,
                                        const FreshVarMap &map1,
                                        const FreshVarMap &map2,
                                        ClauseInserter* ins )
{
#if 0
std::cerr << "Use cl ";
e->ToString(std::cerr);
std::cerr << std::endl;
#endif
  for( size_t i = 0; i < e->GetPositives().GetChildCount(); i++ )
  {
    Literal *l = e->GetPositives().GetAt(i);
    AtomFormula *af = l->GetAtomFormula();

    if( af->GetSymbol()->GetName() != L"=" )
      continue;

    ASSERT( af->GetArity() == 2 );
#if 0
std::cerr << "Use Eq ";
af->ToString(std::cerr);
std::cerr << std::endl;
#endif
    // FIXME: do not cast!
    InferenceInternal( dynamic_cast<Term*>( af->GetTerm(0) ), dynamic_cast<Term*>( af->GetTerm(1) ), l, e, c,
                       map1, map2, ins );
    InferenceInternal( dynamic_cast<Term*>( af->GetTerm(1) ), dynamic_cast<Term*>( af->GetTerm(0) ), l, e, c,
                       map1, map2, ins );
  }
}

void Paramodulation::InferenceInternal( Term *s, Term *t,
                                        Literal *l,
                                        Clause *e, Clause *c,
                                        const FreshVarMap &map1,
                                        const FreshVarMap &map2,
                                        ClauseInserter* ins )
{
  if( s == t ) return; // do not do reflexivity
  InferenceInternal( s, t, l, &(c->GetNegatives()), e, c, map1, map2, ins );
  InferenceInternal( s, t, l, &(c->GetPositives()), e, c, map1, map2, ins );
}

class OperationFindParaPos : public ExpressionOperation
{
  NodePointer m_np;
  Term *m_search;
  FreshVariableGenerator *m_fg;

  typedef std::pair< NodePointer, OperationMGU > Match;
  typedef std::vector< Match > Matches;

  Matches m_matches;

  const FreshVarMap &m_map1;
  const FreshVarMap &m_map2;

public:
  OperationFindParaPos( FreshVariableGenerator *fg,
                        const FreshVarMap &map1,
                        const FreshVarMap &map2,
                        Term *s )
    : m_map1(map1), m_map2(map2)
  {
    m_search = s;
    m_search->Attach();
    m_fg = fg;
  }

  virtual ~OperationFindParaPos()
  {
    m_search->Detach();
  }

  NodeDatabase* GetNodeDatabase() { return m_fg->GetNodeDatabase(); }
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  size_t GetCount() const { return m_matches.size(); }
  const OperationMGU& GetMGU( size_t index ) const
  {
    ASSERT( index < GetCount() );
    return m_matches[index].second;
  }
  const NodePointer& GetPos( size_t index ) const
  {
    ASSERT( index < GetCount() );
    return m_matches[index].first;
  }

protected:
  virtual void Push( size_t index ) { m_np.Push(index); }
  virtual size_t Pop() { return m_np.Pop(); }

  void CheckTerm( Term *t )
  {
    OperationMGU mgu( m_fg, &m_map1, &m_map2 );
  
    if( mgu.Unify( m_search, t ) )
    {
      m_matches.push_back( Match( m_np, mgu ) );
    }
  }

  DECLARE_APPLY_FUNCTION( Variable, Term );
  DECLARE_APPLY_FUNCTION( Constant, Term );
  DECLARE_APPLY_FUNCTION( Function, Term );
};

DECLARE_APPLY_FUNCTION_HEAD( Variable, Term, OperationFindParaPos, v )
{
//  CheckTerm(v);
  return v;
}

DECLARE_APPLY_FUNCTION_HEAD( Constant, Term, OperationFindParaPos, c )
{
  CheckTerm(c);
  return c;
}

DECLARE_APPLY_FUNCTION_HEAD( Function, Term, OperationFindParaPos, f )
{
  CheckTerm(f);
  return ExpressionOperation::ApplyToFunction( f );
}

void Paramodulation::InferenceInternal( Term *s, Term *t,
                                        Literal *l,
                                        LiteralList *list,
                                        Clause *e, Clause *c,
                                        const FreshVarMap &map1,
                                        const FreshVarMap &map2,
                                        ClauseInserter* ins )
{
  for( size_t i = 0; i < list->GetChildCount(); i++ )
  {
    Literal *l2 = list->GetAt(i);
    AtomFormula *af = l2->GetAtomFormula();

    OperationFindParaPos op( m_fg, map1, map2, s );
    af->Apply( op );

#if 0
std::cerr << "Check ";
e->ToString(std::cerr);
std::cerr << std::endl;
std::cerr << " with ";
c->ToString(std::cerr);
std::cerr << std::endl;
std::cerr << " using ";
s->ToString(std::cerr);
std::cerr << " <- ";
t->ToString(std::cerr);
std::cerr << std::endl;
std::cerr << " hits: " << op.GetCount() << std::endl;
#endif
    for( size_t j = 0; j < op.GetCount(); j++ )
    {
      OperationVariableSubstitution vs1( GetNodeDatabase() );
      op.GetMGU(j).GetSubstitution( OperationMGU::Left, vs1 );
      OperationVariableSubstitution vs2( GetNodeDatabase() );
      op.GetMGU(j).GetSubstitution( OperationMGU::Right, vs2 );

      Literal *l3 = DataNode::ApplyTo( l2, vs2 );
      Term *u = DataNode::ApplyTo( s, vs1 );
      u->Attach();
      Term *v = DataNode::ApplyTo( t, vs1 );
      v->Attach();

      NodePointer np;
      np.Push(0);
      NodePointers pointers;
      pointers.Add( np + op.GetPos(j) );
      OperationTermSubstitution ts( GetNodeDatabase(), u, v, &pointers );
#if 0
      OperationTermSubstitution ts( GetNodeDatabase(), u, v, NULL );
#endif

      l3 = DataNode::ApplyTo( l3, ts );

#if 0
std::cerr << "NEW LIT ";
l3->ToString(std::cerr);
std::cerr << " @ ";
op.GetPos(j).Dump(std::cerr);
std::cerr << " using ";
u->ToString(std::cerr);
std::cerr << " <- ";
v->ToString(std::cerr);
std::cerr << std::endl;
op.GetMGU(j).ToString( std::cerr );
std::cerr << std::endl;
#endif
      
      CreateNewClause( l, l2, NULL, l3, op.GetMGU(j), e, c,
                       new ParamodClause( u, v, e, c, op.GetMGU(j) ), ins );

      u->Detach();
      v->Detach();
    }
  }
}

////////////////////////////////////////////////////////////////////////////////

bool TautologyElimination::IsRedundant( ClauseList *usable, Clause *c )
{
  for( size_t i = 0; i < c->GetNegatives().GetChildCount(); i++ )
    if( c->GetPositives().Contains( c->GetNegatives().GetAt(i) ) )
    {
      m_hit_count++;
      return true;
    }
  
  return false;
}

std::ostream& TautologyElimination::DumpStats( std::ostream &out ) const
{
  out << "Detected tautologies: " << m_hit_count << std::endl;
  return out;
}

////////////////////////////////////////////////////////////////////////////////

class OperationCheckSubsumption : protected ExpressionCompareOperation
{
protected:
#if 0
  class Substitution
  {
  public:
    const Variable *var;
    const Term     *by;

  public:
    Substitution() { var = NULL; by = NULL; }
    Substitution( const Variable *v, const Term *r ) { var = v; by = r; }

    void Dump( std::ostream &out ) const
    {
/*
      WriteTerm( (Term*) var, out, true );
      out << " <-- ";
      WriteTerm( (Term*) by, out, true );
*/
    }
  };

  class Substitutions : public std::vector<Substitution>
  {
  public:
    const Term* GetReplacementFor( const Variable *var ) const
    {
      for( size_t i = 0; i < size(); i++ )
      {
        if( (*this)[i].var == var )
          return (*this)[i].by;
      }

      return NULL;
    }

    bool CheckSubstitution( const Variable *var, const Term *rep )
    {
      const Term *ex = GetReplacementFor( var );

      if( ex == NULL )
      {
        push_back( Substitution( var, rep ) );
        return true;
      }

      if( ex == rep )
        return true;

#if 0
WriteTerm( (Term*) t, std::cerr, true );

  if( pos < 0 )
  {
    vars.AddRight( (Term*) t );
    reps.AddRight( (Term*) rep );
std::cerr << " <-- ";
WriteTerm( (Term*) rep, std::cerr, true );
std::cerr << std::endl;
  
    return true;
  }

std::cerr << " = ";
WriteTerm( (Term*) rep, std::cerr, true );

  if( reps.GetTerm( pos ) == rep )
    return true;

std::cerr << " WRONG!";
std::cerr << std::endl;

#endif
      return false;
    }

    // subset of s?
    bool operator < ( const Substitutions &s ) const
    {
      if( size() > s.size() )
        return false;

      for( size_t i = 0; i < size(); i++ )
        if( (*this)[i].by != s.GetReplacementFor( (*this)[i].var ) )
          return false;

      return true;
    }

    void Dump( std::ostream &out ) const
    {
      for( size_t i = 0; i < size(); i++ )
      {
        (*this)[i].Dump( out );
        out << std::endl;
      }
    }
  };

  class Possibilities : public std::vector<Substitutions>
  {
  public:
    void Add( const Substitutions &s )
    {
      for( size_t i = 0; i < size(); i++ )
      {
        if( s < (*this)[i] )
          return; // s is a subset of an already existing subst
        if( (*this)[i] < s )
        {
          (*this)[i] = s; // s is a super set of an ex subst.
          return;
        }
      }

      push_back( s );
    }

    bool Merge( const Possibilities &o )
    {
      Possibilities ex( *this );

      bool was_empty = empty();

      clear();

      for( size_t i = 0; i < o.size(); i++ )
      {
        if( ex.empty() )
        {
          push_back( o[i] );
          continue;
        }

        for( size_t j = 0; j < ex.size(); j++ )
        {
          Substitutions s( ex[j] );

          bool failed = false;

          for( size_t k = 0; k < o[i].size(); k++ )
          {
            if( !s.CheckSubstitution( o[i][k].GetVar(), o[i][k].GetSubst() ) )
            {
              failed = true;
              break;
            }
          }

          if( !failed )
          {
            push_back( s );
          }
        }
      }
      
      return was_empty || !empty();
    }

    std::ostream& ToString( std::ostream &out ) const
    {
      for( size_t i = 0; i < size(); i++ )
      {
        out << i << "." << std::endl;
        (*this)[i].ToString( out );
      }

      return out;
    }
  };
#endif

protected:
  Substitutions current;
  SubstPossibilities possibilities;

public:
  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

public:
  OperationCheckSubsumption() {}
  virtual ~OperationCheckSubsumption() {}

  void ClearCurrent()
  {
  //  while( vars.GetChildCount() > 0 ) vars.RemoveAt(0);
  //  while( reps.GetChildCount() > 0 ) reps.RemoveAt(0);
    current.clear();
  }

  void Clear()
  {
    ClearCurrent();
    possibilities.clear();
  }

  bool Subsumes( Sequent *original, Sequent *modified )
  {
    Clear();
    return original->ApplyCompare( *this, modified );
  }

  bool Subsumes( Clause *original, Clause *modified )
  {
    Clear();
    return original->ApplyCompare( *this, modified );
  }

protected:
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Variable );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( UnaryConFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( QuantorFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( BinaryConFormula );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( SElementList );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Sequent );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Clause );
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( LiteralList );
/*
  DECLARE_APPLY_FUNCTION_FOR_COMPARE( Literal );
*/
};

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Variable, OperationCheckSubsumption, t )
{
  const Term *rep = o->CastTo<const Term>();

  return current.CheckSubstitution( t, rep );
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( UnaryConFormula, OperationCheckSubsumption, f )
{
  ASSERT( false );
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( QuantorFormula, OperationCheckSubsumption, f )
{
  ASSERT( false );
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( BinaryConFormula, OperationCheckSubsumption, f )
{
  ASSERT( false );
  return false;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( SElementList, OperationCheckSubsumption, l )
{
  const SElementList *l_o = o->CastTo<const SElementList>();

  if( l_o == NULL )
    return false;

  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    SubstPossibilities p;

#if 0
std::cerr << "Process: ";
SFormula *sf = l->GetSElement(i)->CastTo<SFormula>();
WriteFormula( sf->GetFormula(), std::cerr );
std::cerr << std::endl;
#endif

    for( size_t j = 0; j < l_o->GetChildCount(); j++ )
    {
      ClearCurrent();

#if 0
std::cerr << "Check: ";
SFormula *sf = l_o->GetSElement(j)->CastTo<SFormula>();
WriteFormula( sf->GetFormula(), std::cerr );
std::cerr << std::endl;
#endif

      Push(i);
      bool ok = l->GetSElement(i)->ApplyCompare( *this, l_o->GetSElement(j) );
      Pop();
      
      if( ok )
{
#if 0
std::cerr << "OK: ";
current.Dump( std::cerr );
std::cerr << std::endl;
#endif
        p.Add( current );
}
    }

    if( p.empty() )
      return false;

#if 0
std::cerr << "before MERGE:" << std::endl;
p.Dump( std::cerr );
possibilities.Dump( std::cerr );
std::cerr << std::endl;
#endif
    if( !possibilities.Merge(p) )
      return false;
  }

  return true;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Sequent, OperationCheckSubsumption, s )
{
  const Sequent *s_o = o->CastTo<const Sequent>();

  if( s_o == NULL )
    return false;

  bool ok = false;

  if( s->GetLeft() != NULL ) // the empty list subsumes all other lists!
  {
    if( s_o->GetLeft() == NULL )
      return false;

    Push(0);
    ok = s->GetLeft()->ApplyCompare( *this, s_o->GetLeft() );
    Pop();

    if( !ok )
      return false;
  }

  ok = false; // reset

  if( s->GetRight() != NULL ) // the empty list subsumes all other lists!
  {
    if( s_o->GetRight() == NULL )
      return false;

    Push(1);
    ok = s->GetRight()->ApplyCompare( *this, s_o->GetRight() );
    Pop();
  }

  return ok;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( LiteralList, OperationCheckSubsumption, l )
{
  const LiteralList *l_o = o->CastTo<const LiteralList>();

  if( l_o == NULL )
    return false;

  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    SubstPossibilities p;

#if 0
std::cerr << "Process: ";
SFormula *sf = l->GetSElement(i)->CastTo<SFormula>();
WriteFormula( sf->GetFormula(), std::cerr );
std::cerr << std::endl;
#endif

    for( size_t j = 0; j < l_o->GetChildCount(); j++ )
    {
      ClearCurrent();

#if 0
std::cerr << "Check: ";
SFormula *sf = l_o->GetSElement(j)->CastTo<SFormula>();
WriteFormula( sf->GetFormula(), std::cerr );
std::cerr << std::endl;
#endif

      Push(i);
      bool ok = l->GetAt(i)->ApplyCompare( *this, l_o->GetAt(j) );
      Pop();
      
      if( ok )
{
#if 0
std::cerr << "OK: ";
current.Dump( std::cerr );
std::cerr << std::endl;
#endif
        p.Add( current );
}
    }

    if( p.empty() )
      return false;

#if 0
std::cerr << "before MERGE:" << std::endl;
p.Dump( std::cerr );
possibilities.Dump( std::cerr );
std::cerr << std::endl;
#endif
    if( !possibilities.Merge(p) )
      return false;
  }

  return true;
}

DECLARE_APPLY_FUNCTION_FOR_COMPARE_HEAD
  ( Clause, OperationCheckSubsumption, c )
{
  const Clause *c_o = o->CastTo<const Clause>();

  if( c_o == NULL )
    return false;

  bool ok = false;

  Push(0);
  ok = c->GetNegatives().ApplyCompare( *this, &(c_o->GetNegatives()) );
  Pop();

  if( !ok )
    return false;

  ok = false; // reset

  Push(1);
  ok = c->GetPositives().ApplyCompare( *this, &(c_o->GetPositives()) );
  Pop();

  return ok;
}

////////////////////////////////////////////////////////////////////////////////

bool ForwardSubsumption::IsRedundant( ClauseList *usable, Clause *c )
{
  for( size_t i = 0; i < usable->GetCount(); i++ )
  {
    OperationCheckSubsumption op;

    if( op.Subsumes( usable->GetAt(i), c ) )
    {
#if 0
  std::cerr << "Subsumed by ";
  usable->GetAt(i)->ToString( std::cerr );
  std::cerr << std::endl;
#endif
      m_hit_count++;
      return true;
    }
  }

  return false;
}

std::ostream& ForwardSubsumption::DumpStats( std::ostream &out ) const
{
  out << "Clauses removed by forward subsumption: " << m_hit_count << std::endl;
  return out;
}

////////////////////////////////////////////////////////////////////////////////

bool BackwardSubsumption::IsRedundant( ClauseList *usable, Clause *c )
{
  for( size_t i = 0; i < usable->GetCount(); i++ )
  {
    OperationCheckSubsumption op;

#if 0
  std::cerr << "Check backward subsumption for ";
  c->ToString( std::cerr );
  std::cerr << " by ";
  usable->GetAt(i)->ToString( std::cerr );
  std::cerr << std::endl;
#endif
    if( op.Subsumes( usable->GetAt(i), c ) )
    {
#if 0
  std::cerr << "Backward subsumed by ";
  usable->GetAt(i)->ToString( std::cerr );
  std::cerr << std::endl;
#endif
      m_hit_count++;
      return true;
    }
  }

  return false;
}

std::ostream& BackwardSubsumption::DumpStats( std::ostream &out ) const
{
  out << "Clauses removed by backward subsumption: " << m_hit_count
      << std::endl;
  return out;
}

////////////////////////////////////////////////////////////////////////////////

OperationNormVars::OperationNormVars( FreshVariableGenerator *fg,
                                      size_t start_index )
{
  ASSERT( fg != NULL );
  m_fg = fg;
  m_index = start_index;
}

NodeDatabase* OperationNormVars::GetNodeDatabase()
{
  return m_fg->GetNodeDatabase();
}

DECLARE_APPLY_FUNCTION_HEAD( Variable, Term, OperationNormVars, v )
{
  int i = list.FindTerm( v );

  if( i < 0 )
  {
    i = list.GetChildCount();
    list.AddRight( v );
  }

  return m_fg->GetVariableNum(i+m_index);
}

////////////////////////////////////////////////////////////////////////////////

void ResolverLogger::AddClause( Resolver *resolver, Clause *c )
{
  resolver->m_list_new.AddClause( c );
  if( c->GetId() > resolver->m_last_clause_id )
    resolver->m_last_clause_id = c->GetId();
}

void ResolverLogger::AddClauseList( Resolver *resolver, const ClauseList *c )
{
  resolver->m_list_new.Add( *c );

  for( size_t i = 0; i < c->GetCount(); i++ )
  {
    if( c->GetAt(i)->GetId() > resolver->m_last_clause_id )
      resolver->m_last_clause_id = c->GetAt(i)->GetId();
  }
}

void ResolverLogger::InactivateClause( Resolver *resolver, size_t id )
{
}

void ResolverLogger::ProcessLevel( Resolver *resolver, size_t level )
{
  resolver->m_list_active.Add( resolver->m_list_fresh );
  resolver->m_list_fresh.Clear();
  resolver->m_list_fresh.Add( resolver->m_list_new );
  resolver->m_list_new.Clear();
  
  if( resolver->m_level != level )
  {
    GenericException * e = new GenericException;
    e->Stream() << "Should process level " << level << ", but have level "
                << resolver->m_level << ".";
    throw e;
  }

  resolver->m_level++;
}

////////////////////////////////////////////////////////////////////////////////

Resolver::Resolver( NodeDatabase *ndb, SymbolTable *symtab )
{
  m_level = 0;
  m_last_clause_id = 0;
  m_gen_count = 0;
  m_ndb = ndb;
  m_symtbl = symtab;
  m_logger = NULL;
}

void Resolver::Init( ClauseList *given )
{
  ASSERT( given != NULL );

  while( !given->IsEmpty() )
  {
    Clause *c = given->Pop();

    bool insert = true;

    for( size_t i = 0; i < m_insert_checkers.size(); i++ )
    {
      insert = !m_insert_checkers[i]->IsRedundant( given, c );
      if( !insert ) break;
      insert = !m_insert_checkers[i]->IsRedundant( &m_list_new, c );
      if( !insert ) break;
    }

    c->SetId( ++m_last_clause_id );
    c->Finalize();

    if( insert )
      m_list_new.AddClause( c );
    else
      m_list_inactive.AddClause( c );
    
    if( m_logger != NULL )
    {
      m_logger->WriteClause( c );
      if( !insert )
        m_logger->WriteInactiveClause( c );
      m_logger->Commit();
    }
    //c->Detach();
  }
}

void Resolver::Init( ResolverLogger *logger )
{
  m_logger = logger;

  m_logger->InitResolver( this );
}

class max_proofs_reached {};

bool Resolver::AddClause( Clause *c )
{
  m_gen_count++;
  bool insert = true;

  if( c->IsBox() )
  {
    c->SetId( ++m_last_clause_id );
    m_list_proofs.AddClause( c );
    c->Finalize();
#if 1
    std::cerr << "*** PROOF ***: " << std::endl;
    c->ToString( std::cerr );
    std::cerr << std::endl;
#endif
    if( m_logger != NULL )
    {
      m_logger->WriteProof( c );
      m_logger->Commit();
    }

    if( m_max_proofs > 0 && m_list_proofs.GetCount() >= m_max_proofs )
      throw max_proofs_reached();

    return true;
  }

  for( size_t i = 0; i < m_insert_checkers.size(); i++ )
  {
    insert = !m_insert_checkers[i]->IsRedundant( &m_list_active, c );
    if( !insert ) break;
    insert = !m_insert_checkers[i]->IsRedundant( &m_list_fresh, c );
    if( !insert ) break;
    insert = !m_insert_checkers[i]->IsRedundant( &m_list_new, c );
    if( !insert ) break;
  }

#if 0
    std::cerr << "GEN: ";
    c->ToString( std::cerr );
    std::cerr << std::endl;
#endif
  if( insert )
  {
    c->SetId( ++m_last_clause_id );
    m_list_new.AddClause( c );
    c->Finalize();
#if 1
    std::cerr << "GEN: ";
    c->ToString( std::cerr );
    std::cerr << std::endl;
#endif
    if( m_logger != NULL )
    {
      m_logger->WriteClause( c );
      m_logger->Commit();
    }
  }

  return insert;
}

void Resolver::Run()
{
  while( !m_list_new.IsEmpty() )
  {
    try
    {
      RunResStep();
    }
    catch( max_proofs_reached e )
    {
      break;
    }
  }

  if( m_logger != NULL )
  {
    m_logger->WriteLevelStatus( this );
    m_logger->WriteTerminationStatus( this );
    m_logger->Commit();
  }
}

void Resolver::RunResStep()
{
  if( m_logger != NULL )
  {
    m_logger->WriteLevelStatus( this );
    m_logger->Commit();
  }

  m_level++;

  ASSERT( m_list_fresh.IsEmpty() );
  m_list_fresh.Add( m_list_new );
  m_list_new.Clear();

  ClauseList::const_iterator i,j;

  for( i = m_list_active.begin(); i != m_list_active.end(); i++ )
    for( j = m_list_fresh.begin(); j != m_list_fresh.end(); j++ )
      for( size_t k = 0; k < m_inferences.size(); k++ )
        m_inferences[k]->Inference( *i, *j, this );

  for( i = m_list_fresh.begin(); i != m_list_fresh.end(); i++ )
    for( j = i; j != m_list_fresh.end(); j++ )
      for( size_t k = 0; k < m_inferences.size(); k++ )
        m_inferences[k]->Inference( *i, *j, this );

  m_list_active.Add( m_list_fresh );
  m_list_fresh.Clear();

  ClauseList::iterator k;

  if( m_backward_checkers.size() > 0 )
    for( k = m_list_active.begin(); k != m_list_active.end(); )
    {
      bool remove = false;

      for( size_t j = 0; j < m_backward_checkers.size(); j++ )
        if( m_backward_checkers[j]->IsRedundant( &m_list_new, *k ) )
        {
          remove = true;
          break;
        }

      if( remove )
      {
        m_list_inactive.AddClause( *k );

        if( m_logger != NULL )
        {
          m_logger->WriteInactiveClause( *k );
          m_logger->Commit();
        }

        k = m_list_active.erase(k);
      }
      else
        k++;
    }
}

Variable* Resolver::GetVariableNum( size_t index )
{
  while( index >= m_fresh_db.size() )
  {
    char firsts[] = { 'x', 'y', 'z', 'u', 'v', 'w' };
    const Symbol *s = NULL;
    size_t num = m_fresh_db.size();
    USTRING var;
    do
    {
      std::wostringstream str;

      if( num < sizeof( firsts ) )
        str << firsts[num++];
      else
        str << "v_{" << (num++ - sizeof( firsts )) << "}";
 
      var = str.str().c_str();
      
      s = m_symtbl->Get( var, 0, SYMBOL_TYPE_VARIABLE );
    }
    while( s != NULL );

    s = m_symtbl->GetOrAdd( var, 0, SYMBOL_TYPE_VARIABLE );
    Variable *v = Variable::New( m_ndb, s );
    v->Attach(); // TODO Detach sometime (at destruction!)
    m_fresh_db.push_back( v );
  }

  return m_fresh_db[index];
}

void Resolver::AddInsertChecker( InsertionCheck *check )
{
  m_insert_checkers.push_back( check );
}

void Resolver::AddBackwardChecker( BackwardCheck *check )
{
  m_backward_checkers.push_back( check );
}

void Resolver::AddInference( ResInference *inf )
{
  m_inferences.push_back( inf );
}

std::ostream& Resolver::ToString( std::ostream &out ) const
{
  out << "** at level: " << m_level << std::endl;
  out << "** inactive (";
  out << m_list_inactive.GetCount() << "): " << std::endl;
  m_list_inactive.ToString(out);
  out << "** active: (";
  out << m_list_active.GetCount() << "): " << std::endl;
  m_list_active.ToString(out);
  out << "** fresh: (";
  out << m_list_fresh.GetCount() << "): " << std::endl;
  m_list_fresh.ToString(out);
  out << "** new: (";
  out << m_list_new.GetCount() << "): " << std::endl;
  m_list_new.ToString(out);
  return out;
}

std::ostream& Resolver::DumpStats( std::ostream &out ) const
{
  out << "At level: " << m_level << std::endl;
  out << "Generated clauses: " << m_gen_count << std::endl;
  out << "Kept clauses (incl. given clauses): "
      << m_list_active.GetCount() + m_list_fresh.GetCount() +
         m_list_new.GetCount()
      << std::endl;
  out << "Found proofs: " << m_list_proofs.size() << std::endl;

  for( size_t i = 0; i < m_insert_checkers.size(); i++ )
    m_insert_checkers[i]->DumpStats( out );

  for( size_t i = 0; i < m_backward_checkers.size(); i++ )
    m_backward_checkers[i]->DumpStats( out );

  for( size_t i = 0; i < m_inferences.size(); i++ )
    m_inferences[i]->DumpStats( out );

  return out;
}

////////////////////////////////////////////////////////////////////////////////

}
