////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: resolution.h
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////

#ifndef _RESOLUTION_H_
#define _RESOLUTION_H_

#include <set>
#include <ostream>

#include <libproof.h>
#include <operations.h>
#include <expression_operations.h>

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

class VariableMapping : public VariableSubstitution
{
public:
  virtual size_t GetCount() const;

  virtual const Variable* GetVar( size_t index ) const;
  virtual const Variable* GetSubst( size_t index ) const;
};

////////////////////////////////////////////////////////////////////////////////

class Literal : public DataNode, public TreeInternalNode
{
private:
  AtomFormula               *m_formula;

  void SetAtomFormula( AtomFormula *f );

public:
  Literal( AtomFormula* f );
  virtual ~Literal();

  AtomFormula* GetAtomFormula() const { return m_formula; }

  bool operator==( const Literal &o ) const;

  DECLARE_APPLY

//  virtual Literal* Clone() const;

  virtual size_t GetChildCount() const { return 1; }
  virtual DataNode* GetChild( size_t index ) const;

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static AtomFormula* Normalize( AtomFormula* f, VariableMapping *m );
};

////////////////////////////////////////////////////////////////////////////////

class LiteralList : private std::vector<Literal*>, public TreeInternalNode,
                    public DataNode
{
public:
  void AddLiteral( Literal *l );
  void AddAtomFormula( AtomFormula *f );

  Literal* GetAt( size_t index );
  const Literal* GetAt( size_t index ) const;

  void ReplaceAt( size_t index, Literal *l );

  bool Contains( const Literal *l ) const;

  void BuildFromSElementList( SElementList *l );

  bool IsEmpty() const { return empty(); }

  DECLARE_APPLY

  virtual size_t GetChildCount() const { return size(); }
  virtual DataNode* GetChild( size_t index ) const;

  bool HaveSameLiterals( const LiteralList &o ) const;

  std::ostream& ToString( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////

class Clause : public TreeInternalNode, public DataNode
{
protected:
  LiteralList                 m_negatives;
  LiteralList                 m_positives;

  size_t                      m_id;

public:
  Clause() { m_id = 0; }
  virtual ~Clause() {}

  LiteralList& GetNegatives() { return m_negatives; }
  LiteralList& GetPositives() { return m_positives; }
  const LiteralList& GetNegatives() const { return m_negatives; }
  const LiteralList& GetPositives() const { return m_positives; }

  bool IsBox() const
    { return  GetNegatives().IsEmpty() && GetPositives().IsEmpty(); }

  void SetId( size_t id ) { m_id = id; }
  size_t GetId() const { return m_id; }

  DECLARE_APPLY

  virtual size_t GetChildCount() const { return 2; }
  virtual DataNode* GetChild( size_t index ) const;

  bool HaveSameLiterals( const Clause &o ) const;

  virtual std::ostream& ToString( std::ostream &out ) const;

  virtual void Finalize() {}
};

////////////////////////////////////////////////////////////////////////////////

class GivenClause : public Clause
{
public:
  virtual ~GivenClause() {}

  virtual std::ostream& ToString( std::ostream &out ) const;

public:
  static GivenClause* BuildFromSequent( Sequent *s );
};

////////////////////////////////////////////////////////////////////////////////

class ResolvedClause : public Clause
{
  const Clause               *m_left_parent;
  const Clause               *m_right_parent;
  OperationMGU                m_mgu;

public:
  ResolvedClause( const Clause *l, const Clause *r, const OperationMGU &mgu );
  virtual ~ResolvedClause() {}

  virtual std::ostream& ToString( std::ostream &out ) const;
  virtual std::ostream& ToStringDetailed( std::ostream &out ) const;

  virtual const OperationMGU& GetMGU() const { return m_mgu; }
  virtual const Clause* GetParentLeft() const { return m_left_parent; }
  virtual const Clause* GetParentRight() const { return m_right_parent; }

  virtual void Finalize();
};

////////////////////////////////////////////////////////////////////////////////

class ParamodClause : public Clause
{
  const Clause               *m_left_parent;
  const Clause               *m_right_parent;
  Term                       *m_s;
  Term                       *m_t;
  OperationMGU                m_mgu;

public:
  ParamodClause( Term *s, Term *t, 
                 const Clause *l, const Clause *r, const OperationMGU &mgu );
  virtual ~ParamodClause();

  virtual std::ostream& ToString( std::ostream &out ) const;
  virtual std::ostream& ToStringDetailed( std::ostream &out ) const;

  virtual const OperationMGU& GetMGU() const { return m_mgu; }
  virtual const Clause* GetParentLeft() const { return m_left_parent; }
  virtual const Clause* GetParentRight() const { return m_right_parent; }

  virtual void Finalize();
};

////////////////////////////////////////////////////////////////////////////////

class ClauseList : public std::vector<Clause*>
{
public:
  void Add( const ClauseList &o );
  void AddClause( Clause *c );

  Clause* Pop();
  Clause* GetAt( size_t index );
  const Clause* GetAt( size_t index ) const;
  
  void Clear() { clear(); }

  bool IsEmpty() const { return empty(); }
  size_t GetCount() const { return size(); }

  std::ostream& ToString( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////

class ClauseInserter
{
public:
  virtual ~ClauseInserter() {}

  virtual bool AddClause( Clause *c ) = 0;
};

////////////////////////////////////////////////////////////////////////////////

class OperationNormVars :
  public ExpressionOperation
{
protected:
  TermList list;
  FreshVariableGenerator *m_fg;
  size_t m_index;

public:
  NodeDatabase* GetNodeDatabase();
  ProofDatabase* GetProofDatabase() { ASSERT(false); return NULL; }

  virtual void Push( size_t index ) {}
  virtual size_t Pop() { return 0; }

  size_t GetVarCount() const { return list.GetChildCount() + m_index; }

public:
  OperationNormVars( FreshVariableGenerator *fg, size_t start_index = 0 );

  DECLARE_APPLY_FUNCTION( Variable, Term );
};

////////////////////////////////////////////////////////////////////////////////

class ResInference
{
public:
  virtual ~ResInference() {}

  virtual void Inference( Clause *c1, Clause *c2, ClauseInserter* ins ) = 0;
  virtual std::ostream& DumpStats( std::ostream &out ) const = 0;
};

class BinaryGen : public ResInference
{
protected:
  size_t m_hit_count;

public:
  virtual ~BinaryGen() {}

  virtual NodeDatabase* GetNodeDatabase() = 0;

protected:
  void CreateNewClause( const Literal *l1, const Literal *l2,
                        Literal *r1, Literal *r2,
                        const OperationMGU &mgu,
                        const Clause *c1, const Clause *c2,
                        Clause *c,
                        ClauseInserter* ins );
};

class BinaryResolution : public BinaryGen
{
  FreshVariableGenerator  *m_fg;

public:
  BinaryResolution( FreshVariableGenerator *fg );
  virtual ~BinaryResolution() {}

  virtual NodeDatabase* GetNodeDatabase() { return m_fg->GetNodeDatabase(); }

  virtual void Inference( Clause *c1, Clause *c2, ClauseInserter* ins );
  virtual std::ostream& DumpStats( std::ostream &out ) const;
};

class Paramodulation : public BinaryGen
{
  FreshVariableGenerator  *m_fg;

public:
  Paramodulation( FreshVariableGenerator *fg );
  virtual ~Paramodulation() {}

  virtual NodeDatabase* GetNodeDatabase() { return m_fg->GetNodeDatabase(); }

  virtual void Inference( Clause *c1, Clause *c2, ClauseInserter* ins );
  virtual std::ostream& DumpStats( std::ostream &out ) const;

protected:
  void InferenceInternal( Clause *e, Clause *c,
                          const FreshVarMap &map1, const FreshVarMap &map2,
                          ClauseInserter* ins );
  void InferenceInternal( Term *s, Term *t, Literal *l,
                          Clause *e, Clause *c,
                          const FreshVarMap &map1, const FreshVarMap &map2,
                          ClauseInserter* ins );
  void InferenceInternal( Term *s, Term *t, Literal *l, LiteralList *list,
                          Clause *e, Clause *c,
                          const FreshVarMap &map1, const FreshVarMap &map2,
                          ClauseInserter* ins );
};

////////////////////////////////////////////////////////////////////////////////

class InsertionCheck
{
public:
  virtual ~InsertionCheck() {}

  virtual bool IsRedundant( ClauseList *usable, Clause *c ) = 0;
  virtual std::ostream& DumpStats( std::ostream &out ) const = 0;
};

class TautologyElimination : public InsertionCheck
{
  size_t m_hit_count;

public:
  TautologyElimination() { m_hit_count = 0; }
  virtual ~TautologyElimination() {}

  virtual bool IsRedundant( ClauseList *usable, Clause *c );
  virtual std::ostream& DumpStats( std::ostream &out ) const;
};

class ForwardSubsumption : public InsertionCheck
{
  size_t m_hit_count;

public:
  ForwardSubsumption() { m_hit_count = 0; }
  virtual ~ForwardSubsumption() {}

  virtual bool IsRedundant( ClauseList *usable, Clause *c );
  virtual std::ostream& DumpStats( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////

class BackwardCheck
{
public:
  virtual ~BackwardCheck() {}

  virtual bool IsRedundant( ClauseList *usable, Clause *c ) = 0;
  virtual std::ostream& DumpStats( std::ostream &out ) const = 0;
};

class BackwardSubsumption : public BackwardCheck
{
  size_t m_hit_count;

public:
  BackwardSubsumption() { m_hit_count = 0; }
  virtual ~BackwardSubsumption() {}

  virtual bool IsRedundant( ClauseList *usable, Clause *c );
  virtual std::ostream& DumpStats( std::ostream &out ) const;
};

////////////////////////////////////////////////////////////////////////////////

class Resolver;

class ResolverLogger
{
public:
  virtual ~ResolverLogger() {}

  virtual void Commit() = 0;
  virtual void Rollback() = 0;

  virtual void WriteLevelStatus( Resolver *resolver ) = 0;
  virtual void WriteTerminationStatus( Resolver *resolver ) = 0;
  virtual void WriteClause( Clause *clause ) = 0;
  virtual void WriteProof( Clause *clause ) = 0;
  virtual void WriteInactiveClause( Clause *clause ) = 0;
  virtual void WriteClauseList( ClauseList *clauselist ) = 0;

  virtual void InitResolver( Resolver *resolver ) = 0;

public:
  void AddClause( Resolver *resolver, Clause *c );
  void AddClauseList( Resolver *resolver, const ClauseList *c );
  void InactivateClause( Resolver *resolver, size_t id );
  void ProcessLevel( Resolver *resolver, size_t level );
};

////////////////////////////////////////////////////////////////////////////////

class Resolver : public ClauseInserter, public FreshVariableGenerator
{
private:
  size_t                      m_level;
  size_t                      m_last_clause_id;
  size_t                      m_gen_count;
  size_t                      m_max_proofs;

  ClauseList                  m_list_inactive;
  ClauseList                  m_list_active;
  ClauseList                  m_list_fresh;
  ClauseList                  m_list_new;
  ClauseList                  m_list_proofs;

  std::vector<ResInference*>  m_inferences;
  std::vector<InsertionCheck*> m_insert_checkers;
  std::vector<BackwardCheck*> m_backward_checkers;

  NodeDatabase               *m_ndb;
  SymbolTable                *m_symtbl;
  std::vector<Variable*>      m_fresh_db;

  ResolverLogger             *m_logger;

  friend class ResolverLogger;

public:
  Resolver( NodeDatabase *ndb, SymbolTable *symtab );
  virtual ~Resolver() {}

  void Init( ResolverLogger *logger );
  void Init( ClauseList *given );
  void Run();

  void SetMaxProofs( size_t max_proofs ) { m_max_proofs = max_proofs; }

  void AddInsertChecker( InsertionCheck *check );
  void AddBackwardChecker( BackwardCheck *check );
  void AddInference( ResInference *inf );

  virtual bool AddClause( Clause *c );

  virtual NodeDatabase* GetNodeDatabase() { return m_ndb; }
  virtual Variable* GetVariableNum( size_t );

  size_t GetLevel() const { return m_level; }
  std::ostream& ToString( std::ostream &out ) const;
  std::ostream& DumpStats( std::ostream &out ) const;

protected:
  void RunResStep();
};

////////////////////////////////////////////////////////////////////////////////

}
#endif
