////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: sequents.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Sequent Implemetation
////////////////////////////////////////////////////////////////////////////////
#include "util/myassert.h"
#include "libproof.h"
#include "operations.h"

//#include "printhlk.h"
#include <typeinfo>

#include <iostream>

namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

SElementList::~SElementList()
{
  for( iterator i = begin(); i != end(); i++ )
    delete *i;
}

void SElementList::AddLeft( SElement *f )
{
  ASSERT( f != NULL );
  insert( begin(), f );
  FireUpdateEvent();
}
void SElementList::AddRight( SElement *f )
{
  ASSERT( f != NULL );
  push_back( f );
  FireUpdateEvent();
}

void SElementList::AddLeft( SElementList *o )
{
  if( o == NULL )
    return;

  insert( begin(), o->begin(), o->end() );
  FireUpdateEvent();
}

void SElementList::InsertAt( size_t index, SElement *e )
{
  ASSERT( index <= size() );
  ASSERT( e != NULL );
  insert( begin() + index, e );
  FireUpdateEvent();
}

void SElementList::InsertAt( size_t index, SElementList *l )
{
  ASSERT( index <= size() );
  ASSERT( l != NULL );
  insert( begin() + index, l->begin(), l->end() );
  FireUpdateEvent();
}

void SElementList::InsertAt( size_t index, const FormulaList *l )
{
  ASSERT( index <= size() );
  ASSERT( l != NULL );

  iterator pos = begin() + index;

  for( size_t i = 0; i < l->GetChildCount(); i++ )
  {
    l->GetElement(i)->InsertIntoSElementList( this, i + index );
#if 0
    Formula *f = l->GetChild(i)->CastTo<Formula>();

    if( f != NULL )
    {
      pos = insert( pos, new SFormula( f ) ) + 1;
      continue;
    }

    FormulaListVariable *fv =
      l->GetChild(i)->CastTo<FormulaListVariable>();

    if( fv != NULL )
    {
      pos = insert( pos, new SFormulaListVariable( fv ) ) + 1;
      continue;
    }

    ASSERT( false ); // unknown s-element
#endif
  }
  FireUpdateEvent();
}

SElement* SElementList::GetSElement( size_t index ) const
{
  ASSERT( index < size() );
  return (*this)[index];
}

void SElementList::RemoveAt( size_t index )
{
  ASSERT( index < size() );
  erase( begin() + index );
  FireUpdateEvent();
}

void SElementList::ReplaceSElement( size_t index, SElement *e )
{
  ASSERT( e != NULL );
  ASSERT( index < size() );
  SElement *o = (*this)[index];
  (*this)[index] = e;
  delete o;
  FireUpdateEvent();
}

void SElementList::ExchangeSElements( size_t a, size_t b )
{
  ASSERT( a < size() );
  ASSERT( b < size() );

  SElement *t = (*this)[a];
  (*this)[a] = (*this)[b];
  (*this)[b] = t;
  FireUpdateEvent();
}

bool SElementList::IsSubListOf( const SElementList* o, FormulaList *miss ) const
{
  if( o == NULL )
  {
    // only when empty it is the sub list of another list
    if( GetChildCount() == 0 )
      return true;

    if( miss != NULL ) for( size_t i = 0; i < GetChildCount(); i++ )
    {
      SFormula *sf = GetSElement(i)->CastTo<SFormula>();
      ASSERT( sf != NULL );
  
      miss->AddRight( sf->GetFormula() );
    }

    return false;
  }

  bool ok = true;

  for( size_t i = 0; i < GetChildCount(); i++ )
  {
    SFormula *sf = GetSElement(i)->CastTo<SFormula>();
    ASSERT( sf != NULL );

    Formula *f = sf->GetFormula();

    bool found = false;
    for( size_t j = 0; j < o->GetChildCount(); j++ )
    {
      SFormula *sg = o->GetSElement(j)->CastTo<SFormula>();
      ASSERT( sg != NULL );

      Formula *g = sg->GetFormula();

      if( g == f )
      {
        found = true;
        break;
      }
    }

    if( !found )
    {
      if( miss == NULL )
        return false;

      miss->AddRight( f );
      ok = false;
    }
  }

  return ok;
}

bool SElementList::ContainsFormula( const Formula *f ) const
{
  for( size_t i = 0; i < GetChildCount(); i++ )
  {
    SFormula *sf = GetSElement(i)->CastTo<SFormula>();
    ASSERT( sf != NULL );

    if( f == sf->GetFormula() )
      return true;
  }

  return false;
}

size_t SElementList::CountOfFormula( const Formula *f ) const
{
  size_t count = 0;

  for( size_t i = 0; i < GetChildCount(); i++ )
  {
    SFormula *sf = GetSElement(i)->CastTo<SFormula>();
    ASSERT( sf != NULL );

    if( f == sf->GetFormula() )
      count++;
  }

  return count;
}

bool SElementList::FindFormula( Formula *f, size_t &index ) const
{
  for( ; index < GetChildCount(); index++ )
  {
    SFormula *sf = GetSElement(index)->CastTo<SFormula>();
    ASSERT( sf != NULL );

    if( f == sf->GetFormula() )
      return true;
  }

  return false;
}

bool SElementList::IsSubListAtIndex( const FormulaList *list,
                                     size_t index ) const
{
  ASSERT( list != NULL );
  ASSERT( list->GetChildCount() > 0 );

  size_t last_index = index + list->GetChildCount() - 1;

  if( last_index >= GetChildCount() )
    return false;

  for( size_t j = 0, i = index; i < last_index + 1; i++, j++ )
  {
    DataNode *oitem = list->GetElement(j);
    DataNode *myitem = GetSElement(i)->GetChild(0);

    if( myitem != oitem )
      return false;
  }

  return true;
}

bool SElementList::FindSubListAtIndex( const FormulaList *list,
                                       size_t &index ) const
{
  ASSERT( list != NULL );
  ASSERT( list->GetChildCount() > 0 );

  for( size_t i = index; i < GetChildCount(); i++ )
  {
    if( IsSubListAtIndex( list, i ) )
    {
      index = i;
      return true;
    }
  }

  return false;
}

#if 0
/// Verify index before calling this function!
void SElementList::ReplaceSFormulas( FormulaListVariable *fv,
                                     size_t start_index )
{
  iterator i_start = begin() + start_index;
  iterator i_end = i_start + fv->GetChildCount();

  for( iterator i = i_start; i != i_end; i++ )
    delete (*i);

  iterator i_new_pos = erase( i_start, i_end );

  SFormulaListVariable *sfv =
    new SFormulaListVariable( fv );

  insert( i_new_pos, sfv );
}

/// Verify index before calling this function!
void SElementList::UnsubstSFormulaListVar( FormulaListVariable *fv,
                                           size_t index )
{
#if 0
  SFormulaListVariable *fv_old =
    dynamic_cast<SFormulaListVariable*>( *( begin() + index ) );

  ASSERT( fv_old != NULL );

  iterator pos = erase( begin() + index );

  for( size_t i = 0; i < fv->GetChildCount(); i++ )
  {
    Formula *f = fv->GetChild(i)->CastTo<Formula>();

    if( f != NULL )
    {
      pos = insert( pos, new SFormula( f ) ) + 1;
      continue;
    }

    FormulaListVariable *fv_ =
      fv->GetChild(i)->CastTo<FormulaListVariable>();

    if( fv_ != NULL )
    {
      pos = insert( pos, new SFormulaListVariable( fv_ ) ) + 1;
      continue;
    }

    ASSERT( false );
  }

  delete fv_old;
#endif
}
#endif

IMPLEMENT_APPLY( SElementList, Sequent )

SElementList* SElementList::Clone() const
{
  if( GetChildCount() == 0 )
    return NULL;

  SElementList* l = new SElementList;

  for( size_t i = 0; i < GetChildCount(); i++ )
    l->AddRight( GetSElement(i)->Clone() );

  return l;
}

std::ostream& SElementList::ToString( std::ostream &out ) const
{
  out << "SElementList" <<  std::endl;
  return out;
}

////////////////////////////////////////////////////////////////////////////////

SFormula::SFormula( Formula* f )
{
  ASSERT( f != NULL );

  m_formula = f;
  m_formula->Attach();
}

SFormula::~SFormula()
{
  m_formula->Detach();
}

IMPLEMENT_APPLY( SFormula, Sequent )

DataNode* SFormula::GetChild( size_t index ) const
{
  ASSERT( index == 0 );
  return m_formula;
}

SElement* SFormula::Clone() const
{
  return new SFormula( GetFormula() );
}

#if 0
void SFormula::SetFormula( Formula *f )
{
  ASSERT( f != NULL );
  m_formula->Detach();
  m_formula = f;
  m_formula->Attach();
}
#endif

void SFormula::InsertIntoFormulaList( FormulaList *list, size_t index )
{
  list->InsertAt( index, GetFormula() );
}

void SFormula::InsertIntoSElementList( SElementList *list, size_t index )
{
  list->InsertAt( index, new SFormula( GetFormula() ) );
}


////////////////////////////////////////////////////////////////////////////////

SFormulaListVariable::SFormulaListVariable( FormulaListVariable *fv )
{
  ASSERT( fv != NULL );
  m_formula_list_var = fv;
  m_formula_list_var->Attach();
}

SFormulaListVariable::~SFormulaListVariable()
{
  m_formula_list_var->Detach();
}

IMPLEMENT_APPLY( SFormulaListVariable, Sequent )

DataNode* SFormulaListVariable::GetChild( size_t index ) const
{
  ASSERT( index == 0 );
  return (FormulaListElement*) m_formula_list_var; // problem?
}

SElement* SFormulaListVariable::Clone() const
{
  return new SFormulaListVariable( GetFormulaListVariable() );
}

void SFormulaListVariable::InsertIntoFormulaList( FormulaList *list,
                                                  size_t index )
{
  list->InsertAt( index, m_formula_list_var );
}

void SFormulaListVariable::InsertIntoSElementList( SElementList *list,
                                                   size_t index )
{
  list->InsertAt( index, new SFormulaListVariable( m_formula_list_var ) );
}

std::ostream& SFormulaListVariable::ToString( std::ostream &out ) const
{
  out << "SFormulaListVariable" <<  std::endl;
  return out;
}

////////////////////////////////////////////////////////////////////////////////

Sequent::Sequent( SElementList *l, SElementList *r )
{
  m_leftside  = l;
  m_rightside = r;

  if( m_leftside != NULL )
    m_leftside->AttachSElementListEventClient( this );
  if( m_rightside != NULL )
    m_rightside->AttachSElementListEventClient( this );
}

void Sequent::SetLeft( SElementList* l )
{
  if( m_leftside != NULL )
  {
    m_leftside->DetachSElementListEventClient( this );
    delete m_leftside;
  }

  m_leftside  = l;

  if( m_leftside != NULL )
    m_leftside->AttachSElementListEventClient( this );
}

void Sequent::SetRight( SElementList* l )
{
  if( m_rightside != NULL )
  {
    m_rightside->DetachSElementListEventClient( this );
    delete m_rightside;
  }

  m_rightside = l;

  if( m_rightside != NULL )
    m_rightside->AttachSElementListEventClient( this );
}

Sequent::~Sequent()
{
  if( m_leftside != NULL )
  {
    m_leftside->DetachSElementListEventClient( this );
    delete m_leftside;
  }

  if( m_rightside != NULL )
  {
    m_rightside->DetachSElementListEventClient( this );
    delete m_rightside;
  }
}

IMPLEMENT_APPLY( Sequent, Sequent )

DataNode* Sequent::GetChild( size_t index ) const
{
  ASSERT( index < 2 );

  if( index == 0 )
    return GetLeft();

  return GetRight();
}

void Sequent::AddLeft( SElement* e )
{
  ASSERT( e != NULL );

  if( GetLeft() == NULL )
    m_leftside = new SElementList;

  GetLeft()->AddLeft( e );

  FireUpdateEvent();
}

void Sequent::AddRight( SElement* e )
{
  ASSERT( e != NULL );

  if( GetRight() == NULL )
    m_rightside = new SElementList;

  GetRight()->AddRight( e );

  FireUpdateEvent();
}

void Sequent::AddLeft( Formula* f )
{
  ASSERT( f != NULL );

  AddLeft( new SFormula( f ) );
}

void Sequent::AddRight( Formula* f )
{
  ASSERT( f != NULL );

  AddRight( new SFormula( f ) );
}

Sequent* Sequent::Clone() const
{
  SElementList* l = NULL;
  SElementList* r = NULL;

  if( GetLeft() != NULL )
    l = GetLeft()->Clone();

  if( GetRight() != NULL )
    r = GetRight()->Clone();

  return new Sequent( l, r );
}

bool Sequent::IsSubSequentOf( const Sequent* o,
                              FormulaList *miss_l,
                              FormulaList *miss_r ) const
{
  bool ok = true;

  ASSERT( o != NULL );

  if( GetLeft() != NULL && !GetLeft()->IsSubListOf( o->GetLeft(), miss_l ) )
  {
    if( miss_r == NULL )
      return false;
  
    ok = false;
  }

  if( GetRight() != NULL && !GetRight()->IsSubListOf( o->GetRight(), miss_r ) )
    return false;

  return ok;
}

bool Sequent::ContainsFormula( const Formula *f, bool left ) const
{
  SElementList *l = ( left ? GetLeft() : GetRight() );

  if( l == NULL )
    return false;

  return l->ContainsFormula( f );
}

size_t Sequent::CountOfFormula( const Formula *f, bool left ) const
{
  SElementList *l = ( left ? GetLeft() : GetRight() );

  if( l == NULL )
    return 0;

  return l->CountOfFormula( f );
}

void Sequent::SElementListUpdateEvent( const SElementList *list )
{
  FireUpdateEvent();
}

std::ostream& Sequent::ToString( std::ostream &out ) const
{
  out << "Sequent" <<  std::endl;
  return out;
}

////////////////////////////////////////////////////////////////////////////////

void SElementListNotifier::
AttachSElementListEventClient( SElementListEventClient *client )
{
  m_clients.insert( client );
}

void SElementListNotifier::
DetachSElementListEventClient( SElementListEventClient *client )
{
  m_clients.erase( client );
}

void SElementListNotifier::FireUpdateEvent() const
{
  std::set<SElementListEventClient*>::const_iterator i;

  for( i = m_clients.begin(); i != m_clients.end(); i++ )
    (*i)->SElementListUpdateEvent( dynamic_cast<const SElementList*>( this ) );
}

////////////////////////////////////////////////////////////////////////////////

void SequentNotifier::AttachSequentEventClient( SequentEventClient *client )
{
  m_clients.insert( client );
}

void SequentNotifier::DetachSequentEventClient( SequentEventClient *client )
{
  m_clients.erase( client );
}

void SequentNotifier::FireUpdateEvent() const
{
  std::set<SequentEventClient*>::const_iterator i;

  for( i = m_clients.begin(); i != m_clients.end(); i++ )
    (*i)->SequentUpdateEvent( dynamic_cast<const Sequent*>( this ) );
}

////////////////////////////////////////////////////////////////////////////////

} // namespace ProofLib

