////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: terms.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// TERMS IMPLEMENTATION
////////////////////////////////////////////////////////////////////////////////
#ifdef DEBUG
#include <iostream>
#include <set>
#include <stack>
#include "util/hash.h"
#endif

#include <typeinfo>
#include "util/myassert.h"
#include "libproof.h"
#include "operations.h"


namespace ProofLib
{

////////////////////////////////////////////////////////////////////////////////

std::ostream& AbstractTermListBase::ToString( std::ostream &out ) const
{
  for( size_t i = 0; i < GetChildCount(); i++ )
  {
    if( i != 0 ) out << ",";

    GetTerm(i)->ToString(out);
  }

  return out;
}

////////////////////////////////////////////////////////////////////////////////

AbstractTermList::AbstractTermList( size_t arity )
{
}

size_t AbstractTermList::Hash() const
{
  size_t hash = 0;

  for( size_t i = 0; i < size(); i++ )
  {
    //size_t h = m_terms[i]->Hash();
    size_t h = size_t((*this)[i]); // >> 8;
    //h = h & 0xffff;

    //hash = (hash << 3) ^ h;
    if( i % 2 )
      hash = (hash << 1) ^ h;
    else
      hash = (hash << 1) ^ (h << 16);
  }

  return hash;
}
AbstractTermList::AbstractTermList( const StaticAbstractTermList& stl )
{
  for( size_t i = 0; i < stl.GetChildCount(); i++ )
    AddRight( stl.GetTerm(i) );
}

void AbstractTermList::AddLeft( AbstractTerm *t )
{
  t->Attach();
  insert( begin(), t );
}

void AbstractTermList::AddRight( AbstractTerm *t )
{
  t->Attach();
  push_back( t );
}

size_t AbstractTermList::GetChildCount() const
{
  return size();
}

AbstractTerm* AbstractTermList::GetTerm( size_t index ) const
{
  ASSERT( index < size() );

  return (*this)[index];
}

void AbstractTermList::SetTerms( AbstractTermList *other )
{
  clear();

  insert( begin(), other->begin(), other->end() );

  other->clear();
}


int AbstractTermList::FindTerm( const AbstractTerm *t ) const
{
  for( size_t i = 0; i < GetChildCount(); i++ )
    if( GetTerm(i) == t )
      return i;

  return -1;
}

void AbstractTermList::RemoveAt( size_t index )
{
  ASSERT( index < size() );

  AbstractTerm *t = *(begin() + index);

  erase( begin() + index );

  t->Detach();
}

////////////////////////////////////////////////////////////////////////////////


TermList::TermList( const StaticTermList& stl )
{
  for( size_t i = 0; i < stl.GetChildCount(); i++ )
    AddRight( stl.GetTerm(i) );
}

void TermList::AddLeft( Term *t )
{
  AbstractTermList::AddLeft( t );
}

void TermList::AddRight( Term *t )
{
  AbstractTermList::AddRight( t );
}

Term* TermList::GetTerm( size_t index ) const
{
  return dynamic_cast<Term*>( AbstractTermList::GetTerm( index ) );
}

void TermList::SetTerms( TermList *other )
{
  AbstractTermList::SetTerms( other );
}


int TermList::FindTerm( const Term *t ) const
{
  return AbstractTermList::FindTerm( t );
}

////////////////////////////////////////////////////////////////////////////////

StaticAbstractTermList::StaticAbstractTermList( size_t arity )
{
  typedef AbstractTerm* AbstractTermPtr;

  m_arity = arity;

  if( m_arity == 0 )
    m_terms = NULL;
  else
    m_terms = new AbstractTermPtr[arity];
}

StaticAbstractTermList::StaticAbstractTermList( const AbstractTermListBase *terms )
{
  typedef AbstractTerm* AbstractTermPtr;

  m_arity = 0;

  if( terms != NULL )
    m_arity = terms->GetChildCount();

  if( m_arity == 0 )
    m_terms = NULL;
  else
  {
    m_terms = new AbstractTermPtr[m_arity];

    SetTerms( terms );
  }
}

StaticAbstractTermList::~StaticAbstractTermList()
{
  if( m_terms != NULL )
    delete m_terms;
}

size_t StaticAbstractTermList::Hash() const
{
  size_t hash = 0;

  for( size_t i = 0; i < m_arity; i++ )
  {
    //size_t h = m_terms[i]->Hash();
    size_t h = size_t(m_terms[i]); // >> 8;
    //h = h & 0xffff;

    //hash = (hash << 3) ^ h;
    if( i % 2 )
      hash = (hash << 1) ^ h;
    else
      hash = (hash << 1) ^ (h << 16);
  }

  return hash;
}

void StaticAbstractTermList::SetTerms( const AbstractTermListBase *tl )
{
  ASSERT( tl->GetChildCount() == GetChildCount() );

  for( size_t i = 0; i < m_arity; i++ )
  {
    m_terms[i] = tl->GetTerm(i);
    m_terms[i]->Attach();
  }
}

AbstractTerm* StaticAbstractTermList::GetTerm( size_t index ) const
{
  ASSERT( index < m_arity );

  ASSERT( m_terms != NULL );

  return m_terms[index];
}

int StaticAbstractTermList::FindTerm( const AbstractTerm *t ) const
{
  for( size_t i = 0; i < GetChildCount(); i++ )
    if( GetTerm(i) == t )
      return i;

  return -1;
}

bool StaticAbstractTermList::DbEqual( const DbNode *n, VarMapping var_map ) const
{
  const StaticAbstractTermList *t = dynamic_cast<const StaticAbstractTermList*>(n);

  if( t == NULL )
    return false;

  if( m_arity != t->m_arity )
    return false;

  for( size_t i = 0; i < m_arity; i++ )
  {
    if( !m_terms[i]->DbEqual( t->m_terms[i], var_map ) )
      return false;
  }

  return true;
}

////////////////////////////////////////////////////////////////////////////////


StaticTermList::StaticTermList( size_t arity ) : StaticAbstractTermList( arity ) {}

StaticTermList::StaticTermList( const TermList *terms ) : StaticAbstractTermList( terms ) {}
StaticTermList::StaticTermList( const StaticTermList *terms ) : StaticAbstractTermList( terms ) {}

StaticTermList::~StaticTermList() {}

void StaticTermList::SetTerms( const TermList *tl )
{
  StaticAbstractTermList::SetTerms( tl ) ;
}

void StaticTermList::SetTerms( const StaticTermList *tl )
{
  StaticAbstractTermList::SetTerms( tl ) ;
}

Term* StaticTermList::GetTerm( size_t index ) const
{
  return dynamic_cast<Term*>( StaticAbstractTermList::GetTerm( index ) );
}

int StaticTermList::FindTerm( const Term *t ) const
{
  return StaticAbstractTermList::FindTerm( t );
}

////////////////////////////////////////////////////////////////////////////////

Variable::Variable( const Symbol *e ) : DbSymbolNode( e ) { }

size_t Variable::Hash() const
{
  return GetSymbol()->Hash();
}

DataNode* Variable::Apply( Operation &op_ )
{ 
  try
  {
    ExpressionOperation &op = \
      dynamic_cast<ExpressionOperation&>( op_ );
	return op.ApplyToVariable( this );
  }
  catch( std::bad_cast )
  {
  }
  return this;
  }

bool Variable::ApplyCompare( CompareOperation &op_, const DataNode *other )
const
{
  try
  {
    ExpressionCompareOperation &op = 
      dynamic_cast<ExpressionCompareOperation&>( op_ ); 
      
    return op.ApplyToVariable( this, other ); 
  } 
  catch( std::bad_cast ) 
  { 
  }
  return false; 
}

Variable* Variable::New( NodeDatabase *ndb, const Symbol *e )
{
  Variable *v = new Variable( e );
  return v->AddToNodeDatabase<Variable>( ndb );
}

std::ostream& Variable::ToString( std::ostream &out ) const
{
  SymbolNode::ToString(out);
  return out;
}

////////////////////////////////////////////////////////////////////////////////

Constant::Constant( const Symbol *e ) : DbSymbolNode( e ) { }

size_t Constant::Hash() const
{
  return GetSymbol()->Hash();
}

DataNode* Constant::Apply( Operation &op_ )
{ 
  try
  {
    ExpressionOperation &op = \
      dynamic_cast<ExpressionOperation&>( op_ );
	return op.ApplyToConstant( this );
  }
  catch( std::bad_cast )
  {
  }
  return this;
  }

bool Constant::ApplyCompare( CompareOperation &op_, const DataNode *other )
const
{
  try
  {
    ExpressionCompareOperation &op = 
      dynamic_cast<ExpressionCompareOperation&>( op_ ); 
      
    return op.ApplyToConstant( this, other ); 
  } 
  catch( std::bad_cast ) 
  { 
  }
  return false; 
}

Constant* Constant::New( NodeDatabase *ndb, const Symbol *e )
{
  Constant *c = new Constant( e );
  return c->AddToNodeDatabase<Constant>( ndb );
}

std::ostream& Constant::ToString( std::ostream &out ) const
{
  SymbolNode::ToString(out);
  return out;
}

////////////////////////////////////////////////////////////////////////////////

Function::Function( const Symbol *e, int arity ) :
  DbSymbolNode( e ),
  StaticAbstractTermList( arity )
{
}

bool Function::DbEqual( const DbNode *n, VarMapping var_map ) const
{
  const Function *t = dynamic_cast<const Function*>(n);

  if( t == NULL )
    return false;

  return ( DbSymbolNode::DbEqual( n, var_map ) && StaticAbstractTermList::DbEqual( n, var_map ) );
}

size_t Function::Hash() const
{
  return (StaticAbstractTermList::Hash() << 1) ^ GetSymbol()->Hash();
}

DataNode* Function::Apply( Operation &op_ )
{ 
  try
  {
    ExpressionOperation &op = \
      dynamic_cast<ExpressionOperation&>( op_ );
	return op.ApplyToFunction( this );
  }
  catch( std::bad_cast )
  {
  }
  return this;
  }

bool Function::ApplyCompare( CompareOperation &op_, const DataNode *other )
const
{
  try
  {
    ExpressionCompareOperation &op = 
      dynamic_cast<ExpressionCompareOperation&>( op_ ); 
      
    return op.ApplyToFunction( this, other ); 
  } 
  catch( std::bad_cast ) 
  { 
  }
  return false; 
}

Function* Function::ReplaceTerms( AbstractTermList &terms ) const
{
  return New( GetNodeDatabase(), GetSymbol(), &terms );
}

Function* Function::New( NodeDatabase *ndb, const Symbol *e,
                         AbstractTermList *terms )
{
  ASSERT( terms != NULL );

  Function *f = new Function( e, terms->GetChildCount() );

  f->SetTerms( terms );

  return f->AddToNodeDatabase<Function>( ndb );
}

bool Function::HaveEqualSymbol( const Function *o ) const
{
  return GetSymbol() == o->GetSymbol();
}

std::ostream& Function::ToString( std::ostream &out ) const
{
  SymbolNode::ToString(out);

  if( !IsEmpty() )
  {
    out << "(";
    StaticAbstractTermList::ToString(out);
    out << ")";
  }

  return out;
}

////////////////////////////////////////////////////////////////////////////////

} // namespace ProofLib

