////////////////////////////////////////////////////////////////////////////////
//
// Proof Library for C++ -------------------------------------------------------
//
// File: util/libxml2/xmlreadimpl.cpp
// Copyright 2005-2006 by Hendrik Spohr, Vienna, AUSTRIA
//
////////////////////////////////////////////////////////////////////////////////
#include "util/myassert.h"
#include "util/fileexceptions.h"
#include "util/xmlread.h"
#include "chars/latexchars.h"
#include "util/proofdbdtd.h"

#include <libxml/xmlreader.h>
#include <libxml/catalog.h>

////////////////////////////////////////////////////////////////////////////////

class XmlReadException : public Exception
{
  xmlTextReaderPtr       m_xml_reader;

public:
  XmlReadException( xmlTextReaderPtr ptr ) { m_xml_reader = ptr; }

  virtual void Print( std::ostream &out ) const
  {
    using namespace std;

    out << "XML error";

    if( xmlTextReaderCurrentNode(m_xml_reader) != NULL )
      out << " at line " << xmlTextReaderCurrentNode(m_xml_reader)->line;

    if( xmlTextReaderConstName(m_xml_reader) != NULL )
      out << " and node "
          << '\'' << (const char*) xmlTextReaderConstName(m_xml_reader) << "' ";

    out << " (depth: " << xmlTextReaderDepth(m_xml_reader) << "): ";
  }
};

class XmlTextReaderErrorException : public XmlReadException
{
  std::string            m_msg;
  xmlTextReaderLocatorPtr m_locator;

public:
  XmlTextReaderErrorException( xmlTextReaderPtr ptr, const char *msg,
                               xmlParserSeverities severity,
                               xmlTextReaderLocatorPtr locator )
    : XmlReadException( ptr )
  {
    m_locator = locator;

    if( msg != NULL )
      m_msg = msg;
  }

  virtual void Print( std::ostream &out ) const
  {
    using namespace std;

    XmlReadException::Print( out );

    if( m_locator != NULL )
      out << xmlTextReaderLocatorLineNumber( m_locator ) << "::";

    if( !m_msg.empty() )
      out << m_msg << ".";
  }
};

class XmlMissingChildException : public XmlReadException
{
  std::string            m_name;

public:
  XmlMissingChildException( xmlTextReaderPtr ptr, const char *name )
    : XmlReadException( ptr )
  {
    if( name != NULL )
      m_name = name;
  }

  virtual void Print( std::ostream &out ) const
  {
    using namespace std;

    XmlReadException::Print( out );

    out << "missing child";

    if( !m_name.empty() )
      out << " '" << m_name << '\'';

    out << '.';
  }
};

class XmlInvalidUTF8Exception : public XmlReadException
{
  std::string            m_name;

public:
  XmlInvalidUTF8Exception( xmlTextReaderPtr ptr, const char *name )
    : XmlReadException( ptr )
  {
    if( name != NULL )
      m_name = name;
  }

  virtual void Print( std::ostream &out ) const
  {
    using namespace std;

    XmlReadException::Print( out );

    out << "invalid utf8 in attribute: ";

    if( !m_name.empty() )
      out << " '" << m_name << '\'';

    out << '.';
  }
};

class XmlUnexpectedElementException : public XmlReadException
{
  std::string            m_name;
  std::string            m_name_expected;

public:
  XmlUnexpectedElementException( xmlTextReaderPtr ptr, const char *name )
    : XmlReadException( ptr )
  {
    if( name != NULL )
      m_name = name;
  }

  XmlUnexpectedElementException( xmlTextReaderPtr ptr, const char *name,
                        const char *name_expected )
    : XmlReadException( ptr )
  {
    if( name != NULL )
      m_name = name;
    if( name_expected != NULL )
      m_name_expected = name_expected;
  }

  virtual void Print( std::ostream &out ) const
  {
    using namespace std;

    XmlReadException::Print( out );

    out << "unexpected element";

    if( !m_name.empty() )
      out << " '" << m_name << '\'';

    if( !m_name_expected.empty() )
      out << " expected '" << m_name_expected << '\'';

    out << '.';
  }
};

class XmlMissingAttribute : public XmlReadException
{
  std::string            m_name;

public:
  XmlMissingAttribute( xmlTextReaderPtr ptr, const char *name )
    : XmlReadException( ptr )
  {
    if( name != NULL )
      m_name = name;
  }

  virtual void Print( std::ostream &out ) const
  {
    using namespace std;

    XmlReadException::Print( out );

    out << "XmlMissingAttribute";
  }
};

////////////////////////////////////////////////////////////////////////////////

class LibXml2XmlReader : public XmlReader
{
  xmlTextReaderPtr       m_xml_reader;
  Exception             *m_waiting_exception;

protected:
  LibXml2XmlReader()
  {
    m_xml_reader = NULL;
    m_waiting_exception = NULL;
  }

  friend class XmlReader;

public:
  operator xmlTextReaderPtr() const
  {
    return m_xml_reader;
  }

  /// \return false if end of input is reached.
  virtual bool SeekToElement()
  {
// cerr << "seeking... " << endl;
    if( xmlTextReaderRead( m_xml_reader ) <= 0 )
    {
      CheckForWaitingException();
      return false;
    }

    while( XML_ELEMENT_NODE != xmlTextReaderNodeType( m_xml_reader ) )
    {
      if( xmlTextReaderNodeType( m_xml_reader ) == 0 ||
          xmlTextReaderRead( m_xml_reader ) <= 0 )
          
      {
        CheckForWaitingException();
        return false;
      }
    }

// cerr << "seeked to: " << xmlTextReaderConstName(m_xml_reader)
//      << " at depth: " << xmlTextReaderDepth(m_xml_reader) << endl;

    return true;
  }

  /// \param  depth the depth of the parent node
  /// \return false if no child was found
  virtual bool SeekToChildElement( size_t depth )
  {
    if( !SeekToElement() )
      return false;

    return ( CurrentDepth() == depth + 1 );
  }

  /// \param  depth the depth of the previous node
  /// \return false if no next element exists
  virtual bool SeekToNextElement( size_t depth )
  {
    if( !SeekToElement() )
      return false;

    while( CurrentDepth() > depth )
      if( !SeekToElement() )
        return false;

    return ( CurrentDepth() == depth );
  }

  /// \return depth of the current node
  virtual size_t CurrentDepth()
  {
    return xmlTextReaderDepth( m_xml_reader );
  }

  virtual char* GetElementName()
  {
    const xmlChar *utf = xmlTextReaderConstName( m_xml_reader );

    CheckForWaitingException();

    return strdup( (const char*) utf );
  }

  virtual void AssertCurrentElement( const char *name )
  {
    const xmlChar *utf =  xmlTextReaderConstName(m_xml_reader);

    CheckForWaitingException();
#if 0
cerr << "assert " << name << ": " << utf << endl;
#endif

    if( xmlStrcasecmp( utf, (xmlChar*) name ) != 0 )
      throw new XmlUnexpectedElementException( m_xml_reader,
                                               (const char*) utf, name );
  }


  virtual char* GetCStrAttribute( const char* attr_name, bool optional = false )
  {
    const xmlChar *utf =
      xmlTextReaderGetAttribute( m_xml_reader, (xmlChar*) attr_name );

    CheckForWaitingException();

    if( !optional && utf == NULL )
      throw new XmlMissingAttribute( m_xml_reader, attr_name );

    return strdup( (const char*) utf );
  }

  virtual USTRING GetAttribute( const char* attr_name, bool optional = false )
  {
    const xmlChar *utf =
      xmlTextReaderGetAttribute( m_xml_reader, (xmlChar*) attr_name );

    CheckForWaitingException();

    if( !optional && utf == NULL )
      throw new XmlMissingAttribute( m_xml_reader, attr_name );

    USTRING val;

    if( !UStringFromUTF8( (const char*) utf, val ) )
      throw new XmlInvalidUTF8Exception( m_xml_reader, attr_name );

    return LatexConvertor.KnownLatexSymbolsToUnicode( val );
  }

  virtual void Open( const char *sourcename )
  {
    m_xml_reader = xmlReaderForFile( sourcename, NULL,
                                     XML_PARSE_NOENT |
                                     XML_PARSE_DTDATTR |
                                     XML_PARSE_DTDVALID |
				     XML_PARSE_HUGE); 

    if( m_xml_reader == NULL )
      throw new FileOpenFailException( sourcename, errno );

    xmlTextReaderSetErrorHandler( m_xml_reader,
      LibXml2XmlReader::XmlTextReaderErrorFunc, this );
  }

  virtual void Close()
  {
    ASSERT( m_xml_reader != NULL );

    xmlFreeTextReader( m_xml_reader );
  }

  virtual void ThrowMissingChildException( const char* name )
  {
    throw new XmlMissingChildException( m_xml_reader, name );
  }

  virtual void ThrowUnexpectedElement( const char* name )
  {
    throw new XmlUnexpectedElementException( m_xml_reader, name );
  }


private:
  static void XmlTextReaderErrorFunc( void *arg, const char *msg, 
                                      xmlParserSeverities severity, 
                                      xmlTextReaderLocatorPtr locator )
  {
    ASSERT( arg != NULL );

    ((LibXml2XmlReader*) arg)->XmlTextReaderError( msg, severity, locator );
  }

  void XmlTextReaderError( const char *msg, xmlParserSeverities severity,
                           xmlTextReaderLocatorPtr locator )
  {
    m_waiting_exception =
      new XmlTextReaderErrorException( m_xml_reader, msg, severity, locator );
  }

  void CheckForWaitingException()
  {
    if( m_waiting_exception == NULL )
     return;
    
    throw m_waiting_exception;

    m_waiting_exception = NULL;
  }
};

////////////////////////////////////////////////////////////////////////////////

XmlReader* XmlReader::New()
{
  return new LibXml2XmlReader;
}

void XmlReader::Init()
{
  xmlInitializeCatalog();

  xmlNewCatalog(0);
}

void XmlReader::AddCatalogSystemItem( const char *dtd_name,
                                      const char *file_name )
{
  if( xmlCatalogAdd(
        (const xmlChar*) "system",
        (const xmlChar*) dtd_name,
        (const xmlChar*) file_name
    ) != 0 )
  {
    GenericException *e = new GenericException();
    e->Stream() << "Can not add XML catalog item " << dtd_name << " which "
                << "should be located in " << file_name << ".";
    throw e;
  }
}
